/*
 * csggraph.cc: Routines for class CSGgraph and supporting code.
 *
 * Contains code to create and maintain the CSGgraph representation of
 * an object.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

/*
 *  Routines for objectnode
 */

void objectnode::set_defaults(TransformList *transformlist)
{
  index = modeller->csg_graph->new_index(this);
  last_traverse_id = 0;
  if (transformlist == NULL) {
    this->transformlist = new TransformList();
    Transform::unit(position);
  } else {
    this->transformlist = transformlist;
    transformlist->get_position(this->position);
  }
  parenttracker = new ParentTracker(this);
  copytracker = NULL;
  layoutnode = NULL;
  is_selected = FALSE;
  linkptr = NULL;
  link_type = PLUS;
  linked_by = NULL;
  strcpy(name, "");
}

void objectnode::set_name(char *new_name)
{
  strncpy(name, new_name, OBJECTNODE_NAMELEN);
  if (name[OBJECTNODE_NAMELEN - 1] != '\0')
    name[OBJECTNODE_NAMELEN - 1] = '\0';
}

objectnode::~objectnode(void)
{
  // NB: All objectnodes should be deleted through
  // the CSGgraph's delete_objectnode. This deletes
  // things like copytrackers, etc. automatically.
}

void objectnode::select(void)
{
  is_selected = TRUE;
  copytracker->select_copies();
}

void objectnode::deselect(void)
{
  is_selected = FALSE;
  copytracker->deselect_copies();
}

/*
 *  Routines for groupobjectnode
 */

// Create a groupobjectnode as a wrapper for an existing objectnode.
// Note that since this objectnode is not active, it has no CopyTracker.
groupobjectnode::groupobjectnode(Modeller *m, objectnode *group,
				 TransformList *transformlist)
{
  this->group = group;
  this->modeller = m;

  set_defaults(transformlist);
}

void groupobjectnode::deallocate_csggraph(void)
{
  objectnode *child;
  Tracker *current_tracker;
  primtreenode *current_copy;
  primwrappernode *above_wrapper;
  groupobjectnode *above_groupnode;

  // First: this portion of the CSG tree is being deleted. Find any
  // wrappers that point to this node, and set them to NULL to indicate
  // that the below tree has been deleted.
  for (current_tracker = copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {
    current_copy = current_tracker->copy;
    
    // Find the corresponding primtreenode's parent's original CSGtree node.
    if (current_copy->parent_type == WRAPPER) {
      if (current_copy->parent != NULL) {
	assert(current_copy->parent->type() == PRIMTREENODE_WRAPPER);
	above_wrapper = (primwrappernode *) current_copy->parent;

	assert(above_wrapper->original != NULL);
	assert(above_wrapper->original->type() == OBJECTNODE_GROUP);
	above_groupnode = (groupobjectnode *) above_wrapper->original;

	// Set the above groupnode's child pointer to point to NULL,
	// as we are now deallocating this portion of the tree.
	above_groupnode->group = NULL;
      }
    }
  }
	
  // Delete the child node, unless it's already been deleted.
  child = group;
  if (child != NULL) {
    child->deallocate_csggraph();
    delete child;
  }

  // Delete any nodes we're linked to.
  if (linkptr != NULL) {
    linkptr->deallocate_csggraph();
    delete linkptr;
  }
}

// Print out the CSG graph.
void groupobjectnode::print_csggraph(int indent, int traverse_id)
{
  Tracker *current_tracker;

  if (traverse_id == last_traverse_id) {
    printf("%*sC%03d/Grp already printed\n", indent, "", index);
    return;
  }

  printf("%*sC%03d/Grp T(",
	 indent, "", index);
  
  for (current_tracker = copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {
    printf("P%03d", current_tracker->copy->index);
    if (current_tracker->next != NULL)
      printf(" ");
    else
      printf(") ");
  }

  last_traverse_id = traverse_id;

  if (linked_by == NULL)
    printf("linked_by NULL\n");
  else
    printf("linked_by C%03d\n", linked_by->index);

  printf("%*sC%03d/Grp print parent-tracker\n", indent, "", index);
  parenttracker->print(indent);

  printf("%*sC%03d/Grp traverse wrapped\n", indent, "", index);
  group->print_csggraph(indent+1, traverse_id);

  if (linkptr == NULL)
    printf("%*sC%03d/Grp link NULL\n", indent, "", index);
  else {
    printf("%*sC%03d/Grp traverse \'%c\' link\n", indent, "", index,
	   (link_type == PLUS) ? '+' : 
	   (link_type == MINUS) ? '-' :
	   (link_type == AND) ? '&' : '?');
    linkptr->print_csggraph(indent+1, traverse_id);
  }
}

void groupobjectnode::initiate_copytracker_pass(objectnode *obj_group,
					       Primtree *primtree)
{
  CopyTracker *group_ct;
  primwrappernode *linked_wrappernode;
  primcsgnode *linknode;
  Tracker *current_tracker;
  primwrappernode *parentnode;

  // Find the CopyTracker for this group's above node.
  assert(obj_group != NULL);
  group_ct = obj_group->copytracker;

  // Create a new CopyTracker for this groupobjectnode.
  copytracker = new CopyTracker(this);

  /*
   * For each copy that is now being made of this structure...
   */

  for (current_tracker = group_ct->trackerlist_head; current_tracker != NULL;
       current_tracker = current_tracker->next) {

    /*
     * Create primtreenodes and update the primtree links and linktype.
     */

    // Create a new primcsgnode that will link this group into the
    // graph.
    linknode = new primcsgnode(modeller);
    
    // Find the primwrappernode that is the copy of obj_group.
    assert(current_tracker->copy->type() == PRIMTREENODE_WRAPPER);
    parentnode = (primwrappernode *) current_tracker->copy;
    
    // The new linknode's left side is the primcsgnode structure that
    // parentnode wraps. Note that this primcsgnode is now UNCHANGED,
    // as it is the base for this addition.
    linknode->left = parentnode->child;
    linknode->left->parent = linknode;
    linknode->left->parent_type = CSG_LEFT;
    linknode->left->octree->clear_changed(linknode->left->octree,
					  &(linknode->left->octree->octree));
    
    // Create a new primwrappernode that will represent this copied group.
    linked_wrappernode = new primwrappernode(modeller, NULL);

    // The new linknode's right side is the new primwrappernode.
    linknode->right = linked_wrappernode;
    linknode->right->parent = linknode;
    linknode->right->parent_type = CSG_RIGHT;

    // The linktype is that of this object, the groupobjectnode.
    linknode->csg_type = this->link_type;

    // Update the link from the wrapper node 'parentnode' to point to
    // linknode, which represents the new CSG structure.
    parentnode->child = linknode;
    parentnode->child->parent = parentnode;
    parentnode->child->parent_type = WRAPPER;

    /*
     * Update the links from the Primtree back to the CSGgraph.
     */
    
    linknode->original = NULL;
    linknode->this_tracker = NULL;

    linked_wrappernode->original = this;

    /*
     * Add a new Tracker into this node's CopyTracker for the newly-created
     * primwrappernode.
     */

    linked_wrappernode->this_tracker =
      copytracker->add_new_tracker(linked_wrappernode);

    /*
     * Update the position matrices in the primtree.
     */

    // The position of the new linknode is equal to the position of
    // its left child.
    Transform::set(linknode->position, linknode->left->position);

    // The position of the new group is equal to this position
    // combined with the transform applied to the new groupobjectnode.
    Transform::combine(this->position, linknode->position, 
		       linknode->right->position);


    /*
     * Create the new primtreenodes for the copied group. As this is done,
     * add to their respective CopyTrackers.
     *
     * Note that we could copy any of the copies of this->group, so we
     * arbitrarily choose to copy the first one, which is
     * this->group->copytracker->trackerlist_head->copy.
     */

    assert(group->copytracker != NULL);
    assert(group->copytracker->trackerlist_head != NULL);

    linked_wrappernode->child = group->copytracker->trackerlist_head
      ->copy->copy_and_update_primtree(linked_wrappernode, WRAPPER, primtree);

    assert(linked_wrappernode->child->octree != NULL);
    linked_wrappernode->octree = linked_wrappernode->child->octree;

    /*
     * Update the SolidOctrees further up the graph.
     *
     * The copy_and_update_primtree pass has created new SolidOctrees for
     * the copied group (linked_wrappernode's child) and this has been set
     * for linked_wrappernode, but linknode does not have a SolidOctree yet.
     */

    // First: construct the SolidOctree for linknode. Note that this
    // SolidOctree should have the CHANGED flag wherever linking in the
    // new right-hand-side has caused a change.
    //
    // The RHS will cause changes wherever it's non-empty for any
    // operation except AND, where it will cause changes wherever it's
    // non-full.
    //
    // Note: this change is necessary because it's tough to propagate
    // the AND / non-AND linktypes down through the CSG graph at
    // octree-creation time.
    //
    // After this, create linknode's octree as a combination.
    if (this->link_type == AND)
      linked_wrappernode->octree->set_all_nonequal_to_changed(
						SOLIDINDEXTYPE_FULL);
    else
      linked_wrappernode->octree->set_all_nonequal_to_changed(
						SOLIDINDEXTYPE_EMPTY);
    
    // Call the SolidOctree constructor for primcsgnodes.
    linknode->octree = new SolidOctree(linknode, primtree->owner);

    // This octree needs to be pointed to by every primwrappernode that exists
    // in a chain above linknode. This includes but IS NOT LIMITED TO
    // parentnode.
    primtree->fix_wrapperoctreeptr(parentnode, parentnode->octree,
				   linknode->octree);
    
    // linknode's octree has now been changed, as the result of a topology
    // manipulation operation. It will need to be propagated during the
    // SolidOctree update pass.
    linknode->octree->has_changed = TRUE;
  }
}


/*
 *  Routines for primobjectnode
 */

// Create a new primitive objectnode for a CSGgraph.
// Note that since this objectnode is not active, it has no CopyTracker.
primobjectnode::primobjectnode(Modeller *m, Primitive *prim,
			       TransformList *transformlist)
{
  this->prim = prim;
  this->modeller = m;

  // We don't have any Collections until a primtreeprimnode creates an
  // octree that references us.
  collections_found = FALSE;
  num_collections = 0;
  collection_bufsize = 0;
  collectionbuf = NULL;
  worldcollectionbuf = NULL;

  set_defaults(transformlist);  
}

// Remove this object's Collections.
void primobjectnode::remove_collections(void)
{
  if (collections_found && (num_collections > 0)) {
    assert(collectionbuf != NULL);

    // Remove a reference from each Collection in our buffer.
    for (int i = 0; i < num_collections; i++)
      if (collectionbuf[i]->remove_reference())
        delete collectionbuf[i];

    free(collectionbuf);
    free(worldcollectionbuf);
  }
}

void primobjectnode::deallocate_csggraph(void)
{
  Tracker *current_tracker;
  primtreenode *current_copy;
  primwrappernode *above_wrapper;
  groupobjectnode *above_groupnode;

  // First: this portion of the CSG tree is being deleted. Find any
  // wrappers that point to this node, and set them to NULL to indicate
  // that the below tree has been deleted.
  for (current_tracker = copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {
    current_copy = current_tracker->copy;
    
    // Find the corresponding primtreenode's parent's original CSGtree node.
    if (current_copy->parent_type == WRAPPER) {
      if (current_copy->parent != NULL) {
	assert(current_copy->parent->type() == PRIMTREENODE_WRAPPER);
	above_wrapper = (primwrappernode *) current_copy->parent;

	assert(above_wrapper->original != NULL);
	assert(above_wrapper->original->type() == OBJECTNODE_GROUP);
	above_groupnode = (groupobjectnode *) above_wrapper->original;

	// Set the above groupnode's child pointer to point to NULL,
	// as we are now deallocating this portion of the tree.
	above_groupnode->group = NULL;
      }
    }
  }
	
  // Delete any nodes we're linked to.
  if (linkptr != NULL) {
    linkptr->deallocate_csggraph();
    delete linkptr;
  }
}

// Print out the CSG graph.
void primobjectnode::print_csggraph(int indent, int traverse_id)
{
  Tracker *current_tracker;

  if (traverse_id == last_traverse_id) {
    printf("%*sC%03d/Prim already printed\n", indent, "", index);
    return;
  }
  
  printf("%*sC%03d/Prim T(",
	 indent, "", index);

  assert(copytracker != NULL);
  
  for (current_tracker = copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {
    printf("P%03d", current_tracker->copy->index);
    if (current_tracker->next != NULL)
      printf(" ");
    else
      printf(") ");
  }

  last_traverse_id = traverse_id;

  if (linked_by == NULL)
    printf("linked_by NULL\n");
  else
    printf("linked_by C%03d\n", linked_by->index);

  printf("%*sC%03d/Prim is a %s\n", indent, "", index, prim->prim_name());

  printf("%*sC%03d/Prim print parent-tracker\n", indent, "", index);
  parenttracker->print(indent);

  if (linkptr == NULL)
    printf("%*sC%03d/Prim link NULL\n", indent, "", index);
  else {
    printf("%*sC%03d/Prim traverse \'%c\' link\n", indent, "", index,
	   (link_type == PLUS) ? '+' : 
	   (link_type == MINUS) ? '-' :
	   (link_type == AND) ? '&' : '?');
    linkptr->print_csggraph(indent+1, traverse_id);
  }
}

// This primitive has been inserted into the CSG tree directly
// under obj_group. For each of the copies of obj_group in the Primtree,
// we need to insert a new primcsgnode and primtreeprimnode.
void primobjectnode::initiate_copytracker_pass(objectnode *obj_group,
					       Primtree *primtree)
{
  CopyTracker *group_ct;
  Tracker *current_tracker;
  primcsgnode *linknode;
  primwrappernode *parentnode;
  primtreeprimnode *linked_primnode;

  // Find the CopyTracker for the group's wrapper node.
  assert(obj_group != NULL);
  group_ct = obj_group->copytracker;

  // Create a new CopyTracker for this primobjectnode.
  this->copytracker = new CopyTracker(this);

  // For each copy of this CSG node in the primtree...
  for (current_tracker = group_ct->trackerlist_head; current_tracker != NULL;
       current_tracker = current_tracker->next) {

    /*
     * Create primtreenodes and update the primtree links and linktype.
     */

    // Create a new primcsgnode that will link this primitive into the
    // graph.
    linknode = new primcsgnode(modeller);
    
    // Find the primwrappernode that is the copy of obj_group.
    assert(current_tracker->copy != NULL);
    assert(current_tracker->copy->type() == PRIMTREENODE_WRAPPER);
    parentnode = (primwrappernode *) current_tracker->copy;
    
    // The new linknode's left side is the primcsgnode structure that
    // parentnode wraps.
    linknode->left = parentnode->child;
    linknode->left->parent = linknode;
    linknode->left->parent_type = CSG_LEFT;
    
    // Create a new primtreeprimnode that will represent this primitive.
    linked_primnode = new primtreeprimnode(modeller, prim);

    // The new linknode's right side is the new primtreeprimnode.
    linknode->right = linked_primnode;
    linknode->right->parent = linknode;
    linknode->right->parent_type = CSG_RIGHT;

    // The linktype is that of this object, the primobjectnode.
    linknode->csg_type = this->link_type;

    // Update the link from the wrapper node 'parentnode' to point to
    // linknode, which represents the new CSG structure.
    parentnode->child = linknode;
    parentnode->child->parent = parentnode;
    parentnode->child->parent_type = WRAPPER;

    /*
     * Update the links from the Primtree back to the CSGgraph.
     */
    
    linknode->original = NULL;
    linknode->this_tracker = NULL;

    linked_primnode->original = this;

    /*
     * Add a new Tracker into this node's CopyTracker for the newly-created
     * primtreeprimnode.
     */

    linked_primnode->this_tracker = 
      copytracker->add_new_tracker(linked_primnode);

    /*
     * Update the position matrices in the primtree.
     */

    // The position of the new linknode is equal to the position of
    // its left child.
    Transform::set(linknode->position, linknode->left->position);

    // The position of the new primnode is equal to this position
    // combined with the transform applied to the new primobjectnode.
    Transform::combine(this->position, linknode->position, 
		       linknode->right->position);

    // Create a SolidOctree for the primtreeprimnode.
    linked_primnode->octree = new SolidOctree(linked_primnode,
					      primtree->owner,
					      this->link_type);

    /*
     * Update the SolidOctrees further up the graph.
     *
     * A new SolidOctree has been created for linked_primnode, but linknode
     * does not have a SolidOctree yet.
     */

    // Create a new SolidOctree using the primcsgnode constructor.
    linknode->octree = new SolidOctree(linknode, primtree->owner);

    // This octree needs to be pointed to by every primwrappernode that exists
    // in a chain above linknode. This includes but IS NOT LIMITED TO
    // parentnode.
    primtree->fix_wrapperoctreeptr(parentnode, parentnode->octree,
				   linknode->octree);
    
    // linknode's octree has now been changed, as the result of a topology
    // manipulation operation. It will need to be propagated during the
    // SolidOctree update pass.
    linknode->octree->has_changed = TRUE;
  }
}

/*
 *  Routines for repeatobjectnode
 */

// Create a repeatobjectnode that copies an existing objectnode.
// Note that since this objectnode is not active, it has no CopyTracker.
repeatobjectnode::repeatobjectnode(Modeller *m, objectnode *node, int copies,
				   TransformList *transformlist)
{
  this->repeat = node;
  this->modeller = m;
  this->num_copies = copies;

  set_defaults(transformlist);
}

void repeatobjectnode::deallocate_csggraph(void)
{
  objectnode *child;
  Tracker *current_tracker;
  primtreenode *current_copy;
  primwrappernode *above_wrapper;
  groupobjectnode *above_groupnode;

  // First: this portion of the CSG tree is being deleted. Find any
  // wrappers that point to this node, and set them to NULL to indicate
  // that the below tree has been deleted.
  for (current_tracker = copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {
    current_copy = current_tracker->copy;
    
    // Find the corresponding primtreenode's parent's original CSGtree node.
    if (current_copy->parent_type == WRAPPER) {
      if (current_copy->parent != NULL) {
	assert(current_copy->parent->type() == PRIMTREENODE_WRAPPER);
	above_wrapper = (primwrappernode *) current_copy->parent;

	assert(above_wrapper->original != NULL);
	assert(above_wrapper->original->type() == OBJECTNODE_GROUP);
	above_groupnode = (groupobjectnode *) above_wrapper->original;

	// Set the above groupnode's child pointer to point to NULL,
	// as we are now deallocating this portion of the tree.
	above_groupnode->group = NULL;
      }
    }
  }
	
  // Delete the child node, unless it's already been deleted.
  child = repeat;
  if (child != NULL) {
    child->deallocate_csggraph();
    delete child;
  }

  // Delete any nodes we're linked to.
  if (linkptr != NULL) {
    linkptr->deallocate_csggraph();
    delete linkptr;
  }
}

// Print out the CSG graph.
void repeatobjectnode::print_csggraph(int indent, int traverse_id)
{
  Tracker *current_tracker;

  if (traverse_id == last_traverse_id) {
    printf("%*sC%03d/Rep%d already printed\n", indent, "", index,
	   num_copies);
    return;
  }

  printf("%*sC%03d/Rep%d T(",
	 indent, "", index, num_copies);
  
  for (current_tracker = copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {
    printf("P%03d", current_tracker->copy->index);
    if (current_tracker->next != NULL)
      printf(" ");
    else
      printf(") ");
  }

  last_traverse_id = traverse_id;

  if (linked_by == NULL)
    printf("linked_by NULL\n");
  else
    printf("linked_by C%03d\n", linked_by->index);

  printf("%*sC%03d/Rep%d print parent-tracker\n", indent, "", index,
	 num_copies);
  parenttracker->print(indent);

  printf("%*sC%03d/Rep%d traverse wrapped\n", indent, "", index,
	 num_copies);
  repeat->print_csggraph(indent+1, traverse_id);

  if (linkptr == NULL)
    printf("%*sC%03d/Rep%d link NULL\n", indent, "", index, num_copies);
  else {
    printf("%*sC%03d/Rep%d traverse \'%c\' link\n", indent, "", index,
	   num_copies,
	   (link_type == PLUS) ? '+' : 
	   (link_type == MINUS) ? '-' :
	   (link_type == AND) ? '&' : '?');
    linkptr->print_csggraph(indent+1, traverse_id);
  }
}

void repeatobjectnode::initiate_copytracker_pass(objectnode *obj_group,
						 Primtree *primtree)
{
  CopyTracker *group_ct;
  primwrappernode *linked_wrappernode;
  primcsgnode *linknode;
  Tracker *current_tracker;
  primwrappernode *parentnode;

  // Find the CopyTracker for this group's above node.
  assert(obj_group != NULL);
  group_ct = obj_group->copytracker;

  // Create a new CopyTracker for this repeatobjectnode.
  copytracker = new CopyTracker(this);

  /*
   * For each copy that is now being made of this structure...
   */

  for (current_tracker = group_ct->trackerlist_head; current_tracker != NULL;
       current_tracker = current_tracker->next) {

    /*
     * Create primtreenodes and update the primtree links and linktype.
     */

    // Create a new primcsgnode that will link this group into the
    // graph.
    linknode = new primcsgnode(modeller);
    
    // Find the primwrappernode that is the copy of obj_group.
    assert(current_tracker->copy->type() == PRIMTREENODE_WRAPPER);
    parentnode = (primwrappernode *) current_tracker->copy;
    
    // The new linknode's left side is the primcsgnode structure that
    // parentnode wraps. Note that this primcsgnode is now UNCHANGED,
    // as it is the base for this addition.
    linknode->left = parentnode->child;
    linknode->left->parent = linknode;
    linknode->left->parent_type = CSG_LEFT;
    linknode->left->octree->clear_changed(linknode->left->octree,
					  &(linknode->left->octree->octree));
    
    // Create a new primwrappernode that will represent this copied group.
    linked_wrappernode = new primwrappernode(modeller, NULL);

    // The new linknode's right side is the new primwrappernode.
    linknode->right = linked_wrappernode;
    linknode->right->parent = linknode;
    linknode->right->parent_type = CSG_RIGHT;

    // The linktype is that of this object, the groupobjectnode.
    linknode->csg_type = this->link_type;

    // Update the link from the wrapper node 'parentnode' to point to
    // linknode, which represents the new CSG structure.
    parentnode->child = linknode;
    parentnode->child->parent = parentnode;
    parentnode->child->parent_type = WRAPPER;

    /*
     * Update the links from the Primtree back to the CSGgraph.
     */
    
    linknode->original = NULL;
    linknode->this_tracker = NULL;

    linked_wrappernode->original = this;

    /*
     * Add a new Tracker into this node's CopyTracker for the newly-created
     * primwrappernode.
     */

    linked_wrappernode->this_tracker =
      copytracker->add_new_tracker(linked_wrappernode);

    /*
     * Update the position matrices in the primtree.
     */

    // The position of the new linknode is equal to the position of
    // its left child.
    Transform::set(linknode->position, linknode->left->position);

    // The position of the new group is THE SAME as the 'this->position'
    // matrix specifies the repeat transform, which isn't applied to the
    // first branch.
    Transform::set(linknode->right->position, linknode->position);


    // Construct the primtree copies of the repeated object.
    construct_primtree_copies(linked_wrappernode, primtree);

    /*
     * Update the SolidOctrees further up the graph.
     *
     * The copy_and_update_primtree pass has created new SolidOctrees for
     * the copied group (linked_wrappernode's child) and this has been set
     * for linked_wrappernode, but linknode does not have a SolidOctree yet.
     */

    // First: construct the SolidOctree for linknode. Note that this
    // SolidOctree should have the CHANGED flag wherever linking in the
    // new right-hand-side has caused a change.
    //
    // The RHS will cause changes wherever it's non-empty for any
    // operation except AND, where it will cause changes wherever it's
    // non-full.
    //
    // Note: this change is necessary because it's tough to propagate
    // the AND / non-AND linktypes down through the CSG graph at
    // octree-creation time.
    //
    // After this, create linknode's octree as a combination.
    if (this->link_type == AND)
      linked_wrappernode->octree->set_all_nonequal_to_changed(
						SOLIDINDEXTYPE_FULL);
    else
      linked_wrappernode->octree->set_all_nonequal_to_changed(
						SOLIDINDEXTYPE_EMPTY);
    
    // Call the SolidOctree constructor for primcsgnodes.
    linknode->octree = new SolidOctree(linknode, primtree->owner);

    // This octree needs to be pointed to by every primwrappernode that exists
    // in a chain above linknode. This includes but IS NOT LIMITED TO
    // parentnode.
    primtree->fix_wrapperoctreeptr(parentnode, parentnode->octree,
				   linknode->octree);
    
    // linknode's octree has now been changed, as the result of a topology
    // manipulation operation. It will need to be propagated during the
    // SolidOctree update pass.
    linknode->octree->has_changed = TRUE;
  }
}

// Construct the Primtree copies for a repeatobjectnode.
void repeatobjectnode::construct_primtree_copies(primwrappernode *wrapper,
						 Primtree *primtree)
{
  /*
   * Create the new primtreenodes for each of the copies of the group
   * that's being repeated. As this is done, add to their respective
   * CopyTrackers.
   *
   * Note that we could copy any of the copies of this->group, so we
   * arbitrarily choose to copy the first one, which is
   * this->group->copytracker->trackerlist_head->copy.
   */

  assert(repeat->copytracker != NULL);
  assert(repeat->copytracker->trackerlist_head != NULL);

  /*
   * We have to create a CSG chain for n copies of repeat's primtree
   * graph. It looks like this:
   *
   *        R                R: primwrappernode, linked_wrappernode
   *        |                C: primcsgnodes, all PLUS
   *        C                P: copies of repeat's primtree graph
   *       / \               N: nothing primitive node.
   *      C   P              total of num_copies of P
   *     / \
   *    C   P
   *   / \
   *  N   P
   *
   * The N node is necessary as each one of the C nodes has a different
   * position transform... each repeated by repeat_transform. If we tried
   * to terminate the chain with two P's off the same C, there would be
   * no way to get them to be spaced apart as they should be.
   */
  
  wrapper->child = create_repeat_chain(wrapper, WRAPPER, wrapper->position,
				       this->position, num_copies,
				   repeat->copytracker->trackerlist_head->copy,
				       primtree, modeller);

  // This new octree has been created out of the CSG chain... it replaces the
  // original copy of P. We need to mark it as changed everywhere it differs
  // from P.
  wrapper->child->octree->set_all_nonequal_to_changed(repeat->copytracker->trackerlist_head->copy->octree);
  if (wrapper->child->octree->actual_changes())
    wrapper->child->octree->has_changed = TRUE;

  // This new chain's octree needs to be pointed to by every primwrappernode
  // that exists the chain above. This includes but IS NOT LIMITED TO wrapper.
  primtree->fix_wrapperoctreeptr(wrapper, wrapper->octree,
				 wrapper->child->octree);
}
			       

// Create a CSG chain that repeats a given part of the Primtree.
primtreenode * repeatobjectnode::create_repeat_chain(primtreenode *parent,
						   enum parenttype parent_type,
						     transform position,
						    transform repeat_transform,
						     int copycount,
						     primtreenode *repeatnode,
						     Primtree *primtree,
						     Modeller *modeller)
{
  primtreeprimnode *nothingnode;
  primcsgnode *linknode;
  transform next_chain_position;

  assert (copycount >= 0);

  if (copycount == 0) {
    // This is the end of the chain. We create a single nothingPrimitive node that
    // is used solely as a placeholder. Note that it has no original or
    // this_tracker fields, which *may* be dangerous. I hope not.
    nothingnode = new primtreeprimnode(modeller, modeller->tailend_nothingprim);
    Transform::set(nothingnode->position, position);
    nothingnode->parent = parent;
    nothingnode->parent_type = parent_type;

    // Create the SolidOctree.
    nothingnode->octree = new SolidOctree(nothingnode, modeller, PLUS);

    return nothingnode;
  } else {
    // Create a CSG node for this part of the chain.
    linknode = new primcsgnode(modeller);
    linknode->csg_type = PLUS;
    linknode->parent = parent;
    linknode->parent_type = parent_type;

    Transform::set(linknode->position, position);
    Transform::combine(repeat_transform, position, next_chain_position);

    // Create a primitive tree for the right-hand side of the chain.
    linknode->right = repeatnode->copy_and_update_primtree(linknode, CSG_RIGHT,
							   primtree);
    // The SolidOctree for the right-hand side must be set to be changed
    // wherever it is non-empty.
    linknode->right->octree->set_all_nonequal_to_changed(SOLIDINDEXTYPE_EMPTY);

    // Create the remainder of the chain for the left-hand side.
    linknode->left = create_repeat_chain(linknode, CSG_LEFT,
					 next_chain_position,
					 repeat_transform, copycount - 1,
					 repeatnode, primtree, modeller);

    // Call the SolidOctree constructor for primcsgnodes.
    linknode->octree = new SolidOctree(linknode, modeller);

    return linknode;
  }
}


/*
 *  Routines for CSGgraph
 */

// Construct a new CSGgraph.
CSGgraph::CSGgraph(Modeller *owner)
{
  this->owner = owner;
  
  // The CSGgraph is initially empty.
  object = NULL;
  selected_object = NULL;
  
  next_index = 0;
  nodearray_size = 0;
  nodearray = NULL;
  next_traverse_id = 1;
  next_update_id = 1;
}

// Select an element of the CSGgraph.
void CSGgraph::select_element(objectnode *obj_node)
{
  // First, is this object already selected? If so, no change.
  if (selected_object == obj_node)
    return;

  deselect();

  selected_object = obj_node;

  if (obj_node != NULL) {
    obj_node->select();

    // This object has been selected, so we need to insert its transformations
    // into the SkitterList.
    owner->skitterlist->new_selection(obj_node);

    // Update the graphview.
    owner->graphlayout->complete_layout();
  }
}

// Select an element, by node ID.
objectnode *CSGgraph::retrieve_element(int node_id)
{
  assert((node_id < nodearray_size) && (node_id >= 0));
  assert(nodearray[node_id] != NULL);

  return nodearray[node_id];
}

// Deselect all elements.
void CSGgraph::deselect(void)
{
  if (selected_object != NULL) {
    selected_object->deselect();

    selected_object = NULL;

    // Nothing is selected, so purge the SkitterList.
    owner->skitterlist->purge();

    // Update the graphview.
    owner->graphlayout->complete_layout();
  }
}

// When a node's TransformList is changed, we must get the new transform
// matrix and update the CSGgraph, Primtree and SolidOctrees accordingly.
void CSGgraph::update_transformed_object(objectnode *obj_node)
{
  Tracker *current_tracker;
  transform parent_position, parent_position_inv, old_trans;
  transform old_trans_inv, update_trans, new_trans, update_pre, update_post;
  primtreenode *current_copy, *first_child;
  int this_update_id = next_update_id++;

  // Get the new transform matrix from the TransformList.
  obj_node->transformlist->get_position(new_trans);

  Transform::set(old_trans, obj_node->position);
  Transform::set(obj_node->position, new_trans);
  Transform::invert(old_trans, old_trans_inv);

  for (current_tracker = obj_node->copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {
    current_copy = current_tracker->copy;

    if (obj_node->type() != OBJECTNODE_REPEAT) {
      /*
       * Calculate the transformation that is applied to every node in the
       * primtree below this point.
       */
      
      // Find the position matrix of this copy's parent.
      if (current_copy->parent == NULL)
	Transform::unit(parent_position);
      else
	Transform::set(parent_position, current_copy->parent->position);
      
      // The CSGgraph matrix at this point M0 is changed to M'. The position
      // matrix at this point is a combination of the parent matrix P0 and M,
      // so P0M0 -> P0M'. Similarly we need to find a transform that will
      // change the position of an arbitrary primtree node below this point
      // from P0M0Q0R0S0 to P0M'Q0R0S0. We do this by multiplying by a
      // transformation matrix T = P0M'M0-1P0-1.
      //
      // Calculate this matrix.
      
      Transform::invert(parent_position, parent_position_inv);
      
      // Multiply T = P0M'M0-1P0-1.
      Transform::combine(new_trans, parent_position, update_trans);
      Transform::combine(old_trans_inv, update_trans, update_trans);
      Transform::combine(parent_position_inv, update_trans, update_trans);
    
      // Recursively update the parameters of the primtree.
      current_copy->update_primtree(update_trans, this_update_id);
    } else {
      // This is a repeatobjectnode that's being manipulated. Requires
      // significantly trickier coding.

      // We need to construct an update_trans for each one of the repeated
      // children. They take the form P_0T'^kT^-kP_0^-1 for k = 0..n-1.
      // We do this through a recursive pass through the repeated children.
      
      // First: construct update_pre and update_post as P_0 and P_0^-1.
      Transform::set(update_pre, current_copy->position);
      Transform::invert(update_pre, update_post);

      // Call the recursive pass.
      assert(current_copy->type() == PRIMTREENODE_WRAPPER);
      first_child = ((primwrappernode *) current_copy)->child;
      update_transformed_repeatchain(first_child, update_pre, update_post,
				     new_trans, old_trans_inv, this_update_id);
    }
  }

  // Initiate a SolidOctree pass through the Primtree. Each octree that
  // has changed information (all of the primitives' octrees that were
  // changed) is marked with the has_changed flag; this information needs
  // to be propagated upwards through all of the above SolidOctrees.
  owner->primtree->initiate_solidoctree_pass();
}

// Update all the repeated children of the repeatobjectnode, with the
// new position matrix of the node.
void CSGgraph::update_transformed_repeatchain(primtreenode *child,
					      transform update_pre,
					      transform update_post,
					      transform new_trans,
					      transform old_trans_inv, int this_update_id)
{
  transform update_trans;
  primcsgnode *csg_node;

  switch(child->type()) {
  case PRIMTREENODE_CSG:
    csg_node = (primcsgnode *) child;

    // This is a CSG node in the chain. We form the update_trans from 
    // the combination of update_pre and update_post.
    Transform::combine(update_post, update_pre, update_trans);

    // We update the position matrix in the CSG node and then update the
    // primtree that it encapsulates (its right-hand child).
    Transform::combine(child->position, update_trans, child->position);
    csg_node->right->update_primtree(update_trans, this_update_id);

    // Now, we need to create the next matrix, for our left-hand child.
    // Note that each one of the repeated copies has a different update
    // transform because it is being changed from T^k to T'^k ... so, we take
    // the update_pre / post pair P_0T'^k-1 / T^-(k-1)P_0^-1 and post / pre
    // multiply by T' / T-1 to get the new pre / post pair
    // P_0T'^k / T^-kP_0^-1 which are combined in the child to form the
    // full update_trans of P_0T'^kT^-kP_0^-1
    Transform::combine(new_trans, update_pre, update_pre);
    Transform::combine(update_post, old_trans_inv, update_post);

    // Call the recursive routine with the left-hand child.
    update_transformed_repeatchain(csg_node->left,
				   update_pre, update_post,
				   new_trans, old_trans_inv, this_update_id);
    break;
  default:
    // We have come to the end of the recursive chain, and hit the
    // nothingPrimitive that is the sentinel.
    assert(child->type() == PRIMTREENODE_PRIM);
    break;
  }
}

// Link an element into a group of the CSGgraph. This element must not
// already be linked into a group - it must be newly created.
void CSGgraph::link_new_element(groupobjectnode *obj_group, linktype link_type,
				objectnode *addthing)
{
  objectnode *chain_head;
  

  assert(addthing != NULL);
  assert(obj_group != NULL);

  // First: construct the link pointer along the chain.
  chain_head = obj_group->group;
  assert(chain_head != NULL);
  assert(chain_head->linked_by == NULL);
  chain_head->linked_by = addthing;
  addthing->linkptr = chain_head;
  addthing->link_type = link_type;

  // Next: Set the chain-head pointer correctly.
  obj_group->group = addthing;

  // Update the parenttrackers correctly.
  //
  // chain_head is no longer linked to by obj_group.
  // addthing is now linked to by obj_group.
  chain_head->parenttracker->remove(obj_group);
  addthing->parenttracker->add(obj_group);

  // Initiate a CopyTracker / Primtree update pass throughout the
  // downlink objectnode. This creates the relevant portions
  // of the Primtree and sets up all of the CopyTrackers accordingly.
  addthing->initiate_copytracker_pass(obj_group, owner->primtree);

  // Initiate a SolidOctree pass through the Primtree. Each octree that
  // has changed information is marked with the has_changed flag; this
  // information needs to be propagated upwards through all of the above
  // SolidOctrees.
  owner->primtree->initiate_solidoctree_pass();
}

// Add a wrapper around an object that's already linked into the CSGgraph.
// This replaces the object's node with a wrappernode, which takes its
// place in the CSGgraph.
groupobjectnode *CSGgraph::add_wrapper_around(objectnode *original)
{
  groupobjectnode *new_wrapper;
  Tracker *current_tracker;
  primtreenode *current_copy;
  primwrappernode *new_primwrapper;
  ParentTracker *new_parenttracker;
  parenttrackernode *ptptr;

  // Create a new groupobjectnode to be the wrapper.
  new_wrapper = new groupobjectnode(owner, original);

  // Set up the up-chain link of the wrapper.
  new_wrapper->linked_by = original->linked_by;
  if (new_wrapper->linked_by != NULL) {
    original->linked_by = NULL;
    new_wrapper->linked_by->linkptr = new_wrapper;
  }
  
  // Set up the down-chain link of the wrapper.
  new_wrapper->link_type = original->link_type;
  new_wrapper->linkptr = original->linkptr;
  if (new_wrapper->linkptr != NULL) {
    original->linkptr = NULL;
    new_wrapper->linkptr->linked_by = new_wrapper;
  }

  new_wrapper->copytracker = new CopyTracker(new_wrapper);

  /*
   * Update the ParentTrackers involved.
   */

  // The new wrapper's parenttracker is empty... keep a pointer to it.
  new_parenttracker = new_wrapper->parenttracker;

  // The new wrapper is linked to by all of the things that linked to the
  // original node.
  new_wrapper->parenttracker = original->parenttracker;
  new_wrapper->parenttracker->owner = new_wrapper;

  // The original node uses the new parenttracker.
  original->parenttracker = new_parenttracker;
  original->parenttracker->owner = original;

  // The original node is linked to by one wrapper node only.
  original->parenttracker->add(new_wrapper);

  /*
   * For each node that linked down to the original node (listed in the
   * ParentTracker), change this link to point to new_wrapper.
   */

  for (ptptr = new_wrapper->parenttracker->head; ptptr != NULL;
       ptptr = ptptr->next) {
    if (ptptr->parent->type() == OBJECTNODE_GROUP) {
      groupobjectnode *group_parent = (groupobjectnode *) ptptr->parent;

      group_parent->group = new_wrapper;
    } else if (ptptr->parent->type() == OBJECTNODE_REPEAT) {
      repeatobjectnode *repeat_parent = (repeatobjectnode *) ptptr->parent;

      repeat_parent->repeat = new_wrapper;
    } else {
      assert(0);
    }
  }

  /*
   * For each copy made of the original node, update the Primtree.
   */

  for (current_tracker = original->copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {

    current_copy = current_tracker->copy;

    // Create a primwrappernode for this copy.
    new_primwrapper = new primwrappernode(owner, current_copy);

    // Set up the wrapper's position details, etc.
    Transform::set(new_primwrapper->position, current_copy->position);
    new_primwrapper->original = new_wrapper;

    // Set up a Tracker for the new CSGgraph wrapper.
    new_primwrapper->this_tracker = new_wrapper->copytracker->
      add_new_tracker(new_primwrapper);

    // Set up the links from the wrapper to the copy.
    new_primwrapper->parent      = current_copy->parent;
    new_primwrapper->parent_type = current_copy->parent_type;
    current_copy->parent = new_primwrapper;
    current_copy->parent_type = WRAPPER;

    // The wrapper has its child's octree... at least, for now.
    new_primwrapper->octree = current_copy->octree;

    // Set up the links down to the new wrapper from the copy's original parent
    // node, both in the primtree and the CSGgraph.
    if (new_primwrapper->parent == NULL) {
      // We must set the CSGgraph and Primtree's toplevel object link(s).
      this->object = new_wrapper;
      this->owner->primtree->root = new_primwrapper;
      this->owner->focused_octree = new_primwrapper->octree;
    } else {
      // There is a node above this one, which must be updated.
      if (new_primwrapper->parent_type == WRAPPER) {
	assert(new_primwrapper->parent->type() == PRIMTREENODE_WRAPPER);
	
	primwrappernode *above_wrapper = (primwrappernode *)
	  new_primwrapper->parent;
	
	above_wrapper->child = new_primwrapper;
	
      } else {
	// The copy is linked in by a CSG node.
	assert(new_primwrapper->parent->type() == PRIMTREENODE_CSG);
	
	primcsgnode *above_csgnode = (primcsgnode *) new_primwrapper->parent;
	
	// Update the appropriate child link to point to the wrapper
	// instead of the underlying node.
	if (new_primwrapper->parent_type == CSG_LEFT)
	  above_csgnode->left = new_primwrapper;
	else
	  above_csgnode->right = new_primwrapper;
      }
    }
  }  

  // We don't have to call initiate_solidoctree_pass because as far as the
  // Octrees of the graph are concerned, we haven't changed anything. We've
  // only added a few wrappernodes in, which don't really affect that much.

  // Update the graph view to reflect this change.
  owner->graphlayout->update_wrapper(new_wrapper);

  // Constructed the wrapper successfully.
  return new_wrapper;
}

// Create a repeatobjectnode around an object that's already linked into the
// CSGgraph. This replaces the object's node with a repeatobjectnode, which
// takes its place in the CSGgraph.
repeatobjectnode *CSGgraph::add_repeat_around(objectnode *original,
					      int num_copies,
					      int num_trans,
				          ObjectTransformation **trans_array)
{
  repeatobjectnode *new_wrapper;
  Tracker *current_tracker, *next_tracker;
  primtreenode *current_copy;
  primwrappernode *new_primwrapper;
  ParentTracker *new_parenttracker;
  parenttrackernode *ptptr;

  // Create a new repeatobjectnode to be the wrapper.
  new_wrapper = new repeatobjectnode(owner, original, num_copies,
		(num_trans) ? new TransformList(num_trans, trans_array)
				     : NULL);

  // Set up the up-chain link of the wrapper.
  new_wrapper->linked_by = original->linked_by;
  if (new_wrapper->linked_by != NULL) {
    original->linked_by = NULL;
    new_wrapper->linked_by->linkptr = new_wrapper;
  }
  
  // Set up the down-chain link of the wrapper.
  new_wrapper->link_type = original->link_type;
  new_wrapper->linkptr = original->linkptr;
  if (new_wrapper->linkptr != NULL) {
    original->linkptr = NULL;
    new_wrapper->linkptr->linked_by = new_wrapper;
  }

  new_wrapper->copytracker = new CopyTracker(new_wrapper);

  /*
   * Update the ParentTrackers involved.
   */

  // The new wrapper's parenttracker is empty... keep a pointer to it.
  new_parenttracker = new_wrapper->parenttracker;

  // The new wrapper is linked to by all of the things that linked to the
  // original node.
  new_wrapper->parenttracker = original->parenttracker;
  new_wrapper->parenttracker->owner = new_wrapper;

  // The original node uses the new parenttracker.
  original->parenttracker = new_parenttracker;
  original->parenttracker->owner = original;

  // The original node is linked to by one wrapper node only.
  original->parenttracker->add(new_wrapper);

  /*
   * For each node that linked down to the original node (listed in the
   * ParentTracker), change this link to point to new_wrapper.
   */

  for (ptptr = new_wrapper->parenttracker->head; ptptr != NULL;
       ptptr = ptptr->next) {
    if (ptptr->parent->type() == OBJECTNODE_GROUP) {
      groupobjectnode *group_parent = (groupobjectnode *) ptptr->parent;

      group_parent->group = new_wrapper;
    } else if (ptptr->parent->type() == OBJECTNODE_REPEAT) {
      repeatobjectnode *repeat_parent = (repeatobjectnode *) ptptr->parent;

      repeat_parent->repeat = new_wrapper;
    } else {
      assert(0);
    }
  }

  /*
   * For each copy made of the original node, update the Primtree.
   */

  // We need to be able to distinguish between the existing copies (which
  // need to be repeated) and the new copies (which should be skipped).
  original->copytracker->advance_timestamp();

  for (current_tracker = original->copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = next_tracker) {

    // Get the next tracker in the copytracker chain. We have to access this
    // now and store the value, as the Tracker that stores it may be deleted
    // later on in the loop.
    next_tracker = current_tracker->next;

    // Skip over any of the 'copy' tracker nodes, we only need to update
    // the originals.
    assert(current_tracker != NULL);
    if (original->copytracker->has_current_timestamp(current_tracker))
      continue;

    current_copy = current_tracker->copy;

    // Create a primwrappernode for this copy.
    new_primwrapper = new primwrappernode(owner, NULL);

    // Set up the wrapper's position details, etc.
    //
    // NB: the primwrapper is taking the place of the copied repeatnode's
    // PARENT so it must start off with the position matrix of the copied
    // node's parent.
    if (current_copy->parent == NULL)
      Transform::unit(new_primwrapper->position);
    else
      Transform::set(new_primwrapper->position,
		     current_copy->parent->position);
    new_primwrapper->original = new_wrapper;

    // Set up a Tracker for the new wrappernode.
    new_primwrapper->this_tracker = new_wrapper->copytracker->
      add_new_tracker(new_primwrapper);

    // Set up the link from the wrapper to its parent.
    new_primwrapper->parent      = current_copy->parent;
    new_primwrapper->parent_type = current_copy->parent_type;

    // Construct the repeat chain.
    //
    // NB: construct_primtree_copies creates an octree for the CSGnodes
    // inside new_wrapper. It then propagates this octree up through
    // new_wrapper to any primwrappernodes that exist above it.
    new_wrapper->construct_primtree_copies(new_primwrapper, owner->primtree);

    // Delete the existing copy of the primtree, as we have just created
    // a whole set of copies.
    current_copy->recursively_delete_children();
    owner->primtree->delete_primtreenode(current_copy);

    // Set up the links down to the new wrapper from the copy's original parent
    // node, in the primtree.
    if (new_primwrapper->parent == NULL) {
      // We must set the CSGgraph and Primtree's toplevel object link(s).
      this->object = new_wrapper;
      this->owner->primtree->root = new_primwrapper;
      this->owner->focused_octree = new_primwrapper->octree;
    } else {
      // There is a node above this one, which must be updated.
      if (new_primwrapper->parent_type == WRAPPER) {
	assert(new_primwrapper->parent->type() == PRIMTREENODE_WRAPPER);
	
	primwrappernode *above_wrapper = (primwrappernode *)
	  new_primwrapper->parent;
	
	above_wrapper->child = new_primwrapper;
	
      } else {
	// The copy is linked in by a CSG node.
	assert(new_primwrapper->parent->type() == PRIMTREENODE_CSG);
	
	primcsgnode *above_csgnode = (primcsgnode *) new_primwrapper->parent;
	
	// Update the appropriate child link to point to the wrapper
	// instead of the underlying node.
	if (new_primwrapper->parent_type == CSG_LEFT)
	  above_csgnode->left = new_primwrapper;
	else
	  above_csgnode->right = new_primwrapper;
      }
    }
  }  

  // We've made significant changes to the structure of the CSGgraph
  // and this requires that we update the SolidOctrees.
  owner->primtree->initiate_solidoctree_pass();

  // Update the graph view to reflect this change.
  owner->graphlayout->update_repeat(new_wrapper);

  // Constructed the repeatobjectnode successfully.
  return new_wrapper;
}

// Print out the CSG graph.
void CSGgraph::print_csggraph(void)
{
  assert(object != NULL);
  object->print_csggraph(0, next_traverse_id++);
}

// PRIMARY UI ACCESS:
// Link a new primitive into the modelled object.
groupobjectnode *CSGgraph::link_in_primitive(objectnode *link_to,
					     linktype link_type,
					     Primitive *prim,
					     primobjectnode **return_primobj,
					     int num_trans,
				ObjectTransformation **trans_array)
{
  groupobjectnode *link_obj;
  primobjectnode *link_in;
  parenttrackernode *ptptr;

  // If this is a primitive or repeat node, then we will need to create
  // a group around it in order to add something to it.
  if (link_to->type() == OBJECTNODE_GROUP) {
    link_obj = (groupobjectnode *) link_to;
    
    // Now, we have to be careful. If this is a groupnode already, it might
    // not be the only pointer to link_obj->group. If that's the case, we will
    // have to add a groupobjectnode between link_obj and its ->group before
    // starting the link operation.
    //
    // Otherwise we are modifying all copies of this node, simply by adding
    // something to one copy of the node.
    ptptr = link_obj->group->parenttracker->head;
    if ((ptptr != NULL) && (ptptr->next != NULL)) {
      // There are at least two parents of link_obj->group.
      add_groupnode_under(link_obj);
    }
  } else {
    link_obj = add_wrapper_around(link_to);
  }

  // Create the primobjectnode that will be linked in.
  if (num_trans == 0)
    link_in = new primobjectnode(owner, prim);
  else
    link_in = new primobjectnode(owner, prim,
				 new TransformList(num_trans, trans_array));

  assert(link_in->layoutnode == NULL);

  // Perform the link operation.
  link_new_element(link_obj, link_type, link_in);

  // Update the graph view to reflect this change. NB: the graph will
  // add a LayoutNode for the new element.
  owner->graphlayout->update_linkinprim(link_obj, link_in);

  if (return_primobj != NULL)
    *return_primobj = link_in;

  return link_obj;
}

// PRIMARY UI ACCESS:
// Link a copy of an object into the modelled object.
groupobjectnode *CSGgraph::link_in_copy(objectnode *link_to,
					linktype link_type,
					objectnode *copy,
					groupobjectnode **return_groupobj,
					int num_trans,
				ObjectTransformation **trans_array)
{
  groupobjectnode *link_obj, *link_in;
  parenttrackernode *ptptr;

  // If this is a primitive or repeat node, then we will need to create
  // a group around it in order to add something to it.
  if (link_to->type() == OBJECTNODE_GROUP) {
    link_obj = (groupobjectnode *) link_to;
    
    // Now, we have to be careful. If this is a groupnode already, it might
    // not be the only pointer to link_obj->group. If that's the case, we will
    // have to add a groupobjectnode between link_obj and its ->group before
    // starting the link operation.
    //
    // Otherwise we are modifying all copies of this node, simply by adding
    // something to one copy of the node.
    ptptr = link_obj->group->parenttracker->head;
    if ((ptptr != NULL) && (ptptr->next != NULL)) {
      // There are at least two parents of link_obj->group.
      add_groupnode_under(link_obj);
    }
  } else {
    link_obj = add_wrapper_around(link_to);
  }

  // We are now going to create a ghost link to copy. We must ensure that we
  // don't create a ghost link into the middle of a chain.
  //
  // NB: if we are linking a copy of an object into a groupnode that already
  // wraps that node, then we would be making a chain with a vertical pointer
  // down into the middle of it. That would be a Bad Thing.
  if ((copy->linkptr != NULL) || (copy->linked_by != NULL) ||
      (link_obj->group == copy)) {
    // We need to make a wrapper around copy, we can then link to the
    // underlying node.
    add_wrapper_around(copy);
  }

  // Create the groupobjectnode that will be linked in.
  link_in = new groupobjectnode(owner, copy,
				new TransformList(num_trans, trans_array));

  // Update the copy's ParentTracker.
  copy->parenttracker->add(link_in);

  // Perform the link operation.
  link_new_element(link_obj, link_type, link_in);

  // Update the graph view to reflect this change. NB: the graph will
  // add a LayoutNode for the new element.
  owner->graphlayout->update_linkincopy(link_obj, link_in);

  if (return_groupobj != NULL)
    *return_groupobj = link_in;

  return link_obj;
}

// Get a new index for an objectnode.
int CSGgraph::new_index(objectnode *node)
{
  int index;

  // Check to make sure that we have enough space in the nodearray.
  while (next_index >= nodearray_size) {
    nodearray_size += csggraph_nodearray_growsize;

    nodearray = (objectnode **) realloc((void *) nodearray,
					nodearray_size * sizeof(objectnode *));
    if (!nodearray)
      fatal_error("CSGgraph::new_index", "Can't grow nodearray!",
		  "Out of memory.");
  }

  // Store the objectnode pointer.
  index = next_index++;
  nodearray[index] = node;

  return index;
}

// Tag that an objectnode has been deleted.
void CSGgraph::delete_index(int index)
{
  // Called when an objectnode is deleted.
  assert((index < nodearray_size) && (index >= 0));
  assert(nodearray[index] != NULL);
  nodearray[index] = NULL;
}

// PRIMARY UI ACCESS:
// Add a groupobjectnode under an object.
void CSGgraph::add_groupnode_under(groupobjectnode *add_under)
{
  groupobjectnode *new_node;
  parenttrackernode *ptptr;
  primwrappernode *prim_add_under, *prim_new_node;
  Tracker *current_tracker;

  // Create the new node.
  new_node = new groupobjectnode(owner, add_under->group);
  add_under->group = new_node;
  new_node->copytracker = new CopyTracker(new_node);

  // Update the ParentTrackers.
  new_node->parenttracker->add(add_under);
  for (ptptr = new_node->group->parenttracker->head; ptptr != NULL;
       ptptr = ptptr->next)
    if (ptptr->parent == add_under)
      ptptr->parent = new_node;

  /*
   * For each copy of add_under, update the Primtree with a new
   * primwrappernode.
   */

  for (current_tracker = add_under->copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = current_tracker->next) {

    assert(current_tracker->copy->type() == PRIMTREENODE_WRAPPER);
    prim_add_under = (primwrappernode *) current_tracker->copy;

    // Create a new primwrappernode.
    prim_new_node = new primwrappernode(owner, prim_add_under->child);

    // Set up its position details.
    Transform::set(prim_new_node->position, prim_add_under->position);
    prim_new_node->original = new_node;

    // Add it to the CopyTracker.
    prim_new_node->this_tracker = new_node->copytracker->
      add_new_tracker(prim_new_node);

    // Set up the primtree links.
    prim_new_node->child              = prim_add_under->child;
    prim_new_node->child->parent      = prim_new_node;
    prim_new_node->child->parent_type = WRAPPER;

    prim_add_under->child = prim_new_node;
    prim_new_node->parent = prim_add_under;
    prim_new_node->parent_type = WRAPPER;

    // The wrapper has its child's octree.
    prim_new_node->octree = prim_new_node->child->octree;
  }

  // Update the graph view to reflect this change.
  owner->graphlayout->update_addgroupnodeunder(add_under);
}

// PRIMARY UI ACCESS:
// Delete an objectnode.
void CSGgraph::delete_object(objectnode *kill_me)
{
  Tracker *current_tracker, *next_tracker;

  assert(kill_me != NULL);
  assert(kill_me->copytracker != NULL);

  // Delete each of the primtree copies of this node.
  for (current_tracker = kill_me->copytracker->trackerlist_head;
       current_tracker != NULL; current_tracker = next_tracker) {
    next_tracker = current_tracker->next;
    owner->primtree->delete_subgraph(current_tracker->copy);
  }

  // Initiate a solidoctree pass to update the SolidOctrees in the scene
  // to remove all traces of the deleted nodes.
  owner->primtree->initiate_solidoctree_pass();

  // If we have deleted everything in the scene, add a placeholder sphere to
  // link against.
  if (object == NULL)
    owner->create_placeholder_object(new spherePrimitive);

  // The layout will need to be updated... significantly.
  owner->graphlayout->complete_layout();
}

// Delete an objectnode in the CSGgraph.
void CSGgraph::delete_objectnode(objectnode *kill_me)
{
  parenttrackernode *ptptr;
  groupobjectnode *parentnode;

  // Delete our CopyTracker.  
  delete kill_me->copytracker;

  // Link the linkchain around this node (if there is one).
  if (kill_me->linkptr != NULL) {
    // We are not at the end of the chain. The next object along the chain has
    // to be linked up to the previous one.
    kill_me->linkptr->linked_by = kill_me->linked_by;

    if (kill_me->linkptr->linked_by == NULL) {
      // kill_me was at the start of the chain. It should have precisely
      // one parent, which must now point to kill_me->linkptr.
      assert(kill_me->parenttracker != NULL);
      ptptr = kill_me->parenttracker->head;
      assert(ptptr != NULL);
      assert(ptptr->next == NULL);
      
      assert(ptptr->parent->type() == OBJECTNODE_GROUP);
      parentnode = (groupobjectnode *) ptptr->parent;

      parentnode->group = kill_me->linkptr;
      kill_me->linkptr->parenttracker->add(parentnode);

      owner->graphlayout->update_chainheadswitch(parentnode,
						 kill_me, kill_me->linkptr);
    }
  }

  if (kill_me->linked_by != NULL) {
    // We are not at the start of the chain. The previous object along the
    // chain has to be linked up to the next one.
    kill_me->linked_by->linkptr = kill_me->linkptr;
  }

  // NB: Any objectnodes that point down to this will be deleted if necessary
  // in the Primtree's recursive_tidyup pass. This includes the csg_graph->
  // object pointer.

  // This objectnode no longer needs to be accessible by its index, as
  // it won't appear on the layout. We must ensure that we don't keep a
  // dangling pointer to it in the CSGgraph->nodearray as well.
  delete_index(kill_me->index);

  // Delete our LayoutNode.
  delete kill_me->layoutnode;

  // Delete our ParentTracker.
  delete kill_me->parenttracker;

  // Delete our TransformList.
  delete kill_me->transformlist;

  // If we are a primobjectnode, we will need to delete our Collections'
  // references, and also our primitive.
  if (kill_me->type() == OBJECTNODE_PRIM) {
    ((primobjectnode *) kill_me)->remove_collections();

    delete ((primobjectnode *) kill_me)->prim;
  }

  delete kill_me;
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
