/*
 * csggraph.h: Definitions for class CSGgraph and supporting code.
 *
 * Contains code to create and maintain the CSGgraph representation of
 * an object.
 */

/* Collection-safe */

const int OBJECTNODE_NAMELEN = 40;

// An objectnode is an abstract class that represents some object
// being modelled by the system. It may be a group object, or a primitive.
class objectnode {
public:
  // The Modeller that owns this objectnode.
  Modeller *modeller;

  // A unique-ID index. Used for debugging.
  int index;

  // The last traversal that printed out this objectnode; if
  // this traversal tries to print out the objectnode again, just
  // say "already printed".
  int last_traverse_id;
  
  // The transform applied to this objectnode.
  TransformList *transformlist;
  transform position;

  // The ParentTracker tracks all groupobjectnodes that copy this objectnode.
  ParentTracker *parenttracker;

  // The CopyTracker tracks all copies that are made of this objectnode.
  CopyTracker *copytracker;

  // Each objectnode has a LayoutNode that tracks its current position
  // and status for the GraphLayout layout engine.
  LayoutNode *layoutnode;

  // The name of the objectnode.
  char name[OBJECTNODE_NAMELEN];

  // Is this objectnode currently selected?
  int is_selected;

  // A pointer to the next objectnode in the group, if there is one, and
  // the kind of link which was applied to this objectnode.
  objectnode *linkptr;
  linktype link_type;

  // The objectnode that chains this object in.
  objectnode *linked_by;

  // Set default values.
  void set_defaults(TransformList *transformlist = NULL);
  
  // Set the name field of the node.
  void set_name(char *new_name);

  // This object has been inserted into the CSGgraph directly under
  // obj_group. Update the Primtree and all appropriate CopyTrackers.
  virtual void initiate_copytracker_pass(objectnode *obj_group,
										Primtree *primtree) = 0;

  // Select or deselect an objectnode.
  void select(void);
  void deselect(void);

  // Deallocate the CSGgraph.
  virtual void deallocate_csggraph(void) = 0;

  // Print the CSGgraph.
  virtual void print_csggraph(int indent, int traverse_id) = 0;

  virtual objectnodetype type(void) = 0;

  virtual ~objectnode(void);
};

// A groupobjectnode is a class of objectnode that represents a CSG combination
// of several objectnodes.
class groupobjectnode : public objectnode {
public:
  // The group that is represented by this node.
  objectnode *group;

  // This object has been inserted into the CSGgraph directly under
  // obj_group. Update the Primtree and all appropriate CopyTrackers.
  virtual void initiate_copytracker_pass(objectnode *obj_group,
					                     Primtree *primtree);

  // Deallocate the CSGgraph.
  virtual void deallocate_csggraph(void);

  // Print the CSGgraph.
  virtual void print_csggraph(int indent, int traverse_id);

  virtual objectnodetype type(void) {
    return OBJECTNODE_GROUP;
  }

  // Create a wrapper for an existing objectnode.
  groupobjectnode(Modeller *m, objectnode *node,
		          TransformList *transformlist = NULL);

  virtual ~groupobjectnode(void) {
  }
};

const int prim_collection_bufsize_increment = 20;

// A primobjectnode is a class of objectnode that represents a single
// primitive. This is the lowest level of the CSG graph.
class primobjectnode : public objectnode {
public:
  // The primitive that is represented by this node.
  Primitive *prim;

  // The array of Collections that contains the triangles which make up
  // the surface of the primitive.
  int collections_found;
  int num_collections, collection_bufsize;
  class Collection **collectionbuf;
  class WorldCollection **worldcollectionbuf;

  // Remove this object's Collections.
  void remove_collections(void);

  // This object has been inserted into the CSGgraph directly under
  // obj_group. Update the Primtree and all appropriate CopyTrackers.
  virtual void initiate_copytracker_pass(objectnode *obj_group,
					                     Primtree *primtree);

  // Deallocate the CSGgraph.
  virtual void deallocate_csggraph(void);

  // Print the CSGgraph.
  virtual void print_csggraph(int indent, int traverse_id);

  virtual objectnodetype type(void) {
    return OBJECTNODE_PRIM;
  }

  // Create a new primobjectnode.
  primobjectnode(Modeller *m, Primitive *prim,
		         TransformList *transformlist = NULL);

  virtual ~primobjectnode(void) {
  }
};

// A repeatobjectnode is a class of objectnode that represents a series of
// copies of an object, each subsequent copy transformed by the same transform.
// NB: repeatobjectnode uses its position matrix as the copy transform,
// and the n copies are placed at 0, T, 2T, 3T, ..., (n-1)T
class repeatobjectnode : public objectnode {
public:
  // The number of copies of the repeated object.
  int num_copies;

  // The object that is repeated by this node.
  objectnode *repeat;

  // This object has been inserted into the CSGgraph directly under
  // obj_group. Update the Primtree and all appropriate CopyTrackers.
  virtual void initiate_copytracker_pass(objectnode *obj_group,
					 Primtree *primtree);

  // Create a CSG chain that repeats a given part of the Primtree.
  class primtreenode *create_repeat_chain(primtreenode *parent,
					  parenttype parent_type,
					  transform position,
					  transform repeat_transform,
					  int num_copies,
					  primtreenode *repeatnode,
					  Primtree *primtree,
					  Modeller *modeller);

  // Construct the Primtree copies for a repeatobjectnode.
  void construct_primtree_copies(primwrappernode *wrapper,
				                 Primtree *primtree);

  // Deallocate the CSGgraph.
  virtual void deallocate_csggraph(void);

  // Print the CSGgraph.
  virtual void print_csggraph(int indent, int traverse_id);

  virtual objectnodetype type(void) {
    return OBJECTNODE_REPEAT;
  }

  // Create a repeated object.
  repeatobjectnode(Modeller *m, objectnode *node, int copies,
		           TransformList *transformlist = NULL);

  virtual ~repeatobjectnode(void) {
  }
};

const int csggraph_nodearray_growsize = 20;

// A CSGgraph class encapsulates a graph of objectnodes, which together
// represent a complex object.
class CSGgraph {
public:
  // A unique-ID traversal counter used for debugging.
  int next_traverse_id;

  // Each objectnode created is given a unique index; this counter is used
  // to give the next ID number.
  int next_index;

  // Each update of a transformed object is given a unique ID so that the octrees
  // can tell if a primitive has been moved in this update pass.
  int next_update_id;

  // An array that is used for tracking objectnode pointers, and converting
  // indices to pointers.
  int nodearray_size;
  objectnode **nodearray;

  // The Modeller that owns this CSGgraph.
  Modeller *owner;

  // The root of the CSG graph.
  objectnode *object;

  // The node in the graph which is currently selected. NULL if none.
  objectnode *selected_object;

  // Get a new index for an objectnode.
  int new_index(objectnode *node);

  // Tag that an objectnode has been deleted.
  void delete_index(int index);

  // Retrieve an element by node ID.
  objectnode *retrieve_element(int node_id);

  // Select an element of the CSGgraph.
  void select_element(objectnode *obj_node);

  // Deselect all elements.
  void deselect(void);

  // Deselect an element.
  void deselect_element(objectnode *obj_node);

  // When a node's TransformList is changed, we must get the new transform
  // matrix and update the CSGgraph, Primtree and SolidOctrees accordingly.
  void update_transformed_object(objectnode *obj_node);

  // Update all the repeated children of a repeatobjectnode with the
  // new position matrix of the node.
  void update_transformed_repeatchain(primtreenode *child,
				      transform update_pre,
				      transform update_post,
				      transform new_trans,
				      transform old_trans_inv, int this_update_id);

  // PRIMARY UI ACCESS:
  // Link a new primitive into the modelled object.
  groupobjectnode *link_in_primitive(objectnode *link_to, linktype link_type,
				     Primitive *prim,
				     primobjectnode **return_primobj,
				     int num_trans = 0,
			         ObjectTransformation **trans_array = NULL);

  // PRIMARY UI ACCESS:
  // Link a copy of an object into the modelled object.
  groupobjectnode *link_in_copy(objectnode *link_to, linktype link_type,
				objectnode *copy,
				groupobjectnode **return_groupobj,
				int num_trans = 0,
			    ObjectTransformation **trans_array = NULL);

  // PRIMARY UI ACCESS:
  // Add a groupobjectnode under an object.
  void add_groupnode_under(groupobjectnode *add_under);

  // PRIMARY UI ACCESS:
  // Delete an objectnode.
  void delete_object(objectnode *kill_me);

  // Link an element into a group of the CSGgraph. This element must not
  // already be linked into a group - it must be newly created.
  void link_new_element(groupobjectnode *obj_group, linktype link_type,
			objectnode *addthing);

  // Add a wrapper around an object that's already linked into the CSGgraph.
  groupobjectnode *add_wrapper_around(objectnode *original);

  // Create a repeatobjectnode around an object that's already linked into the
  // CSGgraph. This replaces the object's node with a repeatobjectnode, which
  // takes its place in the CSGgraph.
  repeatobjectnode *add_repeat_around(objectnode *original,
				      int num_copies,
				      int num_trans = 0,
			          ObjectTransformation **trans_array = NULL);
  // Print the CSGgraph.
  void print_csggraph(void);

  // Delete an objectnode in the CSGgraph.
  void delete_objectnode(objectnode *kill_me);

  // Construct a new CSGgraph.
  CSGgraph(Modeller *owner);

  // Delete the CSGgraph.
  ~CSGgraph(void)
  {
    // The Modeller should have deleted the Primtree before
    // deleting this CSGgraph. That means that all primtreenodes
    // will have been deleted, deleting the objectnodes that
    // represent them as they do so. The CSGgraph should now be
    // empty.
    assert(object == NULL);

    if (nodearray != NULL)
      free(nodearray);
  }
};
