/*
 * dmd.cc: Routines for all the direct manipulation devices.
 *
 * The direct manipulation devices that may be attached to transformations are
 * defined by the behaviour listed in this file; how to select and manipulate
 * them, and their effects on their controlling transformations.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

/*
 * Routines for all DMD subclasses
 */

// Used for ray-line intersections.
double ray_line_intersect_tolerance = 0.01;

// Check to see whether a ray intersects a star. Rays must be given in
// a DMD's coordinate space.
int DMD::star_ray_intersect(vector star_pos, double star_size,
			    vector ray_from, vector ray_dir)
{
  vector real_from;
  double a, b, c;

  // Subtract star_pos from ray_from, transforming the ray into a space with
  // the sphere at the center.
  Vector::sub(ray_from, star_pos, real_from);

  // The quadratic equation that defines t is given by (u + tv)^2 = r^2
  // for a sphere ray-intersection. Compute the coefficients of this eqn.
  a = Vector::dot(ray_dir, ray_dir);
  b = 2 * Vector::dot(real_from, ray_dir);
  c = Vector::dot(real_from, real_from) - SQR(star_size);

  // If the sign of the determinant is positive, there is an intersection.
  if (SQR(b) - 4*a*c > 0)
    return TRUE;
  else
    return FALSE;
}
  
// Check to see whether a ray intersects a line-segment that is part of
// a DMD.
int DMD::linesegment_ray_intersect(vector line_0, vector line_1,
				   vector ray_from, vector ray_dir)
{
  vector line_dir, l_u, vxd, tempvec, closest_pt;
  double vxd_sq, s, t, dist;

  Vector::sub(line_1, line_0, line_dir);

  // The lines may be given by the equations p = u + vt and q = l + ds
  // where u, v are ray_from, ray_dir and l, d are line_0, line_dir.
  //
  // Then the values of t and s that we're looking for are
  // t = (((l - u) x d) . (v x d)) / (v x d)^2
  // s = (((l - u) x v) . (v x d)) / (v x d)^2
  //
  // This gives the distance of closest approach as
  // magnitude(l + ds - u - vt)
  //
  // Note that if s < 0 or s > 1, the intersection point lies beyond the end
  // of the line segment.

  Vector::sub(line_0, ray_from, l_u);
  Vector::cross(ray_dir, line_dir, vxd);
  vxd_sq = Vector::dot(vxd, vxd);
  if (vxd_sq < TINY) {
    // Parallel lines. Check to see if they are nearly-coincident, though.
    Vector::set(closest_pt, line_0);
  } else {
    Vector::cross(l_u, ray_dir, tempvec);
    s = Vector::dot(tempvec, vxd) / vxd_sq;

    if (s < 0) {
      // The closest-approach point is beyond line_0. Check to see if line_0
      // is close enough.
      Vector::set(closest_pt, line_0);
    } else if (s > 1) {
      // The closest-approach point is beyond line_1. Check to see if line_1
      // is close enough.
      Vector::set(closest_pt, line_1);
    } else {
      // Find the coordinates of q.
      Vector::set(closest_pt, line_0);
      Vector::addfrac(s, line_dir, closest_pt);
    }
  }
  
  // Now, check the distance from closest_pt to the ray.
  
  // The t-value along the ray of p, the closest point to closest_pt, is
  // given by
  //
  // t = ((q - u) . v) / (v . v)
  Vector::sub(closest_pt, ray_from, tempvec);
  t = Vector::dot(tempvec, ray_dir) / CHECK0(Vector::dot(ray_dir, ray_dir),
					     "lineseg_r_i:ray_dir");

  // We calculate the distance from closest_pt to p as
  // dist = | q - (u + vt) |
  // Note that we can use (q - u) which is still in tempvec.
  Vector::addfrac(-t, ray_dir, tempvec);
  dist = Vector::dot(tempvec, tempvec);

  if (dist < SQR(ray_line_intersect_tolerance))
    return TRUE;
  else
    return FALSE;
}

/*
 * Routines for TranslationDMD
 */

// Draw the DMD.
void TranslationDMD::draw_dmd(View *view, Skitter *skitter,
			      int use_fractional_alpha)
{
  int i, ax, handle;
  vector lines[8];
  vector axis_vec[6], star_pt;
  color draw_col;

  // First: we draw the central (white) skitter which represents the origin of
  // the translation coordinate system. This is connected by a line to the
  // translation DMD.
  view->set_drawcolor(1.0, 1.0, 1.0);
  view->set_drawalpha(use_fractional_alpha ? 0.5 : 1.0);

  Vector::set(lines[0], zero_vector);
  Vector::assign(lines[1], +xhair_size, 0.0, 0.0);
  Vector::set(lines[2], zero_vector);
  Vector::assign(lines[3], 0.0, +xhair_size, 0.0);
  Vector::set(lines[4], zero_vector);
  Vector::assign(lines[5], 0.0, 0.0, +xhair_size);
  Vector::set(lines[6], zero_vector);
  Vector::set(lines[7], trans->trans_amount);

  for (i = 0; i < 8; i++)
    Matrix::apply(skitter->withouttrans_position[FORE], lines[i], lines[i]);

  view->draw_lines(4, lines);

  // Next: draw the axes for the translation.
  view->set_drawcolor(0.133, 0.545, 0.133);

  Vector::assign(lines[0], -dmd_size, 0.0, 0.0);
  Vector::assign(lines[1], +dmd_size, 0.0, 0.0);
  Vector::assign(lines[2], 0.0, -dmd_size, 0.0);
  Vector::assign(lines[3], 0.0, +dmd_size, 0.0);
  Vector::assign(lines[4], 0.0, 0.0, -dmd_size);
  Vector::assign(lines[5], 0.0, 0.0, +dmd_size);
  
  for (i = 0; i < 6; i++) {
    Vector::add(trans->trans_amount, lines[i], lines[i]);
    Matrix::apply(skitter->withouttrans_position[FORE], lines[i], lines[i]);
  }

  view->draw_lines(3, lines);

  // Set up handy axis vectors. 0 = +X, 1 = -X, 2 = +Y, etc
  for (i = 2*X; i <= 2*Z+1; i++) {
    Vector::set(axis_vec[i], zero_vector);
    axis_vec[i][i/2] = ((i % 2) ? -dmd_size : +dmd_size);
  }

  // Draw the central (free-translation) star on the DMD.
  if (is_being_manipulated && (current_move == W)) {
    view->set_drawcolor(1.0, 1.0, 1.0);
  } else {
    view->set_drawcolor(0.133, 0.545, 0.133);
  }
  view->draw_star(3, star_size, skitter->withouttrans_position,
		  trans->trans_amount, plus_Z, axis_vec[2*X]);

  // For each of the six points on the DMD, draw a star.
  for (ax = X; ax <= Z; ax++) {
    for (handle = 0; handle <= 1; handle++) {
      if (is_being_manipulated && (current_move == ax) &&
	  (handle_selected == handle)) {
	// This handle is being dragged, make it white.
	view->set_drawcolor(1.0, 1.0, 1.0);
      } else {
	// This handle is inactive.
	Color::set(draw_col, zero_color);
	draw_col[ax] = 1.0;         // R, G, B for X, Y, Z
	view->set_drawcolor(draw_col);
      }
      Vector::add(trans->trans_amount, axis_vec[2*ax + handle], star_pt);
      view->draw_star(2, star_size, skitter->withouttrans_position,
		      star_pt, axis_vec[2*ax],
		      axis_vec[2*((ax + 1) % 3)]);
    }
  }
}

// Could this click be a manipulate? The ray is given in the coordinate
// space of the DMD.
int TranslationDMD::check_manipulate(vector ray_from, vector ray_dir)
{
  int ax;
  vector star_pos, line0, line1;

  // Check to see if the ray intersects any of the stars (manipulation points).
  if (star_ray_intersect(trans->trans_amount, star_size, ray_from, ray_dir)) {
    // Selected the free-manipulation star.
    is_being_manipulated = TRUE;
    current_move = W;
    Vector::set(start_translation, trans->trans_amount);
    Vector::set(start_ray_from, ray_from);
    Vector::set(start_ray_dir, ray_dir);
    return TRUE;
  }

  for (ax = X; ax <= Z; ax++) {
    Vector::set(star_pos, trans->trans_amount);
    star_pos[ax] += dmd_size;

    if (star_ray_intersect(star_pos, star_size, ray_from, ray_dir)) {
      // Selected an axis-manipulation star.
      is_being_manipulated = TRUE;
      current_move = (axis) ax;
      handle_selected = 0;
      Vector::set(start_translation, trans->trans_amount);
      Vector::set(start_ray_from, ray_from);
      Vector::set(start_ray_dir, ray_dir);
      return TRUE;
    }
    
    Vector::set(star_pos, trans->trans_amount);
    star_pos[ax] -= dmd_size;

    if (star_ray_intersect(star_pos, star_size, ray_from, ray_dir)) {
      // Selected an axis-manipulation star.
      is_being_manipulated = TRUE;
      current_move = (axis) ax;
      handle_selected = 1;
      Vector::set(start_translation, trans->trans_amount);
      Vector::set(start_ray_from, ray_from);
      Vector::set(start_ray_dir, ray_dir);
      return TRUE;
    }
  }

  // Check to see if this is a selection on any of the DMD's lines. If so,
  // then we keep the DMD selected.
  for (ax = X; ax <= Z; ax++) {
    Vector::set(line0, trans->trans_amount);
    Vector::set(line1, trans->trans_amount);
    line0[ax] += dmd_size;
    line1[ax] -= dmd_size;

    if (linesegment_ray_intersect(line0, line1, ray_from, ray_dir))
      return TRUE;
  }  

  // Check the white guide-line of the DMD.
  Vector::set(line0, zero_vector);
  Vector::set(line1, trans->trans_amount);
  if (linesegment_ray_intersect(line0, line1, ray_from, ray_dir))
    return TRUE;

  // Check the central white skitter.
  Vector::set(line0, zero_vector);
  for (ax = X; ax <= Z; ax++) {
    Vector::set(line1, zero_vector);
    line1[ax] += xhair_size;

    if (linesegment_ray_intersect(line0, line1, ray_from, ray_dir))
      return TRUE;
  }  

  // We haven't intersected any of the components of the DMD.
  return FALSE;
}

// Manipulate the DMD.
void TranslationDMD::manipulate(vector ray_from, vector ray_dir)
{
  double start_t, current_t, u_u, u_v, v_v;
  vector start_transpt, current_transpt, tempvec;
  vector new_translation;
  int other_axis1, other_axis2;

  if (!is_being_manipulated)
    return;

  if (current_move == W) {
    // Perform a free translation.
    
    // We are translating in the plane perpendicular to start_ray_dir,
    // and through the point start_translation in local coordinate space.
    // Construct where the start ray intersected this plane.
    Vector::sub(start_translation, start_ray_from, tempvec);
    start_t = Vector::dot(tempvec, start_ray_dir)
      / Vector::dot(start_ray_dir, start_ray_dir);
    Vector::set(start_transpt, start_ray_from);
    Vector::addfrac(start_t, start_ray_dir, start_transpt);

    // Construct where the current ray intersects this plane.
    Vector::sub(start_translation, ray_from, tempvec);
    current_t = Vector::dot(tempvec, ray_dir)
      / Vector::dot(ray_dir, ray_dir);
    Vector::set(current_transpt, ray_from);
    Vector::addfrac(current_t, ray_dir, current_transpt);
  } else {
    // Perform a translation about an axis.
    other_axis1 = (current_move + 1) % 3;
    other_axis2 = (current_move + 2) % 3;
    
    // First up: find the closest point on the manipulated axis to the
    // original ray.
    u_u = SQR(start_ray_from[other_axis1]) + SQR(start_ray_from[other_axis2]);
    u_v = start_ray_from[other_axis1]*start_ray_dir[other_axis1] +
      start_ray_from[other_axis2]*start_ray_dir[other_axis2];
    v_v = SQR(start_ray_dir[other_axis1]) + SQR(start_ray_dir[other_axis2]);
    if (fabs(v_v) > TINY)
      start_t = - u_v / v_v;
    else
      start_t = 0;
    
    Vector::set(start_transpt, zero_vector);
    start_transpt[current_move] = start_ray_from[current_move]
      + start_t * start_ray_dir[current_move];   

    // Construct the closest point on this axis to the current ray.
    u_u = SQR(ray_from[other_axis1]) + SQR(ray_from[other_axis2]);
    u_v = ray_from[other_axis1]*ray_dir[other_axis1] +
      ray_from[other_axis2]*ray_dir[other_axis2];
    v_v = SQR(ray_dir[other_axis1]) + SQR(ray_dir[other_axis2]);
    if (fabs(v_v) > TINY)
      current_t = - u_v / v_v;
    else
      current_t = 0;
    
    Vector::set(current_transpt, zero_vector);
    current_transpt[current_move] = ray_from[current_move]
      + current_t * ray_dir[current_move];
  }

  // We are translating by the difference between these two points.
  Vector::sub(current_transpt, start_transpt, tempvec);
  Vector::add(tempvec, start_translation, new_translation);
  
  // Set this translation.
  trans->set_translation(new_translation);
}

// Release the DMD.
void TranslationDMD::release(void)
{
  if (!is_being_manipulated)
    return;

  is_being_manipulated = FALSE;
}


/*
 * Routines for RotationDMD
 */

// Draw the DMD.
void RotationDMD::draw_dmd(View *view, Skitter *skitter,
			   int use_fractional_alpha)
{
  int i, ax, other_axis1, other_axis2;
  double theta;
  vector star_pos, star_dir, perp_dir, plusvec[3], axis_untransformed;
  quaternion temp_q;
  color draw_col;

  Vector::assign(plusvec[X], 1, 0, 0);
  Vector::assign(plusvec[Y], 0, 1, 0);
  Vector::assign(plusvec[Z], 0, 0, 1);

  // First, set up the points for the circle-approximation.
  for (i = 0; i < num_circlesegs; i++) {
    theta = 2*i*PI/num_circlesegs;

    Vector::assign(circle_pt[i], cos(theta), sin(theta), 0);
  }

  // Construct the X, Y, Z rotations from the current rotation quaternion.
  quaternion_to_xyz(trans->rotate_q, axis_rotation);

  // Draw the gyroscope.
  for (ax = X; ax <= Z; ax++) {
    other_axis1 = (ax + 1) % 3;
    other_axis2 = (ax + 2) % 3;

    Color::set(draw_col, zero_color);
    draw_col[ax] = 1.0;
    view->set_drawcolor(draw_col);

    // First: a circle that shows the plane of rotation.
    for (i = 0; i < num_circlesegs; i++) {
      Vector::set(lines[2*i], zero_vector);
      lines[2*i][other_axis1] = dmd_size * circle_pt[i][X];
      lines[2*i][other_axis2] = dmd_size * circle_pt[i][Y];

      Vector::set(lines[2*i+1], zero_vector);
      lines[2*i+1][other_axis1] = dmd_size
        * circle_pt[(i+1) % num_circlesegs][X];
      lines[2*i+1][other_axis2] = dmd_size
        * circle_pt[(i+1) % num_circlesegs][Y];
    }

    for (i = 0; i < num_circlesegs * 2; i++)
      Matrix::apply(skitter->withouttrans_position[FORE], lines[i], lines[i]);

    view->draw_lines(num_circlesegs, lines);

    // Next: a star at the current rotation position.
    Vector::set(star_pos, zero_vector);
    star_pos[other_axis1] = -dmd_size * cos(axis_rotation[ax]);
    star_pos[other_axis2] = -dmd_size * sin(axis_rotation[ax]);

    Vector::set(star_dir, zero_vector);
    star_dir[other_axis1] = -sin(axis_rotation[ax]);
    star_dir[other_axis2] =  cos(axis_rotation[ax]);

    Vector::set(perp_dir, zero_vector);
    perp_dir[other_axis1] = cos(axis_rotation[ax]);
    perp_dir[other_axis2] = sin(axis_rotation[ax]);

    // Is the star highlighted?
    if (is_being_manipulated && is_axis_rotation &&
	(current_rotation == ax))
      view->set_drawcolor(1.0, 1.0, 1.0);

    view->draw_star(3, star_size, skitter->withouttrans_position,
		    star_pos, star_dir, perp_dir);

#if 0
    // Next: a tick on this circle, which shows zero-rotation.
    view->set_drawcolor(1.0, 1.0, 1.0);
    Vector::set(lines[0], zero_vector);
    lines[0][other_axis1] = -dmd_size - tick_size;
    Vector::set(lines[1], zero_vector);
    lines[1][other_axis1] = -dmd_size + tick_size;
    view->draw_lines(1, lines);
#endif
  }

  // For each of X, Y and Z, draw the axes and the free-rotation stars.
  for (ax = X; ax <= Z; ax++) {
    // Draw a white line from the zero-position to the current rotated position
    // of this axis.
    for (i = 0; i <= num_circlesegs / 2; i++) {
      Vector::mult(dmd_size, plusvec[ax], lines[2*i]);

      Quaternion::power(trans->rotate_q, (1.0 * i) / (num_circlesegs / 2),
			temp_q);
      Quaternion::apply(temp_q, lines[2*i], lines[2*i]);

      Matrix::apply(skitter->withouttrans_position[FORE], lines[2*i], lines[2*i]);

      if (i > 0)
        Vector::set(lines[2*i - 1], lines[2*i]);
    }
    view->set_drawcolor(1.0, 1.0, 1.0);
    view->draw_lines(num_circlesegs / 2, lines);
    
    // Find where the axis is.
    view->set_drawcolor(0.627, 0.125, 0.941);

    Vector::set(axis_untransformed, zero_vector);
    axis_untransformed[ax] = dmd_size;
    Quaternion::apply(trans->rotate_q, axis_untransformed, axis_untransformed);

    Matrix::apply(skitter->withouttrans_position[FORE], zero_vector,        lines[0]);
    Matrix::apply(skitter->withouttrans_position[FORE], axis_untransformed, lines[1]);
    view->draw_lines(1, lines);

    // Determine the star's direction and orientation.
    Vector::set(star_dir, plusvec[ax]);
    Quaternion::apply(trans->rotate_q, star_dir, star_dir);

    Vector::set(perp_dir, plusvec[(ax+1) % 3]);
    Quaternion::apply(trans->rotate_q, perp_dir, perp_dir);
    
    // Is the star highlighted?
    view->set_drawcolor(0.627, 0.125, 0.941);
    if (is_being_manipulated && (!is_axis_rotation) && 
        (current_rotation == ax))
      view->set_drawcolor(1.0, 1.0, 1.0);

    view->draw_star(3, star_size, skitter->withouttrans_position,
                    axis_untransformed, star_dir, perp_dir);
  }
}

// Could this click be a manipulate? The ray is given in the coordinate
// space of the DMD.
int RotationDMD::check_manipulate(vector ray_from, vector ray_dir)
{
  int ax, other_axis1, other_axis2;
  vector star_pos, line0, line1, plusvec[3];
  vector torus_ray_from, torus_ray_dir;
  int num_torus_intersections;
  rayintersect torus_intersections[6];
  transform torustransform;
  torusPrimitive *torus_prim;

  Vector::assign(plusvec[X], 1, 0, 0);
  Vector::assign(plusvec[Y], 0, 1, 0);
  Vector::assign(plusvec[Z], 0, 0, 1);

  // Construct the X, Y, Z rotations from the current rotation quaternion.
  quaternion_to_xyz(trans->rotate_q, axis_rotation);

  for (ax = X; ax <= Z; ax++) {
    other_axis1 = (ax + 1) % 3;
    other_axis2 = (ax + 2) % 3;

    // Check the axis-manipulation star.
    Vector::set(star_pos, zero_vector);
    star_pos[other_axis1] = -dmd_size * cos(axis_rotation[ax]);
    star_pos[other_axis2] = -dmd_size * sin(axis_rotation[ax]);
    if (star_ray_intersect(star_pos, star_size, ray_from, ray_dir)) {
      is_being_manipulated = TRUE;
      current_rotation = (axis) ax;
      is_axis_rotation = TRUE;
      Quaternion::set(start_rotation, trans->rotate_q);
      Vector::set(start_ray_from, ray_from);
      Vector::set(start_ray_dir, ray_dir);
      return TRUE;
    }
    
    // Check the free-rotation star.
    Vector::mult(dmd_size, plusvec[ax], star_pos);
    Quaternion::apply(trans->rotate_q, star_pos, star_pos);
    if (star_ray_intersect(star_pos, star_size, ray_from, ray_dir)) {
      is_being_manipulated = TRUE;
      current_rotation = (axis) ax;
      is_axis_rotation = FALSE;

      // We are manipulating on the front side of the sphere if the ray-dir
      // points in the opposite direction to the star's position.
      if (Vector::dot(star_pos, ray_dir) < 0)
	manipulate_on_front_side = TRUE;
      else
	manipulate_on_front_side = FALSE;

      Quaternion::set(start_rotation, trans->rotate_q);
      Vector::set(start_ray_from, ray_from);
      Vector::set(start_ray_dir, ray_dir);
      Quaternion::set(last_rotation, trans->rotate_q);
      Vector::set(last_ray_from, ray_from);
      Vector::set(last_ray_dir, ray_dir);
      return TRUE;
    }
  }

  // Check to see if this is a selection on any of the DMD's axes. If so,
  // then we keep the DMD selected.
  for (ax = X; ax <= Z; ax++) {
    Vector::set(line0, zero_vector);
    Vector::mult(dmd_size, plusvec[ax], line1);
    Quaternion::apply(trans->rotate_q, line1, line1);

    if (linesegment_ray_intersect(line0, line1, ray_from, ray_dir))
      return TRUE;
  }  

  // Check to see if this is a selection on any of the DMD's circles. If so,
  // then we keep the DMD selected.
  for (ax = X; ax <= Z; ax++) {
    // We want to intersect with a torus that's aligned along the ax axis.
    // However the TorusPrimitive is aligned along the Z axis.
    Transform::unit(torustransform);
    switch(ax) {
    case X:
      Transform::rotate(torustransform, X, Z, 90);
      break;
    case Y:
      Transform::rotate(torustransform, Y, Z, 90);
      break;
    }

    // Scale the torus so that it has a radius of 1.0.
    Transform::scale(torustransform, 1.0 / dmd_size);

    // Transform the ray's from and dir vectors by this transform,
    // into torus primspace.
    Matrix::apply(torustransform[FORE], ray_from, torus_ray_from);
    Matrix::applynoshift(torustransform[FORE], ray_dir, torus_ray_dir);

    // Create a torus and intersect the ray with it. 
    torus_prim = new torusPrimitive(0.05 / dmd_size);
    num_torus_intersections = torus_prim->ray_intersect(torus_ray_from,
							torus_ray_dir,
							torus_intersections);
    delete torus_prim;

    if (num_torus_intersections)
      return TRUE;
  }

  // We haven't intersected any of the components of the DMD.
  return FALSE;
}

// Manipulate the DMD.
void RotationDMD::manipulate(vector ray_from, vector ray_dir)
{
  int other_axis1, other_axis2;
  double start_t, current_t, start_theta, current_theta, costheta, theta;
  vector start_controlpt, current_controlpt, increment_rotaxis;
  quaternion new_rotation, increment_q;

  if (!is_being_manipulated)
    return;

  if (is_axis_rotation) {
    other_axis1 = (current_rotation + 1) % 3;
    other_axis2 = (current_rotation + 2) % 3;

    // Where does the start ray intersect the plane of the rotation controller?
    if (fabs(start_ray_dir[current_rotation]) < TINY)
      start_t = 0;
    else
      start_t = - start_ray_from[current_rotation]
	/ start_ray_dir[current_rotation];
    Vector::set(start_controlpt, start_ray_from);
    Vector::addfrac(start_t, start_ray_dir, start_controlpt);

    // Find the angle that this corresponds to.
    start_theta = atan2(-start_controlpt[other_axis1],
			start_controlpt[other_axis2]);

    // Where does the current ray intersect the plane of the rotation
    // controller?
    if (fabs(ray_dir[current_rotation]) < TINY)
      current_t = 0;
    else
      current_t = - ray_from[current_rotation] / ray_dir[current_rotation];
    Vector::set(current_controlpt, ray_from);
    Vector::addfrac(current_t, ray_dir, current_controlpt);

    // Find the angle that this corresponds to.
    current_theta = atan2(-current_controlpt[other_axis1],
			  current_controlpt[other_axis2]);

    // Find the original rotation.
    quaternion_to_xyz(start_rotation, axis_rotation);

    // We offset the original rotation by the difference between
    // the original start_theta and current_theta.
    axis_rotation[current_rotation] += current_theta - start_theta;

    // Set the new rotation.
    xyz_to_quaternion(axis_rotation, new_rotation);
    trans->set_rotation(new_rotation);
  } else {
    // Perform a free rotation.

    // Where did the last ray intersect the sphere of the DMD?
    find_sphere_controlpoint(last_ray_from, last_ray_dir, start_controlpt);

    // Where do we currently intersect the sphere of the DMD?
    find_sphere_controlpoint(ray_from, ray_dir, current_controlpt);

    // Find the quaternion that is the increment of rotation.
    Vector::cross(start_controlpt, current_controlpt, increment_rotaxis);

    costheta = Vector::dot(start_controlpt, current_controlpt);
    if (costheta - 1.0 > -TINY)
      theta = 0;
    else if (costheta + 1.0 < TINY)
      theta = PI;
    else
      theta = acos(costheta);

    Vector::mult(sin(theta/2) / CHECK0(Vector::magnitude(increment_rotaxis),
						"RotDMD::manipulate: i_raxis"),
		 increment_rotaxis, increment_q);
    increment_q[W] = cos(theta/2);

    // Add this quaternion to the current quaternion.
    Quaternion::mult(increment_q, trans->rotate_q, new_rotation);
    trans->set_rotation(new_rotation);

    // Set the 'last' variables.
    Vector::set(last_ray_from, ray_from);
    Vector::set(last_ray_dir, ray_dir);
    Quaternion::set(last_rotation, new_rotation);
  }
}

// Intersect a ray with the DMD's sphere, used by the free rotation code.
void RotationDMD::find_sphere_controlpoint(vector ray_from, vector ray_dir,
					   vector controlpt)
{
  vector outside_pt;
  double t_0, t_1, disc, a, b, c;

  // The quadratic equation that defines t is given by (u + tv)^2 = r^2
  // for a sphere ray-intersection. Compute the coefficients of this eqn.
  a = Vector::dot(ray_dir, ray_dir);
  b = 2 * Vector::dot(ray_from, ray_dir);
  c = Vector::dot(ray_from, ray_from) - SQR(dmd_size);
  
  disc = SQR(b) - 4*a*c;
  if (disc < TINY) {
    // The ray doesn't actually intersect the DMD sphere. We just find the
    // closest point on the outside.
    Vector::cross(ray_dir, ray_from, outside_pt);
    Vector::cross(ray_from, outside_pt, outside_pt);

    Vector::mult(1.0 / CHECK0(Vector::magnitude(outside_pt),
			      "outside_pt"),
		 outside_pt, controlpt);
  } else {

    // Code taken from Mirage: prim_sphere.c to find t_0 and t_1.
    if (a >= 0) {
      if (b >= 0) {
	t_0 = (-b - sqrt(disc)) / (2 * a);
	t_1 = c / (a * t_0);
      } else {
	t_1 = (-b + sqrt(disc)) / (2 * a);
	t_0 = c / (a * t_1);
      }
    } else {
      if (b >= 0) {
	t_1 = (-b - sqrt(disc)) / (2 * a);
	t_0 = c / (a * t_1);
      } else {
	t_0 = (-b + sqrt(disc)) / (2 * a);
	t_1 = c / (a * t_0);
      }
    }

    // Take the appropriate int'n point based on the state variable
    // manipulate_on_front_side. Note that we are guaranteed t_0 < t_1.
    Vector::set(controlpt, ray_from);
    Vector::addfrac(manipulate_on_front_side ? t_0 : t_1,
		    ray_dir, controlpt);

    Vector::normalise(controlpt, "find_sphere_controlpoint: controlpt");
  }
}

// Release the DMD.
void RotationDMD::release(void)
{
  if (!is_being_manipulated)
    return;

  is_being_manipulated = FALSE;
}

// Convert from a rotation quaternion to an XYZ rotation triplet.
void RotationDMD::quaternion_to_xyz(quaternion q, vector rotate_vec)
{
  vector plus_X, tempvec;
  quaternion current_q, inverse_rot;

  Vector::assign(plus_X, 1, 0, 0);
  Quaternion::set(current_q, q);

  // Where is plus_X mapped to by the current quaternion?
  Quaternion::apply(current_q, plus_X, tempvec);

  // Determine the rotation about Z (yaw) from atan2 on X and Y.
  rotate_vec[Z] = atan2(tempvec[Y], tempvec[X]);

  // Form a quaternion for the inverse of this rotation, and apply it to
  // the current quaternion to eliminate the contribution of the Z rotation.
  Quaternion::assign(inverse_rot,
		     cos(rotate_vec[Z]/2), 0, 0, -sin(rotate_vec[Z]/2));
  Quaternion::mult(inverse_rot, current_q, current_q);

  // Where is plus_X mapped to by the current quaternion?
  Quaternion::apply(current_q, plus_X, tempvec);
  
  // Determine the rotation about Y (pitch) from atan2 on Z and X.
  rotate_vec[Y] = atan2(-tempvec[Z], tempvec[X]);

  // Form a quaternion for the inverse of this rotation, and apply it to
  // the current quaternion to eliminate the contribution of the Y rotation.
  Quaternion::assign(inverse_rot,
		     cos(rotate_vec[Y]/2), 0, -sin(rotate_vec[Y]/2), 0);
  Quaternion::mult(inverse_rot, current_q, current_q);
  
  // Where is plus_Z mapped to by the current quaternion?
  Quaternion::apply(current_q, plus_Z, tempvec);

  // Determine the rotation about X (roll) from atan2 on Y and Z.
  rotate_vec[X] = atan2(-tempvec[Y], tempvec[Z]);
}

void RotationDMD::xyz_to_quaternion(vector rotate_vec, quaternion q)
{
  quaternion rotation_q[3];
  int ax;

  // Construct the rotations around X, Y, and Z.
  for (ax = X; ax <= Z; ax++) {
    Quaternion::unit(rotation_q[ax]);
    rotation_q[ax][ax] = sin(rotate_vec[ax]/2);
    rotation_q[ax][W] = cos(rotate_vec[ax]/2);
  }

  // Combine these three rotations in the order q_z q_y q_x.
  Quaternion::mult(rotation_q[Z], rotation_q[Y], q);
  Quaternion::mult(q, rotation_q[X], q);
}


/*
 * Routines for ScaleDMD
 */

// Draw the DMD.
void ScaleDMD::draw_dmd(View *view, Skitter *skitter,
			int use_fractional_alpha)
{
  int i, ax, handle;
  vector lines[6];
  vector axis_vec[6];
  
  // First, draw the axes for the scale.
  view->set_drawcolor(1.000, 1.000, 0.000);

  Vector::assign(lines[0], -trans->scale_amount, 0.0, 0.0);
  Vector::assign(lines[1], +trans->scale_amount, 0.0, 0.0);
  Vector::assign(lines[2], 0.0, -trans->scale_amount, 0.0);
  Vector::assign(lines[3], 0.0, +trans->scale_amount, 0.0);
  Vector::assign(lines[4], 0.0, 0.0, -trans->scale_amount);
  Vector::assign(lines[5], 0.0, 0.0, +trans->scale_amount);
  
  for (i = 0; i < 6; i++)
    Matrix::apply(skitter->withouttrans_position[FORE], lines[i], lines[i]);

  view->draw_lines(3, lines);

  // Set up handy axis vectors. 0 = +X, 1 = -X, 2 = +Y, etc
  for (i = 2*X; i <= 2*Z+1; i++) {
    Vector::set(axis_vec[i], zero_vector);
    axis_vec[i][i/2] = ((i % 2) ? -trans->scale_amount : +trans->scale_amount);
  }

  // For each of the six points on the DMD, draw a star.
  for (ax = X; ax <= Z; ax++) {
    for (handle = 0; handle <= 1; handle++) {
      if (is_being_manipulated && (current_move == ax) &&
	  (handle_selected == handle)) {
	// This handle is being dragged, make it white.
	view->set_drawcolor(1.0, 1.0, 1.0);
      } else {
	// This handle is inactive.
	view->set_drawcolor(1.0, 1.0, 0.0);
      }
      view->draw_star(2, star_size, skitter->withouttrans_position,
		      axis_vec[2*ax + handle], axis_vec[2*ax],
		      axis_vec[2*((ax + 1) % 3)]);
    }
  }
}

// Could this click be a manipulate? The ray is given in the coordinate
// space of the DMD.
int ScaleDMD::check_manipulate(vector ray_from, vector ray_dir)
{
  int ax;
  vector star_pos, line0, line1;

  // Check to see if the ray intersects any of the stars (manipulation points).
  for (ax = X; ax <= Z; ax++) {
    Vector::set(star_pos, zero_vector);
    star_pos[ax] += trans->scale_amount;
    if (star_ray_intersect(star_pos, star_size, ray_from, ray_dir)) {
      // Selected an axis-manipulation star.
      is_being_manipulated = TRUE;
      current_move = (axis) ax;
      handle_selected = 0;
      start_scale = trans->scale_amount;
      Vector::set(start_ray_from, ray_from);
      Vector::set(start_ray_dir, ray_dir);
      return TRUE;
    }
    
    Vector::set(star_pos, zero_vector);
    star_pos[ax] -= trans->scale_amount;
    if (star_ray_intersect(star_pos, star_size, ray_from, ray_dir)) {
      // Selected an axis-manipulation star.
      is_being_manipulated = TRUE;
      current_move = (axis) ax;
      handle_selected = 1;
      start_scale = trans->scale_amount;
      Vector::set(start_ray_from, ray_from);
      Vector::set(start_ray_dir, ray_dir);
      return TRUE;
    }
  }

  // Check to see if this is a selection on any of the DMD's lines. If so,
  // then we keep the DMD selected.
  for (ax = X; ax <= Z; ax++) {
    Vector::set(line0, zero_vector);
    Vector::set(line1, zero_vector);
    line0[ax] += trans->scale_amount;
    line1[ax] -= trans->scale_amount;

    if (linesegment_ray_intersect(line0, line1, ray_from, ray_dir))
      return TRUE;
  }  

  // We haven't intersected any of the components of the DMD.
  return FALSE;
}

// Manipulate the DMD.
void ScaleDMD::manipulate(vector ray_from, vector ray_dir)
{
  double start_t, current_t, u_u, u_v, v_v;
  double start_scaleval, current_scaleval, new_scale;
  int other_axis1, other_axis2;

  if (!is_being_manipulated)
    return;

  // We're scaling about one axis. Perform calculations relative to the
  // other two axes.
  other_axis1 = (current_move + 1) % 3;
  other_axis2 = (current_move + 2) % 3;
    
  // First up: find the closest point on the manipulated axis to the
  // original ray.
  u_u = SQR(start_ray_from[other_axis1]) + SQR(start_ray_from[other_axis2]);
  u_v = start_ray_from[other_axis1]*start_ray_dir[other_axis1] +
    start_ray_from[other_axis2]*start_ray_dir[other_axis2];
  v_v = SQR(start_ray_dir[other_axis1]) + SQR(start_ray_dir[other_axis2]);
  if (fabs(v_v) > TINY)
    start_t = - u_v / v_v;
  else
    start_t = 0;
  
  start_scaleval = fabs(start_ray_from[current_move]
			+ start_t * start_ray_dir[current_move]);   

  // Construct the closest point on this axis to the current ray.
  u_u = SQR(ray_from[other_axis1]) + SQR(ray_from[other_axis2]);
  u_v = ray_from[other_axis1]*ray_dir[other_axis1] +
    ray_from[other_axis2]*ray_dir[other_axis2];
  v_v = SQR(ray_dir[other_axis1]) + SQR(ray_dir[other_axis2]);
  if (fabs(v_v) > TINY)
    current_t = - u_v / v_v;
  else
    current_t = 0;
    
  current_scaleval = fabs(ray_from[current_move]
			  + current_t * ray_dir[current_move]);

  // We are scaling by the ratio of these two points.
  new_scale = start_scale * current_scaleval / start_scaleval;
  
  // Set this translation.
  trans->set_scale(new_scale);
}

// Release the DMD.
void ScaleDMD::release(void)
{
  if (!is_being_manipulated)
    return;

  is_being_manipulated = FALSE;
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
