/*
 * dmd.h: Definitions for all the direct manipulation devices that can be
 * attached to transformations.
 */

/* Collection-safe */

// Used for ray-line intersections.
extern double ray_line_intersect_tolerance;

class DMD {
public:
  // Is this DMD currently being manipulated?
  int is_being_manipulated;

  // Draw the DMD.
  virtual void draw_dmd(View *view, Skitter *skitter,
			int use_fractional_alpha) = 0;

  // Could this click be a manipulate? The ray is given in the coordinate
  // space of the DMD.
  virtual int check_manipulate(vector ray_from, vector ray_dir) = 0;

  // Manipulate the DMD.
  virtual void manipulate(vector ray_from, vector ray_dir) = 0;

  // Release the DMD.
  virtual void release(void) = 0;

  // Check to see whether a ray intersects a star. Rays must be given in
  // a DMD's coordinate space.
  int star_ray_intersect(vector star_pos, double star_size,
			 vector ray_from, vector ray_dir);
  
  // Check to see whether a ray intersects a line-segment that is part of
  // a DMD.
  int linesegment_ray_intersect(vector line_0, vector line_1,
				vector ray_from, vector ray_dir);

  // Free the DMD.
  virtual ~DMD(void) {
  }
};

class TranslationDMD : public DMD {
public:
  // DMD parameters.
  double dmd_size;
  double star_size;
  double xhair_size;

  // The ObjectTranslation that owns this DMD.
  ObjectTranslation *trans;

  // The axis along which the DMD is currently being moved. This axis is
  // W if it's a free translation.
  axis current_move;

  // Which handle on the axis is selected? 0 = positive, 1 = negative.
  int handle_selected;

  // The value of the translation at the start of the movement, and the click
  // location that corresponded to this translation.
  vector start_translation;
  vector start_ray_from, start_ray_dir;

  // Draw the DMD.
  virtual void draw_dmd(View *view, Skitter *skitter,
			int use_fractional_alpha);

  // Could this click be a manipulate? The ray is given in the coordinate
  // space of the DMD.
  virtual int check_manipulate(vector ray_from, vector ray_dir);

  // Manipulate the DMD.
  virtual void manipulate(vector ray_from, vector ray_dir);

  // Release the DMD.
  virtual void release(void);

  // Create the DMD... called by the Translation that owns us.
  TranslationDMD(ObjectTranslation *owner) {
    is_being_manipulated = FALSE;
    trans = owner;

    dmd_size = 0.4;
    star_size = 0.05;
    xhair_size = 0.3;
  }

  // Free the DMD.
  virtual ~TranslationDMD(void) {
  }  
};

class RotationDMD : public DMD {
public:
  // DMD parameters.
  double dmd_size;
  double star_size;
  double tick_size;
  int num_circlesegs;

  // The ObjectRotation that owns this DMD.
  ObjectRotation *trans;

  // The current rotation, in terms of X, Y and Z rotations in radians.
  vector axis_rotation;

  // The type of current rotation. TRUE if this is an axis-aligned rotation
  // using the gyroscope, FALSE if this is a free rotation.
  int is_axis_rotation;

  // The axis around which the DMD is currently being rotated.
  // If this is an axis-aligned rotation, this variable holds the axis around
  // which the rotation is being performed. If not, then this variable holds
  // the axis which is being dragged.
  axis current_rotation;

  // The value of the rotation at the start of the movement, and the click
  // location that started it.
  quaternion start_rotation;
  vector start_ray_from, start_ray_dir;

  // Free-rotation variables; whether or not we are manipulating on the
  // front side of the sphere, and the last ray + quaternion.
  int manipulate_on_front_side;
  quaternion last_rotation;
  quaternion last_ray_from, last_ray_dir;

  // Arrays used for drawing the DMD.
  vector *circle_pt, *lines;
  
  // Draw the DMD.
  virtual void draw_dmd(View *view, Skitter *skitter,
			int use_fractional_alpha);

  // Could this click be a manipulate? The ray is given in the coordinate
  // space of the DMD.
  virtual int check_manipulate(vector ray_from, vector ray_dir);

  // Manipulate the DMD.
  virtual void manipulate(vector ray_from, vector ray_dir);

  // Release the DMD.
  virtual void release(void);

  // Convert from a rotation quaternion to an XYZ rotation triplet.
  void quaternion_to_xyz(quaternion q, vector rotate_vec);
  void xyz_to_quaternion(vector rotate_vec, quaternion q);

  // Intersect a ray with the DMD's sphere, used by the free rotation code.
  void find_sphere_controlpoint(vector ray_from, vector ray_dir,
				vector controlpt);
  // Create the DMD... called by the Rotation that owns us.
  RotationDMD(ObjectRotation *owner) {
    is_being_manipulated = FALSE;
    trans = owner;

    dmd_size = 0.8;
    star_size = 0.08;
    tick_size = 0.1;
    num_circlesegs = 36;

    circle_pt = (vector *) calloc(num_circlesegs, sizeof(vector));
    lines = (vector *) calloc(2*num_circlesegs, sizeof(vector));
    if (!circle_pt || !lines)
      fatal_error("RotationDMD::RotationDMD", "Can't allocate circlepoints",
				  "Out of memory.");
  }

  // Free the DMD.
  virtual ~RotationDMD(void) {
    free(circle_pt);
    free(lines);
  }  
};

class ScaleDMD : public DMD {
public:
  // DMD parameters.
  double star_size;

  // The ObjectScale that owns this DMD.
  ObjectScale *trans;

  // The axis along which the DMD is currently being scaled.
  axis current_move;

  // Which handle on the axis is selected? 0 = positive, 1 = negative.
  int handle_selected;

  // The value of the scale at the start of the movement, and the click
  // location that corresponded to this translation.
  double start_scale;
  vector start_ray_from, start_ray_dir;

  // Draw the DMD.
  virtual void draw_dmd(View *view, Skitter *skitter,
			int use_fractional_alpha);

  // Could this click be a manipulate? The ray is given in the coordinate
  // space of the DMD.
  virtual int check_manipulate(vector ray_from, vector ray_dir);

  // Manipulate the DMD.
  virtual void manipulate(vector ray_from, vector ray_dir);

  // Release the DMD.
  virtual void release(void);

  // Create the DMD... called by the Scale that owns us.
  ScaleDMD(ObjectScale *owner) {
    star_size = 0.05;
    is_being_manipulated = FALSE;
    trans = owner;
  }

  // Free the DMD.
  virtual ~ScaleDMD(void) {
  }  
};
