/*
 * layout.cc: Routines for GraphLayout and associated classes.
 *
 * Contains code to create and maintain the graph-view of the CSGgraph, and
 * lay it out in a human-readable format.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Perform a complete layout operation, from scratch.
void HorizLayout::complete_layout(void)
{
  recompute_layout();

  // Draw the graph on each view.
  for (int i = 0; i < modeller->num_graphviews; i++)
    draw_on_view(modeller->graphview[i]);
}

// Draw the graph on a particular view.
void HorizLayout::draw_on_view(GraphView *v)
{
  v->clear_canvas();
  draw_node(v, modeller->csg_graph->object);
  v->finish_drawing();
}

// Update the layout in response to a new link-in-copy operation.
void HorizLayout::update_linkincopy(groupobjectnode *linkchain_parent,
				    groupobjectnode *newcopy)
{
  HorizLayoutNode *new_lnode, *childcopy_lnode;

  // A copy-node has been added to the graph. It's a newly constructed node,
  // so it will need a layoutnode.
  new_lnode = new HorizLayoutNode(newcopy);
  newcopy->layoutnode = new_lnode;

  // This node is newly created, and linked to by just one parent.
  new_lnode->parent = (HorizLayoutNode *) linkchain_parent->layoutnode;
  new_lnode->layer = new_lnode->parent->layer + 1;
  new_lnode->num_ghostlinks = 0;
  new_lnode->childlink_is_ghost = TRUE;

  // The object that's being linked to by newcopy now has an additional
  // ghost link pointing to it.
  childcopy_lnode = ((HorizLayoutNode *) newcopy->group->layoutnode);
  childcopy_lnode->num_ghostlinks++;

  // We have to check the subgraph of newcopy->group, to see if it needs
  // to be moved down the tree based on this new link.
  if (childcopy_lnode->layer <= new_lnode->layer)
    check_node_shift(new_lnode, newcopy->group,
		     (new_lnode->layer + 1) - (childcopy_lnode->layer));

  // We have to recompute the layout of the graph.
  recompute_layout();
}

// Update the layout in response to a new link-in-primitive operation.
void HorizLayout::update_linkinprim(groupobjectnode *linkchain_parent,
				    primobjectnode *newprim)
{
  HorizLayoutNode *new_lnode;

  // A primnode has been added to the graph. It's a newly constructed node,
  // so it will need a layoutnode.
  new_lnode = new HorizLayoutNode(newprim);
  newprim->layoutnode = new_lnode;

  // This node is newly created, and linked to by just one parent.
  new_lnode->parent = (HorizLayoutNode *) linkchain_parent->layoutnode;
  new_lnode->layer = new_lnode->parent->layer + 1;
  new_lnode->num_ghostlinks = 0;
  new_lnode->childlink_is_ghost = FALSE;

  // We have to recompute the layout of the graph.
  recompute_layout();
}

// Update the layout in response to a new wrapper-around operation.
void HorizLayout::update_wrapper(groupobjectnode *wrapper)
{
  HorizLayoutNode *new_lnode, *child_lnode;

  assert(wrapper != NULL);
  assert(wrapper->group != NULL);
  assert(wrapper->layoutnode == NULL);
  assert(wrapper->group->layoutnode != NULL);

  // This wrapper is newly created, it needs a LayoutNode.
  new_lnode = new HorizLayoutNode(wrapper);
  wrapper->layoutnode = new_lnode;
 
  // Set up the new layoutnode... it is largely the same as the wrapped
  // layoutnode. Modify the child layoutnode as necessary.
  child_lnode = (HorizLayoutNode *) wrapper->group->layoutnode;

  new_lnode->layer = child_lnode->layer;
  child_lnode->layer++;
  
  new_lnode->parent = child_lnode->parent;
  child_lnode->parent = new_lnode;

  new_lnode->num_ghostlinks = child_lnode->num_ghostlinks;
  child_lnode->num_ghostlinks = 0;
  
  new_lnode->childlink_is_ghost = FALSE;

  // Check all the children of wrapper->group to see if they need to be
  // shifted down a layer as well.
  check_children_for_shift(wrapper->group, +1);

  // If there are any ghost links anywhere further down the subgraph
  // enclosed by wrapper->group, then they will need to be shifted down
  // to the next layer etc, which will cause changes in the layer widths,
  // which necessitates a complete recompute of the layout.
  recompute_layout();
}

// Update the layout in response to a new repeat-around operation.
void HorizLayout::update_repeat(repeatobjectnode *repeat)
{
  HorizLayoutNode *new_lnode, *child_lnode;

  assert(repeat != NULL);
  assert(repeat->repeat != NULL);
  assert(repeat->layoutnode == NULL);
  assert(repeat->repeat->layoutnode != NULL);

  // This repeat is newly created, it needs a LayoutNode.
  new_lnode = new HorizLayoutNode(repeat);
  repeat->layoutnode = new_lnode;
 
  // Set up the new layoutnode... it is largely the same as the repeated
  // layoutnode. Modify the child layoutnode as necessary.
  child_lnode = (HorizLayoutNode *) repeat->repeat->layoutnode;

  new_lnode->layer = child_lnode->layer;
  child_lnode->layer++;
  
  new_lnode->parent = child_lnode->parent;
  child_lnode->parent = new_lnode;

  new_lnode->num_ghostlinks = child_lnode->num_ghostlinks;
  child_lnode->num_ghostlinks = 0;
  
  new_lnode->childlink_is_ghost = FALSE;

  // Check all the children of repeat->repeat to see if they need to be
  // shifted down as well.
  check_children_for_shift(repeat->repeat, +1);

  // If there are any ghost links anywhere further down the subgraph
  // enclosed by repeat->repeat, then they will need to be shifted down
  // to the next layer etc, which will cause changes in the layer widths,
  // which necessitates a complete recompute of the layout.
  recompute_layout();
}

// Check a node to see if it needs to be shifted.
void HorizLayout::check_node_shift(HorizLayoutNode *parent_lnode,
				   objectnode *child, int layer_offset)
{
  HorizLayoutNode *child_lnode, *nextparent_lnode, *greatestparent;
  parenttrackernode *ptptr;
  int greatest_parent_layer, shifted = FALSE;

  child_lnode = (HorizLayoutNode *) child->layoutnode;
  assert(child_lnode != NULL);

  if (layer_offset > 0) {
    // The parent layoutnode is moving down the tree.
    if (parent_lnode == child_lnode->parent) {
      // The child depends off the parent (i.e. the parent's link is not
      // a ghost link). Thus it must be shifted.
      child_lnode->layer = parent_lnode->layer + 1;
      shifted = TRUE;
    } else {
      // The child doesn't presently depend off the parent node.
      if (parent_lnode->layer > child_lnode->parent->layer) {
	// The child must be switched from its current parent to the new
	// parent.
	parent_lnode->childlink_is_ghost = FALSE;
	child_lnode->parent->childlink_is_ghost = TRUE;
	child_lnode->parent->ghostlink_index_within_layer =
	  parent_lnode->ghostlink_index_within_layer;
	child_lnode->parent = parent_lnode;

	// Make sure that the child is below its new parent node.
	if (child_lnode->layer <= child_lnode->parent->layer) {
	  child_lnode->layer = child_lnode->parent->layer + 1;
	  shifted = TRUE;
	}
      } else {
	// The child doesn't depend off the shifting node, and it's not
	// going to be a problem. Stop the pass.
      }
    }
  } else if (layer_offset < 0) {
    // The parent node is shrinking back towards the root of the tree.
    //
    // We need to compute the current greatest-layer among those that link
    // to the child node.
    greatest_parent_layer = -1;
    greatestparent = NULL;
    for (ptptr = child->parenttracker->head; ptptr != NULL;
	 ptptr = ptptr->next) {
      nextparent_lnode = (HorizLayoutNode *) ptptr->parent->layoutnode;
      assert(nextparent_lnode != NULL);

      // We set every parent to be a ghost, initially. One of these (the
      // greatest layer-count) will be set to be non-ghost after we terminate
      // the loop.
      nextparent_lnode->childlink_is_ghost = TRUE;

      if (nextparent_lnode->layer > greatest_parent_layer) {
	greatestparent = nextparent_lnode;
	greatest_parent_layer = greatestparent->layer;
      }
    }

    // Attach the child node to the furthest parent down the tree.
    if (child_lnode->layer - greatest_parent_layer > 1) {
      child_lnode->layer = greatest_parent_layer + 1;
      shifted = TRUE;
      greatestparent->childlink_is_ghost = FALSE;
      child_lnode->parent = greatestparent;
    }
  }

  // If the node has been shifted, it will need to have its subgraph
  // shifted as well.
  if (shifted)
    check_children_for_shift(child, layer_offset);
}

// Check the children of a node to see if they need to be shifted.
void HorizLayout::check_children_for_shift(objectnode *graph, int layer_offset)
{
  HorizLayoutNode *this_lnode = (HorizLayoutNode *) graph->layoutnode;
  objectnode *objptr;

  switch(graph->type()) {
  case OBJECTNODE_GROUP:
    // Check every one of the nodes in the link-chain that dangles off
    // the group pointer.
    for (objptr = ((groupobjectnode *) graph)->group;
	     objptr != NULL; objptr = objptr->linkptr)
      check_node_shift(this_lnode, objptr, layer_offset);

    break;
  case OBJECTNODE_REPEAT:
    // Check the single child.
    check_node_shift(this_lnode, ((repeatobjectnode *) graph)->repeat,
		     layer_offset);
    break;
  case OBJECTNODE_PRIM:
    // There are no child nodes below this one. Nothing to do.
    break;
  default:
    assert(0);
  }
}

// Recompute the layout from the root of the tree, in a complete pass.
// This requires that num_ghostlinks, layer and childlink_is_ghost be set
// correctly in all the layoutnodes. It sets the pos and box variables.
void HorizLayout::recompute_layout(void)
{
  HorizLayoutNode *root_lnode;
  int i;

  // First, perform a depth-first search of the CSGgraph, tracking the
  // number of childlink_is_ghost that occur to each layer. Also, calculate
  // the width of each layer.
  for (i = 0; i < LAYOUT_MAX_LAYERS; i++)
    ghostlinks_in_layer[i] = 0;

  count_ghostlinks(modeller->csg_graph->object);
  for (i = 0; i < LAYOUT_MAX_LAYERS; i++)
    layer_width[i] = base_layer_width +
      ghostlink_offset * ghostlinks_in_layer[i];

  // We start the traverse at the root of the CSGgraph.
  root_lnode = (HorizLayoutNode *) modeller->csg_graph->object->layoutnode;
  assert(root_lnode->layer == 0);
  assert(!root_lnode->childlink_is_ghost);

  root_lnode->pos[X] = node_size;
  root_lnode->pos[Y] = node_size;
  layout_children(modeller->csg_graph->object, root_lnode);
}

// Count the ghostlinks to each layer in the CSGgraph.
void HorizLayout::count_ghostlinks(objectnode *node)
{
  HorizLayoutNode *this_lnode, *child_lnode;
  objectnode *child;

  assert(node != NULL);

  // Search horizontally.
  if (node->linkptr != NULL)
    count_ghostlinks(node->linkptr);

  // Search vertically.
  switch(node->type()) {
  case OBJECTNODE_PRIM:
    return;
  case OBJECTNODE_REPEAT:
    child = ((repeatobjectnode *) node)->repeat;
    this_lnode = (HorizLayoutNode *) node->layoutnode;

    // If this is a ghost link, record that there is a ghost link
    // leading to that layer. Give it an index that determines where it
    // links into the child when drawn.
    if (this_lnode->childlink_is_ghost) {
      child_lnode = (HorizLayoutNode *) child->layoutnode;

      assert(child_lnode->layer < LAYOUT_MAX_LAYERS);
      this_lnode->ghostlink_index_within_layer =
	ghostlinks_in_layer[child_lnode->layer]++;
    } else {
      // If not, continue searching down the (non-ghost) vertical pointer.
      assert(child != NULL);
      count_ghostlinks(child);
    }
    break;
  case OBJECTNODE_GROUP:
    child = ((groupobjectnode *) node)->group;
    this_lnode = (HorizLayoutNode *) node->layoutnode;

    // If this is a ghost link, record that there is a ghost link
    // leading to that layer. Give it an index that determines where it
    // links into the child when drawn.
    if (this_lnode->childlink_is_ghost) {
      child_lnode = (HorizLayoutNode *) child->layoutnode;

      assert(child_lnode->layer < LAYOUT_MAX_LAYERS);
      this_lnode->ghostlink_index_within_layer =
	ghostlinks_in_layer[child_lnode->layer]++;
    } else {    
      // If not, continue searching down the (non-ghost) vertical pointer.
      assert(child != NULL);
      count_ghostlinks(child);
    }
    break;
  default:
    assert(0);
  }
}

// Lay out all the children of a node.
void HorizLayout::layout_children(objectnode *parentnode,
				  HorizLayoutNode *parent_lnode)
{
  groupobjectnode *group_node;
  repeatobjectnode *repeat_node;
  objectnode *child, *childptr;
  HorizLayoutNode *child_lnode;
  int next_offset[2], largest_horiz_box;
      
  if (parent_lnode->childlink_is_ghost) {
    // This parent doesn't need any laying out, as it is merely a ghost-type
    // wrapper for its child. It occupies no space (measured from center to
    // center).
    parent_lnode->box[X] = parent_lnode->box[Y] = 0;
  } else {
    assert(parentnode->layoutnode == parent_lnode);

    // We need to traverse the child(ren) of this node and lay them out.
    switch(parentnode->type()) {
    case OBJECTNODE_PRIM:
      // The primitive doesn't need any further laying out, it has no
      // children.
      parent_lnode->box[X] = parent_lnode->box[Y] = 0;
      break;
    case OBJECTNODE_REPEAT:
      repeat_node = (repeatobjectnode *) parentnode;
      child = repeat_node->repeat;
      child_lnode = (HorizLayoutNode *) child->layoutnode;

      // The repeatobjectnode's single child will be offset by a distance
      // according to its layer.
      child_lnode->pos[X] = parent_lnode->pos[X]
	+ layer_width[child_lnode->layer];
      child_lnode->pos[Y] = parent_lnode->pos[Y];

      // Lay out the child node.
      layout_children(child, child_lnode);

      // The parent's bounding box is equal to the child's plus the offset
      // horizontally.
      parent_lnode->box[X] = child_lnode->box[X]
	+ layer_width[child_lnode->layer];
      parent_lnode->box[Y] = child_lnode->box[Y];
      break;
    case OBJECTNODE_GROUP:
      // Each child of the groupobjectnode's link chain will be offset
      // horizontally by a distance according to its layer, and vertically
      // by a distance that corresponds to the bounding box of the one above
      // it. NB: nodes are displayed top to bottom in the OPPOSITE order
      // that the link chain is ordered.
      group_node = (groupobjectnode *) parentnode;
      
      // Find the end of the linkchain, which will be the topmost node.
      for (childptr = group_node->group; childptr->linkptr != NULL;
	   childptr = childptr->linkptr)
	;
      child_lnode = (HorizLayoutNode *) childptr->layoutnode;

      // The groupnode's bounding box is initially just the width of the
      // layer, and the first child position is this offset horizontally.
      next_offset[X] = layer_width[child_lnode->layer];
      next_offset[Y] = 0;
      largest_horiz_box = 0;
      
      // Lay out each child in turn.
      for ( ; childptr != NULL; ) {
	// Calculate the child's position.
	child_lnode->pos[X] = parent_lnode->pos[X] + next_offset[X];
	child_lnode->pos[Y] = parent_lnode->pos[Y] + next_offset[Y];

	// Lay out the child (and its subtree).
	layout_children(childptr, child_lnode);

	// Is this child the widest that we've had so far?
	if (child_lnode->box[X] > largest_horiz_box)
	  largest_horiz_box = child_lnode->box[X];

	// The next child will be placed down by the vertical size of the
	// child's bounding box, plus the vertical link_height.
	next_offset[Y] += child_lnode->box[Y] + link_height;

	childptr = childptr->linked_by;

	// Are we at the end of the chain?
	if (childptr == NULL) {
	  // We are at the end of the chain. The parent's bounding box is
	  // given horizontally by the largest horizontal bounding box
	  // plus the link offset. And vertically, it is equal to the
	  // next offset minus one link_height.
	  assert(child_lnode->layer < LAYOUT_MAX_LAYERS);
	  parent_lnode->box[X] = largest_horiz_box
	    + layer_width[child_lnode->layer];
	  parent_lnode->box[Y] = next_offset[Y] - link_height;
	  break;
	} else {
	  // We need to keep updating child_lnode.
	  child_lnode = (HorizLayoutNode *) childptr->layoutnode;
	}
      }
      break;
    default:
      assert(0);
    }
  }
}

HorizLayout::HorizLayout(Modeller *m)
{
  modeller = m;

  // Set up the layout engine's parameters.
  node_size = 20;
  base_layer_width = 70;
  link_height = 70;
  linktypebox_size = 10;
  ghostlink_offset = 4;  
}

// Draw a node and its subgraph.
void HorizLayout::draw_node(GraphView *gv, objectnode *node)
{
  HorizLayoutNode *lnode = (HorizLayoutNode *) node->layoutnode;
  primobjectnode *prim_node;
  repeatobjectnode *repeat_node;
  groupobjectnode *group_node;
  char tempbuf[10];

  switch(node->type()) {
  case OBJECTNODE_PRIM:
    prim_node = (primobjectnode *) node;

    // Draw the primitive node.
    sprintf(tempbuf, "%c", (prim_node->prim->prim_name())[0]);
    gv->add_node(lnode->pos[X], lnode->pos[Y],
				  node_size, node->is_selected,
				  tempbuf, node->name, node);
    break;
  case OBJECTNODE_REPEAT:
    repeat_node = (repeatobjectnode *) node;
    
    // Draw the repeat node.
    sprintf(tempbuf, "%d", repeat_node->num_copies);
    gv->add_node(lnode->pos[X], lnode->pos[Y],
				  node_size, node->is_selected,
				  tempbuf, node->name, node);

    // Draw the children out of this node.
    draw_children(gv, node, lnode, repeat_node->repeat);
    break;
  case OBJECTNODE_GROUP:
    group_node = (groupobjectnode *) node;
    
    // Draw the groupnode.
    gv->add_node(lnode->pos[X], lnode->pos[Y],
				  node_size, node->is_selected,
				  NULL, node->name, node);

    // Draw the children out of this node.
    draw_children(gv, node, lnode, group_node->group);
    break;
  default:
    assert(0);
  }
}

// Draw the children of a node (a groupnode or repeatnode).
void HorizLayout::draw_children(GraphView *gv, objectnode *node, HorizLayoutNode *lnode,
				objectnode *chain_head)
{
  objectnode *childptr;
  int current_x, current_y, last_y;
  int xoffset_lhs, xoffset_rhs;
  int linkcoords[6], start_pt[2], end_pt[2];
  char tempbuf[2];
  HorizLayoutNode *child_lnode;

  if (!lnode->childlink_is_ghost) {
    // What are the left and right-hand offsets?
    xoffset_lhs = base_layer_width / 2;
    assert(lnode->layer < LAYOUT_MAX_LAYERS - 1);
    xoffset_rhs = layer_width[lnode->layer + 1] - xoffset_lhs;
    
    current_x = lnode->pos[X] + xoffset_lhs;
    current_y = lnode->pos[Y];
    last_y = current_y;

    // We have to lay out the children in order from top to bottom, which is
    // in the reverse order to the chain.

    // Find the end of the linkchain, which will be the topmost node.
    for (childptr = chain_head; childptr->linkptr != NULL;
	 childptr = childptr->linkptr)
      ;

    // Draw each one of the nodes in the chain.
    for ( ; childptr != NULL; ) {
      // If this node is not the topmost (the last in the chain), then draw the
      // link leading down to it.
      if (childptr->linkptr != NULL) {
	linkcoords[0] = current_x;
	linkcoords[1] = last_y;
	linkcoords[2] = current_x;
	linkcoords[3] = current_y;
	gv->add_link(2, linkcoords, FALSE, FALSE);

	// Draw the node that represents the linktype.
	switch(childptr->link_type) {
	case PLUS:
	  strcpy(tempbuf, "+");
	  break;
	case MINUS:
	  strcpy(tempbuf, "-");
	  break;
	case AND:
	  strcpy(tempbuf, "&");
	  break;
	default:
	  assert(0);
	}
	gv->add_node(current_x, (current_y + last_y) / 2,
				      linktypebox_size, FALSE, tempbuf, NULL,
				      NULL);
      }

      // Draw the link that points to the child node. This starts at current_x
      // for all nodes except the first.
      if (childptr->linkptr == NULL)
	linkcoords[0] = current_x + (node_size/2) - xoffset_lhs;
      else
	linkcoords[0] = current_x;
      linkcoords[1] = current_y;
      linkcoords[2] = current_x - (node_size/2) + xoffset_rhs;
      linkcoords[3] = current_y;
      gv->add_link(2, linkcoords, FALSE, TRUE);

      // Draw the child node.
      draw_node(gv, childptr);

      // Set up for the next node.
      child_lnode = (HorizLayoutNode *) childptr->layoutnode;
      last_y = current_y;
      current_y += child_lnode->box[Y] + link_height;
      childptr = childptr->linked_by;
    }
  } else {
    assert(chain_head != NULL);
    child_lnode = (HorizLayoutNode *) chain_head->layoutnode;

    // Where does the ghost link start and end?
    start_pt[X] = lnode->pos[X] + (node_size/2);
    start_pt[Y] = lnode->pos[Y];
    end_pt[X] = child_lnode->pos[X] - (node_size/2) - (base_layer_width/2)
      - lnode->ghostlink_index_within_layer * ghostlink_offset;
    end_pt[Y] = child_lnode->pos[Y];

    // Draw the ghost link to the child node.
    linkcoords[0] = start_pt[X];
    linkcoords[1] = start_pt[Y];
    linkcoords[2] = end_pt[X];
    linkcoords[3] = start_pt[Y];
    linkcoords[4] = end_pt[X];
    linkcoords[5] = end_pt[Y];
    gv->add_link(3, linkcoords, TRUE, FALSE);
  }
}

// Update the layout in response to a add-groupnode-under operation.
void HorizLayout::update_addgroupnodeunder(groupobjectnode *add_under)
{
  HorizLayoutNode *addunder_lnode, *new_lnode, *child_lnode;
  groupobjectnode *new_node;

  assert(add_under->group->type() == OBJECTNODE_GROUP);
  new_node = (groupobjectnode *) add_under->group;

  addunder_lnode = (HorizLayoutNode *) add_under->layoutnode;

  // Create a new HorizLayoutNode for the new node.
  new_lnode = new HorizLayoutNode(new_node);
  new_node->layoutnode = new_lnode;

  new_lnode->parent = addunder_lnode;
  new_lnode->layer = new_lnode->parent->layer + 1;
  new_lnode->num_ghostlinks = 0;

  // Set up the lnode's child link.
  new_lnode->childlink_is_ghost = addunder_lnode->childlink_is_ghost;
  addunder_lnode->childlink_is_ghost = FALSE;

  // Get the layoutnode of add_under's new groupnode's child.
  child_lnode = (HorizLayoutNode *) new_node->group->layoutnode;
  
  // If this child layoutnode's parent was addunder's layoutnode, it is now
  // new_node's layoutnode, and needs to be moved down.
  if (child_lnode->parent == addunder_lnode) {
    child_lnode->parent = new_lnode;
    child_lnode->layer++;
  }

  // See if the remainder of the graph needs to be moved down.
  check_node_shift(new_lnode, new_node->group, +1);
}

// Switch a layoutnode around in response to a chain-head switch.
void HorizLayout::update_chainheadswitch(groupobjectnode *parentnode,
					 objectnode *prev_head,
					 objectnode *new_head)
{
  HorizLayoutNode *parent_lnode, *prev_lnode, *new_lnode;

  parent_lnode = (HorizLayoutNode *) parentnode->layoutnode;
  prev_lnode   = (HorizLayoutNode *) prev_head->layoutnode;
  new_lnode    = (HorizLayoutNode *) new_head->layoutnode;

  assert(prev_lnode->parent == parent_lnode);
  assert(prev_lnode->num_ghostlinks == 0);

  new_lnode->parent = parent_lnode;
  new_lnode->num_ghostlinks = 0;
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
