/*
 * layout.h: Definitions for GraphLayout and associated classes.
 *
 * Contains code to create and maintain the graph-view of the CSGgraph, and
 * lay it out in a human-readable format.
 */

// LayoutNode is subclassed by a type of layout node for each different
// layout engine. A layout node contains information regarding the position
// of a particular objectnode in the graph view.
class LayoutNode {
public:
  objectnode *owner;
};

// GraphLayout is subclassed by each layout engine, and provides an entrance
// point only.
class GraphLayout {
public:
  // Update the layout in response to a new link-in-copy operation.
  virtual void update_linkincopy(groupobjectnode *linkchain_parent,
				 groupobjectnode *newcopy) = 0;

  // Update the layout in response to a new link-in-primitive operation.
  virtual void update_linkinprim(groupobjectnode *linkchain_parent,
				 primobjectnode *newprim) = 0;

  // Update the layout in response to a new wrapper-around operation.
  virtual void update_wrapper(groupobjectnode *wrapper) = 0;

  // Update the layout in response to a new repeat-around operation.
  virtual void update_repeat(repeatobjectnode *repeat) = 0;

  // Update the layout in response to a add-groupnode-under operation.
  virtual void update_addgroupnodeunder(groupobjectnode *add_under) = 0;

  // Switch a layoutnode around in response to a chain-head switch.
  virtual void update_chainheadswitch(groupobjectnode *parentnode,
				      objectnode *prev_head,
				      objectnode *new_head) = 0;

  // Perform a complete layout operation, from scratch.
  virtual void complete_layout(void) = 0;

  // Draw the graph on a particular view.
  virtual void draw_on_view(GraphView *v) = 0;
};


/*
 * Definitions for the horizontal layout-engine.
 */

class HorizLayoutNode : public LayoutNode {
public:
  // The position of the node.
  int pos[2];

  // The bounding-box around the node's subgraph tree. Note that this is
  // the distance-between-centers, so if there are no children (or only
  // ghost links) then the bounding box will be zero.
  int box[2];

  // The layer of the node. The root of the tree is layer 0.
  int layer;

  // A pointer to the layoutnode of the node's immediate parent (the one
  // that determines its layer).
  HorizLayoutNode *parent;

  // The number of ghost links coming into the node.
  int num_ghostlinks;

  // For a nodetype that has children (group, or repeat)... is the child-link
  // a ghost?
  int childlink_is_ghost;
  int ghostlink_index_within_layer;

  // Create a new node.
  HorizLayoutNode(objectnode *owner) {
    this->owner = owner;
  }
};

// The maximum number of layers.
const int LAYOUT_MAX_LAYERS = 20;

class HorizLayout : public GraphLayout {
public:
  // Parameters for the layout engine.
  int node_size, base_layer_width, link_height, linktypebox_size;
  int ghostlink_offset;

  // The Modeller that owns this layout engine.
  Modeller *modeller;

  // The number of ghost links associated with each layer of the graph,
  // and the resulting layer width.
  int ghostlinks_in_layer[LAYOUT_MAX_LAYERS];
  int layer_width[LAYOUT_MAX_LAYERS];

  /*
   * Entrance points from GraphLayout.
   */

  // Update the layout in response to a new link-in-copy operation.
  virtual void update_linkincopy(groupobjectnode *linkchain_parent,
				 groupobjectnode *newcopy);

  // Update the layout in response to a new link-in-primitive operation.
  virtual void update_linkinprim(groupobjectnode *linkchain_parent,
				 primobjectnode *newprim);

  // Update the layout in response to a new wrapper-around operation.
  virtual void update_wrapper(groupobjectnode *wrapper);

  // Update the layout in response to a new repeat-around operation.
  virtual void update_repeat(repeatobjectnode *repeat);
  
  // Update the layout in response to a add-groupnode-under operation.
  virtual void update_addgroupnodeunder(groupobjectnode *add_under);

  // Switch a layoutnode around in response to a chain-head switch.
  virtual void update_chainheadswitch(groupobjectnode *parentnode,
				      objectnode *prev_head,
				      objectnode *new_head);

  // Perform a complete layout operation, from scratch.
  virtual void complete_layout(void);

  // Draw the graph on a particular view.
  virtual void draw_on_view(GraphView *v);

  /*
   * End entrance points.
   */

  // Check a node to see if it needs to be shifted.
  void check_node_shift(HorizLayoutNode *parent_lnode,
			objectnode *child, int layer_offset);

  // Check the children of a node to see if they need to be shifted.
  void check_children_for_shift(objectnode *graph, int layer_offset);

  // Recompute the layout from the root of the tree, in a complete pass.
  // This requires that num_ghostlinks, layer and childlink_is_ghost be set
  // correctly in all the layoutnodes. It sets the pos and box variables.
  void recompute_layout(void);

  // Lay out all the children of a node.
  void layout_children(objectnode *parentnode, HorizLayoutNode *parent_lnode);

  // Count the ghostlinks to each layer in the CSGgraph.
  void count_ghostlinks(objectnode *node);

  // Draw a node and its subgraph.
  void draw_node(GraphView *gv, objectnode *node);

  // Draw the children of a node (a groupnode or repeatnode).
  void draw_children(GraphView *gv, objectnode *node, HorizLayoutNode *lnode,
		                 objectnode *chain_head);
 
  HorizLayout(Modeller *m);
};
