/* Collection-safe */

/*
 * memory.h: Definitions for templated class MemoryManager.
 *
 * Contains code to construct a MemoryManager for some class M. The Memory
 * Manager will then set up a Mirage-style blocklist that may be used to
 * allocate and deallocate objects of type M.
 */

// The default values for max_memblocks and blocksize that
// are used if no values are specified in the constructor.
const int memorymanager_default_blocksize = 1000;
const int memorymanager_default_maxblocks = 1000;

template <class M>
class MemoryManager {
private:
  // Does this memory manager return indices?
  int returns_indices;

  // Print the percentage of memory usage on deallocation?
  int print_usage;

  int max_memblocks, blocksize;
  int blocks_used, items_used, items_freed;
  M *next_free;
  unsigned int next_free_index;
  M **storage;

  // A struct freelist is used to store which is the next free list.
  struct freelist {
    M *next_free;
    unsigned int next_free_index;
  };

public:
  // Constructor: specified blocksize and maxblocks.
  MemoryManager(int returns_indices,
		int blocksize = memorymanager_default_blocksize,
		int max_memblocks = memorymanager_default_maxblocks,
		int print_usage = FALSE)
  {
    this->returns_indices = returns_indices;
    this->blocksize = blocksize;
    this->max_memblocks = max_memblocks;
    this->print_usage = print_usage;

    // The object that we're being created for must be at least a (M *) + (int)
    // in length, as we will want to store a (M *) and an
    // (unsigned int) in it for freelist purposes.
    assert(sizeof(M) >= sizeof(M *) + sizeof(unsigned int));

    // Allocate the storage pointers.
    storage = (M **) calloc(max_memblocks, sizeof(M *));
    assert(storage != NULL);
    for (int i = 0; i < max_memblocks; i++)
      storage[i] = NULL;

    // Set up memory management values.
    blocks_used = 0;
    items_used = 0;
    items_freed = 0;
    next_free = NULL;
    next_free_index = 0;
  }

  // Destructor; deallocate all memory and exit.
  ~MemoryManager(void)
  {
    if (print_usage) {
      // Print out blocks used, bytes allocated, number of items stored in this
      // space. Also print out percentage of memory allocated that was used.
      printf("MemoryManager: %3d blocks, %9d bytes, %7d items (%02d%%)\n",
	     blocks_used, blocks_used * blocksize * sizeof(M), items_used,
	     (blocks_used == 0) ? 100 :
	     (100 * items_used) / (blocks_used * blocksize));
    }
    
    // Deallocate each block.
    for (int i = 0; i < blocks_used; i++) {
      assert(storage[i] != NULL);
      free(storage[i]);
    }

    // Deallocate the storage memory.
    free(storage);
  }

  // Allocate a new M from the storage, and return a pointer to it.
  M *allocate(void)
  {
    return allocate(NULL);
  }

  // Allocate a new M from the storage and return a pointer to it, and also
  // the M's index in the memory.
  M *allocate(unsigned int *index_return)
  {
    M *the_block, *the_object;
    int block_no, index_no;
    unsigned int obj_index;
    struct freelist *freelist_ptr;
    
    // Check to see if we need to allocate a new block.
    if (next_free == NULL) {

      //      printf("max_memblocks %d\n", max_memblocks);

      if (blocks_used >= max_memblocks)
	fatal_error("MemoryManager::allocate", "maximum memory size exceeded",
		    "Each memory manager has a fixed limit on the number of\n"
		    "blocks that can be allocated. Increase this, or reduce\n"
		    "memory usage.");

      the_block = (M *) calloc(blocksize, sizeof(M));

      if (the_block == NULL)
	fatal_error("MemoryManager::allocate", "can't allocate new block!",
		    "Tried to dynamically expand memory and failed. This\n"
		    "could indicate that you are running severely low on\n"
		    "memory or that there is a memory leak.");
      
      // Update the array of blocks.
      block_no = blocks_used++;
      storage[block_no] = the_block;

      // Now for each unused element in the block (which is all of them)
      // we set it to point to the next free element. We start from the
      // zero element, each one pointing to the next. The last one in the block
      // points to NULL.
      //
      // This pointer is stored in a freelist struct at the start of the
      // element as there is no data being stored in it right now anyway.

      // Allocate one object for immediate use.
      the_object = &the_block[0];
      obj_index = blocksize * block_no;

      // The next one is the next_free object.
      next_free = &the_block[1];
      next_free_index = blocksize * block_no + 1;

      // Tricky pointer cast... get a pointer to the current element
      // and then cast this to struct freelist. Then we can store a freelist
      // in the storage which would ordinarily be used for an M... but since
      // it's a free object the storage can be used for whatever.
      for (index_no = 1; index_no < blocksize - 1; index_no++) {
	freelist_ptr = (struct freelist *) &the_block[index_no];

	freelist_ptr->next_free = &the_block[index_no + 1];
	freelist_ptr->next_free_index = blocksize * block_no + index_no + 1;
      }

      // The last element is free, but there's no free memory after it.
      freelist_ptr = (struct freelist *) &the_block[blocksize - 1];
      freelist_ptr->next_free = NULL;
      freelist_ptr->next_free_index = 0;

    } else {
      // Just use the next free object.
      the_object = next_free;
      obj_index = next_free_index;

      // The next element in the free list is stored in the space that
      // the currently free node occupies.
      freelist_ptr = (struct freelist *) the_object;
      next_free = freelist_ptr->next_free;
      next_free_index = freelist_ptr->next_free_index;

#if _DEBUG
      if (next_free != NULL) {
        // Check that the free list pointer is valid.
        for (int i = 0; i < blocks_used; i++) {
          if ((next_free >= &storage[i][0]) &&
              (next_free < &storage[i][blocksize]))
            break;
        }

        assert(i < blocks_used);
      }
#endif
    }

    items_used++;

    // Return the node that we allocated.
    if (index_return != NULL) {
      assert(returns_indices);
      *index_return = obj_index;
      //      printf("Returned index %d\n", obj_index);
    }
    return the_object;
  }

  // Deallocate a previously allocated object. This is the version
  // for a memory manager that doesn't return indices.
  void deallocate(M *the_object)
  {
    struct freelist *freelist_ptr;
    
    freelist_ptr = (struct freelist *) the_object;

    assert(!returns_indices);

#ifndef NDEBUG
    int i;
    
    // Check that the object is part of our memory, and that
    // the object isn't already free.
    for (i = 0; i < blocks_used; i++) {
      assert(storage[i] != NULL);

      // Was the object allocated from this block?
      if ((the_object >= &storage[i][0]) &&
	  (the_object <  &storage[i][blocksize])) {
	// This is a pointer to inside block i.
	break;
      }
    }
    
    // The object should have satisfied the break condition above,
    // otherwise it isn't part of our memory.
    assert(i < blocks_used);
#endif

    // Free the element.
    freelist_ptr->next_free = next_free;
    
    // This object will be next to be allocated.
    next_free = the_object;

    items_used--;
  }

  // Deallocate a previously allocated object. This is the version
  // for a memory manager that does return indices.
  void deallocate(M *the_object, unsigned int obj_index)
  {
    struct freelist *freelist_ptr;
    
    freelist_ptr = (struct freelist *) the_object;

    assert(returns_indices);

#ifndef NDEBUG
    int i;
    
    // Check that the object is part of our memory, and that
    // the object isn't already free.
    for (i = 0; i < blocks_used; i++) {
      assert(storage[i] != NULL);

      // Was the object allocated from this block?
      if ((the_object >= &storage[i][0]) &&
	  (the_object <  &storage[i][blocksize])) {
	// This is a pointer to inside block i.
	break;
      }
    }
    
    // The object should have satisfied the break condition above,
    // otherwise it isn't part of our memory.
    assert(i < blocks_used);
#endif

    // Free the element.
    freelist_ptr->next_free = next_free;
    freelist_ptr->next_free_index = next_free_index;
    
    // This object will be next to be allocated.
    next_free = the_object;
    next_free_index = obj_index;

    items_used--;
  }

  // Retrieve an element given its index.
  M *retrieve_element(unsigned int obj_index)
  {
    assert(returns_indices);
    //    printf("retrieve %d (%d / %d)\n", obj_index, obj_index / blocksize,
    //	   obj_index % blocksize);
    assert(obj_index < (unsigned int) (blocks_used * blocksize));
    assert(storage != NULL);
    assert(storage[obj_index / blocksize] != NULL);

    return &storage[obj_index / blocksize][obj_index % blocksize];
  }
};
