/*
 * modeller.cc: Routines for class Modeller.
 *
 * The Modeller class contains the current state of the CSG modeller,
 * including the object being modelled and any views of the object.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

Modeller::Modeller(void)
{
  int array_side;

  // Set the bounding box up.
#if 1
  Vector::assign(minbound, -2.01, -2.02, -3.03);
  Vector::assign(maxbound, +2.001, +2.002, +5.003);
#else
  Vector::assign(minbound, -1.51, -1.52, -2.53);
  Vector::assign(maxbound, +1.501, +1.502, +2.503);
#endif
  // Set detail parameters.
  octree_level = 3;
  terminal_res = 4;
  use_gouraud = TRUE;
  draw_inout_pts = FALSE;

  // Construct the octree's terminal corner positions.
  array_side = 1 << octree_level;
  terminalcorner = (vector *) malloc((CUBE(array_side) + 1) * sizeof(vector));
  Vector::sub(maxbound, minbound, terminalsize);
  Vector::mult(1.0 / array_side, terminalsize, terminalsize);

  calculate_terminal_corners(0, 0, 0, 0, 1 << octree_level);

  // Set up the lighting.
  ambient_light = 0.3;
  num_lights = 2;
  light_dir = (vector *) calloc(num_lights, sizeof(vector));
  if (!light_dir)
    fatal_error("Modeller::Modeller", "Can't allocate light_dir",
                "Out of memory.");
  light_strength = (double *) calloc(num_lights, sizeof(double));
  if (!light_strength)
    fatal_error("Modeller::Modeller", "Can't allocate light_strength",
                "Out of memory.");

  light_strength[0] = 0.6;
  Vector::assign(light_dir[0], +1.3, -0.5, +1);
  Vector::normalise(light_dir[0], "Modeller::Modeller, light_dir");

  light_strength[1] = 0.4;
  Vector::assign(light_dir[1], -1, +1, -1);
  Vector::normalise(light_dir[1], "Modeller::Modeller, light_dir");

  primsplitinfo_counter = 0;

  /*
   * Debugging stuff
   */
  debug_line = FALSE;
  debug_all_splits = FALSE;
#if 0
  terminal_highlight = FALSE;
  debug_terminal_count = 0;
  terminal_highlight = FALSE;
  highlighted_count = 0;
  show_only_highlighted = FALSE;
  highlight_terminal = NULL;
#endif

  // Initialise the triangle and string memory.
  trimem = new TriMemory;
  stringmem = new StringMemory(this);
  worldstringmem = new WorldStringMemory(this);

  // Create a TriSplitter.
  trisplitter = new TriSplitter(this);

  // Create a blank CSGgraph and Primtree.
  csg_graph = new CSGgraph(this);
  primtree = new Primtree(this);

  // Create the GraphLayout (will eventually be bound to a window).
  graphlayout = new HorizLayout(this);

  // Set up the modeller's starting state.
  init_objectgraphs();

  focused_octree = primtree->root->octree;

  // There are no modelviews bound to this Modeller yet.
  num_modelviews = 0;

  // There are no graphviews bound to this Modeller yet.
  num_graphviews = 0;

  // There are no transviews bound to this Modeller yet.
  num_transviews = 0;

  // Nothing has been selected.
  last_click = NULL;
  selected_node = NULL;
  skitterlist = new SkitterList(this);
}

// Destructor.
Modeller::~Modeller(void)
{
  int i;

  free(terminalcorner);
  delete primtree;
  delete csg_graph;
  delete trimem;
  delete stringmem;
  delete worldstringmem;
  delete trisplitter;
  delete graphlayout;
  delete skitterlist;

  for (i = 0; i < num_modelviews; i++)
    delete modelview[i];

  for (i = 0; i < num_graphviews; i++)
    delete graphview[i];

  for (i = 0; i < num_transviews; i++)
    delete transview[i];

  delete tailend_nothingprim;

  if (light_dir != NULL)
    free(light_dir);
  if (light_strength != NULL)
    free(light_strength);

  assert(globalhackvar_solidoctreecount == 0);
}

// Calculate the positions of the octree's terminal corners.
void Modeller::calculate_terminal_corners(int terminal_id, int x, int y, int z, int size)
{
  if (size == 1) {
    // Calculate the bottom-level terminal corner.
    terminalcorner[terminal_id][X] = minbound[X] + x * terminalsize[X];
    terminalcorner[terminal_id][Y] = minbound[Y] + y * terminalsize[Y];
    terminalcorner[terminal_id][Z] = minbound[Z] + z * terminalsize[Z];
//    TRACE("corner id %d is at %d %d %d = (%f %f %f)\n", terminal_id, x, y, z,
//          terminalcorner[terminal_id][X], terminalcorner[terminal_id][Y], terminalcorner[terminal_id][Z]);
  } else {
    // Continue recursing down the octree.
    size /= 2;
    for (int i = 0; i < 8; i++)
      calculate_terminal_corners(8 * terminal_id + i,
                                 x + ((i & 0x1) ? size : 0),
                                 y + ((i & 0x2) ? size : 0),
                                 z + ((i & 0x4) ? size : 0), size);                                 
  }
}

// Get the bounds of the object being modelled.
void Modeller::get_bounds(box b)
{
  Vector::assign(b[0], minbound[X], minbound[Y], minbound[Z]);
  Vector::assign(b[1], maxbound[X], minbound[Y], minbound[Z]);
  Vector::assign(b[2], minbound[X], maxbound[Y], minbound[Z]);
  Vector::assign(b[3], maxbound[X], maxbound[Y], minbound[Z]);
  Vector::assign(b[4], minbound[X], minbound[Y], maxbound[Z]);
  Vector::assign(b[5], maxbound[X], minbound[Y], maxbound[Z]);
  Vector::assign(b[6], minbound[X], maxbound[Y], maxbound[Z]);
  Vector::assign(b[7], maxbound[X], maxbound[Y], maxbound[Z]);
}

// Set up the CSGgraph and Primtree to their initial state.
void Modeller::init_objectgraphs(void)
{
  // Create a nothingPrimitive for use in the tail-ends of
  // repeatobjectnode primitive graphs.
  tailend_nothingprim = new nothingPrimitive;

#if 1
  // Set up the objectgraphs to contain a single primitive.
  create_placeholder_object(new planePrimitive);

  graphlayout->complete_layout();

  make_piston();
#else
  // Set up the objectgraphs to contain a single primitive.
  create_placeholder_object(new spherePrimitive);

  graphlayout->complete_layout();


#if 1
  // Set up an object to look at.
  ObjectTransformation *trans_array[10];
  primobjectnode *newsphere_node;
  objectnode *chainhead;

  trans_array[0] = new ObjectScale(0.4);
  trans_array[1] = new ObjectTranslation(0.9, 0.0, 0.0);
  csg_graph->link_in_primitive(csg_graph->object,
			       MINUS, new spherePrimitive,
			       &newsphere_node, 2, trans_array);
#endif
#if 0
  chainhead = csg_graph->object;
  for (int i = 1; i <= 4; i++) {
    trans_array[0] = new ObjectRotation(X, Y, PI/6 * i);
    chainhead = csg_graph->link_in_copy(chainhead, MINUS, newsphere_node, NULL, 1, trans_array);
  }
#endif
#if 0
  trans_array[0] = new ObjectRotation(X, Y, 0.01 * PI/2);
  csg_graph->add_repeat_around(newsphere_node, 4, 1, trans_array);
#endif
#endif
}

// Bind a View to the modeller.
void Modeller::bind_modelview(View *v)
{
  if (num_modelviews >= MODELLER_MAX_VIEWS)
    fatal_error("Modeller::bind_modelview", "MODELLER_MAX_VIEWS exceeded!",
		"Change static limit and recompile if you want more views.");

  modelview[num_modelviews++] = v;
}

// Unbind a View from the modeller.
void Modeller::unbind_modelview(View *v)
{
  int i;

  // Find this view's entry in the modelview list.
  for (i = 0; i < num_modelviews; i++)
    if (modelview[i] == v)
      break;

  // We cannot unbind a view that isn't already in the view list.
  assert(i < num_modelviews);

  // Copy the last view into this storage location (which is now unused).
  modelview[i] = modelview[--num_modelviews];
}

// Bind a GraphView to the modeller.
void Modeller::bind_graphview(GraphView *v)
{
  if (num_graphviews >= MODELLER_MAX_VIEWS)
    fatal_error("Modeller::bind_graphview", "MODELLER_MAX_VIEWS exceeded!",
		"Change static limit and recompile if you want more views.");

  graphview[num_graphviews++] = v;

  graphlayout->draw_on_view(v);
}

// Unbind a GraphView from the modeller.
void Modeller::unbind_graphview(GraphView *v)
{
  int i;

  // Find this view's entry in the view list.
  for (i = 0; i < num_graphviews; i++)
    if (graphview[i] == v)
      break;

  // We cannot unbind a graphview that isn't already in the graphview list.
  assert(i < num_graphviews);

  // Copy the last view into this storage location (which is now unused).
  graphview[i] = graphview[--num_graphviews];
}

// Bind a TransView to the modeller.
void Modeller::bind_transview(TransView *v)
{
  if (num_transviews >= MODELLER_MAX_VIEWS)
    fatal_error("Modeller::bind_transview", "MODELLER_MAX_VIEWS exceeded!",
		"Change static limit and recompile if you want more views.");

  transview[num_transviews++] = v;
}

// Unbind a TransView from the modeller.
void Modeller::unbind_transview(TransView *v)
{
  int i;

  // Find this view's entry in the view list.
  for (i = 0; i < num_transviews; i++)
    if (transview[i] == v)
      break;

  // We cannot unbind a transview that isn't already in the transview list.
  assert(i < num_transviews);

  // Copy the last view into this storage location (which is now unused).
  transview[i] = transview[--num_transviews];
}

// Select the object along a ray cast into the scene.
void Modeller::select_along_ray(vector ray_from, vector ray_dir)
{
  primtreeprimnode *hit_prim;
  double hit_t;
  vector hit_norm;

  // Cast the ray against the world's octree.
  hit_prim = primtree->root->octree->trace_ray(ray_from, ray_dir,
					       &hit_t, hit_norm);
  
  if (hit_prim == NULL) {
    selected_node = NULL;
    last_click = NULL;
    csg_graph->deselect();
  } else {
    if (last_click == hit_prim) {
      // Advance the selection one step up the primtree. Skip over
      // CSG nodes, stop at PRIM or WRAPPER nodes (which correspond to a
      // node somewhere in the CSG graph).
      assert(selected_node != NULL);
      selected_node = selected_node->parent;

      while ((selected_node != NULL) &&
	     (selected_node->type() == PRIMTREENODE_CSG))
        selected_node = selected_node->parent;
      
      if (selected_node == NULL) {
	// There isn't anywhere further up the tree to go to. So, deselect.
	last_click = NULL;
	csg_graph->deselect();
      } else {
	csg_graph->select_element(selected_node->original);
      }
    } else {
      selected_node = hit_prim;
      last_click = hit_prim;
      csg_graph->select_element(selected_node->original);
    }
  }

  update_all_views();
}

#if 0
// Cycle the selected terminal along the debugging ray.
void Modeller::cycle_debugging_terminal(void)
{
  int i, bi, intersect_count;
  terminaltrackernode *pptr;
  transform ray_transform;
  box rayspace_box;
  cylinderPrimitive *cylprim = new cylinderPrimitive;

  // Form a matrix which transforms a point from worldspace to ray-centric
  // space (ray_from = 0, ray_dir along +Z).
  Transform::unit(ray_transform);
  Transform::stretch(ray_transform, 0.00001, 0.00001, 1.0);
  Transform::rotate(ray_transform, Z, X, 90);
  Transform::rotate(ray_transform, debug_linedir, plus_Z);
  Transform::shift(ray_transform,
		   debug_linefrom[X], debug_linefrom[Y], debug_linefrom[Z]);

  // Look at the ray; which terminals does it intersect?
  for (i = 0, intersect_count = 0, pptr = terminaltracker->head;
       pptr != NULL; pptr = pptr->next, i++) {
    for (bi = 0; bi < 8; bi++)
      Matrix::apply(ray_transform[AFT], pptr->terminal_extent[bi],
		    rayspace_box[bi]);

    if (cylprim->voxel_intersect(rayspace_box) != OUT) {
      intersect_count++;
      if (intersect_count > debug_terminal_count) {
	// Highlight this terminal, and stop.
	terminal_highlight = TRUE;
	highlighted_count = i;
	highlight_terminal = pptr;
	break;
      }
    }
  }
  if (intersect_count <= debug_terminal_count) {
    // We didn't find a terminal to highlight.
    terminal_highlight = FALSE;
    debug_terminal_count = 0;
  } else {
    // Found a terminal to highlight.
    debug_terminal_count++;
  }

  delete cylprim;

  update_all_views();
}

// Toggle the terminal-only highlighting.
void Modeller::toggle_debugging_highlight(void)
{
  if (terminal_highlight) {
    if (show_only_highlighted) {
      show_only_highlighted = FALSE;
    } else {
      printf("=================\n");
      printf("Selected: Terminal %d\n", highlighted_count);
      for (int bi = 0; bi < 8; bi++)
	printf("%s (%d) %f %f %f\n", (bi == 0) ? "Extent:" : "       ", bi,
	       highlight_terminal->terminal_extent[bi][X],
	       highlight_terminal->terminal_extent[bi][Y],
	       highlight_terminal->terminal_extent[bi][Z]);
      printf("*** Print Terminal...\n");
      highlight_terminal->solidterm->terminal->print();
      
      printf("\n\n*** Polygonize Terminal / Print...\n");
      highlight_terminal->solidterm->terminal->
	polygonize_terminal(highlight_terminal->terminal_extent, TRUE);
      
      show_only_highlighted = TRUE;
    }
  }
  update_all_views();
}
#endif

// Set gouraud-shading.
void Modeller::set_gouraud(int new_value)
{
  if (new_value != use_gouraud) {
    use_gouraud = new_value;
    update_all_views();
  }
}

// Set inout-point drawing.
void Modeller::set_inout_pts(int new_value)
{
  draw_inout_pts = new_value;
  update_all_views();
}

// Update all Views that are bound to this Modeller.
void Modeller::update_all_views(void)
{
  for (int i = 0; i < num_modelviews; i++)
    modelview[i]->update();
}

// Update all TransViews that are bound to this Modeller.
void Modeller::update_all_transviews(void)
{
  for (int i = 0; i < num_transviews; i++)
    transview[i]->update();
}

// Debugging routine.
void Modeller::debugcall(char *mode, char *args)
{
  if (strcmp(mode, "version") == 0) {
    printf("Chris's Modelling application, version 0.0.1a01\n");
  } else if (strcmp(mode, "octree") == 0) {
    if (primtree->root)
      primtree->root->octree->print(0);
    else
      printf("Primtree is empty.\n");
  } else if (strcmp(mode, "printgraphs") == 0) {
    csg_graph->print_csggraph();
    primtree->print_primtree();
  } else {
    printf("Unrecognized debugging mode \"%s\".\n", mode);
  }
}

// Create a placeholder object (used if the entire objectgraph is deleted).
void Modeller::create_placeholder_object(Primitive *placeholder_prim)
{
  HorizLayoutNode *lnode;

  assert(csg_graph->object == NULL);
  assert(primtree->root == NULL);

  // Initialise the CSGgraph.
  csg_graph->object = new primobjectnode(this, placeholder_prim);

  // Initialise the Primtree.
  primtree->root = new primtreeprimnode(this,
					((primobjectnode *) csg_graph->object)
					->prim);
  primtree->root->parent = NULL;
  primtree->root->original = csg_graph->object;
  
  // Create the first CopyTracker.
  csg_graph->object->copytracker = new CopyTracker(csg_graph->object);
  primtree->root->this_tracker = csg_graph->object->copytracker->
    add_new_tracker(primtree->root);
  
  // Set up the layout view.
  lnode = new HorizLayoutNode(csg_graph->object);
  lnode->layer = 0;
  lnode->num_ghostlinks = 0;
  lnode->childlink_is_ghost = FALSE;
  csg_graph->object->layoutnode = lnode;

  // Set up the initial SolidOctree.
  primtree->root->octree = new SolidOctree((primtreeprimnode *) primtree->root,
					   this, PLUS);
}

// Select a node.
void Modeller::do_select(objectnode *node)
{
  if (node == NULL)
    csg_graph->deselect();
  else
    csg_graph->select_element(node);

  // Update the views.
  update_all_views();
}

// Delete the selected node..
void Modeller::do_delete(void)
{
  objectnode *delete_obj = csg_graph->selected_object;

  if (delete_obj != NULL) {
    // Deselect this object.
    csg_graph->deselect();

    // Delete the object.
    csg_graph->delete_object(delete_obj);

    // Update the 3D view.
    update_all_views();
  }
}

// Set the state of the graph views.
void Modeller::setgraphviewstate(graphviewstate state)
{
  int i;

  for (i = 0; i < num_graphviews; i++)
    graphview[i]->viewstate = state;
}

// Called by the graphview classes.
void Modeller::do_link_copy(objectnode *copynode, linktype link_type)
{
  objectnode *link_to;
  ObjectTransformation *obj_trans;

  // Place a link to the object in question.
  link_to = csg_graph->selected_object;
  csg_graph->deselect();
  obj_trans = new ObjectTranslation(0.01, 0.01, 0.01);

  // Make the link.
  link_to = csg_graph->link_in_copy(link_to, link_type,
  		                             copynode, NULL,
                                   1, &obj_trans);
  csg_graph->select_element(link_to);

  // Update the polygonization to change the 3D view.
  update_all_views();
}

// Called by the graphview classes.
void Modeller::do_link_prim(Primitive *new_prim, linktype link_type)
{
  objectnode *link_to;
  ObjectTransformation *obj_trans;

  // Place a link to the object in question.
  link_to = csg_graph->selected_object;
  csg_graph->deselect();
  obj_trans = new ObjectTranslation(0.01, 0.01, 0.01);

  // Make the link.
  link_to = csg_graph->link_in_primitive(link_to, link_type,
  		                                   new_prim, NULL,
                                         1, &obj_trans);
  csg_graph->select_element(link_to);

  // Update the polygonization to change the 3D view.
  update_all_views();
}

extern int debug_split_breakid;
extern int debug_octree_print;

int debug_incidentstatus = -1;

primobjectnode *manipulatenode = NULL, *downcutnode = NULL;

// Debugging hack.
void Modeller::hack(int hackval)
{
  int foobar = 0, foobar2 = 0, foobar3 = 0, foobar99 = 0;

  _CrtDbgBreak();

  if (foobar2) {
    update_all_views();
    return;
  }

  if (foobar3) {
    primtree->root->octree->debug_find_collx(debug_split_breakid, primtree->root->octree->octree);
    return;
  }

  if (foobar99) {

#if 1
    /*
     * translates the subtracted sphere out of modeller space
     */

    if (debug_incidentstatus == 0) {
      // Translate the subtracted sphere.
      csg_graph->select_element(((groupobjectnode *) csg_graph->object)->group);
      skitterlist->dmd_select(skitterlist->copies[0].skitter[0]);

      ((ObjectTranslation *) skitterlist->current_dmd->transformation->trans)->set_translation(0.25, 0.75, -0.68);
      csg_graph->selected_object->transformlist->update_transform(skitterlist->current_dmd->transformation);
      csg_graph->selected_object->transformlist->propagate_matrices();
      skitterlist->update_all_skitters();
      update_all_transviews();
      csg_graph->update_transformed_object(csg_graph->selected_object);
      update_all_views();
    }

    if (debug_incidentstatus == 1) {
      ((ObjectTranslation *) skitterlist->current_dmd->transformation->trans)->set_translation(2.3, 0.0, 0.0);
      csg_graph->selected_object->transformlist->update_transform(skitterlist->current_dmd->transformation);
      csg_graph->selected_object->transformlist->propagate_matrices();
      skitterlist->update_all_skitters();
      update_all_transviews();
      csg_graph->update_transformed_object(csg_graph->selected_object);
      update_all_views();
    }
#endif

#if 0
    debug_incidentstatus++;

    if ((debug_incidentstatus >= 0) && (debug_incidentstatus < 2)) {      
      if (debug_incidentstatus == 0) {      
        /*
         * translates the main cutting plane to (0, 0, 0.69)
         */     
        csg_graph->select_element(downcutnode);
        skitterlist->dmd_select(skitterlist->copies[0].skitter[0]);
        ((ObjectTranslation *) skitterlist->current_dmd->transformation->trans)->set_translation(0, 0, 0.69);
      } else {
        /*
         *  rotates the secondary cutting plane by 114.0 degrees about (1 0 0)
         */
//        debug_split_breakid = 5640;

        csg_graph->select_element(manipulatenode);
        skitterlist->dmd_select(skitterlist->copies[0].skitter[0]);

        vector tempaxis;
        Vector::assign(tempaxis, 1.00, 0.00, 0.00);
        ((ObjectRotation *) skitterlist->current_dmd->transformation->trans)->set_rotation(tempaxis, DEG2RAD(114.0));
      }
      
      csg_graph->selected_object->transformlist->update_transform(skitterlist->current_dmd->transformation);
      csg_graph->selected_object->transformlist->propagate_matrices();
      skitterlist->update_all_skitters();
      update_all_transviews();
      csg_graph->update_transformed_object(csg_graph->selected_object);
      csg_graph->deselect();
      update_all_views();
    } else if (debug_incidentstatus == 2) {
      transview[0]->add_trans(new ObjectTranslation(0, 0, -0.62));
    }
#endif


#if 0
    debug_incidentstatus++;

    if (debug_incidentstatus == 0) {      
      /*
       * translates the main cutting plane to (0, 0, -1.06).
       */     
      csg_graph->select_element(downcutnode);
      skitterlist->dmd_select(skitterlist->copies[0].skitter[0]);
      ((ObjectTranslation *) skitterlist->current_dmd->transformation->trans)->set_translation(0, 0, -1.1);
      csg_graph->selected_object->transformlist->update_transform(skitterlist->current_dmd->transformation);
      csg_graph->selected_object->transformlist->propagate_matrices();
      skitterlist->update_all_skitters();
      update_all_transviews();
      csg_graph->update_transformed_object(csg_graph->selected_object);
      csg_graph->deselect();
      update_all_views();
    }
#endif

#if 0
    debug_incidentstatus++;

    if ((debug_incidentstatus >= 0) && (debug_incidentstatus < 2)) {      
      /*
       * rotates the cutting plane's top section
       */     
      csg_graph->select_element(manipulatenode);
      skitterlist->dmd_select(skitterlist->copies[0].skitter[0]);
      ((ObjectRotation *) skitterlist->current_dmd->transformation->trans)->
        set_rotation(Z, Y, DEG2RAD((debug_incidentstatus == 0) ? 160 : 150));
      csg_graph->selected_object->transformlist->update_transform(skitterlist->current_dmd->transformation);
      csg_graph->selected_object->transformlist->propagate_matrices();
      skitterlist->update_all_skitters();
      update_all_transviews();
      csg_graph->update_transformed_object(csg_graph->selected_object);
      csg_graph->deselect();
      update_all_views();
    }
#endif

  }
}

// Test object: a piston.
void Modeller::make_piston(void)
{
  ObjectTransformation *trans_array[10];
  primobjectnode *plane_prim, *sheathnode, *cutternode, *cutplanenode;
  groupobjectnode *cuttergroup;
  primobjectnode *ringobj, *armobj, *wingobj;
  groupobjectnode *ringgrp, *armgrp, *winggrp;

  double piston_thickness = 0.1;
  double piston_height = 1.8;
  double conrod_connect_height = 0.8;
  double conrod_sheath_rad = 0.5;
  double conrod_inside_rad = 0.3;
  double conrod_cutout_rad = 0.35;
  double cut_plane_height = 0.9;
  double ring_base = 1.4;
  double ring_thickness = 0.08;
  double second_ring = 0.2;
  double ring_inside_rad = 1.0 - piston_thickness / 2;
  double crank_sheath_outer_rad = 0.8;
  double conrod_rad = 0.26;
  double conrod_dist = 2.4;
  double arm_thickness = conrod_cutout_rad - 0.04;
  double wing_thickness = arm_thickness / 2;
  double arm_end_inside_rad = 0.6;


  // We start from a plane... everything in +Z is full.
  
#if 1
  trans_array[0] = new ObjectScale(1.0 - piston_thickness);
  csg_graph->link_in_primitive(csg_graph->object, MINUS,
                               new cylinderPrimitive, NULL, 1, trans_array);
  
#endif

#if 1
  trans_array[0] = new ObjectTranslation(0, 0, piston_height - piston_thickness);
  csg_graph->link_in_primitive(csg_graph->object, PLUS, new planePrimitive, &plane_prim,
                               1, trans_array);
#endif

#if 1
  trans_array[0] = new ObjectScale(conrod_sheath_rad);
  trans_array[1] = new ObjectRotation(Z, X, PI/2);
  trans_array[2] = new ObjectTranslation(0, 0, conrod_connect_height);
  csg_graph->link_in_primitive(csg_graph->object, PLUS, new cylinderPrimitive, &sheathnode,
                               3, trans_array);

  trans_array[0] = new ObjectTranslation(0, 0, -conrod_cutout_rad / conrod_sheath_rad);
  csg_graph->link_in_primitive(sheathnode, MINUS, new planePrimitive, &cutternode,
                               1, trans_array);

  trans_array[0] = new ObjectTranslation(0, 0, 2*conrod_cutout_rad / conrod_sheath_rad);
  csg_graph->link_in_primitive(cutternode, MINUS, new planePrimitive, NULL,
                               1, trans_array);
#endif

#if 0
  trans_array[0] = new ObjectScale(conrod_rad);
  trans_array[1] = new ObjectRotation(Z, X, PI/2);
  trans_array[2] = new ObjectTranslation(0, 0, conrod_connect_height);
  csg_graph->link_in_primitive(csg_graph->object, PLUS, new cylinderPrimitive, NULL, 3, trans_array);
#endif

#if 1
  csg_graph->link_in_primitive(csg_graph->object, AND, new cylinderPrimitive, NULL, 0, NULL);
#endif

#if 1
  trans_array[0] = new ObjectScale(conrod_inside_rad);
  trans_array[1] = new ObjectRotation(Z, X, PI/2);
  trans_array[2] = new ObjectTranslation(0, 0, conrod_connect_height);
  csg_graph->link_in_primitive(csg_graph->object, MINUS, new cylinderPrimitive, NULL,
                               3, trans_array);
#endif

#if 1
  trans_array[0] = new ObjectTranslation(0, 0, piston_height);
  csg_graph->link_in_primitive(csg_graph->object, MINUS, new planePrimitive, NULL, 1, trans_array);
#endif

  trans_array[0] = new ObjectTranslation(0, 0, ring_base);
  csg_graph->link_in_primitive(csg_graph->object, MINUS, new planePrimitive, &ringobj, 1, trans_array);

  trans_array[0] = new ObjectTranslation(0, 0, ring_thickness);
  ringgrp = csg_graph->link_in_primitive(ringobj, MINUS, new planePrimitive, NULL, 1, trans_array);

  ringgrp = csg_graph->add_wrapper_around(ringgrp);

  trans_array[0] = new ObjectTranslation(0, 0, second_ring);
  csg_graph->link_in_primitive(ringgrp, PLUS, new planePrimitive, &ringobj, 1, trans_array);

  trans_array[0] = new ObjectTranslation(0, 0, ring_thickness);
  csg_graph->link_in_primitive(ringobj, MINUS, new planePrimitive, NULL, 1, trans_array);

  trans_array[0] = new ObjectScale(ring_inside_rad);
  csg_graph->link_in_primitive(ringgrp, MINUS, new cylinderPrimitive, NULL, 1, trans_array);



  trans_array[0] = new ObjectRotation(Z, X, PI/2);
  trans_array[1] = new ObjectTranslation(0, 0, conrod_connect_height);
  csg_graph->link_in_primitive(csg_graph->object, PLUS, new nothingPrimitive, &armobj, 2, trans_array);

  trans_array[0] = new ObjectScale(conrod_sheath_rad);
  armgrp = csg_graph->link_in_primitive(armobj, PLUS, new cylinderPrimitive, NULL, 1, trans_array);

  trans_array[0] = new ObjectScale(crank_sheath_outer_rad);
  trans_array[1] = new ObjectTranslation(conrod_dist, 0, 0);
  csg_graph->link_in_primitive(armgrp, PLUS, new cylinderPrimitive, NULL, 2, trans_array);


  trans_array[0] = new ObjectTranslation(0, 0, arm_thickness);
  csg_graph->link_in_primitive(armgrp, MINUS, new planePrimitive, NULL, 1, trans_array);

  trans_array[0] = new ObjectRotation(Z, X, PI);
  trans_array[1] = new ObjectTranslation(0, 0, -arm_thickness);
  csg_graph->link_in_primitive(armgrp, MINUS, new planePrimitive, NULL, 2, trans_array);




  csg_graph->link_in_primitive(armgrp, PLUS, new nothingPrimitive, &wingobj, 0, NULL);

  trans_array[0] = new ObjectRotation(Z, Y, PI/2);
  trans_array[1] = new ObjectTranslation(0, -conrod_inside_rad, 0);
  trans_array[2] = new ObjectRotation(Y, X, PI / 18);
  winggrp = csg_graph->link_in_primitive(wingobj, PLUS, new planePrimitive, NULL, 3, trans_array);

  trans_array[0] = new ObjectRotation(Z, Y, PI/2);
  trans_array[1] = new ObjectTranslation(0, +conrod_inside_rad, 0);
  trans_array[2] = new ObjectRotation(X, Y, PI / 18);
  csg_graph->link_in_primitive(winggrp, MINUS, new planePrimitive, NULL, 3, trans_array);

  trans_array[0] = new ObjectTranslation(0, 0, wing_thickness);
  csg_graph->link_in_primitive(winggrp, MINUS, new planePrimitive, NULL, 1, trans_array);

  trans_array[0] = new ObjectTranslation(0, 0, wing_thickness);
  trans_array[1] = new ObjectRotation(Z, Y, PI);
  csg_graph->link_in_primitive(winggrp, MINUS, new planePrimitive, NULL, 2, trans_array);

#if 1
  trans_array[0] = new ObjectRotation(Z, X, -PI/2);
  csg_graph->link_in_primitive(winggrp, MINUS, new planePrimitive, NULL, 1, trans_array);
#endif

  trans_array[0] = new ObjectRotation(Z, X, PI/2);
  trans_array[1] = new ObjectTranslation(conrod_dist, 0, 0);
  csg_graph->link_in_primitive(winggrp, MINUS, new planePrimitive, NULL, 2, trans_array);

#if 1
  trans_array[0] = new ObjectScale(arm_end_inside_rad);
  trans_array[1] = new ObjectTranslation(conrod_dist, 0, 0);
  csg_graph->link_in_primitive(armgrp, MINUS, new cylinderPrimitive, NULL, 2, trans_array);
#endif




#if 1
  csg_graph->add_wrapper_around(csg_graph->object);

//  vector ax = {1.0, 1.0, 0.0};
//  trans_array[0] = new ObjectRotation(ax, PI/4);
  trans_array[0] = new ObjectTranslation(0, 0, 2.0);
  csg_graph->link_in_primitive(csg_graph->object, MINUS, new planePrimitive, &cutplanenode, 1, trans_array);
  downcutnode = cutplanenode;

  trans_array[0] = new ObjectTranslation(0, 0, 0.2);
  trans_array[1] = new ObjectRotation(Z, Y, 0.2 * PI/2);
  cuttergroup = csg_graph->link_in_primitive(cutplanenode, MINUS, new planePrimitive, &manipulatenode, 2, trans_array);
#endif



#if 0
  // Set up an object to look at.
  ObjectTransformation *trans_array[10];
  primobjectnode *newsphere_node;
  objectnode *chainhead;

  trans_array[0] = new ObjectScale(0.4);
  trans_array[1] = new ObjectTranslation(0.9, 0.0, 0.0);
  csg_graph->link_in_primitive(csg_graph->object,
			       MINUS, new spherePrimitive,
			       &newsphere_node, 2, trans_array);
#endif
#if 0
  chainhead = csg_graph->object;
  for (int i = 1; i <= 1; i++) {
    trans_array[0] = new ObjectRotation(X, Y, PI/6 * i);
    chainhead = csg_graph->link_in_copy(chainhead, MINUS, newsphere_node, NULL, 1, trans_array);
  }
#endif
#if 0
  trans_array[0] = new ObjectRotation(X, Y, 0.01 * PI/2);
  csg_graph->add_repeat_around(newsphere_node, 4, 1, trans_array);
#endif


}

// Allocate or deallocate a space in the primsplitinfo array.
int Modeller::alloc_primsplitinfo(void)
{
  int next_psi = primsplitinfo_counter++;

  if (next_psi >= max_primsplitinfo)
    fatal_error("Modeller::alloc_primsplitinfo", "primsplitinfo array overflow",
                "Change hardcoded max_primsplitinfo and recompile");

  return next_psi;
}
void Modeller::dealloc_primsplitinfo(int dealloc_psi)
{
  primsplitinfo_counter--;

  // The primsplitinfo array is basically a stack. We must deallocate from
  // the far end always.
  assert(primsplitinfo_counter >= 0);
  assert(dealloc_psi == primsplitinfo_counter);
}



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif