/*
 * modeller.h: Definitions for class Modeller.
 *
 * The Modeller class contains the current state of the CSG modeller,
 * including the object being modelled and any views of the object.
 */

/* Collection-safe */

#define MODELLER_MAX_VIEWS  10

// Global variables
extern int nasty_hack_debugging_flag;

class Modeller {
public:
  // Bounding box around the object, and the size of the lowest-level
  // (terminal) nodes of the octree.
  vector minbound, maxbound, terminalsize;

  // The positions of the zero-corners of the terminal nodes of the octree.
  // Indexed by terminal_id.
  vector *terminalcorner;

  // The current detail levels and toggles.
  int octree_level, terminal_res, use_gouraud, draw_inout_pts;

  // The object being modelled.
  CSGgraph *csg_graph;
  Primtree *primtree;

  SolidOctree *focused_octree;

  // The class that lays out the view of the CSGgraph.
  GraphLayout *graphlayout;

  // Triangle and string storage memory.
  TriMemory *trimem;
  StringMemory *stringmem;
  WorldStringMemory *worldstringmem;

  // The TriSplitter that is used to split cells' triangles.
  TriSplitter *trisplitter;

  // The model view windows that are open on the scene.
  int num_modelviews;
  View *modelview[MODELLER_MAX_VIEWS];

  // The graph view windows that are open on the scene.
  int num_graphviews;
  GraphView *graphview[MODELLER_MAX_VIEWS];

  // The transformation view windows that are open on the scene.
  int num_transviews;
  TransView *transview[MODELLER_MAX_VIEWS];

  // The last object that was clicked for selection, and the object that was
  // selected because of this.
  primtreeprimnode *last_click;
  primtreenode *selected_node;

  // The Skitters that are currently visible on the object. These are little
  // sets of axes that are displayed for each ObjectTransformation on the
  // selected objectnode. This also contains information about the currently
  // selected DMD.
  SkitterList *skitterlist;

  // A nothingPrimitive for use in the tail-ends of
  // repeatobjectnode primitive graphs.
  nothingPrimitive *tailend_nothingprim;

  // Lighting parameters.
  double ambient_light;
  int num_lights;
  double *light_strength;
  vector *light_dir;

  // The next entry in the primsplitinfo arrays to be used.
  int primsplitinfo_counter;

  /*
   * Debugging stuff
   */

  int debug_line;
  int debug_all_splits;
  vector debug_linefrom, debug_linedir;
#if 0
  int debug_terminal_count;
  int terminal_highlight, highlighted_count, show_only_highlighted;
  terminaltrackernode *highlight_terminal;
  void cycle_debugging_terminal(void);
  void toggle_debugging_highlight(void);
#endif
  void debugcall(char *mode, char *args);


  // Set up the objectgraphs
  void init_objectgraphs(void);
  void create_placeholder_object(Primitive *placeholder_prim);

  // Get the bounds of the object.
  void get_bounds(box b);

  // Calculate the positions of the octree's terminal corners.
  void calculate_terminal_corners(int terminal_id, int x, int y, int z, int size);

  // Bind a View to the modeller.
  void bind_modelview(View *v);

  // Unbind a View from the modeller.
  void unbind_modelview(View *v);

  // Bind a GraphView to the modeller.
  void bind_graphview(GraphView *v);

  // Unbind a GraphView from the modeller.
  void unbind_graphview(GraphView *v);

  // Bind a TransView to the modeller.
  void bind_transview(TransView *v);

  // Unbind a TransView from the modeller.
  void unbind_transview(TransView *v);

  // Select the object along a ray cast into the scene.
  void select_along_ray(vector ray_from, vector ray_dir);

  // Set the state of the graph views.
  void setgraphviewstate(graphviewstate state);

  // Select a node.
  void do_select(objectnode *node);

  // Delete the selected node..
  void do_delete(void);

  // Set gouraud-shading.
  void set_gouraud(int new_value);

  // Set point-drawing.
  void set_inout_pts(int new_value);

  // Update all Views that are bound to this Modeller.
  void update_all_views(void);

  // Update all TransViews that are bound to this Modeller.
  void update_all_transviews(void);

  // Debugging hack.
  void hack(int hackval);

  // Called by the graphview classes.
  void do_link_copy(objectnode *copynode, linktype link_type);
  void do_link_prim(Primitive *new_prim, linktype link_type);

  // Test object: a piston.
  void make_piston(void);

  // Allocate or deallocate a space in the primsplitinfo array.
  int alloc_primsplitinfo(void);
  void dealloc_primsplitinfo(int psi_index);

  // Create the Modeller.
  Modeller(void);

  // Delete the Modeller.
  ~Modeller(void);
};
