/*
 * octree.cc: Routines for class SolidOctree and supporting code.
 *
 * Contains code to create and maintain the SolidOctree representation of
 * an object.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

extern splittrackernode *debug_splittrackernode;
#define TERMINAL_FACETOUCH_TOLERANCE  1e-04

int globalhackvar_solidoctreecount = 0;

/*
 * Routines for SolidOctree
 */

// Constructor that creates a copy of an old SolidOctree.
SolidOctree::SolidOctree(SolidOctree *original)
{
  primitive_changed_this_update = FALSE;
  octree_deleted = FALSE;

  splittracker = new SplitTracker(this);

  // Initialise the memory managers.
  mm_octreenode = new MemoryManager<solidoctreenode>(TRUE, 10);
  mm_terminal   = new MemoryManager<solidterminal>  (TRUE, 10);
  mm_rcsgnode   = new MemoryManager<solidrcsgnode>  (FALSE, 10);
  mm_leafnode   = new MemoryManager<solidleafnode>  (FALSE, 10);

  // Copy the parameters of the original SolidOctree.
  represented_object = original->represented_object;
  modeller = original->modeller;

  assert(modeller != NULL);
  
  // Copy the octree of the original SolidOctree.
  octree = duplicate_octree(original, &(original->octree));

  left_octree = right_octree = NULL;

  // Allocate arrays for the ray tracing.
  cell_size      = (vector *) calloc(modeller->octree_level + 1,
				     sizeof(vector));
  delta_t_cell_x = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
  delta_t_cell_y = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
  delta_t_cell_z = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
    
  if (cell_size == NULL || delta_t_cell_x == NULL ||
      delta_t_cell_y == NULL || delta_t_cell_z == NULL)
    fatal_error("SolidOctree::SolidOctree", "Can't allocate arrays!",
		"Out of memory.");

  has_changed = FALSE;

  globalhackvar_solidoctreecount++;
}

// Destructor deallocates the SolidOctree.
SolidOctree::~SolidOctree(void)
{
  assert(!octree_deleted);

  // Delete the SolidOctree. This will take care of de-referencing
  // WorldCollections and Collections.
  delete_solidindex(octree);

  // Delete the SplitTracker.
  delete splittracker;

  // Delete the memory managers.
  delete mm_octreenode;
  delete mm_rcsgnode;
  delete mm_leafnode;
  delete mm_terminal;

  // Free the ray-tracing arrays.
  free(cell_size);
  free(delta_t_cell_x);
  free(delta_t_cell_y);
  free(delta_t_cell_z);

  octree_deleted = TRUE;

  globalhackvar_solidoctreecount--;
}

// Duplicates some sub-graph of an octree.
solidindex SolidOctree::duplicate_octree(SolidOctree *original_solidoctree,
                                         solidindex *original,
                                         int invert_octree,
                                         int set_changed,
                                         int clear_changed,
                                         int store_in_splittracker,
                                         SolidOctree *splittracker_opponent_octree,
                                         int splittracker_is_left)
{
  solidindex new_index;

  int foobar = 0;

  if (foobar)
    original_solidoctree->debug_solidindex(*original);

  if (SI_ISEMPTY(*original)) {
    if (invert_octree) {
      new_index = SOLIDINDEXTYPE_FULL;
    } else {
      new_index = *original;
    }
  } else if (SI_ISFULL(*original)) {
    if (invert_octree) {
      new_index = SOLIDINDEXTYPE_EMPTY;
    } else {
      new_index = *original;
    }

  } else if (SI_ISOCTREE(*original)) {
    solidoctreenode *new_node = mm_octreenode->allocate(&new_index),
      *old_node = original_solidoctree->mm_octreenode->
                      retrieve_element(SI_GETDATA(*original));

    new_index |= SOLIDINDEXTYPE_OCTREE;
    
    for (int i = 0; i < 8; i++)
      new_node->child[i] = duplicate_octree(original_solidoctree,
					    &(old_node->child[i]),
              invert_octree, set_changed, clear_changed,
              store_in_splittracker,
              splittracker_opponent_octree,
              splittracker_is_left);

  } else if (SI_ISTERMINAL(*original)) {
    solidterminal *old_node = original_solidoctree->mm_terminal->
                      retrieve_element(SI_GETDATA(*original));

#if 0
    if ((old_node->terminal_id == 392) &&
        ((original_solidoctree->represented_object->index == 65) ||
         (this->represented_object->index == 65)))
      _CrtDbgBreak();
#endif

    new_index = duplicate_terminal(original_solidoctree, old_node,
                                   invert_octree, store_in_splittracker,
                                   splittracker_opponent_octree, splittracker_is_left);

  } else {
    // Never get here.
    assert(0);
  }

  if (set_changed || SI_ISCHANGED(*original))
    new_index |= SOLIDINDEX_CHANGED;

  if (clear_changed)
    *original &= ~SOLIDINDEX_CHANGED;

  return new_index;
}

// Clear the changed flags of an octree.
void SolidOctree::clear_changed(SolidOctree *original_solidoctree,
				solidindex *original)
{
  if (SI_ISOCTREE(*original)) {
    solidoctreenode *old_node = original_solidoctree->mm_octreenode->
                      retrieve_element(SI_GETDATA(*original));

    for (int i = 0; i < 8; i++) {
      clear_changed(original_solidoctree,
		    &(old_node->child[i]));
    }  
  }

  *original &= ~SOLIDINDEX_CHANGED;
}

// Duplicate a terminal node of an octree.
solidindex SolidOctree::duplicate_terminal(SolidOctree *original_solidoctree,
                                           solidterminal *original,
                                           int invert_rcsggraph,
                                           int store_in_splittracker,
                                           SolidOctree *splittracker_opponent_octree,
                                           int splittracker_is_left)
{
  solidindex new_index;
  solidterminal *new_node = mm_terminal->allocate(&new_index);

  new_node->terminal_id = original->terminal_id;
  new_node->rcsggraph = duplicate_rcsggraph(original_solidoctree, new_node, original->rcsggraph,
                                            invert_rcsggraph, FALSE, new_node->terminal_id, store_in_splittracker,
                                            splittracker_opponent_octree, splittracker_is_left);

  new_index |= SOLIDINDEXTYPE_TERMINAL;
  return new_index;
}

// Duplicate an RCSG graph in the terminal node.
solidrcsggraph SolidOctree::duplicate_rcsggraph(SolidOctree *original_solidoctree,
                                                solidterminal *newlycreated_rcsg, 
                                                solidrcsggraph rcsggraph,
                                                int invert_rcsggraph, int is_flipped,
                                                int terminal_id, int store_in_splittracker,
                                                SolidOctree *splittracker_opponent_octree,
                                                int splittracker_is_left)
{
  solidrcsggraph newnode;
  solidrcsgnode *new_rcsg;
  solidleafnode *new_leaf;
  int i;

  newnode.flags = rcsggraph.flags;
  if (invert_rcsggraph)
    newnode.flags ^= RCSGGRAPH_FLIP_MASK;
  
  if (RCSG_ISFLIPPED(newnode.flags))
    is_flipped = !is_flipped;

  if (RCSG_ISLEAF(rcsggraph.flags)) {
    // Duplicate this leafnode.
    new_leaf = mm_leafnode->allocate();

    new_leaf->is_deleted = FALSE;
    new_leaf->prim = rcsggraph.leaf->prim;
    for (i = 0; i < 8; i++)
      Vector::set(new_leaf->terminalbox[i], rcsggraph.leaf->terminalbox[i]);

    if (store_in_splittracker) {
      // We are NOT going to duplicate all the WorldCollections just yet.
      // Instead we create an empty leafnode and register with the Octree's
      // SplitTracker that all these WorldCollections want to be added here.
      // It will sort it out.
      new_leaf->num_collections = 0;
      new_leaf->collection_bufsize = 0;
      new_leaf->collectionbuf = NULL;

      for (i = 0; i < rcsggraph.leaf->num_collections; i++) {
        assert(rcsggraph.leaf->collectionbuf[i].collection->owner == new_leaf->prim);

        // Register that this collection is here being KEPT (opposing_terminal == NULL)
        splittracker->register_terminal(rcsggraph.leaf->collectionbuf[i].collection, 
                                        is_flipped, this, terminal_id, newlycreated_rcsg, new_leaf,
                                        splittracker_opponent_octree, splittracker_is_left, NULL);
      }

    } else {
      // Duplicate the leafnode.
      new_leaf->num_collections = rcsggraph.leaf->num_collections;
      new_leaf->collection_bufsize = new_leaf->num_collections;
      new_leaf->collectionbuf = (leafcollx *) calloc(new_leaf->collection_bufsize,
                                                     sizeof(leafcollx));

      for (i = 0; i < new_leaf->num_collections; i++) {
        new_leaf->collectionbuf[i].collection = rcsggraph.leaf->collectionbuf[i].collection;
        new_leaf->collectionbuf[i].splitptr = NULL;
        new_leaf->collectionbuf[i].collection->add_reference();
      } 
    }

    newnode.leaf = new_leaf;
  } else {
    // Duplicate this solidrcsgnode.
    new_rcsg = mm_rcsgnode->allocate();

    new_rcsg->link = rcsggraph.rcsg->link;
    new_rcsg->left  = duplicate_rcsggraph(original_solidoctree, newlycreated_rcsg, rcsggraph.rcsg->left,
                                          FALSE, is_flipped, terminal_id, store_in_splittracker,
                                          splittracker_opponent_octree, splittracker_is_left);

    // Subtracted surfaces are flipped.
    if (new_rcsg->link == MINUS)
      is_flipped = !is_flipped;

    new_rcsg->right = duplicate_rcsggraph(original_solidoctree, newlycreated_rcsg, rcsggraph.rcsg->right, 
                                          FALSE, is_flipped, terminal_id, store_in_splittracker,
                                          splittracker_opponent_octree, splittracker_is_left);

    newnode.rcsg = new_rcsg;
  }

  return newnode;
}
				       
int debug_octree_print = -1;

// Create a new SolidOctree for a CSGnode that is part of a section of
// the Primtree that has been newly created.
// 
// This routine is called when new objects have
// been linked into the Primtree, and is called in a recursive pass from
// the bottom of the Primtree up to the linkpoint.
SolidOctree::SolidOctree(primcsgnode *represented_object, Modeller *modeller)
{
  primcsgnode *owner_csgnode;
  int dummy;
    
  primitive_changed_this_update = FALSE;
  octree_deleted = FALSE;

  this->represented_object = represented_object;
  this->modeller = modeller;

  splittracker = new SplitTracker(this);

  assert(modeller != NULL);
  
  // Initialise the memory managers.
  mm_octreenode = new MemoryManager<solidoctreenode>(TRUE, 10);
  mm_terminal   = new MemoryManager<solidterminal>  (TRUE, 10);
  mm_rcsgnode   = new MemoryManager<solidrcsgnode>  (FALSE, 10);
  mm_leafnode   = new MemoryManager<solidleafnode>  (FALSE, 10);

  // Allocate arrays for the ray tracing.
  cell_size      = (vector *) calloc(modeller->octree_level + 1,
				     sizeof(vector));
  delta_t_cell_x = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
  delta_t_cell_y = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
  delta_t_cell_z = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
    
  if (cell_size == NULL || delta_t_cell_x == NULL ||
      delta_t_cell_y == NULL || delta_t_cell_z == NULL)
    fatal_error("SolidOctree::SolidOctree", "Can't allocate arrays!",
		"Out of memory.");

  assert(represented_object->type() == PRIMTREENODE_CSG);

  owner_csgnode = (primcsgnode *) represented_object;
   
  left_octree = owner_csgnode->left->octree;
  right_octree = owner_csgnode->right->octree;
  
  assert(left_octree != NULL);
  assert(right_octree != NULL);

  switch(owner_csgnode->csg_type) {
  case PLUS:
    // The OUT sections of both objects are active.
    left_in_is_active = right_in_is_active = FALSE;
    break;
  case MINUS:
    // The OUT section of left and the IN section of right are active.
    left_in_is_active = FALSE;
    right_in_is_active = TRUE;
    break;
  case AND:
    // The IN sections of both objects are active.
    left_in_is_active = right_in_is_active = TRUE;
    break;
  default:
    assert(0);
  }

#if 1
  if (represented_object->index == debug_octree_print) {
    _CrtDbgBreak();
    left_octree->print(0);
    right_octree->print(0);
  }
#endif

  octree = create_csg_octree(owner_csgnode->csg_type,
			     left_octree,  &(left_octree->octree),
			     right_octree, &(right_octree->octree), &dummy, 0);

#if 1
  if (represented_object->index == debug_octree_print) {
    TRACE("csg post-left\n");
    left_octree->print(0);
    TRACE("csg post-right\n");
    right_octree->print(0);
    TRACE("csg done\n");
    print(0);
  }
#endif

  // Perform all splits that are necessary.
  splittracker->go_split();

  has_changed = FALSE;

  globalhackvar_solidoctreecount++;
}

// Create a new SolidOctree for a primtreeprimnode that is part of a section of
// the Primtree that has been newly created.
// 
// This routine is called when new primitives have
// been linked into the Primtree (on their own or as part of a group), and is
// called in a recursive pass from the bottom of the Primtree up to
// the linkpoint.
SolidOctree::SolidOctree(primtreeprimnode *represented_object,
			 Modeller *modeller,
			 linktype prim_linktype)
{
  primtreeprimnode *owner_prim;
  box worldspace_box, primspace_box, collectionspace_box;
  int i;
  solidindex ethertype;
  primobjectnode *primobject;
  transform collectionspace;

  primitive_changed_this_update = FALSE;
  octree_deleted = FALSE;

  this->represented_object = represented_object;
  this->modeller = modeller;

  assert(modeller != NULL);
  
  splittracker = new SplitTracker(this);

  // Initialise the memory managers.
  mm_octreenode = new MemoryManager<solidoctreenode>(TRUE, 10);
  mm_terminal   = new MemoryManager<solidterminal>  (TRUE, 10);
  mm_rcsgnode   = new MemoryManager<solidrcsgnode>  (FALSE, 10);
  mm_leafnode   = new MemoryManager<solidleafnode>  (FALSE, 10);

  // Allocate arrays for the ray tracing.
  cell_size      = (vector *) calloc(modeller->octree_level + 1,
				     sizeof(vector));
  delta_t_cell_x = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
  delta_t_cell_y = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
  delta_t_cell_z = (double *) calloc(modeller->octree_level + 1,
				     sizeof(double));
    
  if (cell_size == NULL || delta_t_cell_x == NULL ||
      delta_t_cell_y == NULL || delta_t_cell_z == NULL)
    fatal_error("SolidOctree::SolidOctree", "Can't allocate arrays!",
		"Out of memory.");

  assert(represented_object->type() == PRIMTREENODE_PRIM);  
  owner_prim = (primtreeprimnode *) represented_object;

  // Remember, the original ptr can be NULL if we are a tail-end
  // repeatobjectnode.
  if (owner_prim->original == NULL) {
    // We need an empty octree, it's just that simple.
    octree = SOLIDINDEXTYPE_EMPTY;
  } else {

    assert(owner_prim->original->type() == OBJECTNODE_PRIM);
    primobject = (primobjectnode *) owner_prim->original;

    assert(modeller != NULL);
  
    // Form the corners of the modeller's bounding box.
    modeller->get_bounds(worldspace_box);
  
    // Transform the corners of the bounding box into primitivespace,
    // to form primspace_box. Also find collectionspace for this primobject.
    for (i = 0; i < 8; i++) {
      Matrix::apply(represented_object->position[AFT], worldspace_box[i], primspace_box[i]);
      Matrix::apply(represented_object->collection_trans[AFT], worldspace_box[i], collectionspace_box[i]);
    }
  
    // Find the transformation from primitivespace to collectionspace.
    // represented_object has two transformations; collectiontrans from collectionspace
    // to worldspace, and position from primitivespace to worldspace. We want primitivespace to
    // collectionspace.
    Transform::invert(represented_object->collection_trans, collectionspace);
    Transform::combine(represented_object->position, collectionspace, collectionspace);

    ethertype = (prim_linktype == AND) ? SOLIDINDEXTYPE_FULL 
                                       : SOLIDINDEXTYPE_EMPTY;

    if (primobject->collections_found) {
      // This primitive already has collections. We need to transform them into
      // the new worldspace and thus set up the new worldcollections for this primitive.
      for (i = 0; i < primobject->num_collections; i++) {
        primobject->worldcollectionbuf[i] = new WorldCollection(represented_object, primobject->collectionbuf[i]);
        primobject->worldcollectionbuf[i]->add_reference();
      }
    }

    // Create the octree structure for this primitive.
    octree = create_primitiveoctree(primobject->collections_found, primobject,
                                    collectionspace, primspace_box, collectionspace_box,
			  	                          modeller->octree_level, ethertype, 0);

    if (primobject->collections_found) {
      // Clear out the worldcollectionbuf.
      for (i = 0; i < primobject->num_collections; i++) {
        if (primobject->worldcollectionbuf[i]->remove_reference())
          delete primobject->worldcollectionbuf[i];
        primobject->worldcollectionbuf[i] = NULL;
      }
    } else {
      // We have now createdd all of the collections for this object.
      primobject->collections_found = TRUE;
    }
  }

  has_changed = FALSE;

  globalhackvar_solidoctreecount++;
}

// Creates a SolidOctree structure from a primitive.
//
// If use_collections is FALSE, then we are the first octree to
// polygonize a primobjectnode. Create new Collections at the bottom
// level and store them in the CSGgraph.
//
// If use_collections is TRUE, we have already polygonized the
// primobjectnode. Use the existing Collections at the bottom level.
solidindex SolidOctree::create_primitiveoctree(int use_collections, primobjectnode *primobject,
                                  transform collectionspace, box primbox, box collectionbox, int octree_level,
                                  solidindex ethertype, int terminal_id)
{
  intersect_result prim_result;
  solidindex new_index;
  int i;
  
  // First: test the box versus the Primitive.
  prim_result = primobject->prim->voxel_intersect(primbox);

  if (prim_result == IN) {
    // This box is FULL of the primitive.
    new_index = SOLIDINDEXTYPE_FULL;
    if (new_index != ethertype)
      new_index |= SOLIDINDEX_CHANGED;
    return new_index;

  } else if (prim_result == OUT) {
    // This box is EMPTY of the primitive.
    new_index = SOLIDINDEXTYPE_EMPTY;
    if (new_index != ethertype)
      new_index |= SOLIDINDEX_CHANGED;
    return new_index;

  } else if (prim_result == SURFACE) {
    // Should this box be subdivided?
    if (octree_level == 0) {
      // Create a terminal node that holds the primitive.
      solidterminal *new_node = mm_terminal->allocate(&new_index);
      solidleafnode *new_leaf = mm_leafnode->allocate();;

      new_leaf->is_deleted = FALSE;
      new_node->rcsggraph.flags = RCSGGRAPH_LEAF_MASK;
      new_node->rcsggraph.leaf = new_leaf;
      new_node->terminal_id = terminal_id;

      new_leaf->prim = (primtreeprimnode *) represented_object;
      for (i = 0; i < 8; i++)
        Vector::set(new_leaf->terminalbox[i], collectionbox[i]);
      
      if (use_collections) {
        box transformedbox;
        vector trans_shift, trans_scales;

#if 0
        if (((this->represented_object->index == 63) ||
             (this->represented_object->index == 65) ||
             (this->represented_object->index == 66)) &&
            (terminal_id == 392))
              _CrtDbgBreak();
#endif

        // We have a number of Collections, each of which has collection->bounds
        // in collectionspace. And we also have the bounds of this leafnode in
        // collectionspace, as collectionbox. Intersect these. Any Collections that
        // intersect this leafnode must be placed here.
        new_leaf->num_collections = new_leaf->collection_bufsize = 0;
        new_leaf->collectionbuf = NULL;

        for (i = 0; i < primobject->num_collections; i++) {
          // Transform collectionbox into a space where the
          // bounds of the Collection lie at -1, +1 in X, Y and Z.
          //
          // Note that because we want the transformedbox to return OUT even
          // if it lies with an adjacent face, we actually scale the transformedbox
          // so that the Collection's boundaries lie at 1 + TERMINAL_FACETOUCH_TOLERANCE.

          trans_shift[X] = (primobject->collectionbuf[i]->bounds[0][X] + primobject->collectionbuf[i]->bounds[1][X]) / 2.0;
          trans_shift[Y] = (primobject->collectionbuf[i]->bounds[0][Y] + primobject->collectionbuf[i]->bounds[2][Y]) / 2.0;
          trans_shift[Z] = (primobject->collectionbuf[i]->bounds[0][Z] + primobject->collectionbuf[i]->bounds[4][Z]) / 2.0;

          trans_scales[X] = primobject->collectionbuf[i]->bounds[1][X] - primobject->collectionbuf[i]->bounds[0][X];
          trans_scales[Y] = primobject->collectionbuf[i]->bounds[2][Y] - primobject->collectionbuf[i]->bounds[0][Y];
          trans_scales[Z] = primobject->collectionbuf[i]->bounds[4][Z] - primobject->collectionbuf[i]->bounds[0][Z];

          assert(fabs(trans_scales[X]) > TINY);
          assert(fabs(primobject->collectionbuf[i]->bounds[1][Y] - primobject->collectionbuf[i]->bounds[0][Y]) < TINY);
          assert(fabs(primobject->collectionbuf[i]->bounds[1][Z] - primobject->collectionbuf[i]->bounds[0][Z]) < TINY);
          assert(fabs(trans_scales[Y]) > TINY);
          assert(fabs(primobject->collectionbuf[i]->bounds[2][X] - primobject->collectionbuf[i]->bounds[0][X]) < TINY);
          assert(fabs(primobject->collectionbuf[i]->bounds[2][Z] - primobject->collectionbuf[i]->bounds[0][Z]) < TINY);
          assert(fabs(trans_scales[Z]) > TINY);
          assert(fabs(primobject->collectionbuf[i]->bounds[4][X] - primobject->collectionbuf[i]->bounds[0][X]) < TINY);
          assert(fabs(primobject->collectionbuf[i]->bounds[4][Y] - primobject->collectionbuf[i]->bounds[0][Y]) < TINY);

          trans_scales[X] = 2.0 / trans_scales[X];
          trans_scales[Y] = 2.0  / trans_scales[Y];
          trans_scales[Z] = 2.0 / trans_scales[Z];

          for (int j = 0; j < 8; j++) {
            Vector::sub(collectionbox[j], trans_shift, transformedbox[j]);
            transformedbox[j][X] *= trans_scales[X] * (1 + TERMINAL_FACETOUCH_TOLERANCE);
            transformedbox[j][Y] *= trans_scales[Y] * (1 + TERMINAL_FACETOUCH_TOLERANCE);
            transformedbox[j][Z] *= trans_scales[Z] * (1 + TERMINAL_FACETOUCH_TOLERANCE);
          }

          // If there is an intersection, create a WorldCollection which is this Collection in the
          // worldspace of this primnode.
          if (cube_voxel_intersect(transformedbox) != OUT)
            add_collection(primobject->worldcollectionbuf[i], new_leaf);
        }
              
      } else {
        // We must create a new Collection for this terminal node of the octree.
        new_leaf->num_collections = 1;
        new_leaf->collection_bufsize = 1;
        new_leaf->collectionbuf = (leafcollx *) malloc(sizeof(leafcollx));
        if (!new_leaf->collectionbuf)
          fatal_error("SolidOctree::create_primitiveoctree",
                      "Can't allocate 1-entry collectionbuf!", "Out of memory.");

        new_leaf->collectionbuf[0].collection = new WorldCollection((primtreeprimnode *) represented_object, primobject,
                                                                    modeller->terminal_res, collectionspace, collectionbox,
                                                                    modeller);
        new_leaf->collectionbuf[0].splitptr = NULL;
        new_leaf->collectionbuf[0].collection->add_reference();

        // We have now created this WorldCollection. We must add its Collection to the primobject's collection
        // buffer so that future objects may use it.
        if (primobject->num_collections >= primobject->collection_bufsize) {
          // Grow the collection buffer.
          primobject->collection_bufsize += prim_collection_bufsize_increment;
          primobject->collectionbuf = (Collection **) realloc((void *) primobject->collectionbuf,
                                           primobject->collection_bufsize * sizeof(Collection *));
          primobject->worldcollectionbuf = (WorldCollection **) realloc((void *) primobject->worldcollectionbuf,
                                           primobject->collection_bufsize * sizeof(WorldCollection *));
        }

        // Note: add a reference as well for this reference (in the primobject collectionbuf).
        primobject->collectionbuf[primobject->num_collections] = new_leaf->collectionbuf[0].collection->collx;
        primobject->collectionbuf[primobject->num_collections]->add_reference();
        primobject->num_collections++;
      }
      
      new_index |= SOLIDINDEXTYPE_TERMINAL | SOLIDINDEX_CHANGED;
    } else {
      solidoctreenode *new_node;
      box childprimbox[8], childcollectionbox[8];
      int i;

      // Create an octree node.
      new_node = mm_octreenode->allocate(&new_index);
      
      // Subdivide the boxes. [This is a Mirage routine: subdivide.c]
      make_subvoxels(childprimbox, primbox);
      make_subvoxels(childcollectionbox, collectionbox);

      // Subdivide the primitive.
      for (i = 0; i < 8; i++)
        new_node->child[i] = create_primitiveoctree(use_collections, primobject,
                                                    collectionspace,
                                                    childprimbox[i],
                                                    childcollectionbox[i],
                                                    octree_level - 1,
                                                    ethertype, 8*terminal_id + i);

      new_index |= SOLIDINDEXTYPE_OCTREE | SOLIDINDEX_CHANGED;
    }

    return new_index;
  } else {
    // Never get here.
    assert(0);
    return SOLIDINDEXTYPE_EMPTY;
  }
}

// Combine two SolidOctree structures with a CSG operation. Used when
// creating a SolidOctree from scratch.
solidindex SolidOctree::create_csg_octree(linktype csg_type,
					  SolidOctree *left_owner,
					  solidindex *left,
					  SolidOctree *right_owner,
					  solidindex *right,
					  int *is_changed, int terminal_id)
{
  solidindex returnval = MAXINT;

  switch(csg_type) {
  case PLUS:
    if (SI_ISEMPTY(*left)) {
      if (SI_ISCHANGED(*left) || SI_ISCHANGED(*right)) {
        *is_changed = TRUE;
        returnval = duplicate_octree(right_owner, right, FALSE, SI_ISCHANGED(*left), TRUE,
                                     TRUE, left_owner, FALSE);
      } else {
        returnval = duplicate_octree(right_owner, right, FALSE, FALSE, TRUE,
                                     TRUE, left_owner, FALSE);
      }

    } else if (SI_ISEMPTY(*right)) {
      if (SI_ISCHANGED(*left) || SI_ISCHANGED(*right)) {
        *is_changed = TRUE;
        returnval = duplicate_octree(left_owner, left, FALSE, SI_ISCHANGED(*right), TRUE,
                                     TRUE, right_owner, TRUE);
      } else {
        returnval = duplicate_octree(left_owner, left, FALSE, FALSE, TRUE,
                                     TRUE, right_owner, TRUE);
      }

    } else if (SI_ISFULL(*left)) {
      
      if (SI_ISCHANGED(*left))
        *is_changed = TRUE;
      returnval = *left;

    } else if (SI_ISFULL(*right)) {

      if (SI_ISCHANGED(*right))
        *is_changed = TRUE;
      returnval = *right;        // NB: there is an inside-material
                                 // surface here, which Mirage would consider,
                                 // but which we discard.
    }

    // Other cases: left and right are both terminals, or both octrees,
    // in which case we fall out of switch statement and handle generically.
    break;

  case MINUS:
    if (SI_ISEMPTY(*left)) {

      if (SI_ISCHANGED(*left))
        *is_changed = TRUE;
      returnval = *left;

    } else if (SI_ISEMPTY(*right)) {
      if (SI_ISCHANGED(*left) || SI_ISCHANGED(*right)) {
        *is_changed = TRUE;
        returnval = duplicate_octree(left_owner, left, FALSE, SI_ISCHANGED(*right), TRUE,
                                     TRUE, right_owner, TRUE);
      } else {
        returnval = duplicate_octree(left_owner, left, FALSE, FALSE, TRUE,
                                     TRUE, right_owner, TRUE);
      }

    } else if (SI_ISFULL(*right)) {

      returnval = SOLIDINDEXTYPE_EMPTY;
      if (SI_ISCHANGED(*right)) {
        returnval |= SOLIDINDEX_CHANGED;
        *is_changed = TRUE;
      }

    } else if (SI_ISFULL(*left)) {
      // Surfaces in here are those of the right octree, but all surfaces
      // have normals reversed...
      if (SI_ISCHANGED(*left) || SI_ISCHANGED(*right)) {
        *is_changed = TRUE;
        returnval = duplicate_octree(right_owner, right, TRUE, SI_ISCHANGED(*left), TRUE,
                                     TRUE, left_owner, FALSE);
      } else {
        returnval = duplicate_octree(right_owner, right, TRUE, FALSE, TRUE,
                                     TRUE, left_owner, FALSE);
      }
    }

    // Other cases: left and right are both terminals, or both octrees,
    // in which case we fall out of switch statement and handle generically.
    break;

  case AND:
    if (SI_ISEMPTY(*left)) {

      if ((SI_ISCHANGED(*left)))
        *is_changed = TRUE;
      returnval = *left;

    } else if (SI_ISEMPTY(*right)) {

      if (SI_ISCHANGED(*right))
        *is_changed = TRUE;
      returnval = *right;

    } else if (SI_ISFULL(*left)) {
      // Surfaces in here are those of the right octree.
      if (SI_ISCHANGED(*left) || SI_ISCHANGED(*right)) {
        *is_changed = TRUE;
        returnval = duplicate_octree(right_owner, right, FALSE, SI_ISCHANGED(*left), TRUE,
                                     TRUE, left_owner, FALSE);
      } else {
        returnval = duplicate_octree(right_owner, right, FALSE, FALSE, TRUE,
                                     TRUE, left_owner, FALSE);
      }

    } else if (SI_ISFULL(*right)) {
      // Surfaces in here are those of the left octree.
      if (SI_ISCHANGED(*left) || SI_ISCHANGED(*right)) {
        *is_changed = TRUE;
        returnval = duplicate_octree(left_owner, left, FALSE, SI_ISCHANGED(*right), TRUE,
                                     TRUE, right_owner, TRUE);
      } else {
        returnval = duplicate_octree(left_owner, left, FALSE, FALSE, TRUE,
                                     TRUE, right_owner, TRUE);
      }
    }

    // Other cases: left and right are both terminals, or both octrees,
    // in which case we fall out of switch statement and handle generically.
    break;

  default:
    // No other linktypes are implemented in the modeller, as there is
    // no concept of materials.
    assert(0);
  }

  *left  &= ~SOLIDINDEX_CHANGED;
  *right &= ~SOLIDINDEX_CHANGED;

  // Have we found an answer yet?
  if (returnval != MAXINT)
    return returnval;

  // We have fallen out of the above switch statement, so either
  // both sides are terminals, or both sides are octrees.
  if (SI_ISTERMINAL(*left)) {
    solidterminal *left_term, *right_term, *new_node;
    solidrcsgnode *new_rcsg;
    solidindex new_index;

    assert(SI_ISTERMINAL(*right));
    
    // Get the terminal node pointers.
    left_term =  left_owner->mm_terminal->retrieve_element(SI_GETDATA(*left));
    right_term= right_owner->mm_terminal->retrieve_element(SI_GETDATA(*right));

    assert(left_term ->terminal_id == terminal_id);
    assert(right_term->terminal_id == terminal_id);

    // Create a new solidterminal for us, and a solidrcsgnode too.
    new_node = mm_terminal->allocate(&new_index);

    new_rcsg = mm_rcsgnode->allocate();

    // Set up the terminal.
    new_node->rcsggraph.flags = 0;
    new_node->rcsggraph.rcsg = new_rcsg;
    new_node->terminal_id = terminal_id;

    new_rcsg->link = csg_type;

    // Create the RCSG graph that is the CSG combination of these two child
    // terminals' RCSG graphs.
    create_linked_rcsg(new_node, new_rcsg, left_owner, left_term, right_owner, right_term);

    // Return the new terminal.
    new_index |= SOLIDINDEXTYPE_TERMINAL | SOLIDINDEX_CHANGED;
    *is_changed = TRUE;
    return new_index;

  } else {
    // Both sides are octrees.
    solidoctreenode *left_oct, *right_oct, *new_node;
    solidindex new_index;
    int i, children_changed = FALSE;

    ASSERT(SI_ISOCTREE(*left));
    ASSERT(SI_ISOCTREE(*right));
    left_oct =left_owner ->mm_octreenode->retrieve_element(SI_GETDATA(*left));
    right_oct=right_owner->mm_octreenode->retrieve_element(SI_GETDATA(*right));

    // Create a new octree for this combination.
    new_node = mm_octreenode->allocate(&new_index);

    for (i = 0; i < 8; i++)
      new_node->child[i] = create_csg_octree(csg_type, 
					     left_owner,  
					     &(left_oct ->child[i]),
					     right_owner, 
					     &(right_oct->child[i]),
					     &children_changed, 8*terminal_id + i);

    // Something to think about... should we keep a track of the number of 
    // empty / full results and collapse the octree to a single EMPTY or FULL
    // as a result?

    new_index |= SOLIDINDEXTYPE_OCTREE;
    if (children_changed) {
      new_index |= SOLIDINDEX_CHANGED;
      *is_changed = TRUE;
    }

    return new_index;
  }
}

// Combine two RCSG graphs with a CSG operation. Registers that all of the
// Collections in the rcsggraphs need to be split.
void SolidOctree::create_linked_rcsg(solidterminal *our_terminal, solidrcsgnode *new_rcsgnode,
                                     SolidOctree *left_owner,  solidterminal *left_term,
                                     SolidOctree *right_owner, solidterminal *right_term)
{
  // The left surface is never flipped.
  new_rcsgnode->left  = create_shell_rcsg(our_terminal, left_owner, left_term->rcsggraph, left_in_is_active,
                                          FALSE, right_term, right_owner, TRUE);

  // The right surace is flipped if the linktype is MINUS.
  new_rcsgnode->right = create_shell_rcsg(our_terminal, right_owner, right_term->rcsggraph, right_in_is_active, 
                                          (new_rcsgnode->link == MINUS), left_term, left_owner, FALSE);
}

// Create a shell RCSG (all leafnodes empty) and register that all the
// leafnodes will need to have split child collections added to them.
solidrcsggraph SolidOctree::create_shell_rcsg(solidterminal *our_terminal, SolidOctree *owner_octree, 
                                              solidrcsggraph original,
                                              int inside_child_active,
                                              int is_flipped, solidterminal *split_vs,
                                              SolidOctree *opponent_octree, int is_left)
{
  solidrcsggraph returnval;
  int i;

  if (RCSG_ISFLIPPED(original.flags))
    is_flipped = !is_flipped;

  if (RCSG_ISLEAF(original.flags)) {
    // Make a shell leafnode. It has no Collections in it.
    solidleafnode *new_leaf = mm_leafnode->allocate();

    new_leaf->is_deleted = FALSE;
    new_leaf->prim = original.leaf->prim;
    for (i = 0; i < 8; i++)
      Vector::set(new_leaf->terminalbox[i], original.leaf->terminalbox[i]);

    new_leaf->num_collections = 0;
    new_leaf->collection_bufsize = 0;
    new_leaf->collectionbuf = NULL;

    // All of the WorldCollections in the original leafnode will
    // need to be split against the split_vs terminal.
    // Their children will then be inserted into new_leaf.
    for (i = 0; i < original.leaf->num_collections; i++) {
      assert(original.leaf->collectionbuf[i].collection->owner == new_leaf->prim);

      // Register that this collection is here being SPLIT (opposing_terminal != NULL)
      splittracker->register_terminal(original.leaf->collectionbuf[i].collection,
                                      is_flipped, this, split_vs->terminal_id, our_terminal, new_leaf,
                                      opponent_octree, is_left, split_vs);
    }

    returnval.leaf = new_leaf;

  } else {
    // Duplicate the rcsgnode and make two shell children.
    solidrcsgnode *new_rcsg = mm_rcsgnode->allocate();

    new_rcsg->link = original.rcsg->link;
    new_rcsg->left  = create_shell_rcsg(our_terminal, owner_octree, original.rcsg->left,
                                        inside_child_active, is_flipped, split_vs,
                                        opponent_octree, is_left);

    // Subtracted surfaces are flipped.
    if (new_rcsg->link == MINUS)
      is_flipped = !is_flipped;

    new_rcsg->right = create_shell_rcsg(our_terminal, owner_octree, original.rcsg->right,
                                        inside_child_active, is_flipped, split_vs,
                                        opponent_octree, is_left);
    returnval.rcsg = new_rcsg;
  }

  returnval.flags = original.flags;

  return returnval;
}


// Set everything nonequal to the value in question to changed.
// This value is either SOLIDINDEXTYPE_EMPTY or SOLIDINDEXTYPE_FULL.
void SolidOctree::set_all_nonequal_to_changed(solidindex equal_value)
{
  int dummy;

  recursive_set_all_nonequal_to_changed(&octree, this, equal_value, &dummy);
}

// Set everything nonequal to specified octree to changed.
// Calls recursive_set_all_nonequal_to_changed.
void SolidOctree::set_all_nonequal_to_changed(SolidOctree *equal_oct)
{
  int dummy;

  recursive_set_all_nonequal_to_changed(&octree, equal_oct, equal_oct->octree, &dummy);
}

// Set everything nonequal to the value in question to be changed.
// This value is either SOLIDINDEXTYPE_EMPTY or SOLIDINDEXTYPE_FULL.
void SolidOctree::recursive_set_all_nonequal_to_changed(solidindex *oct,
						       SolidOctree *equal_octree, solidindex equal_value,
						       int *is_changed)
{
  assert(oct != NULL);

  if (SI_ISEMPTY(*oct)) {
    if (SI_ISEMPTY(equal_value)) {
      *oct = SOLIDINDEXTYPE_EMPTY;
    } else {
      *oct = SOLIDINDEXTYPE_EMPTY | SOLIDINDEX_CHANGED;
      *is_changed = TRUE;
    }

  } else if (SI_ISFULL(*oct)) {
    if (SI_ISFULL(equal_value)) {
      *oct = SOLIDINDEXTYPE_FULL;
    } else {
      *oct = SOLIDINDEXTYPE_FULL | SOLIDINDEX_CHANGED;
      *is_changed = TRUE;
    }

  } else if (SI_ISTERMINAL(*oct)) {

    *oct |= SOLIDINDEX_CHANGED;
    *is_changed = TRUE;
    
  } else if (SI_ISOCTREE(*oct)) {
    solidoctreenode *oct_node = mm_octreenode->
      retrieve_element(SI_GETDATA(*oct)), *equaloct_node;
    solidindex childequal_val;
    int children_changed = FALSE;

    if (SI_ISOCTREE(equal_value)) {
      equaloct_node = equal_octree->mm_octreenode->retrieve_element(SI_GETDATA(equal_value));
      childequal_val = SOLIDINDEXTYPE_EMPTY;         // Will be set for every child.
    } else {
      // We are at a leafnode, test against the same value for each
      // child.
      equaloct_node = NULL;
      childequal_val = equal_value;
    }

    for (int i = 0; i < 8; i++) {
      if (equaloct_node)
        childequal_val = equaloct_node->child[i];

      recursive_set_all_nonequal_to_changed(&(oct_node->child[i]), 
					    equal_octree, childequal_val, &children_changed);
    }
    
    if (children_changed) {
      *oct |= SOLIDINDEX_CHANGED;
      *is_changed = TRUE;
    }

  } else {
    // Never get here.
    assert(0);
  }
}

// Update the SolidOctree for a primitive, after its position changes.
void SolidOctree::new_prim_position(void)
{
  box worldspace_box, primspace_box, collectionspace_box;
  primobjectnode *primobject;
  primtreeprimnode *primptr;
  int children_changed = 0, i;

  // Form the corners of the modeller's bounding box.
  modeller->get_bounds(worldspace_box);

  // Find our primobjectnode and primtreeprimnode.
  assert(represented_object->original->type() == OBJECTNODE_PRIM);
  assert(represented_object->type() == PRIMTREENODE_PRIM);
  primobject = (primobjectnode *) represented_object->original;
  primptr = (primtreeprimnode *) represented_object;
  
  // Transform the corners of the bounding box into primitivespace,
  // to form primspace_box. Also find collectionspace for this primobject.
  for (i = 0; i < 8; i++) {
    Matrix::apply(primptr->position[AFT],
                  worldspace_box[i], primspace_box[i]);
    Matrix::apply(primptr->collection_trans[AFT],
                  worldspace_box[i], collectionspace_box[i]);
  }  

  // Set up the new worldcollections for this primitive.
  for (i = 0; i < primobject->num_collections; i++) {
    primobject->worldcollectionbuf[i] = new WorldCollection(primptr, primobject->collectionbuf[i]);
    primobject->worldcollectionbuf[i]->add_reference();
  }

  // Recursively update the primitive's octree.
  octree = update_primitiveoctree(octree, primobject,
                                  primspace_box, collectionspace_box,
                                  modeller->octree_level, &children_changed, 0);

  // Clear out the worldcollectionbuf.
  for (i = 0; i < primobject->num_collections; i++) {
    if (primobject->worldcollectionbuf[i]->remove_reference())
      delete primobject->worldcollectionbuf[i];
    primobject->worldcollectionbuf[i] = NULL;
  }

  if (children_changed > 0)
    octree |= SOLIDINDEX_CHANGED;
}

// Used for printing out the octree.
char SolidOctree::get_char(solidindex oct, int levels, int x, int y, int z)
{
  int childindex, newx, newy, newz;
  int res = (1 << levels), res2 = res / 2;

  if (SI_ISEMPTY(oct)) {
    if ((x == 0) && (y == 0) && (z == 0))
      if (SI_ISCHANGED(oct))
	return '>';
      else
	return '.';
    else
      if (SI_ISCHANGED(oct))
	return '-';
      else
	return ',';
  } else if (SI_ISFULL(oct)) {
    if ((x == 0) && (y == 0) && (z == 0))
      if (SI_ISCHANGED(oct))
	return 'F';
      else
	return 'f';
    else
      if (SI_ISCHANGED(oct))
	return 'G';
      else
	return 'g';
  } else if (SI_ISTERMINAL(oct)) {
    if ((x == 0) && (y == 0) && (z == 0))
      if (SI_ISCHANGED(oct))
	return 'L';
      else
	return 'l';
    else
      if (SI_ISCHANGED(oct))
	return 'T';
      else
	return 't';
  } else if (SI_ISOCTREE(oct)) {
    if (levels == 0)
      if (SI_ISCHANGED(oct))
	return 'O';
      else
	return 'o';

    solidoctreenode *oct_node = mm_octreenode->
      retrieve_element(SI_GETDATA(oct));

    childindex = ((x / res2) ? 1 : 0)
      + ((y / res2) ? 2 : 0)
      + ((z / res2) ? 4 : 0);

    newx = x % res2;
    newy = y % res2;
    newz = z % res2;
    
    return get_char(oct_node->child[childindex], levels - 1, newx, newy, newz);
  } else {
    return '?';
  } 
}

// Print out the octree.
void SolidOctree::print(int indent, int details, solidindex oct)
{
  int levels, res, x, y, z, copy_param;

  if (details == -1)
    levels = modeller->octree_level;
  else
    levels = details;

  if (oct == MAXINT)
    oct = octree;

  res = 1 << levels;

  TRACE("** Begin SolidOctree print:\n");
  for (z = 0; z < res; z++) {
    for (y = 0; y < res; y++) {
      TRACE("%*s", indent, "");

      for (copy_param = 0; copy_param < res; copy_param++) {

	if ((y < (res >> copy_param)) &&
	    ((z & ((1 << copy_param) - 1)) == 0)) {

	  for (x = 0; x < (res >> copy_param); x++) {
	    TRACE("%c", get_char(oct, levels - copy_param, x, y, z >> copy_param));
	  }

	  TRACE(" ");
	}
      }

      TRACE("\n");
    }
    TRACE("\n");
  }
  TRACE("** End SolidOctree print\n");
}

// Debug the octree.
void SolidOctree::debug_octree(void)
{
  printf("** Debug SolidOctree\n");
  debug_solidindex(octree);
  printf("** End Debug SolidOctree\n");
}

// Debug the octree.
void SolidOctree::debug_solidindex(solidindex oct)
{
  if (SI_ISEMPTY(oct))
    TRACE("Empty\n");
  else if (SI_ISFULL(oct))
    TRACE("Full\n");
  else if (SI_ISTERMINAL(oct)) {
    TRACE("Terminal\n");
  } else if (SI_ISOCTREE(oct)) {
    solidoctreenode *oct_node = mm_octreenode->retrieve_element(SI_GETDATA(oct));
    TRACE("Octree...\n");
    for (int i = 0; i < 8; i++) {
      TRACE("Child %d:\n", i);
      debug_solidindex(oct_node->child[i]);
    }
  }
}

solidindex SolidOctree::update_primitiveoctree(solidindex old_octree, primobjectnode *primobject,
                                               box primbox, box collectionbox, int level,
                                               int *child_changed, int terminal_id,
                                               intersect_result known_result)
{
  intersect_result prim_result;
  solidindex new_index;
  int i;

  // First: test the box versus the Primitive. If this == IN or OUT, then
  // we already know what the result will be, and this is just a
  // check to see which areas need to be flagged as changed.
  if (known_result == SURFACE)
    prim_result = primobject->prim->voxel_intersect(primbox);
  else
    prim_result = known_result;

  if (prim_result == IN) {
    // This box is FULL of the primitive.

    // If we were already full, just return full.
    if (SI_ISFULL(old_octree)) {
      return SOLIDINDEXTYPE_FULL;

    } else if (SI_ISEMPTY(old_octree)) {
      // This has changed.
      (*child_changed)++;
      return SOLIDINDEXTYPE_FULL | SOLIDINDEX_CHANGED;

    } else if (SI_ISTERMINAL(old_octree)) {
      // This has changed.
      delete_solidindex(old_octree);
      (*child_changed)++;
      return SOLIDINDEXTYPE_FULL | SOLIDINDEX_CHANGED;

    } else {
      // Otherwise, we were an octree. Check each side to see if it changes.
      solidoctreenode *old_node = mm_octreenode->retrieve_element(
						SI_GETDATA(old_octree));
      int children_changed = 0;
      box childprimbox[8], childcollectionbox[8];

      // Subdivide the box. [This is a Mirage routine: subdivide.c]
      make_subvoxels(childprimbox, primbox);
      make_subvoxels(childcollectionbox, collectionbox);

      // Check each child to see if it was not-IN and thus changes.
      for (int i = 0; i < 8; i++)
        old_node->child[i] = update_primitiveoctree(old_node->child[i], primobject,
                                                    childprimbox[i],
                                                    childcollectionbox[i],
                                                    level - 1, &children_changed,
                                                    8*terminal_id + i, IN);

      if (children_changed == 0) {
        // This is FULL and not changed.
        mm_octreenode->deallocate(old_node, SI_GETDATA(old_octree));
        return SOLIDINDEXTYPE_FULL;

      } else if (children_changed == 8) {
        // This is FULL and has changed in each octant.
        mm_octreenode->deallocate(old_node, SI_GETDATA(old_octree));
        (*child_changed)++;
        return SOLIDINDEXTYPE_FULL | SOLIDINDEX_CHANGED;

      } else {
        // This has changed; it's not a simple EMPTY or FULL though,
        // but needs to be represented by an octree. So increment by
        // 100 so that the total won't be exactly 0 or 8.
        (*child_changed) += 100;

        return old_octree | SOLIDINDEX_CHANGED;
      }
    }

  } else if (prim_result == OUT) {
    // This box is EMPTY of the primitive.

    // If we were already empty, just return empty.
    if (SI_ISEMPTY(old_octree)) {
      return SOLIDINDEXTYPE_EMPTY;

    } else if (SI_ISFULL(old_octree)) {
      // This was full, now it is empty.
      (*child_changed)++;
      return SOLIDINDEXTYPE_EMPTY | SOLIDINDEX_CHANGED;

    } else if (SI_ISTERMINAL(old_octree)) {
      // This has changed from a terminal to empty.
      delete_solidindex(old_octree);
      (*child_changed)++;
      return SOLIDINDEXTYPE_EMPTY | SOLIDINDEX_CHANGED;

    } else {
      // Otherwise, we were an octree. Check each side to see if it changes.
      solidoctreenode *old_node = mm_octreenode->retrieve_element(
						SI_GETDATA(old_octree));
      int children_changed = 0;
      box childprimbox[8], childcollectionbox[8];

      // Subdivide the box. [This is a Mirage routine: subdivide.c]
      make_subvoxels(childprimbox, primbox);
      make_subvoxels(childcollectionbox, collectionbox);
	
      // Check each child to see if it changes.
      for (int i = 0; i < 8; i++)
        old_node->child[i] = update_primitiveoctree(old_node->child[i], primobject,
                                                    childprimbox[i],
                                                    childcollectionbox[i],
                                                    level - 1, &children_changed,
                                                    8*terminal_id + i, OUT);

      if (children_changed == 0) {
        // This is EMPTY and not changed.
        mm_octreenode->deallocate(old_node, SI_GETDATA(old_octree));
        return SOLIDINDEXTYPE_EMPTY;

      } else if (children_changed == 8) {
      	// This is EMPTY and has changed in each octant.
      	mm_octreenode->deallocate(old_node, SI_GETDATA(old_octree));
      	(*child_changed)++;
      	return SOLIDINDEXTYPE_EMPTY | SOLIDINDEX_CHANGED;

      } else {
        // This has changed; it's not a simple EMPTY or FULL though,
        // but needs to be represented by an octree. So increment by
        // 100 so that the total won't be exactly 0 or 8.
        (*child_changed) += 100;
  
        return old_octree | SOLIDINDEX_CHANGED;
      }
    }


  } else if (prim_result == SURFACE) {
    // Should this box be subdivided?
    if (level == 0) {
      // We have to create a new terminal node.
      solidterminal *new_node = mm_terminal->allocate(&new_index);
      solidleafnode *new_leaf = mm_leafnode->allocate();
      box transformedbox;
      vector trans_shift, trans_scales;
      int i;

#if 0
      if (((this->represented_object->index == 63) ||
           (this->represented_object->index == 65) ||
           (this->represented_object->index == 66)) &&
          (terminal_id == 392))
            _CrtDbgBreak();
#endif

      new_node->rcsggraph.flags = RCSGGRAPH_LEAF_MASK;
      new_node->rcsggraph.leaf = new_leaf;
      new_node->terminal_id = terminal_id;

      new_leaf->is_deleted = FALSE;
      new_leaf->prim = (primtreeprimnode *) represented_object;
      for (i = 0; i < 8; i++)
        Vector::set(new_leaf->terminalbox[i], collectionbox[i]);

      // We have a number of Collections, each of which has collection->bounds
      // in collectionspace. And we also have the bounds of this leafnode in
      // collectionspace, as collectionbox. Intersect these. Any Collections that
      // intersect this leafnode must be placed here.
      new_leaf->num_collections = new_leaf->collection_bufsize = 0;
      new_leaf->collectionbuf = NULL;

      for (i = 0; i < primobject->num_collections; i++) {
        // Transform collectionbox into a space where the
        // bounds of the Collection lie at -1, +1 in X, Y and Z.
        //
        // Note that because we want the transformedbox to return OUT even
        // if it lies with an adjacent face, we actually scale the transformedbox
        // so that the Collection's boundaries lie at 1 + TERMINAL_FACETOUCH_TOLERANCE.

        trans_shift[X] = (primobject->collectionbuf[i]->bounds[0][X] + primobject->collectionbuf[i]->bounds[1][X]) / 2.0;
        trans_shift[Y] = (primobject->collectionbuf[i]->bounds[0][Y] + primobject->collectionbuf[i]->bounds[2][Y]) / 2.0;
        trans_shift[Z] = (primobject->collectionbuf[i]->bounds[0][Z] + primobject->collectionbuf[i]->bounds[4][Z]) / 2.0;

        trans_scales[X] = primobject->collectionbuf[i]->bounds[1][X] - primobject->collectionbuf[i]->bounds[0][X];
        trans_scales[Y] = primobject->collectionbuf[i]->bounds[2][Y] - primobject->collectionbuf[i]->bounds[0][Y];
        trans_scales[Z] = primobject->collectionbuf[i]->bounds[4][Z] - primobject->collectionbuf[i]->bounds[0][Z];

        assert(fabs(trans_scales[X]) > TINY);
        assert(fabs(primobject->collectionbuf[i]->bounds[1][Y] - primobject->collectionbuf[i]->bounds[0][Y]) < TINY);
        assert(fabs(primobject->collectionbuf[i]->bounds[1][Z] - primobject->collectionbuf[i]->bounds[0][Z]) < TINY);
        assert(fabs(trans_scales[Y]) > TINY);
        assert(fabs(primobject->collectionbuf[i]->bounds[2][X] - primobject->collectionbuf[i]->bounds[0][X]) < TINY);
        assert(fabs(primobject->collectionbuf[i]->bounds[2][Z] - primobject->collectionbuf[i]->bounds[0][Z]) < TINY);
        assert(fabs(trans_scales[Z]) > TINY);
        assert(fabs(primobject->collectionbuf[i]->bounds[4][X] - primobject->collectionbuf[i]->bounds[0][X]) < TINY);
        assert(fabs(primobject->collectionbuf[i]->bounds[4][Y] - primobject->collectionbuf[i]->bounds[0][Y]) < TINY);

        trans_scales[X] = 2.0 / trans_scales[X];
        trans_scales[Y] = 2.0 / trans_scales[Y];
        trans_scales[Z] = 2.0 / trans_scales[Z];

        for (int j = 0; j < 8; j++) {
          Vector::sub(collectionbox[j], trans_shift, transformedbox[j]);
          transformedbox[j][X] *= trans_scales[X] * (1 + TERMINAL_FACETOUCH_TOLERANCE);
          transformedbox[j][Y] *= trans_scales[Y] * (1 + TERMINAL_FACETOUCH_TOLERANCE);
          transformedbox[j][Z] *= trans_scales[Z] * (1 + TERMINAL_FACETOUCH_TOLERANCE);
        }

        // If there is an intersection, add the WorldCollection which is this Collection in the
        // worldspace of this primnode.
        if (cube_voxel_intersect(transformedbox) != OUT)
          add_collection(primobject->worldcollectionbuf[i], new_leaf);
      }

      // NB: We can't delete the old octree until after we have copied the Collections
      // and added another reference to them. Otherwise they might get deleted because
      // their reference count hit zero, etc.
       
      if (SI_ISTERMINAL(old_octree)) {
        // The old terminal must be deleted.
        delete_solidindex(old_octree);
      }

      (*child_changed)++;

      return (new_index | SOLIDINDEXTYPE_TERMINAL | SOLIDINDEX_CHANGED);

    } else {
      // This should be an octree node.
      if (SI_ISOCTREE(old_octree)) {
        box childprimbox[8], childcollectionbox[8];
        int children_changed = 0;

        // It was last time as well; get the octreenode.
        solidoctreenode *old_node = mm_octreenode->
          retrieve_element(SI_GETDATA(old_octree));

        // Subdivide the box. [This is a Mirage routine: subdivide.c]
        make_subvoxels(childprimbox, primbox);
        make_subvoxels(childcollectionbox, collectionbox);
	
        // Subdivide the primitive. For each cell in the octree, look at
        // whether or not it needs to be changed.
        for (i = 0; i < 8; i++)
          old_node->child[i] = update_primitiveoctree(old_node->child[i], primobject,
                                                      childprimbox[i],
                                                      childcollectionbox[i],
                                                      level - 1, &children_changed,
                                                      8*terminal_id + i);
        // Return the octree.
        if (children_changed) {
          (*child_changed)++;
          return old_octree | SOLIDINDEX_CHANGED;

        } else {
          return old_octree & ~SOLIDINDEX_CHANGED;
        }

      } else {
        // This wasn't an octree. Thus, we will need to completely
        // rebuild this section of the octree.	
        solidoctreenode *new_node;
        box childprimbox[8], childcollectionbox[8];
        solidindex ethertype;

        // So, it wasn't an octree. It can't be a terminal node, as levels > 0.
        // Therefore it must be eithe empty or full.
        assert(!SI_ISTERMINAL(old_octree));
        if (SI_ISEMPTY(old_octree))
          ethertype = SOLIDINDEXTYPE_EMPTY;
        else if (SI_ISFULL(old_octree))
          ethertype = SOLIDINDEXTYPE_FULL;
        else
          assert(0);

        // Create an octree node.
        new_node = mm_octreenode->allocate(&new_index);
      
        // Subdivide the box. [This is a Mirage routine: subdivide.c]
        make_subvoxels(childprimbox, primbox);
        make_subvoxels(childcollectionbox, collectionbox);

        // Subdivide the primitive. Set the primitive to be changed in those
        // areas where it's not empty or full according to the previous
        // ethertype result.
        //
        // This routine requires a collectionspace transformation, which is
        // only used when use_collections is true. We just use a dummy transform
        // that is never accessed.
        transform dummytrans;

        for (i = 0; i < 8; i++)
          new_node->child[i] = create_primitiveoctree(TRUE, primobject,
                                                      dummytrans, childprimbox[i],
                                                      childcollectionbox[i],
				                                              level - 1, ethertype, 8*terminal_id + i);

        (*child_changed)++;
        return (new_index | SOLIDINDEXTYPE_OCTREE | SOLIDINDEX_CHANGED);
      }
    }
  } else {
    // Unknown prim_result! Should never get here.
    assert(0);
    return SOLIDINDEXTYPE_EMPTY;
  }
}

// Delete a portion of the octree.
void SolidOctree::delete_solidindex(solidindex oct)
{
  if ((SI_ISEMPTY(oct)) || (SI_ISFULL(oct)))
    return;

  if (SI_ISOCTREE(oct)) {
    solidoctreenode *node = mm_octreenode->retrieve_element(SI_GETDATA(oct));

    for (int i = 0; i < 8; i++)
      delete_solidindex(node->child[i]);

    mm_octreenode->deallocate(node, SI_GETDATA(oct));

  } else if (SI_ISTERMINAL(oct)) {
    solidterminal *term = mm_terminal->retrieve_element(SI_GETDATA(oct));

    delete_rcsggraph(term->rcsggraph, term);

    mm_terminal->deallocate(term, SI_GETDATA(oct));

  } else {
    // Unknown kind of index.
    assert(0);
  }
}

// Delete a RCSG graph from the octree.
void SolidOctree::delete_rcsggraph(solidrcsggraph kill_me,
                                   solidterminal *our_terminal)
{
  if (RCSG_ISLEAF(kill_me.flags)) {
    int i, j;
    splittrackernode *splitptr;

    assert(kill_me.leaf->is_deleted == FALSE);

    // These Collections are no longer referred to here.
    for (i = 0; i < kill_me.leaf->num_collections; i++) {
      // Was this Collection being tracked by our splittracker?
      splitptr = kill_me.leaf->collectionbuf[i].splitptr;
      if (splitptr) {
        assert(splittracker != NULL);

        if (splitptr == debug_splittrackernode)
          _CrtDbgBreak();

        for (j = 0; j < splitptr->num_terminals; j++) {
          if (splitptr->terminalbuf[j].is_deleted)
            continue;

          if (splitptr->terminalbuf[j].our_terminal == our_terminal) {
            if (splitptr == debug_splittrackernode)
              _CrtDbgBreak();

            splitptr->terminalbuf[j].is_deleted = TRUE;        
            splitptr->num_deleted_terminals++;
          }
        }
      }

      // Remove the reference from the Collection. If this was
      // the last reference, it returns TRUE so we delete it.
      if (kill_me.leaf->collectionbuf[i].collection->remove_reference())
        delete kill_me.leaf->collectionbuf[i].collection;
    }

    // Free the Collection array.
    free(kill_me.leaf->collectionbuf);

    kill_me.leaf->is_deleted = TRUE;

    // Delete the leafnode.
    mm_leafnode->deallocate(kill_me.leaf);

  } else {
    delete_rcsggraph(kill_me.rcsg->left, our_terminal);
    delete_rcsggraph(kill_me.rcsg->right, our_terminal);

    // Delete the RCSGnode.
    mm_rcsgnode->deallocate(kill_me.rcsg);
  }
}

extern int debug_incidentstatus;

// Perform an incremental update of the SolidOctree, used during
// solidoctree_pass.
void SolidOctree::incremental_update(linktype csg_type, SolidOctree *left,
				     SolidOctree *right)
{
  int dummy;
#if 0
  int dump_left, dump_right;
#endif

  assert(splittracker != NULL);

//  if (debug_incidentstatus == 1)
//    splittracker->print();

  // Register with the splittracker that we are starting a new update pass.
  splittracker->start_new_update();

  assert(left != NULL);
  assert(right != NULL);

  if (represented_object->index == debug_octree_print) {
    _CrtDbgBreak();
    TRACE("left\n");
    left->print(0);
    TRACE("right\n");
    right->print(0);
    TRACE("old\n");
    this->print(0);
  }

#if 0
  // If our CSG child octrees have totally changed since the last update,
  // then we will need to dump all splittrackernodes related to them.
  //
  // This is the case if the octrees themselves change (rare) or if
  // they are primitive octrees that have had a new position set (not implemented yet).
  dump_left  = (left_octree  != left);
  dump_right = (right_octree != right);

  if (left->primitive_changed_this_update) {
    dump_left = TRUE;
    left->primitive_changed_this_update = FALSE;
  }
  if (right->primitive_changed_this_update) {
    dump_right = TRUE;
    right->primitive_changed_this_update = FALSE;
  }

  splittracker->dump_all_nodes(dump_left, dump_right);
#endif

  left_octree = left;
  right_octree = right;
  
  switch(csg_type) {
  case PLUS:
    // The OUT sections of both objects are active.
    left_in_is_active = right_in_is_active = FALSE;
    break;
  case MINUS:
    // The OUT section of left and the IN section of right are active.
    left_in_is_active = FALSE;
    right_in_is_active = TRUE;
    break;
  case AND:
    // The IN sections of both objects are active.
    left_in_is_active = right_in_is_active = TRUE;
    break;
  default:
    assert(0);
  }

  if (represented_object->index == debug_octree_print)
    debug_find_collx(640, octree);

  octree = update_octree(octree, csg_type, modeller->octree_level,
			 left,  &(left->octree),
			 right, &(right->octree), &dummy, 0);

  if (represented_object->index == debug_octree_print) {
    TRACE("new\n");
    print(0);
  }

//  if (debug_incidentstatus == 1) {
//    TRACE("PRINTING SPLITTRACKER AFTER UPDATE\n");
//    splittracker->print();
//    _CrtDbgBreak();
//  }

  // Perform all splits that are necessary, and create the new legacy nodes.
  splittracker->go_split();

  if (represented_object->index == debug_octree_print)
    debug_find_collx(640, octree);

//  if (debug_incidentstatus == 1)
//    splittracker->print();
}

// Perform an incremental update of a portion of the octree structure.
solidindex SolidOctree::update_octree(solidindex current_octree,
				      linktype csg_type, int levels,
				      SolidOctree *left_owner,
				      solidindex *left,
				      SolidOctree *right_owner,
				      solidindex *right,
				      int *is_changed, int terminal_id)
{
  solidindex returnval = MAXINT;

  // NB: all calls to duplicate_octree must store in the splittracker
  // instead! If they do not then we get wacky ghosts appearing!

  if ((!SI_ISCHANGED(*left)) && (!SI_ISCHANGED(*right))) {
    int fubar = 0;

    if (fubar)
      debug_find_collx(376, current_octree);

    clear_changed(this, &current_octree);

    return current_octree;
  }

  switch(csg_type) {
  case PLUS:
    if (SI_ISEMPTY(*left)) {
      // If this left EMPTY condition is new, then this is changed.
      // Otherwise, just return the right side, with its changed flags
      // intact.
      delete_solidindex(current_octree);

      *is_changed = TRUE;
      returnval = duplicate_octree(right_owner, right,
                                   FALSE, SI_ISCHANGED(*left), TRUE,
                                   TRUE, left_owner, FALSE);

    } else if (SI_ISEMPTY(*right)) {
      // If this right EMPTY condition is new, then this is changed.
      // Otherwise, just return the left side, with its changed flags
      // intact.
      delete_solidindex(current_octree);

      *is_changed = TRUE;
      returnval = duplicate_octree(left_owner, left, 
 				                           FALSE, SI_ISCHANGED(*right), TRUE,
                                   TRUE, right_owner, TRUE);

    } else if (SI_ISFULL(*left)) {
      delete_solidindex(current_octree);

      clear_changed(right_owner, right);

      returnval = *left;

    } else if (SI_ISFULL(*right)) {
      delete_solidindex(current_octree);

      clear_changed(left_owner, left);

      returnval = *right;
    }

    // Other cases: left and right are both terminals, or both
    // octrees, in which case we fall out of switch statement
    // and handle generically.
    break;


  case MINUS:
    if (SI_ISEMPTY(*left)) {
      delete_solidindex(current_octree);

      clear_changed(right_owner, right);

      returnval = *left;

    } else if (SI_ISEMPTY(*right)) {
      // If this right EMPTY condition is new, then this is changed.
      // Otherwise, just return the left side, with its changed flags
      // intact.
      delete_solidindex(current_octree);

      *is_changed = TRUE;
      returnval = duplicate_octree(left_owner, left, 
				                           FALSE, SI_ISCHANGED(*right), TRUE,
                                   TRUE, right_owner, TRUE);

    } else if (SI_ISFULL(*left)) {
      // If this left FULL condition is new, then this is changed. 
      // Otherwise, just return the right side inverted, with its
      // changed flags intact.
      delete_solidindex(current_octree);

      *is_changed = TRUE;
      returnval = duplicate_octree(right_owner, right,
                                   TRUE, SI_ISCHANGED(*left), TRUE,
                                   TRUE, left_owner, FALSE);

    } else if (SI_ISFULL(*right)) {
      delete_solidindex(current_octree);

      clear_changed(left_owner, left);

      // Return EMPTY.
      returnval = SOLIDINDEXTYPE_EMPTY;
      if (SI_ISCHANGED(*right))
        returnval |= SOLIDINDEX_CHANGED;
    }

    // Other cases: left and right are both terminals, or both
    // octrees, in which case we fall out of switch statement
    // and handle generically.
    break;


  case AND:
    if (SI_ISEMPTY(*left)) {
      delete_solidindex(current_octree);

      clear_changed(right_owner, right);

      returnval = *left;

    } else if (SI_ISEMPTY(*right)) {
      delete_solidindex(current_octree);

      clear_changed(left_owner, left);

      returnval = *right;

    } else if (SI_ISFULL(*left)) {
      // If this left FULL condition is new, then this is changed. 
      // Otherwise, just return the right side, with its
      // changed flags intact.
      delete_solidindex(current_octree);

      *is_changed = TRUE;
      returnval = duplicate_octree(right_owner, right,
                                   FALSE, SI_ISCHANGED(*left), TRUE,
                                   TRUE, left_owner, FALSE);

    } else if (SI_ISFULL(*right)) {
      // If this right FULL condition is new, then this is changed. 
      // Otherwise, just return the left side, with its
      // changed flags intact.
      delete_solidindex(current_octree);

      *is_changed = TRUE;
      returnval = duplicate_octree(left_owner, left,
                                   FALSE, SI_ISCHANGED(*right), TRUE,
                                   TRUE, right_owner, TRUE);
    }

    // Other cases: left and right are both terminals, or both
    // octrees, in which case we fall out of switch statement
    // and handle generically.
    break;


  default:
    // Unknown linktype.
    assert(0);
  }

  *left  &= ~SOLIDINDEX_CHANGED;
  *right &= ~SOLIDINDEX_CHANGED;

  if (returnval != MAXINT) {
    // We've found a value to return. Note that we visited this
    // section of the octree.
    splittracker->register_visit(terminal_id, levels);

    if (SI_ISCHANGED(returnval))
      *is_changed = TRUE;

    return returnval;
  }
  

  // We have fallen out of the switch statement; both of the nodes
  // are either terminals or octrees.
  if (SI_ISTERMINAL(*left)) {
    solidterminal *left_term, *right_term, *new_node;
    solidrcsgnode *new_rcsg;
    solidindex new_index;

#if 0
    if (((this->represented_object->index == 63) ||
         (this->represented_object->index == 65) ||
         (this->represented_object->index == 66)) &&
        (terminal_id == 392))
      _CrtDbgBreak();
#endif
    

    assert(SI_ISTERMINAL(*right));
    
    // Get the terminal node pointers.
    left_term =  left_owner->mm_terminal->retrieve_element(SI_GETDATA(*left));
    right_term= right_owner->mm_terminal->retrieve_element(SI_GETDATA(*right));

    assert(left_term ->terminal_id == terminal_id);
    assert(right_term->terminal_id == terminal_id);

    // Create a new solidterminal for us, and a solidrcsgnode too.
    new_node = mm_terminal->allocate(&new_index);
    new_rcsg = mm_rcsgnode->allocate();

    // Set up the terminal.
    new_node->rcsggraph.flags = 0;
    new_node->rcsggraph.rcsg = new_rcsg;
    new_node->terminal_id = terminal_id;

    new_rcsg->link = csg_type;

    // Create the RCSG graph that is the CSG combination of these two child
    // terminals' RCSG graphs.
    create_linked_rcsg(new_node, new_rcsg, left_owner, left_term, right_owner, right_term);

    // We don't need the existing whatever in this cell.
    delete_solidindex(current_octree);

    // Note that we visited this terminal of the octree.
    splittracker->register_visit(terminal_id, 0);

    // Return the new terminal.
    *is_changed = TRUE;
    new_index |= SOLIDINDEXTYPE_TERMINAL | SOLIDINDEX_CHANGED;
    return new_index;

  } else {
    // Both sides are octrees.
    solidoctreenode *old_node = NULL, *left_oct, *right_oct, *new_node;
    solidindex new_index;
    int i, current_is_an_octree, children_changed = FALSE;

    ASSERT(SI_ISOCTREE(*left));
    ASSERT(SI_ISOCTREE(*right));
    left_oct =left_owner ->mm_octreenode->retrieve_element(SI_GETDATA(*left));
    right_oct=right_owner->mm_octreenode->retrieve_element(SI_GETDATA(*right));

    // Create a new octree for this combination.
    new_node = mm_octreenode->allocate(&new_index);

    // Is the existing structure an octree?
    current_is_an_octree = SI_ISOCTREE(current_octree);
    if (current_is_an_octree)
      old_node = mm_octreenode->retrieve_element(SI_GETDATA(current_octree));
    else {
      // We know this current node can't be a terminal. It must be one of the
      // empty or full values.
      assert(SI_ISEMPTY(current_octree) || SI_ISFULL(current_octree));
    }

    new_index |= SOLIDINDEXTYPE_OCTREE;
    for (i = 0; i < 8; i++)
      new_node->child[i] = update_octree((current_is_an_octree)
					 ? old_node->child[i]
					 : current_octree,
					 csg_type, levels - 1,
					 left_owner,  &(left_oct ->child[i]),
					 right_owner, &(right_oct->child[i]),
					 &children_changed, 8*terminal_id + i);

    if (current_is_an_octree) {
      // Destroy the existing octreenode. The children have been destroyed
      // if necessary, in update_octree.
      mm_octreenode->deallocate(old_node, SI_GETDATA(current_octree));
    }

    // Should we keep a track of the number of empty / full results
    // and collapse the octree to a single EMPTY or FULL as a result?

    new_index |= SOLIDINDEXTYPE_OCTREE;
    if (children_changed) {
      *is_changed = TRUE;
      new_index |= SOLIDINDEX_CHANGED;
    }

    return new_index;
  }
}

// Add a WorldCollection to a leafnode. Called from the SplitTracker after
// child WorldCollections have been created / found.
void SolidOctree::add_collection(WorldCollection *collection, solidleafnode *target,
                                 splittrackernode *splitptr)
{
  int i, j;
  splittrackernode *old_splitptr;

  assert(!target->is_deleted);

  // Check to see if there's already a collection in this leafnode that needs
  // to be overwritten.
  for (i = 0; i < target->num_collections; i++)
    if (target->collectionbuf[i].collection->collx->sample_id == collection->collx->sample_id)
      break;

  if (i < target->num_collections) {
    // There is already a collection of this sample ID here.
    old_splitptr = target->collectionbuf[i].splitptr;
    if ((old_splitptr != splitptr) && (old_splitptr != NULL)) {
      // We need to tell the splittrackernode that it shouldn't be
      // pointing to us any further; we are now being updated by
      // another splittrackernode.
      for (j = 0; j < old_splitptr->num_terminals; j++) {
        if ((!old_splitptr->terminalbuf[j].is_deleted) &&
          (old_splitptr->terminalbuf[j].our_leafnode == target))
          break;
      }
      
      assert(j < old_splitptr->num_terminals);
      
      old_splitptr->num_deleted_terminals++;
      old_splitptr->terminalbuf[j].is_deleted = TRUE;
    }

    // Delete the collection that's already here.
    if (target->collectionbuf[i].collection->remove_reference())
      delete target->collectionbuf[i].collection;
  }

  // i may now point to overwriting an old collection, or may point beyond the
  // end of the buffer, indicating that the buffer needs to be extended.
  if (i >= target->collection_bufsize) {
    target->collection_bufsize += collection_bufsize_increment;

    target->collectionbuf = (leafcollx *) realloc((void *) target->collectionbuf,
                                                    target->collection_bufsize * sizeof(leafcollx));
    if (!target->collectionbuf)
      fatal_error("SolidOctree::add_collection", "Can't expand collectionbuf",
                  "Out of memory.");
  }

  target->collectionbuf[i].collection = collection;
  target->collectionbuf[i].splitptr = splitptr;

  if (i == target->num_collections)
    target->num_collections++;

  collection->add_reference();
}

// Remove any Collections from a leafnode. Called from the SplitTracker after
// a split has been performed, creating an empty activechild.
void SolidOctree::remove_collections(int sample_id, solidleafnode *target,
                                     splittrackernode *splitptr)
{
  int i;

  assert(!target->is_deleted);

  // Check to see if there's a collection in this leafnode that needs
  // to be deleted.
  for (i = 0; i < target->num_collections; i++)
    if (target->collectionbuf[i].collection->collx->sample_id == sample_id)
      break;

  if (i >= target->num_collections) {
    // There is not.
    return;
  }

  assert(target->collectionbuf[i].splitptr == splitptr);

  // Delete the collection that's already here.
  if (target->collectionbuf[i].collection->remove_reference())
    delete target->collectionbuf[i].collection;
  // FIXME: should be unnecessary.
  target->collectionbuf[i].collection = NULL;

  target->num_collections--;

  if (target->num_collections > 0) {
    // Fill in the gap here.
    target->collectionbuf[i].collection = target->collectionbuf[target->num_collections].collection;
    target->collectionbuf[i].splitptr   = target->collectionbuf[target->num_collections].splitptr;

    // FIXME: should be unnecessary.
    target->collectionbuf[target->num_collections].collection = NULL;
  }
}

// debugging
void SolidOctree::debug_find_collx(int collx_id, solidindex octindex)
{
  if (SI_ISOCTREE(octindex)) {
    solidoctreenode *node = mm_octreenode->retrieve_element(SI_GETDATA(octindex));

    for (int i = 0; i < 8; i++)
      debug_find_collx(collx_id, node->child[i]);
  } else if (SI_ISTERMINAL(octindex)) {
    solidterminal *term = mm_terminal->retrieve_element(SI_GETDATA(octindex));

    debug_find_collx(collx_id, term->rcsggraph);
  }
}

void SolidOctree::debug_find_collx(int collx_id, solidrcsggraph rcsggraph)
{
  if (RCSG_ISLEAF(rcsggraph.flags)) {
    for (int i = 0; i < rcsggraph.leaf->num_collections; i++)
      if (rcsggraph.leaf->collectionbuf[i].collection->collx->collx_id == collx_id)
        TRACE("found collection\n");
  } else {
    debug_find_collx(collx_id, rcsggraph.rcsg->left);
    debug_find_collx(collx_id, rcsggraph.rcsg->right);
  }
}

// Draw all of our active Collections.
void SolidOctree::draw_collections(int drawid, View *drawview, drawpass pass)
{
  draw_collections(drawid, drawview, octree, pass);
}

void SolidOctree::draw_collections(int drawid, View *drawview, solidindex drawoct, drawpass pass)
{
  if (SI_ISOCTREE(drawoct)) {
    solidoctreenode *node = mm_octreenode->retrieve_element(SI_GETDATA(drawoct));

    for (int i = 0; i < 8; i++)
      draw_collections(drawid, drawview, node->child[i], pass);
  } else if (SI_ISTERMINAL(drawoct)) {
    solidterminal *term = mm_terminal->retrieve_element(SI_GETDATA(drawoct));

    draw_collections(drawid, drawview, term->rcsggraph, FALSE, pass);
  }
}

void SolidOctree::draw_collections(int drawid, View *drawview, solidrcsggraph rcsggraph,
                                   int is_flipped, drawpass pass)
{
  if (RCSG_ISFLIPPED(rcsggraph.flags))
    is_flipped = !is_flipped;

  if (RCSG_ISLEAF(rcsggraph.flags)) {
    primtreeprimnode *prim = rcsggraph.leaf->prim;
    WorldCollection *collx;
    int i;

    for (i = 0; i < rcsggraph.leaf->num_collections; i++) {
      collx = rcsggraph.leaf->collectionbuf[i].collection;

      if (collx->last_draw_id != drawid) {
        // This collection needs to be drawn this draw pass.

        // Make sure the WorldCollection has been calculated!
        collx->calculate();

        // Draw this collection.
        drawview->draw_collection(collx, is_flipped, pass);

        // Mark it as having been drawn.
        collx->last_draw_id = drawid;
      }
    }
  } else {
    draw_collections(drawid, drawview, rcsggraph.rcsg->left, is_flipped, pass);

    if (rcsggraph.rcsg->link == MINUS)
      is_flipped = !is_flipped;

    draw_collections(drawid, drawview, rcsggraph.rcsg->right, is_flipped, pass);
  }
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
