/*
 * octree.h: Definitions for class SolidOctree and supporting code.
 *
 * Contains code to create and maintain SolidOctrees, which are octrees
 * of solidoctreenode, referencing at the lowest level solidrcsgnode and
 * pointers to primtreeprimnodes.
 */

extern int globalhackvar_solidoctreecount;

// A solidindex represents either a solidoctreenode, a solidterminal,
// FULL or EMPTY. The type is determined by the high two bits.
typedef unsigned int solidindex;

/*
 * Bitflags for the solidindex.
 */

// Has the index changed since the last update? Stored in the top bit.
const solidindex SOLIDINDEX_CHANGED      = (((solidindex) 0x4) << 29);

// The type of index is stored in bits 2-3 from the top.
const solidindex SOLIDINDEXTYPE_MASK     = (((solidindex) 0x3) << 29);
const solidindex SOLIDINDEXTYPE_EMPTY    = (((solidindex) 0x0) << 29);
const solidindex SOLIDINDEXTYPE_FULL     = (((solidindex) 0x1) << 29);
const solidindex SOLIDINDEXTYPE_OCTREE   = (((solidindex) 0x2) << 29);
const solidindex SOLIDINDEXTYPE_TERMINAL = (((solidindex) 0x3) << 29);

#define SI_ISCHANGED(x) ((x) & SOLIDINDEX_CHANGED)
#define SI_ISEMPTY(x)   (((x) & SOLIDINDEXTYPE_MASK) == SOLIDINDEXTYPE_EMPTY)
#define SI_ISFULL(x)    (((x) & SOLIDINDEXTYPE_MASK) == SOLIDINDEXTYPE_FULL)
#define SI_ISOCTREE(x)  (((x) & SOLIDINDEXTYPE_MASK) == SOLIDINDEXTYPE_OCTREE)
#define SI_ISTERMINAL(x) (((x) & SOLIDINDEXTYPE_MASK)==SOLIDINDEXTYPE_TERMINAL)

// The data field is the rest of the solidindex.
const solidindex SOLIDINDEX_DATAMASK    = ~(((solidindex) 0x7) << 29);

#define SI_GETDATA(x)   ((x) & SOLIDINDEX_DATAMASK)

// A solidoctreenode represents a level of the SolidOctree.
typedef struct _solidoctreenode {
  solidindex child[8];
} solidoctreenode;

// A rcsggraph may be a pointer to either a rcsgnode or a leafnode.
const unsigned int RCSGGRAPH_LEAF_MASK = (((unsigned int) 1) << 31);
const unsigned int RCSGGRAPH_FLIP_MASK = (((unsigned int) 1) << 30);

#define RCSG_ISLEAF(x)     ((x) & RCSGGRAPH_LEAF_MASK)
#define RCSG_ISFLIPPED(x)  ((x) & RCSGGRAPH_FLIP_MASK)

typedef struct solidrcsggraph_ {
  unsigned int flags;
  union {
    struct solidrcsgnode_ *rcsg;
    struct solidleafnode_ *leaf;
  };
} solidrcsggraph;

// A solidrcsgnode is part of the RCSG tree that's stored in each
// solidterminal. It stores a linktype, and two rcsggraph pointers.
typedef struct solidrcsgnode_ {
  linktype link;
  solidrcsggraph left;
  solidrcsggraph right;
} solidrcsgnode;

// Each WorldCollection in a leafnode also has a pointer to its splittrackernode.
typedef struct leafcollx_ {
  WorldCollection *collection;
  struct splittrackernode_ *splitptr;
} leafcollx;

// A solidleafnode is a leaf of the RCSG tree that's stored in each
// solidterminal. It stores a primtreeprimnode pointer, and a set of
// WorldCollections that intersect this terminal node.
//
// terminalbox is the collectionspace coordinates of the corners of the
// terminal that we're part of.
typedef struct solidleafnode_ {
  int is_deleted;
  primtreeprimnode *prim;
  box terminalbox;
  int num_collections, collection_bufsize;
  leafcollx *collectionbuf;
} solidleafnode;

// A solidterminal represents a terminal node of the octree. It has a
// rcsggraph inside it.
typedef struct _solidterminal {
  int terminal_id;
  solidrcsggraph rcsggraph;
} solidterminal;

// Size of the intersection list that is used to trace a ray against a
// Terminal node.
#define OCTREE_INTLIST_SIZE         20

// An element in the intersection-list structure that is used to store
// intersections with primitives in a given Terminal node of the octree.
typedef struct _octreeint {
  primtreeprimnode *node;
  double t;
  vector norm;
  int deleted;
} octreeint;

// Collection buffer granularities.
const int collection_bufsize_increment = 10;


// The SolidOctree contains a representation of the object with respect to
// the modeller's space. There is a SolidOctree attached to each object
// in the Primtree.
class SolidOctree {
public:
  primtreenode *represented_object;
  Modeller *modeller;

  solidindex octree;

  // debugging int
  int octree_deleted;

  // Has this octree been changed? Does it need to be updated further up
  // the tree?
  int has_changed;

  // The SplitTracker that we use when performing CSG operations on the
  // octree.
  SplitTracker *splittracker;

  // Only used by primitive octrees; like a more sticky has_changed flag,
  // which is cleared by our parent's update_octree.
  int primitive_changed_this_update;

  // Our CSG children (used in split operations).
  SolidOctree *left_octree, *right_octree;
  int left_in_is_active, right_in_is_active;

  // Temporary variables used in traversing a ray through the octree structure.
  vector *cell_size, exit_value, ray_from, ray_dir;
  unsigned char ray_entry_octant;
  double ray_dir_ratio_x_y, ray_dir_ratio_x_z, ray_dir_ratio_y_x,
    ray_dir_ratio_y_z, ray_dir_ratio_z_x, ray_dir_ratio_z_y;
  double t_next_x, t_next_y, t_next_z;
  double *delta_t_cell_x, *delta_t_cell_y, *delta_t_cell_z;
  double cell_entry_t;
  int num_ints;
  octreeint intlist[OCTREE_INTLIST_SIZE];

  MemoryManager<solidoctreenode> *mm_octreenode;
  MemoryManager<solidrcsgnode>   *mm_rcsgnode;
  MemoryManager<solidleafnode>   *mm_leafnode;
  MemoryManager<solidterminal>   *mm_terminal;

  // Duplicate some sub-graph of an octree.
  //
  // NB: This is a multi-function routine. When called with
  // store_in_splittracker as FALSE, it just duplicates a section
  // of the octree. However when store_in_splittracker is TRUE, it
  // creates purely empty leafnodes. Each worldcollection is then stored
  // as a keep-all in the SplitTracker to see whether or not the original
  // or its child will be stored there.
  solidindex duplicate_octree(SolidOctree *original_solidoctree,
                              solidindex *original,
                              int invert_octree = FALSE,
                              int set_changed = FALSE,
                              int clear_changed = FALSE,
                              int store_in_splittracker = FALSE,
                              SolidOctree *splittracker_opponent_octree = NULL,
                              int splittracker_is_left = FALSE);

  // Duplicate a terminal node of an octree.
  solidindex duplicate_terminal(SolidOctree *original_solidoctree,
                                solidterminal *original,
                                int invert_rcsgraph,
                                int store_in_splittracker,
                                SolidOctree *splittracker_opponent_octree,
                                int splittracker_is_left);

  // Duplicate an RCSG graph in the terminal node.
  solidrcsggraph duplicate_rcsggraph(SolidOctree *original_solidoctree,
                                solidterminal *newlycreated_rcsg, solidrcsggraph original,
                                int invert_rcsggraph, int is_flipped, int terminal_id,
                                int store_in_splittracker,
                                SolidOctree *splittracker_opponent_octree,
                                int splittracker_is_left);

  // Register that a sub-graph of an octree is being discarded.
  // This causes it to be registered in the splittracker.
  void register_discard(SolidOctree *original_solidoctree,
                        solidindex *original, int clear_changed,
                        SolidOctree *opponent_octree, int is_left);
  void register_discard(SolidOctree *original_solidoctree, solidrcsggraph original, int is_flipped,
                        int our_terminal_id, solidterminal *our_terminal,
                        SolidOctree *opponent_octree, int is_left);

  // Creates a SolidOctree structure from a primitive.
  //
  // If use_collections is FALSE, then we are the first octree to
  // polygonize a primobjectnode. Create new Collections at the bottom
  // level and store them in the CSGgraph.
  //
  // If use_collections is TRUE, we have already polygonized the
  // primobjectnode. Use the existing Collections at the bottom level.
  solidindex create_primitiveoctree(int use_collections, primobjectnode *primobj,
                                    transform collectionspace,
                                    box primbox, box collectionbox, int octree_level,
                                    solidindex ethertype, int terminal_id);

  // Combine two SolidOctree structures with a CSG operation. Used when
  // creating a SolidOctree from scratch.
  solidindex create_csg_octree(linktype csg_type,
			                         SolidOctree *left_owner,  solidindex *left,
			                         SolidOctree *right_owner, solidindex *right,
			                         int *is_changed, int terminal_id);

  // Combine two terminals with a CSG operation. Registers that all of the
  // WorldCollections in the rcsggraphs need to be split.
  void create_linked_rcsg(solidterminal *our_terminal, solidrcsgnode *new_rcsgnode,
                          SolidOctree *left_owner,  solidterminal *left_term,
                          SolidOctree *right_owner, solidterminal *right_term);

  // Create a shell RCSG (all leafnodes empty) and register that all the
  // leafnodes will need to have split child collections added to them.
  solidrcsggraph create_shell_rcsg(solidterminal *our_terminal, SolidOctree *owner_octree,
                                   solidrcsggraph original,
                                   int inside_child_active,
                                   int flip, solidterminal *split_vs,
                                   SolidOctree *opponent_octree,
                                   int is_left);

  // Set everything nonequal to the EMPTY / FULL value specified
  // to changed.
  // Calls recursive_set_all_nonequal_to_changed.
  void set_all_nonequal_to_changed(solidindex equal_value);

  // Set everything nonequal to specified octree to changed.
  // Calls recursive_set_all_nonequal_to_changed.
  void set_all_nonequal_to_changed(SolidOctree *equal_oct);

  // Set everything nonempty or nonfull in the SolidOctree to changed
  // (behaviour specified by equal_value).
  // This is a routine for internal use only.
  void recursive_set_all_nonequal_to_changed(solidindex *oct,
                              					     SolidOctree *equal_octree, solidindex equal_value,
                              					     int *is_changed);

  // Update the SolidOctree for a primitive, after its position changes.
  void new_prim_position(void);
  solidindex update_primitiveoctree(solidindex octree, primobjectnode *primobject,
                                    box primbox, box collectionbox, int level,
                                    int *child_changed,
                                    int terminal_id, intersect_result known_result = SURFACE);

  // Does the SolidOctree actually have changes? This is called during
  // solidoctree_pass to see if the above node needs to check for changes.
  int actual_changes(void) {
    return SI_ISCHANGED(octree);
  }

  // Perform an incremental update of the SolidOctree, used during
  // solidoctree_pass.
  void incremental_update(linktype csg_type, SolidOctree *left,
                          SolidOctree *right);
  solidindex update_octree(solidindex current_octree,
                           linktype csg_type, int num_levels,
                           SolidOctree *left_owner, solidindex *left,
                           SolidOctree *right_owner, solidindex *right,
                           int *is_changed, int terminal_id);

  // Clear the changed flags of an octree.
  void clear_changed(SolidOctree *original_solidoctree,
                     solidindex *original);

  // Print out the octree.
  void print(int indent, int detail = -1, solidindex oct = MAXINT);

  // Used for printing out the octree.
  char get_char(solidindex oct, int levels, int x, int y, int z);

  // Debug the octree.
  void debug_octree(void);
  void debug_solidindex(solidindex oct);

  // Delete a section of the octree.
  void delete_solidindex(solidindex oct);
  void delete_rcsggraph(solidrcsggraph kill_me, solidterminal *our_terminal);

  // Add a WorldCollection to a leafnode. Called from the SplitTracker after
  // child WorldCollections have been created / found.
  void add_collection(WorldCollection *collection, solidleafnode *target,
                      struct splittrackernode_ *splitptr = NULL);

  // Remove any Collections from a leafnode. Called from the SplitTracker after
  // a split has been performed, creating an empty activechild.
  void remove_collections(int sample_id, solidleafnode *target,
                          struct splittrackernode_ *splitptr);

  // Draw all of our active Collections.
  void draw_collections(int drawid, View *drawview, enum drawpass pass);
  void draw_collections(int drawid, View *drawview, solidindex drawoct, enum drawpass pass);
  void draw_collections(int drawid, View *drawview, solidrcsggraph drawoct, int is_flipped, enum drawpass pass);

  // Render the SolidOctree as a Mirage object.
  object render(void);
  object recursive_render(solidindex octindex, double size);

  // Cast a ray against the SolidOctree, and return the first object
  // that the ray strikes.
  primtreeprimnode *trace_ray(vector ray_from, vector ray_dir,
			      double *t, vector norm);
  primtreeprimnode *ray_trace_octree(int current_level, solidindex the_octree,
				     vector parent_cell_exit, double t_next_x,
				     double t_next_y, double t_next_z,
				     double *t, vector norm);
  primtreeprimnode *ray_trace_terminal(solidterminal *node,
				       double *t, vector norm);
  void ray_trace_rcsggraph(solidrcsggraph trace_me, int is_flipped);
  intersect_result octree_classify_intersection(octreeint *intersect,
						vector hitpoint, solidrcsggraph the_rcsg);

  // debugging
  void debug_find_collx(int collx_id, solidindex octindex);
  void debug_find_collx(int collx_id, solidrcsggraph rcsggraph);

  // Constructor: creates a new SolidOctree, by copying an old one.
  SolidOctree(SolidOctree *original);

  // Constructor: creates a new SolidOctree for a primcsgnode, and the
  // modeller's bounds.
  SolidOctree(primcsgnode *object, Modeller *modeller);

  // Constructor: creates a new SolidOctree for a primtreeprimnode that's
  // just been linked into the Primtree.
  SolidOctree(primtreeprimnode *object, Modeller *modeller,
	      linktype prim_linktype);

  // Destructor deallocates the SolidOctree.
  ~SolidOctree(void);
};
