/*
 * primtree.cc: Rountines for class Primtree and supporting code.
 *
 * Contains code to create and maintain the Primtree representation of
 * an object.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

/*
 *  Routines for primtreenode
 */

void primtreenode::set_defaults(void)
{
  index = modeller->primtree->next_index++;

  is_selected = FALSE;
  Transform::unit(position);
  parent = NULL;
  parent_type = WRAPPER;
  octree = NULL;
  this_tracker = NULL;
  original = NULL;
}

primtreenode *primtreenode::copy_and_update_primtree(
                                         primtreenode *newnode_parent,
					 parenttype newnode_parent_type,
					 Primtree *primtree)
{
  primtreenode *new_graph;
  CopyTracker *original_ct;

  // First, make a copy of the primtree graph. The way this is done depends
  // on the type of primtreenode.
  new_graph = this->copy_primtree(newnode_parent, newnode_parent_type,
				  primtree);

  // If this object is tracked by a objectnode in the CSGgraph, then update
  // the CopyTracker for this objectnode.
  if (original != NULL) {
    // Get the CopyTracker which is used by this node's CSGgraph
    // complement to track all the copies.
    original_ct = original->copytracker;

    // Create a new Tracker for this node.
    new_graph->this_tracker = original_ct->add_new_tracker(new_graph);
  }

  // This node of the primtree's been created. Create its SolidOctree
  // (from scratch), unless it is a wrapper node, in which case just take
  // a pointer to the SolidOctree of its child node.
  switch(new_graph->type()) {
  case PRIMTREENODE_WRAPPER:
    new_graph->octree = ((primwrappernode *) new_graph)->child->octree;
    break;

  case PRIMTREENODE_CSG:
    new_graph->octree = new SolidOctree((primcsgnode *) new_graph,
					primtree->owner);
    break;

  case PRIMTREENODE_PRIM:
    if (new_graph->parent_type == WRAPPER)
      new_graph->octree = new SolidOctree((primtreeprimnode *) new_graph,
					  primtree->owner, PLUS);
    else
      new_graph->octree = new SolidOctree((primtreeprimnode *) new_graph,
					  primtree->owner,
					  ((primcsgnode *) parent)->csg_type);
    break;
    
  default:
    assert(0);
  }

  // Return the updated graph.
  return new_graph;
}

primtreenode::~primtreenode(void)
{

}

/*
 *  Routines for primtreeprimnode
 */

primtreeprimnode::primtreeprimnode(Modeller *m, Primitive *prim)
{
  this->modeller = m;
  this->prim = prim;
  for (int i = 0; i < max_primsplitinfo; i++)
    this->splitinfo[i] = NULL;
  this->lastmoved_update_id = -1;

  set_defaults();
  is_transformed = FALSE;
  Transform::unit(collection_trans);
  viewparam = NULL;

  modeller->primtree->register_primnode(this);
}

// Select this portion of the Primtree.
void primtreeprimnode::select(void)
{
  is_selected = TRUE;
  
  // No children to select.
}

// Deselect this portion of the Primtree.
void primtreeprimnode::deselect(void)
{
  is_selected = FALSE;
  
  // No children to deselect.
}

void primtreeprimnode::print_primtree(int indent)
{
  printf("%*sP%03d/Prim ", indent, "", index);
  printf("(C%03d, T/P%03d) ",
	 (original == NULL) ? -1 : original->index,
	 (this_tracker == NULL)
	 ? -1 : (this_tracker->copy == NULL) ? -2
	 : this_tracker->copy->index);
  printf("prim %s ... ",
	 prim->prim_name());

  if (parent == NULL)
    printf("parent NULL\n");
  else
    printf("parent P%03d by %c\n",
	   (parent == NULL) ? -1 : parent->index,
	   (parent_type == WRAPPER)   ? 'W' :
	   (parent_type == CSG_LEFT)  ? 'L' :
	   (parent_type == CSG_RIGHT) ? 'R' : '?');   
  
  if (octree == NULL)
    printf("%*sP%03d/PrimSolidOctree is NULL\n", indent, "", index);
  else
    printf("%*sP%03d/PrimSolidOctree is non-NULL, and is P%03d's\n",
	   indent, "", index,
	   octree->represented_object->index);
}

int primtreeprimnode::solidoctree_pass(void)
{
  assert(octree != NULL);
  if (octree->has_changed) {
    // This is one of the nodes whose octree is being propagated.
    octree->has_changed = FALSE;
    octree->primitive_changed_this_update = TRUE;
    if (octree->actual_changes()) {
      return TRUE;
    } else {
      return FALSE;
    }
  } else {
    // Nothing below that needs to be checked.
    return FALSE;
  }
}

primtreeprimnode::~primtreeprimnode(void)
{
	// Primitive is deleted by the original objectnode.
  modeller->primtree->register_primnode_deletion(this);

  for (int i = 0; i < max_primsplitinfo; i++) {
    if (splitinfo[i]) {
      free(splitinfo[i]);
    }
  }

  if (viewparam != NULL)
    free(viewparam);
}


// Recursively delete all the children of this node. NB: Deletion !=
// deallocation!
void primtreeprimnode::recursively_delete_children(void)
{
  // No children to delete.
}

void primtreeprimnode::update_primtree(transform update_trans, int update_id)
{
  // Update the primtreeprimnode's position matrix and collectionspace transformation.
  is_transformed = TRUE;
  Transform::combine(position,         update_trans, position);
  Transform::combine(collection_trans, update_trans, collection_trans);

  // We need to delete the primitive-specific view parameter, if one has already
  // been allocated.
  if (viewparam != NULL) {
    free(viewparam);
    viewparam = NULL;
  }

  // The SolidOctree at this point in the primtree now needs to be
  // recalculated to take account of the new position matrix.
  octree->new_prim_position();
  octree->has_changed = TRUE;

  this->lastmoved_update_id = update_id;
}

primtreenode *primtreeprimnode::copy_primtree(primtreenode *newnode_parent,
					      parenttype newnode_parent_type,
					      Primtree *primtree)
{
  primtreeprimnode *new_node;
  transform inverse_primtrans;

  new_node = new primtreeprimnode(modeller, this->prim);

  // Set the new node's properties.
  assert(new_node != NULL);
  assert(parent != NULL);

  // NB!!! If we are copying the tail of a repeatobjectnode, then we may
  // have a primnode that has NO ORIGINAL NODE.
  if (original == NULL)
    Transform::set(new_node->position, newnode_parent->position);
  else
    Transform::combine(original->position, newnode_parent->position, 
		       new_node->position);

  // We are taking a copy of this primtreeprimnode, so it will use the
  // same Collections. The new primnode's collection_trans needs to account
  // for the difference in worldspace positions.
  //
  // If our prim -> worldspace0 trans is P0 and the copy's prim -> worldspace1
  // trans is P1, and our collection -> worldspace0 trans is C0, then we can get
  // the copy's collection -> worldspace1 trans C1 as
  //
  // C1 = P1 P0-1 C0
  new_node->is_transformed = TRUE;
  Transform::invert(this->position, inverse_primtrans);
  Transform::combine(this->collection_trans, inverse_primtrans, new_node->collection_trans);
  Transform::combine(new_node->collection_trans, new_node->position, new_node->collection_trans);
  
  new_node->parent = newnode_parent;
  new_node->parent_type = newnode_parent_type;
  new_node->original = original;
  new_node->this_tracker = NULL;    // Note: this will be set in
                                    // copy_and_update_primtree

  // Nothing below to update.
  return new_node;
}


/*
 *  Routines for primcsgnode
 */

primcsgnode::primcsgnode(Modeller *m)
{
  this->modeller = m;
  this->csg_type = PLUS;
  this->left = NULL;
  this->right = NULL;

  set_defaults();
}

// Select this portion of the Primtree.
void primcsgnode::select(void)
{
  is_selected = TRUE;
  
  left->select();
  right->select();
}

// Deselect this portion of the Primtree.
void primcsgnode::deselect(void)
{
  is_selected = FALSE;
  
  left->deselect();
  right->deselect();
}

void primcsgnode::print_primtree(int indent)
{
  printf("%*sP%03d/CSG ",
	 indent, "", index);
  printf("(C%03d, T/P%03d) mode %c ... ",
	 (original == NULL) ? -1 : original->index,
	 (this_tracker == NULL) ? -1 :
	 (this_tracker->copy == NULL) ? -2 : this_tracker->copy->index,
	 (csg_type == PLUS) ? '+' :
	 (csg_type == MINUS) ? '-' :
	 (csg_type == AND) ? '&' : '?');

  if (parent == NULL)
    printf("parent NULL\n");
  else
    printf("parent P%03d by %c\n",
	   (parent == NULL) ? -1 : parent->index,
	   (parent_type == WRAPPER)   ? 'W' :
	   (parent_type == CSG_LEFT)  ? 'L' :
	   (parent_type == CSG_RIGHT) ? 'R' : '?');   
  
  printf("%*sP%03d/CSG traverse L\n",
	 indent, "", index);
  left->print_primtree(indent + 1);
  printf("%*sP%03d/CSG traverse R\n",
	 indent, "", index);
  right->print_primtree(indent + 1);

  if (octree == NULL)
    printf("%*sP%03d/CSG SolidOctree is NULL\n", indent, "", index);
  else
    printf("%*sP%03d/CSG SolidOctree is non-NULL, and is P%03d's\n",
	   indent, "", index,
	   octree->represented_object->index);
}

extern int debug_octree_print;

int primcsgnode::solidoctree_pass(void)
{
  int left_changes, right_changes;

  assert(octree != NULL);
  if (octree->has_changed) {
    // This is one of the nodes whose octree is being propagated.
    octree->has_changed = FALSE;
    if (octree->actual_changes()) {
      return TRUE;
    } else {
      return FALSE;
    }
  } else {
    if (index == debug_octree_print) {
      _CrtDbgBreak();
      TRACE("old left\n");
      left->octree->print(0);
      TRACE("old right\n");
      right->octree->print(0);
    }

    // Check to see if the CSG nodes below this one have changes to
    // be propagated.
    left_changes = left->solidoctree_pass();
    right_changes = right->solidoctree_pass();

    if (left_changes || right_changes) {
      // One of the child nodes' SolidOctrees has changed. We need to
      // perform an incremental update to our SolidOctree.
      octree->incremental_update(csg_type, left->octree, right->octree);

      if (octree->actual_changes()) {
	return TRUE;
      } else {
	return FALSE;
      }
    }

    return FALSE;
  }
}

// Recursively delete all the children of this node. NB: Deletion !=
// deallocation!
void primcsgnode::recursively_delete_children(void)
{
  left->recursively_delete_children();
  modeller->primtree->delete_primtreenode(left);

  right->recursively_delete_children();
  modeller->primtree->delete_primtreenode(right);
}

primcsgnode::~primcsgnode(void)
{
	// Nothing specific to delete.
}


void primcsgnode::update_primtree(transform update_trans, int update_id)
{
  // Update the primcsgnode's position matrix.
  Transform::combine(position, update_trans, position);

  // Update the children of this node.
  assert(left != NULL);
  left->update_primtree(update_trans, update_id);

  assert(right != NULL);
  right->update_primtree(update_trans, update_id);
}

primtreenode *primcsgnode::copy_primtree(primtreenode *newnode_parent,
					 parenttype newnode_parent_type,
					 Primtree *primtree)
{
  primcsgnode *new_node;

  new_node = new primcsgnode(modeller);

  // Set the new node's properties.
  assert(new_node != NULL);
  assert(newnode_parent != NULL);

  // Note that the new CSGnode has the same position matrix as its
  // parent, due to having no 'original' pointer.
  Transform::set(new_node->position, newnode_parent->position);

  new_node->parent = newnode_parent;
  new_node->parent_type = newnode_parent_type;
  new_node->original = original;
  new_node->this_tracker = NULL;    // Note: this will be set in
                                    // copy_and_update_primtree

  new_node->csg_type = csg_type;

  // Copy and update the below nodes.
  new_node->left  = left ->copy_and_update_primtree(new_node,
						    CSG_LEFT , primtree);
  new_node->right = right->copy_and_update_primtree(new_node,
						    CSG_RIGHT, primtree);

  return new_node;
}


/*
 *  Routines for primwrappernode
 */

primwrappernode::primwrappernode(Modeller *m, primtreenode *child)
{
  this->modeller = m;
  this->child = child;

  set_defaults();
}

// Select this portion of the Primtree.
void primwrappernode::select(void)
{
  is_selected = TRUE;
  
  child->select();
}

// Deselect this portion of the Primtree.
void primwrappernode::deselect(void)
{
  is_selected = FALSE;
  
  child->deselect();
}

void primwrappernode::print_primtree(int indent)
{
  printf("%*sP%03d/Wrap (C%03d, T/P%03d) ... ",
	 indent, "", index, original->index, (this_tracker == NULL)
	 ? -1 : (this_tracker->copy == NULL) ? -2
	 : this_tracker->copy->index);

  if (parent == NULL)
    printf("parent NULL\n");
  else
    printf("parent P%03d by %c\n",
	   (parent == NULL) ? -1 : parent->index,
	   (parent_type == WRAPPER)   ? 'W' :
	   (parent_type == CSG_LEFT)  ? 'L' :
	   (parent_type == CSG_RIGHT) ? 'R' : '?');   
  
  if (octree == NULL)
    printf("%*sP%03d/Wrap SolidOctree is NULL\n", indent, "", index);
  else
    printf("%*sP%03d/Wrap SolidOctree is non-NULL, and is P%03d's\n",
	   indent, "", index,
	   octree->represented_object->index);

  printf("%*sP%03d/Wrap traverse child\n",
	 indent, "", index);
  child->print_primtree(indent + 1);
}

int primwrappernode::solidoctree_pass(void)
{
  // Check to see if the CSG nodes below this one have changes to
  // be propagated.
  return child->solidoctree_pass();
}

// Recursively delete all the children of this node. NB: Deletion !=
// deallocation!
void primwrappernode::recursively_delete_children(void)
{
  child->recursively_delete_children();
  modeller->primtree->delete_primtreenode(child);
}

primwrappernode::~primwrappernode(void)
{
	// We don't delete our octree, as it is just a copy of our
  // child's octree pointer.
}

void primwrappernode::update_primtree(transform update_trans, int update_id)
{
  // Update the primwrappernode's position matrix.
  Transform::combine(position, update_trans, position);

  // Update the child of this node.
  assert(child != NULL);
  child->update_primtree(update_trans, update_id);
}

primtreenode *primwrappernode::copy_primtree(primtreenode *newnode_parent,
					     parenttype newnode_parent_type,
					     Primtree *primtree)
{
  primwrappernode *new_node;

  new_node = new primwrappernode(modeller, NULL);

  // Set the new node's properties.
  assert(new_node != NULL);
  assert(parent != NULL);
  Transform::combine(original->position, newnode_parent->position, 
		     new_node->position);
  new_node->parent = newnode_parent;
  new_node->parent_type = newnode_parent_type;
  new_node->original = original;
  new_node->this_tracker = NULL;    // Note: this will be set in
                                    // copy_and_update_primtree

  // Copy and update the child node.
  new_node->child = child->copy_and_update_primtree(new_node,
						    WRAPPER, primtree);

  return new_node;
}


/*
 *  Routines for Primtree
 */

void Primtree::print_primtree(void)
{
  root->print_primtree(0);
}

void Primtree::initiate_solidoctree_pass(void)
{
  root->solidoctree_pass();
}

// Delete a subgraph of the Primtree.
void Primtree::delete_subgraph(primtreenode *delete_node)
{
  primtreenode *parentnode;
  parenttype parentnode_type;

  assert(delete_node != NULL);

  // First: recursively delete the children of this node.
  delete_node->recursively_delete_children();

  parentnode      = delete_node->parent;
  parentnode_type = delete_node->parent_type;

  // Delete this node.
  delete_primtreenode(delete_node);

  // Tidy up any dangling pointers in the parents above this node.
  // This also ensures the integrity of the SolidOctrees in the graph.
  recursive_tidyup(parentnode, parentnode_type);
}

// Delete a primtreenode.
void Primtree::delete_primtreenode(primtreenode *kill_me)
{
  int must_delete_objectnode;

  // If there is a SolidOctree, remove it. Note that primwrappernodes don't
  // have octrees of their own so they don't need to be deleted.
  if ((kill_me->type() != PRIMTREENODE_WRAPPER) && (kill_me->octree != NULL))
    delete kill_me->octree;

  // Tell the CopyTracker that this node is being deleted. NB: primcsgnodes
  // do not have an original link, so we can skip this part.
  if (kill_me->original != NULL) {
    must_delete_objectnode = kill_me->original->copytracker->
      notify_primtreenode_deletion(kill_me->this_tracker);
    
    if (must_delete_objectnode) {
      // The CSGgraph node associated with this primtreenode has to go as well.
      owner->csg_graph->delete_objectnode(kill_me->original);
    }
  }

  delete kill_me;
}

// Recursively tidy up any dangling pointers.
void Primtree::recursive_tidyup(primtreenode *tidy_node,
				parenttype deleted_child)
{
  primtreenode *next_tidy, *csg_child;
  primcsgnode *csg_node, *parent_csgnode;
  primwrappernode *parent_wrapper;
  parenttype next_childtype;

  if (tidy_node == NULL) {
    // We have deleted the entire Primtree. Stop the recursive pass.
    root = NULL;
    owner->csg_graph->object = NULL;
    owner->focused_octree = NULL;

    return;
  }

  // Depending on the kind of primtreenode that is being tidied up,
  // different actions may be appropriate.
  switch(tidy_node->type()) {
  case PRIMTREENODE_WRAPPER:
    // The wrapper node must be deleted. We continue tidying links up
    // the Primtree.
    next_tidy      = tidy_node->parent;
    next_childtype = tidy_node->parent_type;

    delete_primtreenode(tidy_node);
    recursive_tidyup(next_tidy, next_childtype);
    break;

  case PRIMTREENODE_CSG:
    // The CSG node must be deleted, and the links joined together.
    csg_node = (primcsgnode *) tidy_node;
    csg_child = (deleted_child == CSG_LEFT) ? csg_node->right : csg_node->left;

    if (tidy_node->parent == NULL) {
      // Our parent is NULL, we are the root of the tree. Make our csg_child
      // the new root of the tree.
      root = csg_child;
      owner->focused_octree = root->octree;
      csg_child->parent = NULL;

      // There are no octrees above this one that must be updated.
    } else {
      // We must set our parent primtreenode to point to our csg_child.
      switch(tidy_node->parent_type) {
      case WRAPPER:
	assert(tidy_node->parent->type() == PRIMTREENODE_WRAPPER);
	parent_wrapper = (primwrappernode *) tidy_node->parent;
	
	// NB: Repeatobjectnodes create a dangling primNothing that should
	// be deleted when the whole structure unrolls. We check for this here.
	if ((parent_wrapper->original->type() == OBJECTNODE_REPEAT) &&
	    (csg_child->type() == PRIMTREENODE_PRIM) &&
	    (strcmp(((primtreeprimnode *) csg_child)->prim->prim_name(),
		    "Nothing") == 0)) {
	  // Delete the csg_child primNothing node, and ourselves.
	  delete_primtreenode(csg_child);
	  delete_primtreenode(tidy_node);

	  // Continue tidying up the chain.
	  recursive_tidyup(parent_wrapper, WRAPPER);
	  return;
	}

	// Link the parent wrapper down to our undeleted child.
	parent_wrapper->child = csg_child;
	parent_wrapper->child->parent = parent_wrapper;
	parent_wrapper->child->parent_type = WRAPPER;

	// NB: wrappers also share their child's octree! We must propagate
	// this up through any chain of primwrappernodes that might
	// exist above here.
	fix_wrapperoctreeptr(parent_wrapper, parent_wrapper->octree,
			     csg_child->octree);

  // Our csg_child node's octree is now where we were, and it
  // differs from our octree because it doesn't have the deleted
  // child's contribution. Mark it as CHANGED everywhere that it
  // differs from ours.
  csg_child->octree->set_all_nonequal_to_changed(tidy_node->octree);

  // This octree's changes will need to be propagated up the primtree
  // in the solidoctree_pass.
  csg_child->octree->has_changed = TRUE;
	break;

      case CSG_LEFT:
	assert(tidy_node->parent->type() == PRIMTREENODE_CSG);
	parent_csgnode = (primcsgnode *) tidy_node->parent;
	
	// Link the parent CSG node's left branch down to our undeleted child.
	parent_csgnode->left = csg_child;
	parent_csgnode->left->parent = parent_csgnode;
	parent_csgnode->left->parent_type = CSG_LEFT;
	break;

      case CSG_RIGHT:
	assert(tidy_node->parent->type() == PRIMTREENODE_CSG);
	parent_csgnode = (primcsgnode *) tidy_node->parent;
      
	// Link the parent CSG node's right branch down to our undeleted child.
	parent_csgnode->right = csg_child;
	parent_csgnode->right->parent = parent_csgnode;
	parent_csgnode->right->parent_type = CSG_RIGHT;
	break;

      default:
	assert(0);
      }
    }
    // Delete our node. There is no need to tidy further up the tree.
    delete_primtreenode(tidy_node);
    break;

  default:
    assert(0);
  }
}

// Chain up the Primtree, changing the octree pointers of a primwrappernode
// chain accordingly.
void Primtree::fix_wrapperoctreeptr(primwrappernode *wrapper,
				    SolidOctree *old_octree,
				    SolidOctree *new_octree)
{
  assert(wrapper->octree == old_octree);
  wrapper->octree = new_octree;
  if (wrapper->parent == NULL)
    owner->focused_octree = new_octree;

  // Propagate this changed octree pointer up the tree further, if
  // we have to.
  if ((wrapper->parent_type == WRAPPER) && (wrapper->parent != NULL))
    fix_wrapperoctreeptr((primwrappernode *) wrapper->parent,
			 old_octree, new_octree);
}

// Register a new primtreeprimnode.
void Primtree::register_primnode(primtreeprimnode *node)
{
  if (num_primnodes >= primnodebufsize) {
    // Grow the primnodebuf.
    primnodebufsize += primnodebufsize_increment;
    primnodebuf = (primtreeprimnode **) realloc((void *) primnodebuf,
                                                primnodebufsize * sizeof(primtreeprimnode *));
    if (!primnodebuf)
      fatal_error("Primtree::register_primnode", "Can't grow primnodebuf",
                  "Out of memory");
  }

  primnodebuf[num_primnodes++] = node;
}

// Register a primtreeprimnode's deletion.
void Primtree::register_primnode_deletion(primtreeprimnode *node)
{
  int i;

  for (i = 0; i < num_primnodes; i++)
    if (primnodebuf[i] == node)
      break;

  assert(i < num_primnodes);

  num_primnodes--;

  if (i < num_primnodes)
    primnodebuf[i] = primnodebuf[num_primnodes];
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
