/*
 * primtree.h: Definitions for class Primtree and supporting code.
 *
 * Contains code to create and maintain Primtrees, which are trees of
 * primtreenode (an abstract class for primcsgnode and primtreeprimnode). Each
 * primtreenode contains an octree that describes the object up to that point.
 */

/* Collection-safe */

const int primnodebufsize_increment = 10;
const int max_primsplitinfo = 10;

// A primtreenode is a node in the Primtree. Each node contains an octree that
// represents the object at that level of the primtree. The octree contains
// triangles that make a polygonal representation of the object.
class primtreenode {
public: 
  // The Modeller that owns this primtreenode.
  Modeller *modeller;

  // Is this primtreenode selected?
  int is_selected;

  // A unique-ID index. Used for debugging.
  int index;

  // The node in the CSGgraph that this primnode corresponds to.
  objectnode *original;

  // The Tracker that tracks this node.
  Tracker *this_tracker;

  // Position of this node in world space.
  transform position;

  // The SolidOctree that represents this primtreenode's shape.
  SolidOctree *octree;

  // The direct parent of this primtreenode, and which child pointer
  // points to this node.
  primtreenode *parent;
  parenttype parent_type;

  // Precalculated data used while splitting.
  struct primsplitinfo_ *splitinfo[max_primsplitinfo];

  // Set default values.
  void set_defaults(void);
  
  // Update the parameters of the primtree below this point, corresponding
  // to a change in a CSGgraph node's properties.
  virtual void update_primtree(transform update_trans, int update_id) = 0;

  // Make a copy of the primtree below this point and update its CopyTrackers.
  // This is used by groupobjectnode::initiate_copytracker_pass,
  // which is called by CSGgraph::link_new_element(groupobjectnode *, ...).
  //
  // Makes a copy of the primtree and adds a new Tracker to the appropriate
  // CopyTracker.
  primtreenode *copy_and_update_primtree(primtreenode *parent_,
					 parenttype parent_type,
					 Primtree *primtree);

  // Ancillary routine used by the above; just copies the current node,
  // updates all below but does not update self.
  virtual primtreenode *copy_primtree(primtreenode *parent_,
				      parenttype parent_type,
				      Primtree *primtree) = 0;

  // Recurse through the Primtree and incrementally update the
  // SolidOctrees as necessary.
  virtual int solidoctree_pass(void) = 0;

  // Recursively delete all the children of this node. NB: Deletion !=
  // deallocation!
  virtual void recursively_delete_children(void) = 0;

  // Print out the Primtree.
  virtual void print_primtree(int indent) = 0;

  // Select or deselect this portion of the Primtree.
  virtual void select(void) = 0;
  virtual void deselect(void) = 0;

  virtual primtreenodetype type(void) = 0;

  virtual ~primtreenode(void);
};


// A primtreeprimnode is a Primtree node that represents a single primitive.
class primtreeprimnode : public primtreenode {
public:
  Primitive *prim;

  int lastmoved_update_id;

  // This primtreeprimnode will reference the Collections of its
  // corresponding CSGgraph node. It needs a transform that specifies
  // the offset from collectionspace to its worldspace.
  //
  // If this primtreeprimnode is not transformed from the original one which
  // created the collections, then is_transformed will be false.
  int is_transformed;
  transform collection_trans;

  // View-specific transformation parameter.
  void *viewparam;

  // Update the parameters of the primtree below this point, corresponding
  // to a change in a CSGgraph node's properties.
  virtual void update_primtree(transform update_trans, int update_id);

  // Ancillary routine used by the primtreenode::copy_and_update_primtree.
  // Just copies the current node, updates all below but does not update self.
  virtual primtreenode *copy_primtree(primtreenode *parent_,
				      parenttype parent_type,
				      Primtree *primtree);

  // Recurse through the Primtree and incrementally update the
  // SolidOctrees as necessary.
  virtual int solidoctree_pass(void);

  // Recursively delete all the children of this node. NB: Deletion !=
  // deallocation!
  virtual void recursively_delete_children(void);

  // Print out the Primtree.
  virtual void print_primtree(int indent);

  // Select or deselect this portion of the Primtree.
  virtual void select(void);
  virtual void deselect(void);

  virtual primtreenodetype type(void) {
    return PRIMTREENODE_PRIM;
  }

  // Create a new primtreeprimnode.
  primtreeprimnode(Modeller *m, Primitive *prim);

  virtual ~primtreeprimnode(void);
};

// A primcsgnode is a node in the Primtree that represents a CSG combination
// of primitive objects.
class primcsgnode : public primtreenode {
public:
  linktype csg_type;
  primtreenode *left, *right;

  // Update the parameters of the primtree below this point, corresponding
  // to a change in a CSGgraph node's properties.
  virtual void update_primtree(transform update_trans, int update_id);

  // Ancillary routine used by the primtreenode::copy_and_update_primtree.
  // Just copies the current node, updates all below but does not update self.
  virtual primtreenode *copy_primtree(primtreenode *parent_,
				      parenttype parent_type,
				      Primtree *primtree);

  // Recurse through the Primtree and incrementally update the
  // SolidOctrees as necessary.
  virtual int solidoctree_pass(void);

  // Recursively delete all the children of this node. NB: Deletion !=
  // deallocation!
  virtual void recursively_delete_children(void);

  // Print out the Primtree.
  virtual void print_primtree(int indent);

  // Select or deselect this portion of the Primtree.
  virtual void select(void);
  virtual void deselect(void);

  virtual primtreenodetype type(void) {
    return PRIMTREENODE_CSG;
  }

  // Create a new primcsgnode.
  primcsgnode(Modeller *m);

  virtual ~primcsgnode(void);
};

// A primwrappernode is a node in the Primtree that represents a wrapper.
class primwrappernode : public primtreenode {
public:
  primtreenode *child;

  // Update the parameters of the primtree below this point, corresponding
  // to a change in a CSGgraph node's properties.
  virtual void update_primtree(transform update_trans, int update_id);

  // Ancillary routine used by the primtreenode::copy_and_update_primtree.
  // Just copies the current node, updates all below but does not update self.
  virtual primtreenode *copy_primtree(primtreenode *parent_,
				      parenttype parent_type,
				      Primtree *primtree);

  // Recurse through the Primtree and incrementally update the
  // SolidOctrees as necessary.
  virtual int solidoctree_pass(void);

  // Recursively delete all the children of this node. NB: Deletion !=
  // deallocation!
  virtual void recursively_delete_children(void);

  // Print out the Primtree.
  virtual void print_primtree(int indent);

  // Select or deselect this portion of the Primtree.
  virtual void select(void);
  virtual void deselect(void);

  virtual primtreenodetype type(void) {
    return PRIMTREENODE_WRAPPER;
  }

  // Create a new primwrappernode.
  primwrappernode(Modeller *m, primtreenode *child);

  virtual ~primwrappernode(void);
};

// A Primtree is a binary tree of primtreenodes, some of which are joined
// by CSG links.
class Primtree {
public:
  // A unique-ID index counter used for debugging.
  int next_index;

  // The Modeller that owns this Primtree.
  Modeller *owner;

  primtreenode *root;

  // The array of primtreeprimnodes that is used for drawing.
  int num_primnodes, primnodebufsize;
  primtreeprimnode **primnodebuf;

  // Register a new primtreeprimnode.
  void register_primnode(primtreeprimnode *node);

  // Register a primtreeprimnode's deletion.
  void register_primnode_deletion(primtreeprimnode *node);

  // Print out the Primtree.
  void print_primtree(void);

  // Initiate a SolidOctree pass through the Primtree, incrementally updating
  // the octrees as necessary.
  void initiate_solidoctree_pass(void);

  // Delete a subgraph of the Primtree.
  void delete_subgraph(primtreenode *delete_node);

  // Delete a primtreenode.
  void delete_primtreenode(primtreenode *kill_me);

  // Recursively tidy up any dangling pointers.
  void recursive_tidyup(primtreenode *tidy_node, parenttype deleted_child);

  // Chain up the Primtree, changing the octree pointers of a primwrappernode
  // chain accordingly.
  void fix_wrapperoctreeptr(primwrappernode *wrapper, SolidOctree *old_octree,
			    SolidOctree *new_octree);

  // Construct a new Primtree.
  Primtree(Modeller *owner)
  {
    this->owner = owner;

    next_index = 0;
    root = NULL;

    num_primnodes = primnodebufsize = 0;
    primnodebuf = NULL;
  }

  // Delete the Primtree.
  ~Primtree(void)
  {
    delete_subgraph(root);

    if (primnodebuf != NULL)
      free(primnodebuf);
  }
};


