/*
 * octree.cc: Ray-tracing routines for class SolidOctree.
 *
 * Contains code to trace rays against a SolidOctree.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif


// Cast a ray against the SolidOctree, and return the first object
// that the ray strikes.
primtreeprimnode *SolidOctree::trace_ray(vector ray_from, vector ray_dir,
					 double *t, vector norm)
{
  double ray_dir_mag;
  double dir_x, dir_y, dir_z;
  vector bound_size;
  int cell_no;

  Vector::set(this->ray_from, ray_from);
  Vector::set(this->ray_dir, ray_dir);

  dir_x = ray_dir[X];
  dir_y = ray_dir[Y];
  dir_z = ray_dir[Z];
  ray_dir_mag = Vector::magnitude(ray_dir);

  // Find the size of the top-level cell (note that since octrees are always
  // divided at least once this is half of the bound size).
  Vector::sub(modeller->maxbound, modeller->minbound, bound_size);
  Vector::mult(0.5, bound_size, cell_size[0]);
  
  // Work out which octant the ray is coming from and what the exit
  // values are of X, Y and Z. Also find the absolute values of the ray's
  // direction vector (if -ve, make it +ve).
  ray_entry_octant = 0;
  if (dir_x < 0) {
    ray_entry_octant |= (1 << X);
    exit_value[X] = modeller->minbound[X];
    dir_x = -dir_x;
  } else {
    exit_value[X] = modeller->maxbound[X];
  }
  if (dir_y < 0) {
    ray_entry_octant |= (1 << Y);
    exit_value[Y] = modeller->minbound[Y];
    dir_y = -dir_y;
  } else {
    exit_value[Y] = modeller->maxbound[Y];
  }
  if (dir_z < 0) {
    ray_entry_octant |= (1 << Z);
    exit_value[Z] = modeller->minbound[Z];
    dir_z = -dir_z;
  } else {
    exit_value[Z] = modeller->maxbound[Z];
  }

  // Work out the amount that the ray must traverse to reach the
  // far edge of the bounding volume.
  exit_value[X] = fabs(ray_from[X] - exit_value[X]);
  exit_value[Y] = fabs(ray_from[Y] - exit_value[Y]);
  exit_value[Z] = fabs(ray_from[Z] - exit_value[Z]);

  // Determine values that are necessary for the cell-skipping algorithm.
  //
  // delta_t_cell_x[0], delta_t_cell_y[0] and delta_t_cell_z[0] are the
  //  t-values the ray takes to travel a top-level cell width in X, Y or Z.
  //  (a real top-level cell, not subdivided by one like cell_width[0]).
  //
  // t_next_x, t_next_y and t_next_z are the t-values when the ray will hit
  //  the edge of the bounding volume.

  if (dir_x == 0) {
    delta_t_cell_x[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross X boundaries.
    t_next_x = INFINITY;
    ray_dir_ratio_y_x = INFINITY;
    ray_dir_ratio_z_x = INFINITY;
  } else {
    ray_dir_ratio_y_x = dir_y / dir_x;
    ray_dir_ratio_z_x = dir_z / dir_x;

    delta_t_cell_x[0] = 1.0 / dir_x;        // t for ray to cross 1 X unit
    t_next_x = exit_value[X] * delta_t_cell_x[0];

    if (t_next_x >= INFINITY) {
      delta_t_cell_x[0] = 0.0;
      t_next_x = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_x[0] *= bound_size[X];
    }
  }

  if (dir_y == 0) {
    delta_t_cell_y[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross Y boundaries.
    t_next_y = INFINITY;
    ray_dir_ratio_x_y = INFINITY;
    ray_dir_ratio_z_y = INFINITY;
  } else {
    ray_dir_ratio_x_y = dir_x / dir_y;
    ray_dir_ratio_z_y = dir_z / dir_y;

    delta_t_cell_y[0] = 1.0 / dir_y;        // t for ray to cross 1 Y unit
    t_next_y = exit_value[Y] * delta_t_cell_y[0];

    if (t_next_y >= INFINITY) {
      delta_t_cell_y[0] = 0.0;
      t_next_y = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_y[0] *= bound_size[Y];
    }
  }

  if (dir_z == 0) {
    delta_t_cell_z[0] = 0.0;     // Note that this will never be used since
                                 // the ray won't cross Z boundaries.
    t_next_z = INFINITY;
    ray_dir_ratio_x_z = INFINITY;
    ray_dir_ratio_y_z = INFINITY;
  } else {
    ray_dir_ratio_x_z = dir_x / dir_z;
    ray_dir_ratio_y_z = dir_y / dir_z;

    delta_t_cell_z[0] = 1.0 / dir_z;        // t for ray to cross 1 Z unit
    t_next_z = exit_value[Z] * delta_t_cell_z[0];

    if (t_next_z >= INFINITY) {
      delta_t_cell_z[0] = 0.0;
      t_next_z = INFINITY;
    } else {
      // Work out how long for the ray to cross the top-level cell
      delta_t_cell_z[0] *= bound_size[Z];
    }
  }

  // We have calculated cell_size, delta_t_cell_x etc for cell size 0
  // (top-level cell). Now calculate them for all the smaller cells.
  for (cell_no = 1; cell_no <= modeller->octree_level; cell_no++) {
    delta_t_cell_x[cell_no] = (delta_t_cell_x[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_x[cell_no - 1] * 0.5;

    delta_t_cell_y[cell_no] = (delta_t_cell_y[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_y[cell_no - 1] * 0.5;

    delta_t_cell_z[cell_no] = (delta_t_cell_z[cell_no - 1] == INFINITY) ?
      INFINITY : delta_t_cell_z[cell_no - 1] * 0.5;

    Vector::mult(0.5, cell_size[cell_no - 1], cell_size[cell_no]);
  }

  // Traverse the ray through the octree structure.
  return ray_trace_octree(0, octree, exit_value,
			  t_next_x, t_next_y, t_next_z,
			  t, norm);
}

primtreeprimnode *SolidOctree::ray_trace_octree(int current_level,
						solidindex the_octree,
						vector parent_cell_exit,
						double t_next_x,
						double t_next_y,
						double t_next_z,
						double *t, vector norm)
{
  if (SI_ISTERMINAL(the_octree)) {
    return ray_trace_terminal(mm_terminal->
			      retrieve_element(SI_GETDATA(the_octree)),
			      t, norm);

  } else if (SI_ISOCTREE(the_octree)) {
    // Traverse this octree node's children recursively if the ray
    // intersects them.
    solidoctreenode *node = mm_octreenode->
      retrieve_element(SI_GETDATA(the_octree));
    unsigned char child_no = 0;
    solidindex next_child;
    vector child_cell_exit;
    primtreeprimnode *child_result;

    // Get the size of the child cells.
    double cell_x = cell_size[current_level][X],
           cell_y = cell_size[current_level][Y],
           cell_z = cell_size[current_level][Z];

    current_level++;

    // Work out which child of this node the ray starts in.
    Vector::set(child_cell_exit, parent_cell_exit);
    
    if (child_cell_exit[X] > cell_x) {
      // We start in the low-X child cell. This means that the next X crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-X child cell.
      child_cell_exit[X] -= cell_x;
      t_next_x -= delta_t_cell_x[current_level];
    } else {
      // We start in the high-X child cell. The next X crossing is at the
      // edge of the parent cell.
      child_no |= (1 << X);
    }

    if (child_cell_exit[Y] > cell_y) {
      // We start in the low-Y child cell. This means that the next Y crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-Y child cell.
      child_cell_exit[Y] -= cell_y;
      t_next_y -= delta_t_cell_y[current_level];
    } else {
      // We start in the high-Y child cell. The next Y crossing is at the
      // edge of the parent cell.
      child_no |= (1 << Y);
    }

    if (child_cell_exit[Z] > cell_z) {
      // We start in the low-Z child cell. This means that the next Z crossing
      // won't be at the edge of the parent cell but rather at the edge of
      // the low-Z child cell.
      child_cell_exit[Z] -= cell_z;
      t_next_z -= delta_t_cell_z[current_level];
    } else {
      // We start in the high-Z child cell. The next Z crossing is at the
      // edge of the parent cell.
      child_no |= (1 << Z);
    }

    // Now we know which child cell the ray starts in. Get this child and
    // recursively traverse it. Note that all of the above calculations
    // are done with absolute values of the distances, etc. and work if
    // the ray is coming from the low-XYZ octant. If this isn't the case,
    // then we are actually indexing a different child to the one found
    // above; we do this by doing a bitwise XOR with the octant that the
    // ray is coming from.
    next_child = node->child[child_no ^ ray_entry_octant];

    child_result = ray_trace_octree(current_level, next_child,
				    child_cell_exit, t_next_x, t_next_y,
				    t_next_z, t, norm);
    if (child_result != NULL)
      return child_result;

    // We know which child the ray starts in and have intersected with it.
    // Traverse the rest of the child cells, stopping once we reach the
    // high-XYZ cell or we hit an edge of the parent cell.
    while (child_no != ((1 << X) | (1 << Y) | (1 << Z))) {
      // Work out which direction we will go in next (X, Y or Z)
      if (t_next_x <= t_next_y) {
	if (t_next_x <= t_next_z) {

	  // We will hit an X boundary next.
	  if (child_no & (1 << X)) {
	    // We're already in a high-X child, so we've hit a boundary
	    // of the parent cell. Exit.
	    break;
	  }

	  // We are now in the high-X child.
	  child_no |= (1 << X);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next X boundary up by the X
	  // increment.
	  cell_entry_t = t_next_x;
	  t_next_x += delta_t_cell_x[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[Y] -= child_cell_exit[X] * ray_dir_ratio_y_x;
	  child_cell_exit[Z] -= child_cell_exit[X] * ray_dir_ratio_z_x;
	  child_cell_exit[X] = cell_x;
	} else {

	  // We will hit a Z boundary next.
	  if (child_no & (1 << Z)) {
	    // We're already in a high-Z child, so we've hit a boundary
	    // of the parent cell. Exit.
	    break;
	  }

	  // We are now in the high-Z child.
	  child_no |= (1 << Z);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Z boundary up by the Z
	  // increment.
	  cell_entry_t = t_next_z;
	  t_next_z += delta_t_cell_z[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Z] * ray_dir_ratio_x_z;
	  child_cell_exit[Y] -= child_cell_exit[Z] * ray_dir_ratio_y_z;
	  child_cell_exit[Z] = cell_z;
	}
      } else {
	if (t_next_y <= t_next_z) {
	  // We will hit a Y boundary next.
	  if (child_no & (1 << Y)) {
	    // We're already in a high-Y child, so we've hit a boundary
	    // of the parent cell. Exit.
	    break;
	  }

	  // We are now in the high-Y child.
	  child_no |= (1 << Y);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Y boundary up by the Y
	  // increment.
	  cell_entry_t = t_next_y;
	  t_next_y += delta_t_cell_y[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Y] * ray_dir_ratio_x_y;
	  child_cell_exit[Z] -= child_cell_exit[Y] * ray_dir_ratio_z_y;
	  child_cell_exit[Y] = cell_y;
	} else {
	  // We will hit a Z boundary next.
	  if (child_no & (1 << Z)) {
	    // We're already in a high-Z child, so we've hit a boundary
	    // of the parent cell. Exit.
	    break;
	  }

	  // We are now in the high-Z child.
	  child_no |= (1 << Z);

	  // Set cell_entry_t to be the entry of this child cell.
	  // Set the t-value of hitting the next Z boundary up by the Z
	  // increment.
	  cell_entry_t = t_next_z;
	  t_next_z += delta_t_cell_z[current_level];

	  // Find the new distances to go until the X, Y and Z boundaries.
	  child_cell_exit[X] -= child_cell_exit[Z] * ray_dir_ratio_x_z;
	  child_cell_exit[Y] -= child_cell_exit[Z] * ray_dir_ratio_y_z;
	  child_cell_exit[Z] = cell_z;
	}
      }

      // Now we know which child cell the ray will hit next. Get this child and
      // recursively traverse it. Note that all of the above calculations
      // are done with absolute values of the distances, etc. and work if
      // the ray is coming from the low-XYZ octant. If this isn't the case,
      // then we are actually indexing a different child to the one found
      // above; we do this by doing a bitwise XOR with the octant that the
      // ray is coming from.
      next_child = node->child[child_no ^ ray_entry_octant];
      
      child_result = ray_trace_octree(current_level, next_child,
				      child_cell_exit, t_next_x, t_next_y,
				      t_next_z, t, norm);
      if (child_result != NULL)
	return child_result;
    }
  }

  // Either we are an octree and we have traversed all of the child cells,
  // finding nothing. Or, we were EMPTY / FULL.
  return NULL;
}

primtreeprimnode *SolidOctree::ray_trace_terminal(solidterminal *node,
						  double *t, vector norm)
{
  int i, current_closest;
  double current_closest_t, cell_exit_t;
  intersect_result classification;
  vector hit_point;

  // Set up the intersection list to place primitive intersections in.
  num_ints = 0;

  ray_trace_rcsggraph(node->rcsggraph, FALSE);

  // Delete any invalid intersections.
  for (i = 0; i < num_ints; i++) {
    cell_exit_t = MIN3(t_next_x, t_next_y, t_next_z);
    
    // Discard any intersections outside the bounds of the cell.
    if ((intlist[i].t < cell_entry_t) || (intlist[i].t >= cell_exit_t)) {
      intlist[i].deleted = TRUE;
    } else if (!RCSG_ISLEAF(node->rcsggraph.flags)) {
      // This intersection is inside the cell. Calculate its hitpoint,
      // and check that this is a valid intersection with respect to the
      // RCSG graph in this terminal.
      Vector::set(hit_point, ray_from);
      Vector::addfrac(intlist[i].t, ray_dir, hit_point);
	
      classification = octree_classify_intersection(&intlist[i], hit_point,
                                                    node->rcsggraph);
      
      if (classification != SURFACE)
        intlist[i].deleted = TRUE;
    }
  }

  // Find the first non-deleted intersection.
  current_closest = -1;
  current_closest_t = +INFINITY;
  for (i = 0; i < num_ints; i++) {
    if ((!intlist[i].deleted) && (intlist[i].t < current_closest_t)) {
      current_closest = i;
      current_closest_t = intlist[i].t;
    }
  }
  
  if (current_closest == -1) {
    return NULL;
  } else {
    *t = intlist[current_closest].t;
    Vector::set(norm, intlist[current_closest].norm);
    return intlist[current_closest].node;
  }
}

void SolidOctree::ray_trace_rcsggraph(solidrcsggraph rcsggraph, int is_flipped)
{
  if (RCSG_ISFLIPPED(rcsggraph.flags))
    is_flipped = !is_flipped;

  if (RCSG_ISLEAF(rcsggraph.flags)) {
    // Trace the ray against the primitive that is this node.
    int numint_thisprim, i;
    rayintersect int_storage[10];
    vector prim_ray_from, prim_ray_dir;

    Matrix::apply(rcsggraph.leaf->prim->position[AFT], ray_from, prim_ray_from);
    Matrix::applynoshift(rcsggraph.leaf->prim->position[AFT], ray_dir, prim_ray_dir);

    numint_thisprim = rcsggraph.leaf->prim->prim->ray_intersect(prim_ray_from, prim_ray_dir,
                                                                int_storage);

    // Put all of these intersections in the terminal's intersection-list.
    for (i = 0; i < numint_thisprim; i++) {
      intlist[num_ints].node = rcsggraph.leaf->prim;
      intlist[num_ints].t = int_storage[i].t;

      if (is_flipped)
        Vector::mult(-1, int_storage[num_ints].normal, intlist[i].norm);
      else
        Vector::set(intlist[num_ints].norm, int_storage[i].normal);

      intlist[num_ints].deleted = FALSE;
      num_ints++;
    }
  } else {
    // Store the current end of the intersection list.
    int old_numints = num_ints;

    // Trace the ray against the children of this node.
    ray_trace_rcsggraph(rcsggraph.rcsg->left, is_flipped);

    // Subtracted surfaces have reversed normals.
    if (rcsggraph.rcsg->link == MINUS)
      is_flipped = !is_flipped;

    ray_trace_rcsggraph(rcsggraph.rcsg->right, is_flipped);
  }
}

intersect_result SolidOctree::octree_classify_intersection(octreeint *
							   intersect,
							   vector hitpoint,
							   solidrcsggraph rcsggraph)
{
  intersect_result returnval;

  if (RCSG_ISLEAF(rcsggraph.flags)) {
    // Classify with respect to a primitive.

    if (rcsggraph.leaf->prim == intersect->node) {
      // The intersection _is_ with this primitive, return SURFACE.
      returnval = SURFACE;
    } else {
      // Transform the hitpoint into the primitive space, and sample there.
      vector primpoint;

      Matrix::apply(rcsggraph.leaf->prim->position[AFT], hitpoint, primpoint);
      returnval = rcsggraph.leaf->prim->prim->point_intersect(primpoint);
    }
  } else {

    // Classify with respect to a RCSG graph.
    intersect_result left_result, right_result;

    // Whatever linktype this rcsggraph.rcsg is, we need to classify the point with
    // respect to the right branch of the RCSG tree.
    right_result = octree_classify_intersection(intersect, hitpoint, rcsggraph.rcsg->right);

    switch(rcsggraph.rcsg->link) {
    case PLUS : {
      // The CSG table that governs this is ....
      //  A += B
      //
      //  A / B->   | IN     OUT     SURF
      // -----------+--------------------------------
      //   IN       | IN     IN      IN
      //   OUT      | IN     OUT     SURF
      //  SURF      | IN     SURF    impossible
      
      if (right_result == OUT) {
        // We are adding OUT to some object, result is the previous
        // object (the left branch).
        // Recursively classify the left branch and return it.
        returnval = octree_classify_intersection(intersect, hitpoint, rcsggraph.rcsg->left);
      } else if (right_result == IN) {
        // We return the right's IN result as it overwrites the left branch.
        returnval = IN;
      } else if (right_result == SURFACE) {
        // The right branch has a surface here. We must classify the
        // left branch to determine the return value.
        left_result = octree_classify_intersection(intersect, hitpoint, rcsggraph.rcsg->left);
	
        if (left_result == OUT) {
          // Return the surface from the right branch.
          returnval = SURFACE;
        } else if (left_result == IN) {
          // The left side is IN, there is no surface here.
          returnval = IN;
        } else if (left_result == SURFACE) {
          fatal_error("octree_classify_intersection", "SURFACE x2 conflict",
                      "The left and right branches of the RCSG tree both\n"
                      "returned a SURFACE classification of the currently\n"
                      "considered intersection point. This absolutely\n"
                      "should not happen.");
        } else {
          fatal_error("octree_classify_intersection", "unknown classification",
                      "The left branch of the RCSG returned an unknown\n"
                      "classification of the intersection point (not IN, OUT\n"
                      "or SURFACE). This error should not happen.");
        }
      } else {
        fatal_error("octree_classify_intersection", "unknown classification",
                    "The right branch of the RCSG returned an unknown"
                    "classification of the intersection point (not IN, OUT"
                    "or SURFACE). This error should not happen.");
      }
      break;
    }
    case MINUS : {
      // The CSG table that governs this is ....
      //  A -= B
      //
      //  A / B->   | IN    OUT   SURF
      // -----------+-----------------------
      //   IN       | OUT   IN    SURF
      //   OUT      | OUT   OUT   OUT
      //   SURF     | OUT   SURF  impossible
      
      if (right_result == IN) {
        // A is completely subtracted out by B, return OUT.
        returnval = OUT;
      } else if (right_result == OUT) {
        // Recursively classify the left branch.
        returnval = octree_classify_intersection(intersect, hitpoint, rcsggraph.rcsg->left);
      } else if (right_result == SURFACE) {
        // The right branch has a surface here. We must classify the
        // left branch to determine the return value.
        left_result = octree_classify_intersection(intersect, hitpoint, rcsggraph.rcsg->left);
	
        if (left_result == OUT) {
          // There is no A here in the first place to be subtracted from,
          // return OUT.
          returnval = OUT;
        } else if (left_result == IN) {
          // The right branch's surface is being subtracted from the full space
          // of the left branch. Return SURFACE.
          returnval = SURFACE;
        } else if (left_result == SURFACE) {
          fatal_error("octree_classify_intersection", "SURFACE x2 conflict",
                      "The left and right branches of the RCSG tree both\n"
                      "returned a SURFACE classification of the currently\n"
                      "considered intersection point. This absolutely\n"
                      "should not happen.");
        } else {
          fatal_error("octree_classify_intersection", "unknown classification",
                      "The left branch of the RCSG returned an unknown\n"
                      "classification of the intersection point (not IN, OUT\n"
                      "or SURFACE). This error should not happen.");
        }
      } else {
        fatal_error("octree_classify_intersection", "unknown classification",
                    "The right branch of the RCSG returned an unknown\n"
                    "classification of the intersection point (not IN, OUT\n"
                    "or SURFACE). This error should not happen.");
      }
      break;
    }
    case AND : {
      // The CSG table that governs this is ....
      //  A &= B
      //
      //  A / B->   |  IN     OUT    SURF
      // -----------+------------------------
      //   IN       |  IN     OUT    SURF
      //   OUT      |  OUT    OUT    OUT
      //   SURF     |  SURF   OUT    impossible
      
      // We need to classify the left node no matter what the
      // value of the right node.
      left_result = octree_classify_intersection(intersect, hitpoint, rcsggraph.rcsg->left);

      if ((right_result == OUT) || (left_result == OUT)) {
        // No intersection, return OUT.
        returnval = OUT;
      } else if ((right_result == SURFACE) || (left_result == SURFACE)) {
        returnval = SURFACE;
      } else {
        returnval = IN;
      }
      break;
    }
    default :
      fatal_error("octree_classify_intersection", "unknown link type in rcsg",
                  "The link-type in the RCSG graph being traversed does not\n"
                  "fit any of the valid link types (PLUS, MINUS, or AND).\n"
                  "This error should not happen.");
    }

  }

  if (RCSG_ISFLIPPED(rcsggraph.flags))
    if (returnval == IN)
      return OUT;
    else if (returnval == OUT)
      return IN;

  return returnval;
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
