/*
 * split.cc: Code for the split_triangles routine.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Global counter for debugging purposes.
int trisplit_debugindex = 0;

// Trisplit debug number to break on.
int debug_trisplit_index = -1;

// Tolerance for deciding whether or not a vertex lies on the split plane.
#define PLANE_TOLERANCE 1e-06

// Tolerance for determining whether the intervals overlap.
#define INTERVAL_TOLERANCE  1e-05

// Tolerance for finding sliver triangles
#define COLLINEAR_TRIANGLE_TOLERANCE 1e-06

// Constructor initialises arrays to be empty.
TriSplitter::TriSplitter(Modeller *modeller)
{
  this->modeller = modeller;

  workbufsize = 0;
  workbuf = NULL;

  num_split = 0;
}

// Destructor destroys any arrays as necessary.
TriSplitter::~TriSplitter(void)
{
  if (workbufsize > 0) {
    assert(workbuf != NULL);
    free(workbuf);
  }
}

// Set up the triangles that are to be split.
void TriSplitter::setup_input(string *cs_cell, worldstring *ws_cell)
{
  int i, string_no;
  string *cs_string;
  worldstring *ws_string;
  tri *this_tri;

  num_in = cs_cell->num_tri;

  if (num_in > workbufsize) {
    workbufsize = num_in + split_tribuf_growsize;

    workbuf = (splittri *) realloc((void *) workbuf, workbufsize * sizeof(splittri));
    if (workbuf == NULL)
      fatal_error("TriSplitter::setup_input", "Can't allocate work buffer",
                  "Out of memory.");
  }

  num_in_strings = cs_cell->celllist_length;
  assert(num_in_strings == ws_cell->celllist_length);
  assert(num_in_strings <= max_in_strings);

  i = 0;
  cs_string = cs_cell;
  ws_string = ws_cell;
  for (string_no = 0; cs_string != NULL;
       string_no++, cs_string = cs_string->stringptr, ws_string = ws_string->stringptr) {
    assert(ws_string != NULL);

    in_string[string_no].cs_str = cs_string;
    in_string[string_no].ws_str = ws_string;
    in_string[string_no].split_interval_wrt = -1;

    for (this_tri = cs_string->head; this_tri != NULL; this_tri = this_tri->stringptr) {
      workbuf[i].t = this_tri;

      workbuf[i].original = TRUE;
      workbuf[i].status = UNKNOWN;
      workbuf[i].status_inferred = FALSE;
      workbuf[i].traverse_id = -1;
      workbuf[i].tvals_wrt = -1;

      workbuf[i].this_string_index = string_no;
      
      i++;
    }
  }
  assert(ws_string == NULL);
}

// Splits a cell worth of triangles.
int TriSplitter::split_cell(int cell_index, WorldCollection *parent, WorldCollection *child, 
                            int num_split, splitstring *split_string,
                            double keep_normal, int *all_kept)
{
  if (++trisplit_debugindex == debug_trisplit_index)
    _CrtDbgBreak();

  setup_input(parent->collx->cell[cell_index], parent->cell[cell_index]);

  this->parent        = parent;
  this->child         = child;
  this->cell_index    = cell_index;
  this->num_split     = num_split;
  this->split_string  = split_string;
  this->keep_normal   = keep_normal;

  return split_triangles(all_kept);
}


// Performs a split operation of a set of triangles (in) against another set of
// triangles (split). Places the split triangles into the collection's child.
int TriSplitter::split_triangles(int *all_kept)
{
  int q, instring;

#if _DEBUG
  int i, peer_index, peer_edge;
  splittri *current_tri, *peer_tri;
  vector tempvec;
#endif

  assert(num_split != 0);
  assert(num_in != 0);

  current_numtri = num_in;

#if _DEBUG
  // Check that the adjacency information on the input triangles
  // is correct.
  for (i = 0, current_tri = workbuf; i < num_in; i++, current_tri++) {
    for (q = 0; q < 3; q++) {
      peer_index = current_tri->t->adj[q][ADJ_INDEX];
      if (peer_index == -1)
        continue;
      peer_edge = current_tri->t->adj[q][ADJ_VERTEX];
      assert((peer_index >= 0) && (peer_index < num_in));
      assert((peer_edge >= 0) && (peer_edge < 3));

      peer_tri = &workbuf[peer_index];

      Vector::sub(peer_tri->t->v[(peer_edge + 1) % 3], current_tri->t->v[q], tempvec);
      if (Vector::magnitude(tempvec) > 1e-03) {
        TRACE("S%05d WARNING: adjacency info may be incorrect\n", trisplit_debugindex);
        i = current_numtri;
        break;
      }

      Vector::sub(peer_tri->t->v[peer_edge], current_tri->t->v[(q + 1) % 3], tempvec);
      if (Vector::magnitude(tempvec) > 1e-03) {
        TRACE("S%05d WARNING: adjacency info may be incorrect\n", trisplit_debugindex);
        i = current_numtri;
        break;
      }
    }
  }
#endif

  /*
   * Phase One: Take the in triangles, and split them against the other
   *             triangle buffer. Tag those that are adjacent to a split-edge
   *             with KEEP or DISCARD.
   *
   *           Every time a triangle is split, we set its newadj[i][j][k]
   *            array... i is the side of the parent triangle, j is the index
   *            along that side (there may be two child triangles along that
   *            side. If there is only one, then only j=0 is set and j=1
   *            holds ADJ_INDEX of -1), and k is the adjacency type,
   *            ADJ_INDEX pointing to an index *inside the in buffer* and
   *            ADJ_VERTEX as normal.
   *
   *           The parent triangle's status is set to SPLIT, and its adj[][]
   *            array is unchanged.
   *
   *           Child triangles have parentindex set to point back to their
   *            parents' position inside the 'in' buffer.
   *
   *           Child triangles' adjacency adj[i][j] are NOT set to their final
   *            values. i indexes each of its sides. If this side of the
   *            child triangle buts against the side of the parent triangle,
   *            j=ADJ_INDEX gives the index of this parent triangle, and
   *            j=ADJ_VERTEX gives a number that represents what side of the
   *            parent this side is, and also whereabouts along the side the
   *            fragment is (0 = low side, 1 = high side, 2 = whole side).
   *            (3*side_num + fragment_num).
   */

  split_any = FALSE;

  for (current_string = 0; current_string < num_split; current_string++) {

    calculate_plane_dist(current_string);

    for (tosplit_index = 0; tosplit_index < current_numtri; tosplit_index++) {
      if (workbuf[tosplit_index].status == SPLIT)
        continue;

      if (workbuf[tosplit_index].num_negative && workbuf[tosplit_index].num_positive) {
        // Both positive and negative distances from the plane - this triangle lies
        // across the split plane and may potentially have to be split.
        
        // First, make sure that we have all the correct values pre-calculated for this
        // input string.
        instring = workbuf[tosplit_index].this_string_index;
        if (in_string[instring].split_interval_wrt != current_string) {
          // We must calculate the direction of the line of intersection between the
          // split plane and this plane. This is in worldspace.
          Vector::cross(in_string[instring].ws_str->norm, split_string[current_string].ws_str->norm,
                        in_string[instring].ws_line_dir);

          // Determine this line in the input collectionspace.
          in_string[instring].cs_t_offset = 0;
          if (parent->owner->is_transformed) {
            Matrix::applynoshift(parent->owner->collection_trans[AFT],
                                 in_string[instring].ws_line_dir, in_string[instring].cs_line_dir);

            for (q = X; q <= Z; q++)
              in_string[instring].cs_t_offset -= parent->owner->collection_trans[AFT][q][3] * in_string[instring].cs_line_dir[q];

            in_string[instring].cs_t_scale = Vector::dot(in_string[instring].ws_line_dir, in_string[instring].ws_line_dir)
                                           / Vector::dot(in_string[instring].cs_line_dir, in_string[instring].cs_line_dir);
          } else {
            Vector::set(in_string[instring].cs_line_dir, in_string[instring].ws_line_dir);
            in_string[instring].cs_t_scale = 1.0;
          }

          // Now, determine the set of t-intervals along this line where there are split
          // triangles present.
          find_t_intervals(instring, current_string);

          in_string[instring].split_interval_wrt = current_string;
        }

        if (in_string[instring].num_intervals == 0) {
          // No split triangles overlap the line of intersection anywhere. We do not
          // have to check any triangles in this split plane.
          continue;
        } else {
          // Check to see if the current triangle is intersected by any of the intervals along
          // the line of intersection in which there are split triangles present.
          if (check_split(tosplit_index, instring, current_string))
            split_any = TRUE;
        }

      } else if (workbuf[tosplit_index].status == UNKNOWN) {
        
        if (workbuf[tosplit_index].num_positive == 1) {
          // One positive and no negative (we know because it didn't go into the first if
          // clause)... this triangle has its base along the split plane and points in the direction
          // of the normal. We can infer whether it is KEEP or DISCARD from the sign of keep_normal.
          workbuf[tosplit_index].status = (keep_normal > 0) ? KEEP : DISCARD;
          workbuf[tosplit_index].status_inferred = TRUE;
          
          // zero_vertex is set to the first of the two zero vertices... so it indexes the edge
          // that is zero.
          workbuf[tosplit_index].status_edge = workbuf[tosplit_index].zero_vertex;
          workbuf[tosplit_index].t->highlight_edge |= (1 << workbuf[tosplit_index].status_edge);

        } else if (workbuf[tosplit_index].num_negative == 1) {
          // One negative and no positive ditto but points away from the direction of the normal.
          workbuf[tosplit_index].status = (keep_normal > 0) ? DISCARD : KEEP;
          workbuf[tosplit_index].status_inferred = TRUE;
          workbuf[tosplit_index].status_edge = workbuf[tosplit_index].zero_vertex;
        }
      }
    }
  }

  if (split_any)
    recalculate_adjacency();


#if _DEBUG
  // Check that the adjacency information on the input triangles
  // is correct.
  for (i = 0, current_tri = workbuf; i < current_numtri; i++, current_tri++) {
    if (current_tri->status == SPLIT)
      continue;

    for (q = 0; q < 3; q++) {
      peer_index = current_tri->t->adj[q][ADJ_INDEX];
      if (peer_index == -1)
        continue;
      peer_edge = current_tri->t->adj[q][ADJ_VERTEX];
      assert((peer_index >= 0) && (peer_index < current_numtri));
      assert((peer_edge >= 0) && (peer_edge < 3));

      peer_tri = &workbuf[peer_index];

      Vector::sub(peer_tri->t->v[(peer_edge + 1) % 3], current_tri->t->v[q], tempvec);
      if (Vector::magnitude(tempvec) > 1e-03) {
        TRACE("S%05d WARNING: post-split adjacency info may be incorrect\n", trisplit_debugindex);
        i = current_numtri;
        break;
      }

      Vector::sub(peer_tri->t->v[peer_edge], current_tri->t->v[(q + 1) % 3], tempvec);
      if (Vector::magnitude(tempvec) > 1e-03) {
        TRACE("S%05d WARNING: post-split adjacency info may be incorrect\n", trisplit_debugindex);
        i = current_numtri;
        break;
      }
    }
  }
#endif

  index_final_triangles();

  if (!split_any) {
    // Check to see if we can just return the parent buffer.
    if (num_output_tri == 0) {
      // All triangles are discarded.
      *all_kept = FALSE;
      return FALSE;
    } else if ((num_output_tri == num_in) && (num_in == current_numtri)) {
      // All triangles are kept.
      *all_kept = TRUE;
      return FALSE;
    }
  }

  output_triangles();
  return TRUE;
}

// For every triangle, we calculate the distance from the split plane
// for each of its vertices. We also calculate the number of positive and
// negative vertices in each triangle.
void TriSplitter::calculate_plane_dist(int string_index)
{
  int q;
  vector split_norm;
  double split_dot;

  // Transform the split-string's worldspace normal into the input collectionspace.
  split_dot = split_string[string_index].ws_str->plane_dot;
  if (parent->owner->is_transformed) {
    Matrix::applytranspose(parent->owner->collection_trans[FORE],
                           split_string[string_index].ws_str->norm, split_norm);

    for (q = X; q <= Z; q++)
      split_dot += parent->owner->collection_trans[AFT][q][3] * split_norm[q];
  } else {
    Vector::set(split_norm, split_string[string_index].ws_str->norm);
  }

  // FIXME: severe optimizations possible here, possibly. Would require some
  // more sophisticated adjacency and vertex tracking.
  for (tosplit_index = 0; tosplit_index < current_numtri; tosplit_index++) {
    if (workbuf[tosplit_index].status == SPLIT)
      continue;

    workbuf[tosplit_index].num_positive = workbuf[tosplit_index].num_negative = 0;
    workbuf[tosplit_index].zero_vertex = -1;
    for (q = 0; q < 3; q++) {
      workbuf[tosplit_index].plane_dist[q] = Vector::dot(split_norm, workbuf[tosplit_index].t->v[q])
                 - split_dot;

        if (split_string[string_index].flipped)
          workbuf[tosplit_index].plane_dist[q] = -workbuf[tosplit_index].plane_dist[q];

        if (workbuf[tosplit_index].plane_dist[q] > PLANE_TOLERANCE)
          workbuf[tosplit_index].num_positive++;
        else if (workbuf[tosplit_index].plane_dist[q] < -PLANE_TOLERANCE)
          workbuf[tosplit_index].num_negative++;
        else if (workbuf[tosplit_index].zero_vertex == -1)
          workbuf[tosplit_index].zero_vertex = q;
    }
  }
}

// Recalculate adjacency indices for all the triangles.
void TriSplitter::recalculate_adjacency(void)
{
  int i, q, trinum, parent_edge, fragment_num, peer_index, peer_edge;
  int adj_index, adj_vertex, edgept_index, edgept_vertex;
  int disconnect_index, disconnect_vertex, new_index, new_vertex;
  splittri *current_tri, *parent_tri, *peer_tri;
  tri *new_tri;
  vector our_edge, other_edge;
  double edgeval;

  /*
   * Phase Two: Update the child triangles' adj[][] to point to their final
   *             values.
   *
   *           For each triangle, if it's not original, we look at its
   *            adj[i][j]. i indexes each of its sides. If j=ADJ_INDEX gives
   *            its parentindex, then we look at j=ADJ_VERTEX, which gives
   *            some number q. q / 3 is the side of the parent that we are
   *            sharing, and q % 3 is the fragment that we are (0, 1 or 2).
   *
   *           The parent's adj[q / 3] gives us another triangle, which will
   *            have been a peer of the parent. We need to access its newadj
   *            and find out what it was split into (if anything).
   *
   *           Assert that this newtri doesn't have status != SPLIT
   *
   *           newadj[parent->adj[q / 3][ADJ_VERTEX]][k][l] is the side that we
   *            want to look at. If we are (q % 3) of 2, we assert that
   *            k = 1, l = ADJ_INDEX is -1 (this side wasn't split), and we
   *            take our adjacency from k = 0, l = both.
   *            NB here: we must set newtri->adj[] up so that it points back
   *             to us, as otherwise the triangle won't know that our parent
   *             triangle was split up.
   *           If we are (q % 3) of 0 or 1, we assert that k = 1, l = ADJ_INDEX
   *            is not -1 (this side WAS split), and we take our adjacency
   *            from k = (q % 3), l = both.
   *
   *           That's how we work out our adjacencies on a side, if that side
   *            was the edge of the parent triangle. If the side WASN'T an edge
   *            of the parent triangle, then we just leave adj[] alone. While
   *            the triangle adjacent to us may have been split, when it
   *            checks its triangle they'll update our adjacencies for us.
   */
  
  for (trinum = 0, current_tri = workbuf; trinum < current_numtri; trinum++, current_tri++) {  
    if (current_tri->original) {
      // Because some triangles were split, we need to do some adjacency
      // munging. If there are any original triangle pointers in
      // here, we need to make a copy of this triangle so that we can
      // play with its adjacency parameters without damaging the parent
      // collection's data.
      new_tri = modeller->trimem->allocate();
      *new_tri = *(current_tri->t);
      current_tri->t = new_tri;
    }
    
    // For each side of this triangle...
    for (i = 0; i < 3; i++) {
      if ((!current_tri->original) &&
        (current_tri->t->adj[i][ADJ_INDEX] == current_tri->parentindex)) {
        // We don't know what this side of the triangle is against; all we
        // know is that it was a side of the parent triangle.
        q = current_tri->t->adj[i][ADJ_VERTEX];
        parent_edge = q / 3;
        fragment_num = q % 3;
        
        // We need to look at the parent's adjacency matrix on parent_edge to see which
        // one of the parent's peers was on that edge.
        assert((current_tri->parentindex >= 0) && (current_tri->parentindex < trinum));
        parent_tri = &workbuf[current_tri->parentindex];
        peer_index = parent_tri->t->adj[parent_edge][ADJ_INDEX];
        peer_edge  = parent_tri->t->adj[parent_edge][ADJ_VERTEX];
        
        if (peer_index == -1) {
          // The edge is an edge into the void...
          adj_index = -1;
        } else {
          assert((peer_index >= 0) && (peer_index < current_numtri));
          
          if (fragment_num == 2) {
            // We are adjacent to the parent's peer triangle. It's as simple as that.
            adj_index = peer_index;
            adj_vertex = peer_edge;
          } else {
            peer_tri = &workbuf[peer_index];
            // We are only half of the parent's edge. We need to know what the peer
            // was split into; its newadj[][][] array will tell us.
            // The peer triangle must have been split; specifically, it must have
            // been split along this edge..
            if (peer_tri->status != SPLIT) {
              // Okay. Our peer triangle hasn't been split; this must have been
              // caused by numerical errors in the splitting code. We'll just have
              // to use it as our neighbour, even though we are only adjacent to half
              // of it.
              adj_index = peer_index;
              adj_vertex = peer_edge;
            } else if (peer_tri->newadj[peer_edge][1][ADJ_INDEX] == -1) {
              // Our peer triangle hasn't been split along this edge. Use its only
              // child along this edge; same rationale as above.
              adj_index  = peer_tri->newadj[peer_edge][0][ADJ_INDEX ];
              adj_vertex = peer_tri->newadj[peer_edge][0][ADJ_VERTEX];
            } else {
              // We are adjacent to one of our parent's peer's children.
              // Read which triangle this is from peer_tri->newadj.
              //
              // NB: Because the peer's edge is oriented in the opposite
              // direction to the parent's edge, we must read from its
              // newadj[peer_edge][1 - fragment_num]
              adj_index  = peer_tri->newadj[peer_edge][1 - fragment_num][ADJ_INDEX ];
              adj_vertex = peer_tri->newadj[peer_edge][1 - fragment_num][ADJ_VERTEX];
            }
          }
        }
      } else {
        // We know what we're adjacent to; it is another child of our
        // parent triangle.
        adj_index  = current_tri->t->adj[i][ADJ_INDEX];
        adj_vertex = current_tri->t->adj[i][ADJ_VERTEX];
      }
      
      if (adj_index != -1) {
        assert((adj_index >= 0) && (adj_index < current_numtri));
        assert((adj_vertex >= 0) && (adj_vertex < 3));
      }
      
      // Don't try to descend our children if we are
      // adjacent to the void.
      if (adj_index != -1) {
        // Now we have a triangle that we are adjacent to. We know that it can't
        // have been split along its adj_vertex edge in a split operation that
        // relates to us, because otherwise we would have been as well. So, if
        // it's split along there, this must be related to a split of one of our
        // children.
        while(workbuf[adj_index].status == SPLIT) {
          // This triangle has been split. If it's not split along this edge,
          // then we're interested in it (and continue). If it is split along
          // the edge in question, then...
          if ((workbuf[adj_index].status == SPLIT) &&
            (workbuf[adj_index].newadj[adj_vertex][1][ADJ_INDEX] != -1)) {
            // Okay, it has been split along this edge. Let our child deal
            // with this. If we don't have a child, there is a problem.
            if (current_tri->status != SPLIT) {
              // We haven't been split but our neighbour has. We need to set
              // our adjacency to one of its children, and make the other child
              // point off into the void. This is a kludge to get around problems
              // that occur with face segment ends being mistaken for edge segment
              // ends.
              
              // Find the edge vector of the high-child.
              edgept_index  = workbuf[adj_index].newadj[adj_vertex][1][ADJ_INDEX];
              edgept_vertex = workbuf[adj_index].newadj[adj_vertex][1][ADJ_VERTEX];
              
              assert((edgept_index >= 0) && (edgept_index < current_numtri));
              assert((edgept_vertex >= 0) && (edgept_vertex < 3));
              
              // Calculate the size of this edge vector compared to our edge vector.
              // Note that it's going to be negative because it's pointing in the opposite
              // direction.
              Vector::sub(current_tri->t->v[(i + 1) % 3], current_tri->t->v[i], our_edge);
              Vector::sub(workbuf[edgept_index].t->v[(edgept_vertex + 1) % 3],
                workbuf[edgept_index].t->v[edgept_vertex], other_edge);
              
              edgeval = Vector::dot(other_edge, our_edge) / Vector::dot(our_edge, our_edge);
              if ((edgeval > 0) || (edgeval < -1))
                TRACE("S%05d WARNING: split/unsplit mismatch at edgeval outside [-1, 0]\n",
                trisplit_debugindex);
              
              // Connect to one child, and set the other to point to the void.
              if (-edgeval > 0.5) {
                // Disconnect the low child.
                disconnect_index  = workbuf[adj_index].newadj[adj_vertex][0][ADJ_INDEX];
                disconnect_vertex = workbuf[adj_index].newadj[adj_vertex][0][ADJ_VERTEX];
                
                // The high child is the one that we should connect to, as it has
                // the largest portion of the edge. Note that we already know
                // which one this is as we had to access its vertices to find
                // the edgept. We leave it in edgept.
              } else {
                // Disconnect the high child (again, we already know which one this is).
                disconnect_index  = edgept_index;
                disconnect_vertex = edgept_vertex;
                
                // The low child is the one that we should connect to, as it has
                // the largest portion of the edge.
                edgept_index  = workbuf[adj_index].newadj[adj_vertex][0][ADJ_INDEX];
                edgept_vertex = workbuf[adj_index].newadj[adj_vertex][0][ADJ_VERTEX];
              }
              
              workbuf[disconnect_index].t->adj[disconnect_vertex][ADJ_INDEX] = -1;
              adj_index  = edgept_index;
              adj_vertex = edgept_vertex;
              break;
            } else {
              // We do have a child, let them deal with it.
              break;
            }
          }
          
          // The next level down is accessible, Get it.
          new_index  = workbuf[adj_index].newadj[adj_vertex][0][ADJ_INDEX];
          new_vertex = workbuf[adj_index].newadj[adj_vertex][0][ADJ_VERTEX];
          
          assert((new_index > adj_index) && (new_index < current_numtri));
          assert((new_vertex >= 0) && (new_vertex < 3));
          
          adj_index = new_index;
          adj_vertex = new_vertex;
          
          // If we're at the bottom level, fall out of the loop because
          // status will != SPLIT, so we'll fail the while condition.
        }
      }
      
      // Set the new adjacency in our triangle.
      current_tri->t->adj[i][ADJ_INDEX]  = adj_index;
      current_tri->t->adj[i][ADJ_VERTEX] = adj_vertex;
      
#if _DEBUG
      if ((current_tri->status != SPLIT) && (workbuf[adj_index].status == SPLIT)) {
        // There is a split/unsplit mismatch here.
        TRACE("S%05d WARNING: split/unsplit mismatch in recalculate_adj!\n", trisplit_debugindex);
      }
#endif
    }
  }
}

// Calculate final cell-relative indices for all triangles that are kept.
void TriSplitter::index_final_triangles(void)
{
  int found_keep, found_keep_inferred, found_disc, found_disc_inferred, found_leakage;
  int next_traverse_id, this_traverse_id, i, q;
  inoutstatus traverse_result;

 /* Phase Three: Tag all child triangles with KEEP or DISCARD.
  *
  *              In this phase, we do a graph-search across the adjacency
  *               values, looking for triangles which have already been
  *               tagged. If we can't find any adjacent triangles that have
  *               been tagged (i.e. not part of a connected component that was
  *               split), then we cast a ray.
  *
  *              We give each kept triangle a stringrel_outputindex which is
  *               relative to the string that they lie in. These must then be
  *               converted into cell-relative indices for adjacency encoding.
  */
  
  for (i = 0; i < num_in_strings; i++)
    string_length[i] = 0;
  num_output_tri = 0;
  next_traverse_id = 0;
  
  // Classify all triangles that are left as status=UNKNOWN.
  // We do this by looking for adjacent KEEP or DISCARD results, and if none are
  // found then we call the RCSG's point-intersection routine.
  for (tosplit_index = 0; tosplit_index < current_numtri; tosplit_index++) {
    if (workbuf[tosplit_index].status != SPLIT) {
      // We have to find out what side of the object this triangle
      // lies on if it is UNKNOWN. Note that if we haven't split any then
      // we ignore all status_inferred K/D settings.
      if ((workbuf[tosplit_index].status == UNKNOWN) || 
        (workbuf[tosplit_index].status_inferred && !split_any)) {
        // Begin a graph-traverse from this triangle.
        this_traverse_id = next_traverse_id++;
        found_keep = found_keep_inferred = found_disc = found_disc_inferred = 0;
        graph_traverse(tosplit_index, -1, this_traverse_id,
          !split_any,
          &found_keep, &found_keep_inferred,
          &found_disc, &found_disc_inferred);
        assert(workbuf[tosplit_index].traverse_id == this_traverse_id);
        
        found_leakage = FALSE;
        if (found_keep) {
          if (found_disc) {
            found_leakage = TRUE;
          } else {
            traverse_result = KEEP;
          }
        } else if (found_disc) {
          traverse_result = DISCARD;
        } else if (found_keep_inferred) {
          if (found_disc_inferred) {
            found_leakage = TRUE;
          } else {
            traverse_result = KEEP;
          }
        } else if (found_disc_inferred) {
          traverse_result = DISCARD;
        } else {
          // The graph traverse was inconclusive. It must be a non-connected
          // component. We have to perform a point-intersect vs the RCSGgraph.
          if (rcsg_intersect(tosplit_index)) {
            // INSIDE... do we keep?
            traverse_result = (keep_normal < 0) ? KEEP : DISCARD;
          } else {
            // OUTSIDE... do we keep?
            traverse_result = (keep_normal > 0) ? KEEP : DISCARD;
          }
        } 
        
        // We must update all the triangles that we visited during this traverse.
        for (q = tosplit_index; q < current_numtri; q++) {
          if (workbuf[q].traverse_id == this_traverse_id) {
            if (found_leakage) {
              // There is I/O leakage on this traverse. We need to do an RCSG
              // point-intersect for EVERY triangle.
              if (rcsg_intersect(q)) {
                // INSIDE... do we keep?
                traverse_result = (keep_normal < 0) ? KEEP : DISCARD;
              } else {
                // OUTSIDE... do we keep?
                traverse_result = (keep_normal > 0) ? KEEP : DISCARD;
              }
            }
            
            // Don't need to set status_inferred, it isn't used after Phase One.
            workbuf[q].status = traverse_result;
          }
        }
      } 
    }
    
    if (workbuf[tosplit_index].status == KEEP) {
      // Keep this triangle. Increment the length of its string.
      workbuf[tosplit_index].stringrel_outputindex = string_length[workbuf[tosplit_index].this_string_index]++;
      num_output_tri++;
    } else  {
      workbuf[tosplit_index].stringrel_outputindex = -1;
    }
  }
}

// Copy all final triangles to the child Collection.
void TriSplitter::output_triangles(void)
{
  int num_output_strings, peer_index, i, q;
  splittri *current_tri;
  string *last_cs_string, *this_cs_string;
  worldstring *last_ws_string, *this_ws_string;
  tri *this_tri, *new_tri;

  /* Phase Four: Copy all final triangles to the child Collection.
   *
   *             Note that we must update adj[i][ADJ_INDEX] to correspond
   *              to the triangle's finalindex, assigned in the last pass.
   *              This is an index into the cell's tribuf.
   */
  
  // Set up values for the output pass.
  num_output_strings = 0;
  for (i = 0; i < num_in_strings; i++) {
    first_tri[i] = last_tri[i] = NULL;

    if (i == 0)
      stringhead_cellrel_index[i] = 0;
    else
      stringhead_cellrel_index[i] = stringhead_cellrel_index[i - 1] + string_length[i - 1];

    if (string_length[i] == 0)
      final_stringindex[i] = -1;
    else
      final_stringindex[i] = num_output_strings++;
  }

  for (i = 0, current_tri = workbuf; i < current_numtri; i++, current_tri++) {
    if (current_tri->stringrel_outputindex == -1) {
      // This triangle has been split or discarded. If it
      // wasn't an original triangle, or it was an original triangle
      // that was copied because of split_any, then we need to deallocate it.
      if ((!current_tri->original) || (split_any)) {
        modeller->trimem->deallocate(current_tri->t);
      }
      continue;
    }

    if ((current_tri->original) && (!split_any)) {
      // Because no triangles were split, we didn't do the adjacency
      // munging, so there will still be original triangle pointers in
      // here. We need to make a copy of this triangle so that it can
      // be added to the child collection.
      new_tri = modeller->trimem->allocate();
      *new_tri = *(current_tri->t);
      current_tri->t = new_tri;
    }
      
    // Update the triangle's adjacency information.
    for (q = 0; q < 3; q++) {
      peer_index = current_tri->t->adj[q][ADJ_INDEX];
      if (peer_index != -1) {
        // We have a peer triangle next to us. Has it been discarded?
        if (workbuf[peer_index].stringrel_outputindex == -1) {
          // Our peer has been discarded, we are now adjacent
          // to the void.
          current_tri->t->adj[q][ADJ_INDEX] = -1;
        } else {
          // Update our adjacency index to be a final outputindex relative to
          // the cell... this is string-relative index plus string's index relative
          // to the cell.
          current_tri->t->adj[q][ADJ_INDEX] = workbuf[peer_index].stringrel_outputindex
               + stringhead_cellrel_index[workbuf[peer_index].this_string_index];
        }
      }
    }

    assert(current_tri->status == KEEP);

    if (last_tri[current_tri->this_string_index] == NULL) {
      // This is the first triangle in the string.
      first_tri[current_tri->this_string_index] = current_tri;
    } else {
      // There is another triangle before us. Link it up.
      last_tri[current_tri->this_string_index]->t->stringptr = current_tri->t;
    }
    last_tri[current_tri->this_string_index] = current_tri;

    current_tri->t->string_length = string_length[current_tri->this_string_index]
             - current_tri->stringrel_outputindex;
  }

  // Now, set up each of the strings that we're going to store in the cell.
  last_cs_string = NULL;
  last_ws_string = NULL;
  for (i = 0; i < num_in_strings; i++) {
    if (string_length[i] == 0)
      continue;

    this_cs_string = modeller->stringmem->allocate();
    this_ws_string = modeller->worldstringmem->allocate();

    Vector::set(this_cs_string->norm, in_string[i].cs_str->norm);
    this_cs_string->plane_dot = in_string[i].cs_str->plane_dot;
    Vector::set(this_ws_string->norm, in_string[i].ws_str->norm);
    this_ws_string->plane_dot = in_string[i].ws_str->plane_dot;

    if (last_cs_string == NULL) {
      // This is the first of the strings.
      assert(final_stringindex[i] == 0);
      assert(last_ws_string == NULL);
      child->cell[cell_index] = this_ws_string;
      child->collx->cell[cell_index] = this_cs_string;

      this_ws_string->reference_count = this_cs_string->reference_count = 1;
    } else {
      assert(last_ws_string != NULL);
      last_cs_string->stringptr = this_cs_string;
      last_ws_string->stringptr = this_ws_string;

      this_ws_string->reference_count = this_cs_string->reference_count = -1;
    }

    // Set up the start of the string.
    assert(first_tri[i] != NULL);
    this_tri = this_cs_string->head = first_tri[i]->t;

    // Set up the end of the string.
    assert(last_tri[i] != NULL);
    last_tri[i]->t->stringptr = NULL;

    // Number of triangles in all strings past this one = total triangles - triangles before us.
    this_cs_string->num_tri = num_output_tri - stringhead_cellrel_index[i];

    // Set up our position in the cell-list.
    assert((final_stringindex[i] >= 0) && (final_stringindex[i] < num_output_strings));
    this_ws_string->celllist_length = this_cs_string->celllist_length = num_output_strings - final_stringindex[i];

    last_ws_string = this_ws_string;
    last_cs_string = this_cs_string;
  }
  last_cs_string->stringptr = NULL;
  last_ws_string->stringptr = NULL;

  // Done.
  child->collx->total_tri += num_output_tri;
}

// Check to see if the current triangle is intersected by any of the intervals along
// the line of intersection in which there are split triangles present. 
int TriSplitter::check_split(int tri_index, int string_index, int split_string_index)
{
  splittri *t;
  int q, nz0, nz1, exit_is_inside;
  double ev0, ev1, interp_val;
  int search_sign, interval_index, tval_sign;
  int keep_top, top_pos;
  linept entrybound, exitbound;

  t = &workbuf[tri_index];

  if (t->tvals_wrt != split_string_index) {
    // Calculate t-values for the vertices of this triangle.
    for (q = 0; q < 3; q++)
      t->vertex_t[q] = (Vector::dot(t->t->v[q], in_string[string_index].cs_line_dir)
             + in_string[string_index].cs_t_offset) * in_string[string_index].cs_t_scale;

    // Determine where the line of intersection crosses this triangle.
    if (t->num_positive == 2) {
      assert(t->num_negative == 1);
      // Two positive, one negative. Look for the negative vertex.
      search_sign = -1;

    } else if (t->num_negative == 2) {
      assert(t->num_positive == 1);
      // Two negative, one positive. Look for the positive vertex.
      search_sign = +1;

    } else {
      assert((t->num_positive == 1) && (t->num_negative == 1));
      // One negative, one positive, one zero.
      search_sign = 0;
    }

    if (search_sign) {
      // We are looking for an entry and an exit on edges. First,
      // find the odd vertex out.
      if (t->plane_dist[0] * search_sign < 0) {
        if (t->plane_dist[1] * search_sign < 0) {
          assert(t->plane_dist[2] * search_sign > 0);
          // Vertex 2 is the odd one out, entry and exit are on edges 1 and 2.
          ev0 = -t->plane_dist[1] / (t->plane_dist[2] - t->plane_dist[1]);
          ev1 = -t->plane_dist[2] / (t->plane_dist[0] - t->plane_dist[2]);

          assert((ev0 >= 0) && (ev0 <= 1));
          assert((ev1 >= 0) && (ev1 <= 1));
          
          t->entry.t = (1 - ev0) * t->vertex_t[1] + ev0 * t->vertex_t[2];
          t->entry.location = PT_EDGE;
          t->entry.index = 1;
          t->entry.edgeval = ev0;
          Vector::mult(1.0 - ev0, t->t->v[1], t->entry.pt);
          Vector::addfrac(   ev0, t->t->v[2], t->entry.pt);
          if (modeller->use_gouraud) {
            Vector::mult(1.0 - ev0, t->t->n[1], t->entry.n);
            Vector::addfrac(   ev0, t->t->n[2], t->entry.n);
            Vector::normalise(t->entry.n, "TriSplitter::split_triangles: entry normal");
          }
          
          t->exit.t = (1 - ev1) * t->vertex_t[2] + ev1 * t->vertex_t[0];
          t->exit.location = PT_EDGE;
          t->exit.index = 2;
          t->exit.edgeval = ev1;
          Vector::mult(1.0 - ev1, t->t->v[2], t->exit.pt);
          Vector::addfrac(   ev1, t->t->v[0], t->exit.pt);
          if (modeller->use_gouraud) {
            Vector::mult(1.0 - ev1, t->t->n[2], t->exit.n);
            Vector::addfrac(   ev1, t->t->n[0], t->exit.n);
            Vector::normalise(t->exit.n, "TriSplitter::split_triangles: exit normal");
          }
        } else {
          assert(t->plane_dist[2] * search_sign < 0);
          // Vertex 1 is the odd one out, entry and exit are on edges 0 and 1.
          ev0 = -t->plane_dist[0] / (t->plane_dist[1] - t->plane_dist[0]);
          ev1 = -t->plane_dist[1] / (t->plane_dist[2] - t->plane_dist[1]);
          
          assert((ev0 >= 0) && (ev0 <= 1));
          assert((ev1 >= 0) && (ev1 <= 1));
          
          t->entry.t = (1 - ev0) * t->vertex_t[0] + ev0 * t->vertex_t[1];
          t->entry.location = PT_EDGE;
          t->entry.index = 0;
          t->entry.edgeval = ev0;
          Vector::mult(1.0 - ev0, t->t->v[0], t->entry.pt);
          Vector::addfrac(   ev0, t->t->v[1], t->entry.pt);
          if (modeller->use_gouraud) {
            Vector::mult(1.0 - ev0, t->t->n[0], t->entry.n);
            Vector::addfrac(   ev0, t->t->n[1], t->entry.n);
            Vector::normalise(t->entry.n, "TriSplitter::split_triangles: entry normal");
          }
          
          t->exit.t = (1 - ev1) * t->vertex_t[1] + ev1 * t->vertex_t[2];
          t->exit.location = PT_EDGE;
          t->exit.index = 1;
          t->exit.edgeval = ev1;
          Vector::mult(1.0 - ev1, t->t->v[1], t->exit.pt);
          Vector::addfrac(   ev1, t->t->v[2], t->exit.pt);
          if (modeller->use_gouraud) {
            Vector::mult(1.0 - ev1, t->t->n[1], t->exit.n);
            Vector::addfrac(   ev1, t->t->n[2], t->exit.n);
            Vector::normalise(t->exit.n, "TriSplitter::split_triangles: exit normal");
          }
        }
      } else {
        assert((t->plane_dist[1] * search_sign < 0) && (t->plane_dist[2] * search_sign < 0));
        // Vertex 0 is the odd one out, entry and exit are on edges 2 and 0.
        ev0 = -t->plane_dist[2] / (t->plane_dist[0] - t->plane_dist[2]);
        ev1 = -t->plane_dist[0] / (t->plane_dist[1] - t->plane_dist[0]);
        
        assert((ev0 >= 0) && (ev0 <= 1));
        assert((ev1 >= 0) && (ev1 <= 1));
        
        t->entry.t = (1 - ev0) * t->vertex_t[2] + ev0 * t->vertex_t[0];
        t->entry.location = PT_EDGE;
        t->entry.index = 2;
        t->entry.edgeval = ev0;
        Vector::mult(1.0 - ev0, t->t->v[2], t->entry.pt);
        Vector::addfrac(   ev0, t->t->v[0], t->entry.pt);
        if (modeller->use_gouraud) {
          Vector::mult(1.0 - ev0, t->t->n[2], t->entry.n);
          Vector::addfrac(   ev0, t->t->n[0], t->entry.n);
          Vector::normalise(t->entry.n, "TriSplitter::split_triangles: entry normal");
        }
        
        t->exit.t = (1 - ev1) * t->vertex_t[0] + ev1 * t->vertex_t[1];
        t->exit.location = PT_EDGE;
        t->exit.index = 0;
        t->exit.edgeval = ev1;
        Vector::mult(1.0 - ev1, t->t->v[0], t->exit.pt);
        Vector::addfrac(   ev1, t->t->v[1], t->exit.pt);
        if (modeller->use_gouraud) {
          Vector::mult(1.0 - ev1, t->t->n[0], t->exit.n);
          Vector::addfrac(   ev1, t->t->n[1], t->exit.n);
          Vector::normalise(t->exit.n, "TriSplitter::split_triangles: exit normal");
        }
      }
    } else {
      // There is an entry at a zero vertex; the exit lies along the
      // edge between the two non-zero vertices.
      nz0 = (t->zero_vertex + 1) % 3;
      nz1 = (t->zero_vertex + 2) % 3;

      ev0 = -t->plane_dist[nz0] / (t->plane_dist[nz1] - t->plane_dist[nz0]);
      assert((ev0 >= 0) && (ev0 <= 1));

      t->entry.t = t->vertex_t[t->zero_vertex];
      t->entry.location = PT_VERTEX;
      t->entry.index = t->zero_vertex;
      Vector::set(t->entry.pt, t->t->v[t->zero_vertex]);
      if (modeller->use_gouraud)
        Vector::set(t->entry.n, t->t->n[t->zero_vertex]);
      
      t->exit.t = (1 - ev0) * t->vertex_t[nz0] + ev0 * t->vertex_t[nz1];
      t->exit.location = PT_EDGE;
      t->exit.index = nz0;
      t->exit.edgeval = ev0;
      Vector::mult(1.0 - ev0, t->t->v[nz0], t->exit.pt);
      Vector::addfrac(   ev0, t->t->v[nz1], t->exit.pt);
      if (modeller->use_gouraud) {
        Vector::mult(1.0 - ev0, t->t->n[nz0], t->exit.n);
        Vector::addfrac(   ev0, t->t->n[nz1], t->exit.n);
        Vector::normalise(t->exit.n, "TriSplitter::split_triangles: exit normal");
      }
    }

    t->tvals_wrt = split_string_index;
    t->next_interval = 0;
  }

  // We now have t->entry and t->exit determined with respect to the
  // current split plane.

  // For every t-interval of intersection between our current input string and
  // the current split string, check for overlaps. We start from next_interval.
  for (interval_index = t->next_interval;
       interval_index < in_string[string_index].num_intervals; interval_index++) {

    if (t->exit.t > t->entry.t) {
      // The exit intersection is the high-t one. Thus interval[1] bounds exit
      // and interval[0] bounds entry.
      entrybound.t = in_string[string_index].intervals[interval_index][0];
      exitbound.t  = in_string[string_index].intervals[interval_index][1];

      tval_sign = +1;  // t is in ascending order from entry to exit

      if ((t->exit.t - entrybound.t < INTERVAL_TOLERANCE) ||
          (exitbound.t - t->entry.t < INTERVAL_TOLERANCE)) {
        // This interval does not overlap the triangle by at least INTERVAL_TOLERANCE. No split.
        continue;
      }

    } else {
      // The exit intersection is the low-t one. Thus interval[0] bounds exit
      // and interval[1] bounds entry.
      entrybound.t = in_string[string_index].intervals[interval_index][1];
      exitbound.t  = in_string[string_index].intervals[interval_index][0];

      tval_sign = -1;  // t is in descending order from entry to exit

      if ((t->entry.t - exitbound.t < INTERVAL_TOLERANCE) ||
          (entrybound.t - t->exit.t < INTERVAL_TOLERANCE)) {
        // This interval does not overlap the triangle by at least INTERVAL_TOLERANCE. No split.
        continue;
      }
    }

    // Well, there IS an overlap. We must split this triangle.


    // This triangle's children will start checking from the next interval
    // (if they have to).
    t->next_interval = interval_index + 1;

    // We keep the "top" half of the triangle (the part around the low-vertex
    // on the exit edge) if the vertex in that area is on the correct side of the
    // plane.
    top_pos = (t->plane_dist[t->exit.index] > 0);
    keep_top = (keep_normal * t->plane_dist[t->exit.index] > 0);

    if ((t->exit.t - exitbound.t) * tval_sign > INTERVAL_TOLERANCE) {
      // The exit-bound of the interval is before the triangle's exit by more than
      // INTERVAL_TOLERANCE and so we must create a face point.
      exit_is_inside = TRUE;

      // We need to determine whereabouts this point
      // lies in relation to the entry and exit points.
      interp_val = (exitbound.t - t->entry.t) / (t->exit.t - t->entry.t);
      assert((interp_val >= 0) && (interp_val <= 1));

      exitbound.location = PT_FACE;
      
      Vector::mult(1.0 - interp_val, t->entry.pt, exitbound.pt);
      Vector::addfrac(   interp_val, t->exit .pt, exitbound.pt);
      if (modeller->use_gouraud) {
        Vector::mult(1.0 - interp_val, t->entry.n, exitbound.n);
        Vector::addfrac(   interp_val, t->exit .n, exitbound.n);
      }
    } else {
      exit_is_inside = FALSE;
    }

    // Classify based on where the interval STARTS.
    if ((entrybound.t - t->entry.t) * tval_sign > INTERVAL_TOLERANCE) {
      // The entry-bound of the interval is after the triangle's entry by more than
      // INTERVAL_TOLERANCE and so we must create a face point.

      // We need to determine whereabouts this point
      // lies in relation to the entry and exit points.
      interp_val = (entrybound.t - t->entry.t) / (t->exit.t - t->entry.t);
      assert((interp_val >= 0) && (interp_val <= 1));

      entrybound.location = PT_FACE;
      
      Vector::mult(1.0 - interp_val, t->entry.pt, entrybound.pt);
      Vector::addfrac(   interp_val, t->exit .pt, entrybound.pt);
      if (modeller->use_gouraud) {
        Vector::mult(1.0 - interp_val, t->entry.n, entrybound.n);
        Vector::addfrac(   interp_val, t->exit .n, entrybound.n);
      }

      if (exit_is_inside) {
        // Both entry and exit lie inside the bounds of the triangle.
        // Perform a face-face split.
        split_face_face(tri_index, top_pos, keep_top, &entrybound, &exitbound);

      } else {
        // The exit is determined by the bounds of the triangle.
        assert(t->exit.location == PT_EDGE);  // Must be EDGE as the exit should never lie on a vertex.

        // Perform a face-edge split. The exit strikes the edge.
        split_face_edge_exit(tri_index, top_pos, keep_top, &entrybound);
      }
    } else if (t->entry.location == PT_EDGE) {
      // The start of the interval is on the triangle's edge.

      if (exit_is_inside) {
        // Perform a face-edge split. The exit is inside the triangle's bounds, so
        // only the entry strikes the edge.
        split_face_edge_entry(tri_index, top_pos, keep_top, &exitbound);

      } else {
        // The exit is determined by the bounds of the triangle.
        assert(t->exit.location == PT_EDGE);  // Must be EDGE as the exit should never lie on a vertex.

        // Perform an edge-edge split.
        split_edge_edge(tri_index, top_pos, keep_top);
      }
    } else {
      assert(t->entry.location == PT_VERTEX);
      // The start of the interval is on a vertex of the triangle.

      if (exit_is_inside) {
        // The exit of the interval comes before we hit the far edge.
        // Perform a vertex-face split.
        split_face_vertex(tri_index, top_pos, keep_top, &exitbound);

      } else {
        // The exit is determined by the bounds of the triangle.
        assert(t->exit.location == PT_EDGE);  // Must be EDGE as the exit should never lie on a vertex.

        // Perform a vertex-edge split.
        split_edge_vertex(tri_index, top_pos, keep_top);
      }
    }

    // The triangle has been split.
    workbuf[tri_index].status = SPLIT;
    return TRUE;
  }

  // If we reach here, then this triangle did not overlap any
  // of the split intervals.
  return FALSE;
}

// Look at a complete string of splitagainst triangles, test them against
// an input plane, and determine the t-values along the intersection line
// at which there are split triangles present.
void TriSplitter::find_t_intervals(int in_index, int split_index)
{
  int q, num_positive, num_negative, zero_vertex, found, zv1, zv2, i, no_store;
  double plane_dist[3], vertex_t[3], interval[2], edgeval, low_t, high_t, cs_t_offset, cs_t_scale;
  tri *current_tri;
  instringdata *is = &in_string[in_index];
  vector in_norm, cs_line_dir;
  double in_dot;

  // Transform the input string's normal and the direction of the line of
  // intersection from worldspace into the collectionspace of the split string.
  in_dot = is->ws_str->plane_dot;
  cs_t_offset = 0;
  if (split_string[split_index].owner->owner->is_transformed) {
    Matrix::applytranspose(split_string[split_index].owner->owner->collection_trans[FORE],
                           is->ws_str->norm, in_norm);
    Matrix::applynoshift(split_string[split_index].owner->owner->collection_trans[AFT],
                         is->ws_line_dir, cs_line_dir);

    for (q = X; q <= Z; q++) {
      in_dot      += split_string[split_index].owner->owner->collection_trans[AFT][q][3] * in_norm[q];
      cs_t_offset -= split_string[split_index].owner->owner->collection_trans[AFT][q][3] * cs_line_dir[q];
    }
    cs_t_scale = Vector::dot(is->ws_line_dir, is->ws_line_dir) / Vector::dot(cs_line_dir, cs_line_dir);
  } else {
    Vector::set(in_norm, is->ws_str->norm);
    Vector::set(cs_line_dir, is->ws_line_dir);
    cs_t_scale = 1.0;
  }

  is->num_intervals = 0;
  for (current_tri = split_string[split_index].cs_str->head; current_tri != NULL;
       current_tri = current_tri->stringptr) {
    // Where are this triangle's vertices with respect to the input string's plane?
    num_positive = num_negative = 0;
    zero_vertex = -1;
    for (q = 0; q < 3; q++) {
      plane_dist[q] = Vector::dot(current_tri->v[q], in_norm) - in_dot;
      if (plane_dist[q] > PLANE_TOLERANCE) {
        num_positive++;
      } else if (plane_dist[q] < -PLANE_TOLERANCE) {
        num_negative++;
      } else {
        zero_vertex = q;
      }
    }

    // We are only interested in split triangles that cross from one
    // side of the plane to the other.
    if ((num_positive == 0) || (num_negative == 0))
      continue;
    
    // Calculate t-values for every vertex.
    for (q = 0; q < 3; q++)
      vertex_t[q] = (Vector::dot(current_tri->v[q], cs_line_dir) + cs_t_offset) * cs_t_scale;
    
    // Find the interval in t where this triangle is present along the line of intersection.
    if (zero_vertex == -1) {
      // None of the vertices lie _on_ the line of intersection. Calculate the intersection
      // points along the edges of the triangle.
      found = 0;
      for (q = 0; q < 3; q++) {
        if (plane_dist[q] * plane_dist[(q + 1) % 3] > 0)
          continue;
        
        edgeval = -plane_dist[q] / (plane_dist[(q + 1) % 3] - plane_dist[q]);
        assert((edgeval >= 0) && (edgeval <= 1));
        
        interval[found++] = (1.0 - edgeval) * vertex_t[q] + edgeval * vertex_t[(q + 1) % 3];
      }
      assert(found == 2);
    } else {
      interval[0] = vertex_t[zero_vertex];
      
      zv1 = (zero_vertex + 1) % 3;
      zv2 = (zero_vertex + 2) % 3;
      
      assert(plane_dist[zv1] * plane_dist[zv2] < 0);
      
      edgeval = -plane_dist[zv1] / (plane_dist[zv2] - plane_dist[zv1]);
      assert((edgeval >= 0) && (edgeval <= 1));
      
      interval[1] = (1.0 - edgeval) * vertex_t[zv1] + edgeval * vertex_t[zv2];
    }
    
    if (interval[0] < interval[1]) {
      low_t  = interval[0];
      high_t = interval[1];
    } else {
      low_t  = interval[1];
      high_t = interval[0];
    }
      
    no_store = FALSE;
    for (i = 0; i < is->num_intervals; i++) {
      if (is->intervals[i][0] - high_t > INTERVAL_TOLERANCE) {
        // This interval is more than INTERVAL_TOLERANCE above the upper bound of our
        // current interval. Keep looking.
        continue;
      } else if (low_t - is->intervals[i][1] > INTERVAL_TOLERANCE) {
        // This interval is more than INTERVAL_TOLERANCE below the lower bound of our
        // current interval. Keep looking.
        continue;
      }

      // This interval overlaps our current interval and so must be subsumed into it.

      if (is->intervals[i][0] < low_t) {
        // We extend our current interval out to the lower bound of this one.
        low_t = is->intervals[i][0];

        if (is->intervals[i][1] > high_t) {
          // The current interval is completely contained inside this interval
          // and so holds no interest for us. Stop this process.
          no_store = TRUE;
          break;
        }
      } else if (is->intervals[i][1] > high_t) {
        // We extend our current interval out to the upper bound of this one.
        high_t = is->intervals[i][1];
      }

      // Delete this interval.
      if (i < is->num_intervals - 1) {
        is->intervals[i][0] = is->intervals[is->num_intervals - 1][0];
        is->intervals[i][1] = is->intervals[is->num_intervals - 1][1];
      }
      is->num_intervals--;
    }

    if (no_store)
      continue;

    // Store this interval.
    is->intervals[is->num_intervals][0] = low_t;
    is->intervals[is->num_intervals][1] = high_t;
    is->num_intervals++;
  }
}

int TriSplitter::rcsg_intersect(int tri_index)
{
  int i;
  vector w_pt;

  // Find the center of the triangle in question.
  Vector::set(w_pt, zero_vector);
  for (i = 0; i < 3; i++)
    Vector::addfrac(1.0 / 3.0, workbuf[tri_index].t->v[i], w_pt);

  // Transform the center of the triangle from the input collectionspace
  // into worldspace.
  if (parent->owner->is_transformed)
    Matrix::apply(parent->owner->collection_trans[FORE], w_pt, w_pt);

  return parent->world_point_intersect(cell_index, w_pt);
}


// Traverse a connected component of the adjacency graph, looking for triangles whose
// status we know.
void TriSplitter::graph_traverse(int tri_index, int entry_side, int traverse_id,
                                 int ignore_inferred,
                                 int *found_keep, int *found_keep_inferred,
                                 int *found_disc, int *found_disc_inferred)
{
  splittri *t = &workbuf[tri_index];
  int q, nextindex, nextedge;

  // Don't check triangles that we've already visited.
  if (t->traverse_id == traverse_id)
    return;

  // Found one!
  if (t->status == KEEP) {
    if (t->status_inferred) {
      if (!ignore_inferred) {
        (*found_keep_inferred)++;
        return;
      }
    } else {
      (*found_keep)++;
      return;
    }
  } else if (t->status == DISCARD) {
    if (t->status_inferred) {
      if (!ignore_inferred) {
        (*found_disc_inferred)++;
        return;
      }
    } else {
      (*found_disc)++;
      return;
    }
  } else {
    assert(t->status == UNKNOWN);
  }
  
  // Mark that we have visited this triangle.
  t->traverse_id = traverse_id;

  for (q = 0; q < 3; q++) {
    // Don't search out the side we came from.
    if (q == entry_side)
      continue;

    // Continue the traverse.
    nextindex = t->t->adj[q][ADJ_INDEX];
    if (nextindex == -1)
      continue;

    nextedge = t->t->adj[q][ADJ_VERTEX];
    assert((nextindex >= 0) && (nextindex < current_numtri));
    assert(workbuf[nextindex].status != SPLIT);
    assert((nextedge >= 0) && (nextedge < 3));
    graph_traverse(nextindex, nextedge, traverse_id,
                   ignore_inferred,
                   found_keep, found_keep_inferred,
                   found_disc, found_disc_inferred);
  }
}

// Macros to make the code a LOT smaller.
#define PARENTVERTEX(i) workbuf[index].t->v[(i)], workbuf[index].t->n[(i)], workbuf[index].plane_dist[(i)], workbuf[index].vertex_t[(i)]
#define LINEPT(linept) (linept)->pt, (linept)->n, 0, (linept)->t

#define CHILDEDGE_WHOLEPARENT(child, child_index, parent_index) {                 \
  workbuf[(child)].adjfinal[(child_index)] = FALSE;                               \
  workbuf[(child)].t->adj[(child_index)][ADJ_INDEX] = index;                      \
  workbuf[(child)].t->adj[(child_index)][ADJ_VERTEX] = 3 * (parent_index) + 2;    \
  if ((workbuf[index].t->highlight_edge & (1 << (parent_index))) ||               \
      ((workbuf[index].status != UNKNOWN) &&                                      \
       (workbuf[index].status_edge == (parent_index)))) {                         \
    workbuf[(child)].t->highlight_edge |= (1 << (child_index));                   \
  }                                                                               \
}

#define CHILDEDGE_FRAGMENT(child, child_index, parent_index, fragment_index) {    \
  workbuf[(child)].adjfinal[(child_index)] = FALSE;                               \
  workbuf[(child)].t->adj[(child_index)][ADJ_INDEX] = index;                      \
  workbuf[(child)].t->adj[(child_index)][ADJ_VERTEX] = 3 * (parent_index) + (fragment_index); \
  if (workbuf[index].t->highlight_edge & (1 << (parent_index))) {                 \
    workbuf[(child)].t->highlight_edge |= (1 << (child_index));                   \
  }                                                                               \
}

#define CHILDEDGE_NOTHING(child, child_index) {                                   \
  workbuf[(child)].adjfinal[(child_index)] = TRUE;                                \
  workbuf[(child)].t->adj[(child_index)][ADJ_INDEX] = -1;                         \
}

#define CHILDEDGE_CHILD(child, child_index, adj_index, adj_vertex) {              \
  workbuf[(child)].adjfinal[(child_index)] = TRUE;                                \
  workbuf[(child)].t->adj[(child_index)][ADJ_INDEX]  = (adj_index);               \
  workbuf[(child)].t->adj[(child_index)][ADJ_VERTEX] = (adj_vertex);              \
}

#define STATUS_PARENTEDGE(child, child_index, parent_index) {                     \
  /* This triangle lies along the edge of the parent. If this was the             \
     edge that determined the parent's split-status, then we take it, otherwise   \
     we're unknown. */                                                            \
  if ((workbuf[index].t->highlight_edge & (1 << (parent_index))) ||               \
      ((workbuf[index].status != UNKNOWN) &&                                      \
       (workbuf[index].status_edge == (parent_index)))) {                         \
    workbuf[(child)].status = workbuf[index].status;                              \
    workbuf[(child)].status_edge = (child_index);                                 \
    workbuf[(child)].status_inferred = workbuf[index].status_inferred;            \
    workbuf[(child)].t->highlight_edge |= (1 << (child_index));                   \
  } else {                                                                        \
    workbuf[(child)].status = UNKNOWN;                                            \
  }                                                                               \
}

#define STATUS_SPLITEDGE(child, child_index, keep) {                              \
  /* One edge of this triangle lies along the split. */                           \
  workbuf[(child)].status = (keep) ? KEEP : DISCARD;                              \
  workbuf[(child)].status_inferred = FALSE;                                       \
  workbuf[(child)].status_edge = (child_index);                                   \
  workbuf[(child)].t->highlight_edge |= (1 << child_index);                       \
}

#define PARENTEDGE_UNSPLIT(parent_index, child, child_index) {                    \
  workbuf[index].newadj[(parent_index)][0][ADJ_INDEX ] = (child);                 \
  workbuf[index].newadj[(parent_index)][0][ADJ_VERTEX] = (child_index);           \
  workbuf[index].newadj[(parent_index)][1][ADJ_INDEX ] = -1;                      \
}

#define PARENTEDGE_SPLIT(parent_index, lo_child, lo_index, hi_child, hi_index) {  \
  workbuf[index].newadj[(parent_index)][0][ADJ_INDEX ] = (lo_child);              \
  workbuf[index].newadj[(parent_index)][0][ADJ_VERTEX] = (lo_index);              \
  workbuf[index].newadj[(parent_index)][1][ADJ_INDEX ] = (hi_child);              \
  workbuf[index].newadj[(parent_index)][1][ADJ_VERTEX] = (hi_index);              \
}

void TriSplitter::split_face_face(int index, int top_pos, int keep_top, linept *face1, linept *face2)
{
  int entry_index, exit_index, other_index;
  int peak, side1, side2, base1, base2;
  int through_vertex;

  entry_index = workbuf[index].entry.index;
  exit_index = workbuf[index].exit.index;
  other_index = (exit_index + 1) % 3;

  assert(exit_index == (entry_index + 1) % 3);

  // side1 may have one of its edges along the split line, if the entry to this
  // triangle was through a vertex. Otherwise the split line passes through it.
  through_vertex = (workbuf[index].entry.location == PT_VERTEX);


  if (through_vertex) {
    side1 = next_triangle(PARENTVERTEX(entry_index), PARENTVERTEX(exit_index), LINEPT(face1),
                          index, (top_pos) ? 1 : 0, (top_pos) ? 0 : 1, 0);
  } else {
    side1 = next_triangle(PARENTVERTEX(entry_index), PARENTVERTEX(exit_index), LINEPT(face1),
                          index, 1, 1, 2,
                          TRUE, face1, &(workbuf[index].entry));
  }
  peak = next_triangle(PARENTVERTEX(exit_index), LINEPT(face2), LINEPT(face1),
                       index, (top_pos) ? 1 : 0, (top_pos) ? 0 : 1, 1);
  side2 = next_triangle(PARENTVERTEX(exit_index), PARENTVERTEX(other_index), LINEPT(face2),
                        index, 1, 1, 2,
                        TRUE, face2, &(workbuf[index].exit));
  base2 = next_triangle(PARENTVERTEX(other_index), LINEPT(face1), LINEPT(face2),
                        index, (top_pos) ? 0 : 1, (top_pos) ? 1 : 0, 1);
  
  // base1 may also have one of its edges along the split line.
  if (through_vertex) {
    base1 = next_triangle(PARENTVERTEX(other_index), PARENTVERTEX(entry_index), LINEPT(face1),
                          index, (top_pos) ? 0 : 1, (top_pos) ? 1 : 0, 1);
  } else {
    base1 = next_triangle(PARENTVERTEX(other_index), PARENTVERTEX(entry_index), LINEPT(face1),
                          index, (top_pos) ? 0 : 2, (top_pos) ? 2 : 0, 2);
  }

  CHILDEDGE_WHOLEPARENT(side1, 0, entry_index);
  CHILDEDGE_CHILD      (side1, 1, peak, 2);
  if (through_vertex) {
    CHILDEDGE_NOTHING(side1, 2);
  } else {
    CHILDEDGE_CHILD      (side1, 2, base1, 1);
    workbuf[side1].entry.location = PT_VERTEX;
    workbuf[side1].entry.index = 2;
    workbuf[side1].exit.index = 0;
  }
  STATUS_PARENTEDGE(side1, 0, entry_index);

  CHILDEDGE_CHILD  (peak, 0, side2, 2);
  CHILDEDGE_NOTHING(peak, 1);
  CHILDEDGE_CHILD  (peak, 2, side1, 1);
  STATUS_SPLITEDGE (peak, 1, keep_top);
  
  CHILDEDGE_WHOLEPARENT(side2, 0, exit_index);
  CHILDEDGE_CHILD      (side2, 1, base2, 2);
  CHILDEDGE_CHILD      (side2, 2, peak, 0);
  workbuf[side2].entry.location = PT_VERTEX;
  workbuf[side2].entry.index = 2;
  workbuf[side2].exit.index = 0;
  STATUS_PARENTEDGE    (side2, 0, exit_index);

  CHILDEDGE_CHILD  (base2, 0, base1, 2);
  CHILDEDGE_NOTHING(base2, 1);
  CHILDEDGE_CHILD  (base2, 2, side2, 1);
  STATUS_SPLITEDGE (base2, 1, !keep_top);

  CHILDEDGE_WHOLEPARENT(base1, 0, other_index);
  if (through_vertex) {
    CHILDEDGE_NOTHING(base1, 1);
  } else {
    CHILDEDGE_CHILD(base1, 1, side1, 2);
  }
  CHILDEDGE_CHILD(base1, 2, base2, 0);
  STATUS_PARENTEDGE(base1, 0, other_index);

  PARENTEDGE_UNSPLIT(entry_index, side1, 0);
  PARENTEDGE_UNSPLIT(exit_index,  side2, 0);
  PARENTEDGE_UNSPLIT(other_index, base1, 0);
}

void TriSplitter::split_face_edge_entry(int index, int top_pos, int keep_top, linept *face)
{
  int entry_index, exit_index, other_index;
  int top, bottom, side1, side2;

  entry_index = workbuf[index].entry.index;
  exit_index = workbuf[index].exit.index;
  other_index = (exit_index + 1) % 3;

  assert(exit_index == (entry_index + 1) % 3);

  // The split edge touches edge entry_index of the parent
  // triangle. This shouldn't be the edge that the parent
  // was split along (if it was a split-up triangle)
  if ((workbuf[index].status != UNKNOWN) && (!workbuf[index].status_inferred) &&
      (workbuf[index].status_edge == entry_index))
    TRACE("S%05d WARNING: face-edge (entry) split touches parent edge that is a split!\n", trisplit_debugindex);

  bottom = next_triangle(PARENTVERTEX(entry_index), LINEPT(&workbuf[index].entry), LINEPT(face),
                         index, (top_pos) ? 0 : 1, (top_pos) ? 1 : 0, 1);
  side2 = next_triangle(PARENTVERTEX(entry_index), LINEPT(face), PARENTVERTEX(other_index),
                        index, (top_pos) ? 0 : 2, (top_pos) ? 2 : 0, 1);
  top = next_triangle(LINEPT(&workbuf[index].entry), PARENTVERTEX(exit_index), LINEPT(face),
                      index, (top_pos) ? 1 : 0, (top_pos) ? 0 : 1, 2);
  side1 = next_triangle(PARENTVERTEX(exit_index), PARENTVERTEX(other_index), LINEPT(face),
                        index, 1, 1, 2,
                        TRUE, face, &(workbuf[index].exit));

  CHILDEDGE_FRAGMENT(bottom, 0, entry_index, 0);
  CHILDEDGE_NOTHING (bottom, 1);
  CHILDEDGE_CHILD   (bottom, 2, side2, 0);
  STATUS_SPLITEDGE  (bottom, 1, !keep_top);

  CHILDEDGE_CHILD      (side2, 0, bottom, 2);
  CHILDEDGE_CHILD      (side2, 1, side1, 1);
  CHILDEDGE_WHOLEPARENT(side2, 2, other_index);
  STATUS_PARENTEDGE    (side2, 2, other_index);

  CHILDEDGE_FRAGMENT(top, 0, entry_index, 1);
  CHILDEDGE_CHILD   (top, 1, side1, 2);
  CHILDEDGE_NOTHING (top, 2);
  STATUS_SPLITEDGE  (top, 2, keep_top);

  CHILDEDGE_WHOLEPARENT(side1, 0, exit_index);
  CHILDEDGE_CHILD      (side1, 1, side2, 1);
  CHILDEDGE_CHILD      (side1, 2, top, 1);
  workbuf[side1].entry.location = PT_VERTEX;
  workbuf[side1].entry.index = 2;
  workbuf[side1].exit.index = 0;
  STATUS_PARENTEDGE    (side1, 0, exit_index);


  PARENTEDGE_SPLIT  (entry_index, bottom, 0, top, 0);
  PARENTEDGE_UNSPLIT(exit_index,  side1, 0);
  PARENTEDGE_UNSPLIT(other_index, side2, 2);
}

void TriSplitter::split_face_edge_exit(int index, int top_pos, int keep_top, linept *face)
{
  int entry_index, exit_index, other_index;
  int top, bottom, side1, side2, through_vertex;

  entry_index = workbuf[index].entry.index;
  exit_index = workbuf[index].exit.index;
  other_index = (exit_index + 1) % 3;

  assert(exit_index == (entry_index + 1) % 3);

  // side1 may have one of its edges along the split line, if the entry to this
  // triangle was through a vertex. Otherwise the split line passes through it.
  through_vertex = (workbuf[index].entry.location == PT_VERTEX);

  // The split edge touches edge exit_index of the parent
  // triangle. This shouldn't be the edge that the parent
  // was split along (if it was a split-up triangle)
  if ((workbuf[index].status != UNKNOWN) && (!workbuf[index].status_inferred) &&
      (workbuf[index].status_edge == exit_index))
    TRACE("S%05d WARNING: face-edge (exit) split touches parent edge that is a split!\n", trisplit_debugindex);

  bottom = next_triangle(PARENTVERTEX(other_index), LINEPT(face), LINEPT(&workbuf[index].exit),
                         index, (top_pos) ? 0 : 1, (top_pos) ? 1 : 0, 1);
  if (through_vertex) {
    side2 = next_triangle(PARENTVERTEX(other_index), PARENTVERTEX(entry_index), LINEPT(face),
                          index, (top_pos) ? 0 : 1, (top_pos) ? 1 : 0, 1);
  } else {
    side2 = next_triangle(PARENTVERTEX(other_index), PARENTVERTEX(entry_index), LINEPT(face),
                          index, (top_pos) ? 0 : 2, (top_pos) ? 2 : 0, 2);
  }
  top = next_triangle(LINEPT(&workbuf[index].exit), LINEPT(face), PARENTVERTEX(exit_index),
                      index, (top_pos) ? 1 : 0, (top_pos) ? 0 : 1, 0);
  if (through_vertex) {
    side1 = next_triangle(PARENTVERTEX(exit_index), LINEPT(face), PARENTVERTEX(entry_index),
                          index, (top_pos) ? 1 : 0, (top_pos) ? 0 : 1, 1);
  } else {
    side1 = next_triangle(PARENTVERTEX(exit_index), LINEPT(face), PARENTVERTEX(entry_index),
                          index, 1, 1, 1,
                          TRUE, face, &(workbuf[index].entry));
  }

  CHILDEDGE_CHILD   (bottom, 0, side2, 2);
  CHILDEDGE_NOTHING (bottom, 1);
  CHILDEDGE_FRAGMENT(bottom, 2, exit_index, 1);
  STATUS_SPLITEDGE  (bottom, 1, !keep_top);

  CHILDEDGE_WHOLEPARENT(side2, 0, other_index);
  if (through_vertex) {
    CHILDEDGE_NOTHING  (side2, 1);
  } else {
    CHILDEDGE_CHILD    (side2, 1, side1, 1);
  }
  CHILDEDGE_CHILD      (side2, 2, bottom, 0);
  STATUS_PARENTEDGE    (side2, 0, other_index);

  CHILDEDGE_NOTHING (top, 0);
  CHILDEDGE_CHILD   (top, 1, side1, 0);
  CHILDEDGE_FRAGMENT(top, 2, exit_index, 0);
  STATUS_SPLITEDGE  (top, 0, keep_top);

  CHILDEDGE_CHILD      (side1, 0, top, 1);
  if (through_vertex) {
    CHILDEDGE_NOTHING  (side1, 1);
  } else {
    CHILDEDGE_CHILD    (side1, 1, side2, 1);
  }
  CHILDEDGE_WHOLEPARENT(side1, 2, entry_index);
  if (!through_vertex) {
    workbuf[side1].entry.location = PT_VERTEX;
    workbuf[side1].entry.index = 1;
    workbuf[side1].exit.index = 2;
  }
  STATUS_PARENTEDGE    (side1, 2, entry_index);

  PARENTEDGE_UNSPLIT(entry_index,  side1, 2);
  PARENTEDGE_SPLIT  (exit_index, top, 2, bottom, 2);
  PARENTEDGE_UNSPLIT(other_index, side2, 0);
}

void TriSplitter::split_edge_vertex(int index, int top_pos, int keep_top)
{
  int entry_index, exit_index, other_index;
  int top, bottom;

  entry_index = workbuf[index].entry.index;
  exit_index = workbuf[index].exit.index;
  other_index = (exit_index + 1) % 3;

  assert(exit_index == (entry_index + 1) % 3);

#if _DEBUG
  // The split edge touches a vertex at entry_index. If either of the two edges
  // which meet at this vertex are split edges of the parent, then their KEEP/DISCARD
  // results must match with that of the new split.
  //
  // Also, the split edge touches edge exit_index of the parent
  // triangle. This shouldn't be the edge that the parent was split along.
  if ((workbuf[index].status != UNKNOWN) && (!workbuf[index].status_inferred)) {
    if (workbuf[index].status_edge == entry_index) {
      if (keep_top) {
        if (workbuf[index].status != KEEP)
          TRACE("S%05d WARNING: edge-vertex split does not match existing entry_index KEEP result\n", trisplit_debugindex);
      } else {
        if (workbuf[index].status != DISCARD)
          TRACE("S%05d WARNING: edge-vertex split does not match existing entry_index DISCARD result\n", trisplit_debugindex);
      }

    } else if (workbuf[index].status_edge == exit_index) {
      TRACE("S%05d WARNING: edge-vertex split touches parent edge that is a split!\n", trisplit_debugindex);

    } else if (workbuf[index].status_edge == other_index) {
      if (keep_top) {
        if (workbuf[index].status != DISCARD)
          TRACE("S%05d WARNING: edge-vertex split does not match existing other_index DISCARD result\n", trisplit_debugindex);
      } else {
        if (workbuf[index].status != KEEP)
          TRACE("S%05d WARNING: edge-vertex split does not match existing other_index KEEP result\n", trisplit_debugindex);
      }

    } else {
      assert(0);
    }
  }
#endif

  top = next_triangle(PARENTVERTEX(entry_index), PARENTVERTEX(exit_index), LINEPT(&workbuf[index].exit),
                      index, (top_pos) ? 1 : 0, (top_pos) ? 0 : 1, 2);
  bottom = next_triangle(PARENTVERTEX(entry_index), LINEPT(&workbuf[index].exit), PARENTVERTEX(other_index),
                         index, (top_pos) ? 0 : 1, (top_pos) ? 1 : 0, 0);

  CHILDEDGE_WHOLEPARENT(top, 0, entry_index);
  CHILDEDGE_FRAGMENT   (top, 1, exit_index, 0);
  CHILDEDGE_NOTHING    (top, 2);
  STATUS_SPLITEDGE     (top, 2, keep_top);

  CHILDEDGE_NOTHING    (bottom, 0);
  CHILDEDGE_FRAGMENT   (bottom, 1, exit_index, 1);
  CHILDEDGE_WHOLEPARENT(bottom, 2, other_index);
  STATUS_SPLITEDGE     (bottom, 0, !keep_top);

  PARENTEDGE_UNSPLIT(entry_index, top, 0);
  PARENTEDGE_SPLIT  (exit_index,  top, 1, bottom, 1);
  PARENTEDGE_UNSPLIT(other_index, bottom, 2);
}

void TriSplitter::split_face_vertex(int index, int top_pos, int keep_top, linept *face)
{
  int entry_index, exit_index, other_index;
  int top, bottom, side;

  entry_index = workbuf[index].entry.index;
  exit_index = workbuf[index].exit.index;
  other_index = (exit_index + 1) % 3;

  assert(exit_index == (entry_index + 1) % 3);

#if _DEBUG
  // The split edge touches a vertex at entry_index. If either of the two edges
  // which meet at this vertex are split edges of the parent, then their KEEP/DISCARD
  // results must match with that of the new split.
  if ((workbuf[index].status != UNKNOWN) && (!workbuf[index].status_inferred)) {
    if (workbuf[index].status_edge == entry_index) {
      if (keep_top) {
        if (workbuf[index].status != KEEP)
          TRACE("S%05d WARNING: face-vertex split does not match existing entry_index KEEP result\n", trisplit_debugindex);
      } else {
        if (workbuf[index].status != DISCARD)
          TRACE("S%05d WARNING: face-vertex split does not match existing entry_index DISCARD result\n", trisplit_debugindex);
      }

    } else if (workbuf[index].status_edge == other_index) {
      if (keep_top) {
        if (workbuf[index].status != DISCARD)
          TRACE("S%05d WARNING: face-vertex split does not match existing other_index DISCARD result\n", trisplit_debugindex);
      } else {
        if (workbuf[index].status != KEEP)
          TRACE("S%05d WARNING: face-vertex split does not match existing other_index KEEP result\n", trisplit_debugindex);
      }

    } else {
      assert(workbuf[index].status_edge == exit_index);
    }
  }
#endif

  top = next_triangle(PARENTVERTEX(entry_index), PARENTVERTEX(exit_index), LINEPT(face),
                      index, (top_pos) ? 1 : 0, (top_pos) ? 0 : 1, 2);
  bottom = next_triangle(PARENTVERTEX(entry_index), LINEPT(face), PARENTVERTEX(other_index),
                         index, (top_pos) ? 0 : 1, (top_pos) ? 1 : 0, 0);
  side = next_triangle(LINEPT(face), PARENTVERTEX(exit_index), PARENTVERTEX(other_index),
                       index, 1, 1, 0,
                       TRUE, face, &(workbuf[index].exit));

  CHILDEDGE_WHOLEPARENT(top, 0, entry_index);
  CHILDEDGE_CHILD      (top, 1, side, 0);
  CHILDEDGE_NOTHING    (top, 2);
  STATUS_SPLITEDGE     (top, 2, keep_top);

  CHILDEDGE_NOTHING    (bottom, 0);
  CHILDEDGE_CHILD      (bottom, 1, side, 2);
  CHILDEDGE_WHOLEPARENT(bottom, 2, other_index);
  STATUS_SPLITEDGE     (bottom, 0, !keep_top);

  CHILDEDGE_CHILD      (side, 0, top, 1);
  CHILDEDGE_WHOLEPARENT(side, 1, exit_index);
  CHILDEDGE_CHILD      (side, 2, bottom, 1);
  workbuf[side].entry.location = PT_VERTEX;
  workbuf[side].entry.index = 0;
  workbuf[side].exit.index = 1;
  STATUS_PARENTEDGE    (side, 1, exit_index);

  PARENTEDGE_UNSPLIT(entry_index, top, 0);
  PARENTEDGE_UNSPLIT(exit_index,  side, 1);
  PARENTEDGE_UNSPLIT(other_index, bottom, 2);
}

void TriSplitter::split_edge_edge(int index, int top_pos, int keep_top)
{
  int entry_index, exit_index, other_index;
  int peak, base2, base1;

  entry_index = workbuf[index].entry.index;
  exit_index = workbuf[index].exit.index;
  other_index = (exit_index + 1) % 3;

  assert(exit_index == (entry_index + 1) % 3);

#if _DEBUG
  // The split edge touches two parent edges, neither of which should be split.
  if ((workbuf[index].status != UNKNOWN) && (!workbuf[index].status_inferred)) {
    if ((workbuf[index].status_edge == entry_index) || (workbuf[index].status_edge == exit_index)) {
      TRACE("S%05d WARNING: edge-edge split touches parent edge that is a split!\n", trisplit_debugindex);
    }
  }
#endif

  peak = next_triangle(PARENTVERTEX(exit_index), LINEPT(&workbuf[index].exit), LINEPT(&workbuf[index].entry),
                       index, (top_pos) ? 1 : 0, (top_pos) ? 0 : 1, 1);
  base2 = next_triangle(PARENTVERTEX(other_index), LINEPT(&workbuf[index].entry), LINEPT(&workbuf[index].exit),
                        index, (top_pos) ? 0 : 1, (top_pos) ? 1 : 0, 1);
  base1 = next_triangle(PARENTVERTEX(other_index), PARENTVERTEX(entry_index), LINEPT(&workbuf[index].entry),
                        index, (top_pos) ? 0 : 2, (top_pos) ? 2 : 0, 2);

  CHILDEDGE_FRAGMENT(peak, 0, exit_index, 0);
  CHILDEDGE_NOTHING (peak, 1);
  CHILDEDGE_FRAGMENT(peak, 2, entry_index, 1);
  STATUS_SPLITEDGE  (peak, 1, keep_top);

  CHILDEDGE_CHILD   (base2, 0, base1, 2);
  CHILDEDGE_NOTHING (base2, 1);
  CHILDEDGE_FRAGMENT(base2, 2, exit_index, 1);
  STATUS_SPLITEDGE  (base2, 1, !keep_top);

  CHILDEDGE_WHOLEPARENT(base1, 0, other_index);
  CHILDEDGE_FRAGMENT   (base1, 1, entry_index, 0);
  CHILDEDGE_CHILD      (base1, 2, base2, 0);
  STATUS_PARENTEDGE    (base1, 0, other_index);

  PARENTEDGE_SPLIT(entry_index, base1, 1, peak, 2);
  PARENTEDGE_SPLIT(exit_index,  peak, 0, base2, 2);
  PARENTEDGE_UNSPLIT(other_index, base1, 0);
}

// Insert a fresh triangle in the work buffer, enlarging it if necessary.
int TriSplitter::next_triangle(vector v0, vector v0_n, double v0_dist, double v0_t,
                               vector v1, vector v1_n, double v1_dist, double v1_t,
                               vector v2, vector v2_n, double v2_dist, double v2_t,
                               int parentindex, int num_pos, int num_neg, int zero_v,
                               int has_pts, linept *entry, linept *exit)
{
  splittri *p, *t;

  if (current_numtri >= workbufsize) {
    workbufsize = current_numtri + split_tribuf_growsize;
    workbuf = (splittri *) realloc((void *) workbuf, workbufsize * sizeof(splittri));

    if (workbuf == NULL)
      fatal_error("TriSplitter::next_triangle", "Can't grow in buffer", "Out of memory.");
  }

  p = &workbuf[parentindex];
  t = &workbuf[current_numtri];

  t->t = modeller->trimem->allocate();  

  Vector::set(t->t->v[0], v0);
  Vector::set(t->t->v[1], v1);
  Vector::set(t->t->v[2], v2);
  if (modeller->use_gouraud) {
    Vector::set(t->t->n[0], v0_n);
    Vector::set(t->t->n[1], v1_n);
    Vector::set(t->t->n[2], v2_n);
  }
  t->t->highlight_edge = 0;

  t->plane_dist[0] = v0_dist;
  t->plane_dist[1] = v1_dist;
  t->plane_dist[2] = v2_dist;
  t->num_positive = num_pos;
  t->num_negative = num_neg;
  t->zero_vertex = zero_v;

  t->tvals_wrt = p->tvals_wrt;
  t->vertex_t[0] = v0_t;
  t->vertex_t[1] = v1_t;
  t->vertex_t[2] = v2_t;

  if (has_pts) {
    t->entry = *entry;
    t->exit = *exit;
  
    t->next_interval = p->next_interval;
  } else {
    t->entry.location = t->exit.location = PT_NONE;
    t->next_interval = INTERVAL_NEVER;
  }

  t->this_string_index = p->this_string_index;
  t->original = FALSE;
  t->traverse_id = -1;
  t->parentindex = parentindex;

  return current_numtri++;
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
