/*
 * split.h: Definitions for the TriSplitter.
 */

// Global counter for debugging purposes.
extern int trisplit_debugindex;

// The inside/outside status of a triangle.
enum inoutstatus {UNKNOWN, KEEP, DISCARD, SPLIT};

// Location of a point along the intersection line.
enum pt_location {PT_VERTEX, PT_EDGE, PT_FACE, PT_NONE};

#define INTERVAL_NEVER 10000;

// Hardcoded constants... gotta love 'em.
const int split_tribuf_growsize = 200;
const int max_splitcells = 200;
const int max_in_strings = 20;
const int max_strings = 200;
const int max_intervals = 10;

// A point along the intersection line.
typedef struct linept_ {
  double t;
  pt_location location; // type of point
  int index;            // 0-2... refers to vertex or edge
  double edgeval;       // only PT_EDGE
  vector pt, n;
} linept;

// A triangle to be split.
typedef struct splittri_ {
  tri *t;

  // Distance of the three vertices from the current split plane,
  // number of positive and negative vertices. If there is one positive
  // and one negative vertex, then the zero_vertex variable is set to
  // the vertex which lies on the plane.
  double plane_dist[3];
  int num_positive, num_negative, zero_vertex;

  // T-values for each vertex of the triangle, the split string which
  // these are relative to, and the endpoints of the interval.
  int tvals_wrt;
  double vertex_t[3];
  linept entry, exit;

  // The index of the next interval to be checked along the line of
  // intersection with the current split plane.
  int next_interval;

  int this_string_index;
  int original, parentindex, stringrel_outputindex;
  int newadj[3][2][ADJ_LEN];
  int adjfinal[3];
  int traverse_id;
  inoutstatus status;
  int status_edge, status_inferred;
} splittri;

typedef struct instringdata_ {
  string *cs_str;
  worldstring *ws_str;
  int split_interval_wrt;
  double cs_t_offset, cs_t_scale;
  vector cs_line_dir, ws_line_dir;
  int num_intervals;
  double intervals[max_intervals][2];
} instringdata;

// A string of triangles to split against.
typedef struct splitstring_ {
  WorldCollection *owner;
  int flipped;
  string *cs_str;
  worldstring *ws_str;
} splitstring;

class TriSplitter {
public:
  // The Modeller that we are attached to.
  Modeller *modeller;

  // Variables used during the computation.
  int current_string, tosplit_index;
  double keep_normal;

  // Work buffer.
  int current_numtri, workbufsize, split_any;
  splittri *workbuf;

  // Input buffer.
  int num_in, num_in_strings;
  instringdata in_string[max_in_strings];

  // Split buffer.
  int num_split;
  splitstring *split_string;

  // Output pointers.
  int cell_index;
  WorldCollection *parent, *child;

  // Output buffer construction variables.
  int num_output_tri;
  int string_length[max_strings], stringhead_cellrel_index[max_strings];
  int final_stringindex[max_strings];
  splittri *first_tri[max_strings], *last_tri[max_strings];

  /*
   * INTERFACE ROUTINES
   */

  // Splits a cell worth of triangles [by calling split_triangles].
  int split_cell(int cell_index, WorldCollection *parent, WorldCollection *child, 
                 int num_split, splitstring *split_string,
                 double keep_normal, int *all_kept);

  /*
   * INTERNAL ROUTINES
   */

  // Set up the triangles that are to be split.
  void setup_input(string *cs_cell, worldstring *ws_cell);

  // Performs a split operation of a set of triangles (in) against another set of
  // triangles (split). Places the split triangles into tri_in and tri_out.
  //
  // If no intersections are found, then the return value is FALSE,
  // all_inside TRUE/FALSE specifies whether all triangles are KEEP or DISCARD.
  int split_triangles(int *all_kept);

  // For every triangle, we calculate the distance from the split plane
  // for each of its vertices. We also calculate the number of positive and
  // negative vertices in each triangle.
  void calculate_plane_dist(int string_index);

  // Recalculate adjacency indices for all the triangles.
  void recalculate_adjacency(void);

  // Calculate final cell-relative indices for all triangles that are kept.
  void index_final_triangles(void);

  // Copy all final triangles to the child Collection.
  void output_triangles(void);

  // Look at a complete string of splitagainst triangles, test them against
  // an input plane, and determine the t-values along the intersection line
  // at which there are split triangles present.
  void find_t_intervals(int in_index, int split_index);

  // Check to see if the current triangle is intersected by any of the intervals along
  // the line of intersection in which there are split triangles present. 
  int check_split(int tri_index, int string_index, int split_string_index);

  // Routines to perform a split on a triangle.
  void split_face_face      (int index, int top_pos, int keep_top, linept *face1, linept *face2);
  void split_face_edge_entry(int index, int top_pos, int keep_top, linept *face);
  void split_face_edge_exit (int index, int top_pos, int keep_top, linept *face);
  void split_face_vertex    (int index, int top_pos, int keep_top, linept *face);
  void split_edge_edge      (int index, int top_pos, int keep_top);
  void split_edge_vertex    (int index, int top_pos, int keep_top);

  // Traverse a connected component of the adjacency graph, looking for triangles whose
  // status we know.
  void graph_traverse(int tri_index, int entry_side, int traverse_id,
                      int ignore_inferred,
                      int *found_keep, int *found_keep_inferred,
                      int *found_disc, int *found_disc_inferred);

  // Perform a point-intersection against the RCSGgraph.
  int rcsg_intersect(int tri_index);

  // Insert a fresh triangle in the work buffer, enlarging it if necessary.
  int next_triangle(vector v0, vector v0_n, double v0_dist, double v0_t,
                    vector v1, vector v1_n, double v1_dist, double v1_t,
                    vector v2, vector v2_n, double v2_dist, double v2_t,
                    int parentindex, int num_pos, int num_neg, int zero_v,
                    int has_pts = FALSE, linept *entry = NULL, linept *exit = NULL);
  
  // Constructor initialises arrays to be empty.
  TriSplitter(Modeller *modeller);

  // Destructor destroys any arrays as necessary.
  ~TriSplitter(void);
};

