/*
 * topology.cc: Routines to set up the polygon_topology_table.
 *
 * Contains code to initialise the polygon_topology_table.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

/*
 * Tables that are used for the cell rotation and reflection functions.
 */

// These tables specify where each point GOES TO.
int rotatedcell_points[8]    = {0, 2, 4, 6, 1, 3, 5, 7};
int xreflectedcell_points[8] = {1, 0, 3, 2, 5, 4, 7, 6};
int yreflectedcell_points[8] = {2, 3, 0, 1, 6, 7, 4, 5};
int zreflectedcell_points[8] = {4, 5, 6, 7, 0, 1, 2, 3};

// These tables specify where each edge GOES TO.
int rotated_edges[12]    = { 1,  2,  0,  6,  5,  8,  7,  3,  4, 11,  9, 10};
int xreflected_edges[12] = { 0,  3,  4,  1,  2,  5,  9,  7, 10,  6,  8, 11};
int yreflected_edges[12] = { 5,  1,  6,  3,  9,  0,  2, 11,  8,  4, 10,  7};
int zreflected_edges[12] = { 7,  8,  2, 10,  4, 11,  6,  0,  1,  9,  3,  5};

int topology_table_initialised = FALSE;
topology_struct polygon_topology_table[256];

// The patterns that generate the topology table. These patterns are taken
// from Lorenson and Cline's Marching Cubes paper in SIGGRAPH 87.
const int num_topology_patterns = 22;
unsigned char polygon_topology_inout[num_topology_patterns] =
{0x00, 0xFF, 0x01, 0xFE, 0x03, 0xFC, 0x21, 0xDE, 0x81, 0x7E, 0x0E, 0xF1,
 0x83, 0x7C, 0x92, 0x6D, 0x0F, 0x4D, 0x99, 0x8D, 
 0x1E, 0x69};

#define NONE   NONE_ADJACENT,NONE_ADJACENT
#define BLANK  0,0,0,0,0,0,0,0,0
#define END0   TRI_FINISH
#define END1   TRI_FINISH, BLANK
#define END2   TRI_FINISH, BLANK, BLANK
#define END3   TRI_FINISH, BLANK, BLANK, BLANK
#define END4   TRI_FINISH, BLANK, BLANK, BLANK, BLANK
#define END5   TRI_FINISH, BLANK, BLANK, BLANK, BLANK, BLANK

// Face-OUT disambiguated patterns.
tri_topology face_out_topology[num_topology_patterns] = {
  {                                                                                                                                  END5},
  {                                                                                                                                  END5},
  { 0, 1, 2, NONE,NONE,NONE,                                                                                                         END4},
  { 2, 1, 0, NONE,NONE,NONE,                                                                                                         END4},
  { 1, 2, 4, NONE,NONE,1,0,   1, 4, 3, 0,2, NONE,NONE,                                                                               END3},
  { 4, 2, 1, NONE,NONE,1,1,   3, 4, 1, NONE,0,2, NONE,                                                                               END3},
  { 0, 1, 2, NONE,NONE,NONE,  4, 7,10, NONE,NONE,NONE,                                                                               END3},
  { 4,10, 0, NONE,1,2, NONE, 10, 1, 0, 2,2, NONE,0,1,  10, 2, 1, 3,2, NONE,1,0,  10, 7, 2, NONE,NONE,2,0,                            END1},
  { 0, 1, 2, NONE,NONE,NONE,  9,10,11, NONE,NONE,NONE,                                                                               END3},
  { 2, 1, 0, NONE,NONE,NONE, 11,10, 9, NONE,NONE,NONE,                                                                               END3},
  { 0, 4, 1, NONE,1,0, NONE,  1, 4, 6, 0,1, 2,2, NONE,  4, 9, 6, NONE,NONE,1,1,                                                      END2},
  { 1, 4, 0, 1,1, NONE,NONE,  6, 4, 1, 2,2, 0,0, NONE,  6, 9, 4, NONE,NONE,1,0,                                                      END2},
  { 1, 2, 4, NONE,NONE,1,0,   1, 4, 3, 0,2, NONE,NONE,  9,10,11, NONE,NONE,NONE,                                                     END2},
  { 3, 9, 1, NONE,1,2, NONE,  9,11, 1, NONE,2,2, 0,1,  11, 2, 1, 3,2, NONE,1,1,  11,10, 2, NONE,4,2, 2,0,  10, 4, 2, NONE,NONE,3,1,  END0},
  { 0, 4, 3, NONE,NONE,NONE,  2, 8, 7, NONE,NONE,NONE,  9,10,11, NONE,NONE,NONE,                                                     END2},
  { 8, 2, 0, NONE,NONE,1,1,  11, 8, 0, NONE,0,2, 2,0,  11, 0, 3, 1,2, NONE,3,1,   9,11, 3, NONE,2,2, NONE,  7,10, 4, NONE,NONE,NONE, END0},
  { 2, 9, 6, 1,2, NONE,NONE,  2, 4, 9, NONE,NONE,0,0,                                                                                END3},
  { 8, 2,11, NONE,1,2, NONE,  2, 0,11, NONE,2,2, 0,1,   0, 9,11, 3,0, NONE,1,1,   9, 0, 3, 2,0, NONE,NONE,                           END1},
  { 1, 8, 7, NONE,NONE,1,1,   0, 1, 7, NONE,0,2, NONE,  3,10, 5, NONE,3,0, NONE,  5,10,11, 2,1, NONE,NONE,                           END1},
  { 0, 6, 2, 1,2, NONE,NONE,  0,10, 6, 2,2, 3,0, 0,0,   0, 3,10, NONE,NONE,1,0,   6,10,11, 1,1, NONE,NONE,                           END1},
  { 0, 4, 1, NONE,1,0, NONE,  1, 4, 6, 0,1, 2,2, NONE,  4, 9, 6, NONE,NONE,1,1,   2, 8, 7, NONE,NONE,NONE,                           END1},
  { 0, 1, 2, NONE,NONE,NONE,  6,11, 8, NONE,NONE,NONE,  4, 7,10, NONE,NONE,NONE,  3, 9, 5, NONE,NONE,NONE,                           END1}};

// Face-IN disambiguated patterns.
tri_topology face_in_topology[num_topology_patterns] = {
  {                                                                                                                                  END5},
  {                                                                                                                                  END5},
  { 0, 1, 2, NONE,NONE,NONE,                                                                                                         END4},
  { 2, 1, 0, NONE,NONE,NONE,                                                                                                         END4},
  { 1, 2, 4, NONE,NONE,1,0,   1, 4, 3, 0,2, NONE,NONE,                                                                               END3},
  { 4, 2, 1, NONE,NONE,1,1,   3, 4, 1, NONE,0,2, NONE,                                                                               END3},
  { 0,10, 4, 1,2, NONE,NONE,  0, 1,10, NONE,2,2, 0,0,   1, 2,10, NONE,3,2, 1,1,   2, 7,10, NONE,NONE,2,1,                            END1},
  { 2, 1, 0, NONE,NONE,NONE, 10, 7, 4, NONE,NONE,NONE,                                                                               END3},
  { 0, 1, 2, NONE,NONE,NONE,  9,10,11, NONE,NONE,NONE,                                                                               END3},
  { 2, 1, 0, NONE,NONE,NONE, 11,10, 9, NONE,NONE,NONE,                                                                               END3},
  { 0, 4, 1, NONE,1,0, NONE,  1, 4, 6, 0,1, 2,2, NONE,  4, 9, 6, NONE,NONE,1,1,                                                      END2},
  { 1, 4, 0, 1,1, NONE,NONE,  6, 4, 1, 2,2, 0,0, NONE,  6, 9, 4, NONE,NONE,1,0,                                                      END2},
  { 1, 9, 3, 1,2, NONE,NONE,  1,11, 9, 2,2, NONE,0,0,   1, 2,11, NONE,3,2, 1,0,   2,10,11, 4,2, NONE,2,1,   2, 4,10, NONE,NONE,3,0,  END0},
  { 4, 2, 1, NONE,NONE,1,1,   3, 4, 1, NONE,0,2, NONE, 11,10, 9, NONE,NONE,NONE,                                                     END2},
  { 0, 2, 8, NONE,NONE,1,0,   0, 8,11, 0,2, NONE,2,1,   3, 0,11, NONE,1,2, 3,0,   3,11, 9, 2,2, NONE,NONE,  4,10, 7, NONE,NONE,NONE, END0},
  { 3, 4, 0, NONE,NONE,NONE,  7, 8, 2, NONE,NONE,NONE, 11,10, 9, NONE,NONE,NONE,                                                     END2},
  { 2, 9, 6, 1,2, NONE,NONE,  2, 4, 9, NONE,NONE,0,0,                                                                                END3},
  { 8, 2,11, NONE,1,2, NONE,  2, 0,11, NONE,2,2, 0,1,   0, 9,11, 3,0, NONE,1,1,   9, 0, 3, 2,0, NONE,NONE,                           END1},
  { 0, 3,10, NONE,NONE,1,0,   0,10, 7, 0,2, NONE,NONE,  1, 8,11, NONE,NONE,3,0,   1,11, 5, 2,2, NONE,NONE,                           END1},
  { 0, 6, 2, 1,2, NONE,NONE,  0,10, 6, 2,2, 3,0, 0,0,   0, 3,10, NONE,NONE,1,0,   6,10,11, 1,1, NONE,NONE,                           END1},
  { 0, 2, 1, NONE,NONE,NONE,  4, 9, 7, NONE,2,0, NONE,  7, 9, 8, 1,1, 3,0, NONE,  8, 9, 6, 2,1, NONE,NONE,                           END1},
  { 9,11,10, NONE,NONE,NONE,  3, 4, 0, NONE,NONE,NONE,  7, 8, 2, NONE,NONE,NONE,  1, 6, 5, NONE,NONE,NONE,                           END1}};

#undef BLANK
#undef END0
#undef END1
#undef END2
#undef END3
#undef END4
#undef END5
#undef NONE

// The program currently uses the face-IN topology. This is the most plausible
// topology for CSG objects; see logbook 13/9/98 for reasoning.
tri_topology *polygon_topology_edges = face_in_topology;

/*
 * Operations on an inout and tri_topology structure.
 */

// Set an tri_topology to be another.
void set_tri_topology(tri_topology dest, tri_topology src)
{
  int i;

  for (i = 0; i < TRI_TOPOLOGY_LEN; i++)
    if ((dest[i] = src[i]) == TRI_FINISH)
      return;
}

// Reverses an tri_topology.
void reverse_tri_topology(tri_topology e, tri_topology result)
{
  int i = 0, j;
  int edgeswap[3] = {1, 0, 2};

  while((e[i] != TRI_FINISH) && (i < TRI_TOPOLOGY_LEN)) {
    result[i]     = e[i + 2];
    result[i + 1] = e[i + 1];
    result[i + 2] = e[i];

    // We are inverting this triangle by swapping vertices 0 and 2.
    // This means we must swap edges 0 and 1 around. Also,
    // since all our neighbours (if any) in the topology will also be
    // inverted, we must change our ADJ_VERTEX similarly (i.e. swap 0 for 1).
    for (j = 0; j < 3; j++) {
      result[i + 3 + 2*j + 0] =          e[i + 3 + 2*edgeswap[j] + 0];
      result[i + 3 + 2*j + 1] = edgeswap[e[i + 3 + 2*edgeswap[j] + 1]];
    }

    i += 9;
  }
  result[i] = e[i];
}

// Rotates the cell around 120 degrees about the 0-7 diagonal of the cell.
// X -> Y, Y -> Z, Z -> X.
void rotate_cell(unsigned char inout, tri_topology e,
		 unsigned char *new_inout, tri_topology new_e)
{
  int i, j;

  // First: for each bit in inout, rotate it to the correct position in
  // new_inout.
  *new_inout = 0;
  for (i = 0; i < 8; i++)
    if (inout & (1 << i))
      (*new_inout) |= (1 << rotatedcell_points[i]);

  // Second: for each edge in e, rotate it to the correct new edge.
  for (i = 0; i < MAX_TRI_IN_CELL; i++) {
    if (e[9*i] == TRI_FINISH) {
      // End-of-loop sentinel.
      new_e[9*i] = TRI_FINISH;
      break;
    }

    for (j = 0; j < 3; j++) {
      // Rotate the edge.
      assert(e[9*i + j] < 12);
      new_e[9*i + j] = rotated_edges[e[9*i + j]];
    }

    // Copy the adjacency information.
    for (j = 3; j < 9; j++)
      new_e[9*i + j] = e[9*i + j];
  }      
}

// Reflect the cell in X.
void x_reflect_cell(unsigned char inout, tri_topology e,
		    unsigned char *new_inout, tri_topology new_e)
{
  int i, j;

  // First: for each bit in inout, reflect it to the correct position in
  // new_inout.
  *new_inout = 0;
  for (i = 0; i < 8; i++)
    if (inout & (1 << i))
      (*new_inout) |= (1 << xreflectedcell_points[i]);

  // Second: for each edge in old_tri_topology, reflect it to the correct new
  // edge.
  for (i = 0; i < MAX_TRI_IN_CELL; i++) {
    if (e[9*i] == TRI_FINISH) {
      // End-of-loop sentinel.
      new_e[9*i] = TRI_FINISH;
      break;
    }

    for (j = 0; j < 3; j++) {
      // Rotate the edge.
      assert(e[9*i + j] < 12);
      new_e[9*i + j] = xreflected_edges[e[9*i + j]];
    }

    // Copy the adjacency information.
    for (j = 3; j < 9; j++)
      new_e[9*i + j] = e[9*i + j];
  }      
}

// Reflect the cell in Y.
void y_reflect_cell(unsigned char inout, tri_topology e,
		    unsigned char *new_inout, tri_topology new_e)
{
  int i, j;

  // First: for each bit in inout, reflect it to the correct position in
  // new_inout.
  *new_inout = 0;
  for (i = 0; i < 8; i++)
    if (inout & (1 << i))
      (*new_inout) |= (1 << yreflectedcell_points[i]);

  // Second: for each edge in old_tri_topology, reflect it to the correct new
  // edge.
  for (i = 0; i < MAX_TRI_IN_CELL; i++) {
    if (e[9*i] == TRI_FINISH) {
      // End-of-loop sentinel.
      new_e[9*i] = TRI_FINISH;
      break;
    }

    for (j = 0; j < 3; j++) {
      // Rotate the edge.
      assert(e[9*i + j] < 12);
      new_e[9*i + j] = yreflected_edges[e[9*i + j]];
    }

    // Copy the adjacency information.
    for (j = 3; j < 9; j++)
      new_e[9*i + j] = e[9*i + j];
  }      
}

// Reflect the cell in Z.
void z_reflect_cell(unsigned char inout, tri_topology e,
		    unsigned char *new_inout, tri_topology new_e)
{
  int i, j;

  // First: for each bit in inout, reflect it to the correct position in
  // new_inout.
  *new_inout = 0;
  for (i = 0; i < 8; i++)
    if (inout & (1 << i))
      (*new_inout) |= (1 << zreflectedcell_points[i]);

  // Second: for each edge in old_tri_topology, reflect it to the correct new
  // edge.
  for (i = 0; i < MAX_TRI_IN_CELL; i++) {
    if (e[9*i] == TRI_FINISH) {
      // End-of-loop sentinel.
      new_e[9*i] = TRI_FINISH;
      break;
    }

    for (j = 0; j < 3; j++) {
      // Rotate the edge.
      assert(e[9*i + j] < 12);
      new_e[9*i + j] = zreflected_edges[e[9*i + j]];
    }

    // Copy the adjacency information.
    for (j = 3; j < 9; j++)
      new_e[9*i + j] = e[9*i + j];
  }      
}

// Initialise the polygon topology table.
void init_topology_table(void)
{
  int i, j, rotation, reflection, num_reflections;
  unsigned char pattern, new_pattern;
  tri_topology edges, new_edges;

  // First: set up the topology table with blank values.
  for (i = 0; i < 256; i++) {
    polygon_topology_table[i].top_type = -1;
    polygon_topology_table[i].topology[0] = TRI_FINISH;
  }

  // For each of the num_topology_patterns, we take the pattern, rotate
  // and reflect it every way we can, and store the new patterns
  // in the topology table (if it's not already there).
  for (i = 0; i < num_topology_patterns; i++) {
    // For each pattern,

    // For each symmetry of the cube...
    for (rotation = 0; rotation < 3; rotation++) {
      for (reflection = 0; reflection < 8; reflection++) {
	// Copy the original pattern.
	pattern = polygon_topology_inout[i];
	set_tri_topology(edges, polygon_topology_edges[i]);
	
#if 1
  // DEBUG: Check the topology for consistency.
  int peer_index, peer_edge;
  for (int q = 0; q < MAX_TRI_IN_CELL; q++) {
    if (edges[9*q] == TRI_FINISH)
      break;

    for (int r = 0; r < 3; r++) {
      peer_index = edges[9*q + 3 + 2*r + 0];
      peer_edge  = edges[9*q + 3 + 2*r + 1];
      if (peer_index != NONE_ADJACENT) {
        assert(edges[9*peer_index + 3 + 2*peer_edge + 0] == q);
        assert(edges[9*peer_index + 3 + 2*peer_edge + 1] == r);
        assert(edges[9*q + r]             == edges[9*peer_index + ((peer_edge + 1) % 3)]);
        assert(edges[9*q + ((r + 1) % 3)] == edges[9*peer_index + peer_edge]);
      }
    }
  }
#endif



	// Perturb the pattern according to this particular symmetry.
	num_reflections = 0;
	
	// Reflect the pattern.
	if (reflection & (1 << X)) {
	    x_reflect_cell(pattern, edges, &new_pattern, new_edges);
	    pattern = new_pattern;
	    set_tri_topology(edges, new_edges);
	    num_reflections++;
	}
	if (reflection & (1 << Y)) {
	  y_reflect_cell(pattern, edges, &new_pattern, new_edges);
	  pattern = new_pattern;
	  set_tri_topology(edges, new_edges);
	  num_reflections++;
	}
	if (reflection & (1 << Z)) {
	  z_reflect_cell(pattern, edges, &new_pattern, new_edges);
	  pattern = new_pattern;
	  set_tri_topology(edges, new_edges);
	  num_reflections++;
	}
	
	// Rotate the pattern.
	for (j = 0; j < rotation; j++) {
	  rotate_cell(pattern, edges, &new_pattern, new_edges);
	  pattern = new_pattern;
	  set_tri_topology(edges, new_edges);
	}
	
#if 1
    tri_topology orig_edges;
    set_tri_topology(orig_edges, edges);
#endif

	// Odd numbers of reflections require the tri_topologys to be
	// reversed in order to maintain the right-hand ordering.
	if (num_reflections % 2) {
    reverse_tri_topology(edges, new_edges);
	  set_tri_topology(edges, new_edges);
	}
	
	
#if 1
  // DEBUG: Check the topology for consistency (again).
  for (q = 0; q < MAX_TRI_IN_CELL; q++) {
    if (edges[9*q] == TRI_FINISH)
      break;

    for (int r = 0; r < 3; r++) {
      peer_index = edges[9*q + 3 + 2*r + 0];
      peer_edge  = edges[9*q + 3 + 2*r + 1];
      if (peer_index != NONE_ADJACENT) {
        assert(edges[9*peer_index + 3 + 2*peer_edge + 0] == q);
        assert(edges[9*peer_index + 3 + 2*peer_edge + 1] == r);
        assert(edges[9*q + r]             == edges[9*peer_index + ((peer_edge + 1) % 3)]);
        assert(edges[9*q + ((r + 1) % 3)] == edges[9*peer_index + peer_edge]);
      }
    }
  }
#endif

  // Now we have the final pattern. Check to see if this entry
	// in the topology table has been set already; if it hasn't,
	// set it!
#if _DEBUG
	if (polygon_topology_table[pattern].top_type == -1) {
#endif
	  polygon_topology_table[pattern].top_type = i;
	  set_tri_topology(polygon_topology_table[pattern].topology, edges);
#if _DEBUG
	} else if (polygon_topology_table[pattern].top_type != i) {
	  printf("WARNING: topology overlap (%d <- %d) on 0x%02X\n",
		 polygon_topology_table[pattern].top_type, i, pattern);
	}
#endif
      }
    }
  }

  // The num_topology_patterns have now been propagated through all of
  // the symmetries of the cube, and cover all of the 256 possible patternns.
  // The polygon_topology_table is ready for use in the polygonizer.


#ifdef _DEBUG
  // Check the polygon_topology_table from 0 to 255 (excluding patterns 0x00
  // and 0xFF, which we know have no polygons) to see that each pattern
  // has been matched by one of the num_patterns symmetries.
  for (i = 1; i < 255; i++)
    assert(polygon_topology_table[i].top_type != -1);
#endif

  /*
   * Print out the calculated polygon_topology_table; this was just
   * used for debugging, but might become useful in the event of an error
   * (heaven forbid!)

  printf("Print polygon_topology_table\n");
  int loopstart;
  for (i = 0; i < 256; i++) {
    printf("0x%02X: (type %d) ", i, polygon_topology_table[i].top_type);
    loopstart = TRUE;
    for (j = 0; j < TRI_TOPOLOGY_LEN; j += 3) {
      if (polygon_topology_table[i].topology[j] == TRI_FINISH) {
	printf("\n");
	break;
      } else {
	if (!loopstart)
	  printf(" | ");
	printf("%d,%d,%d",
	       polygon_topology_table[i].topology[j],
	       polygon_topology_table[i].topology[j + 1],
	       polygon_topology_table[i].topology[j + 2]);
	loopstart = FALSE;
      }
    }
  }
  */

  topology_table_initialised = TRUE;
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
