/*
 * topology.h: Definitions for the polygon_topology_table.
 *
 * Contains code to initialise the polygon_topology_table.
 */

// An tri_topology is a structure that contains some u_chars. It is used
// for storing the tri_topology that represents the triangles which need to be
// formed from a given cell.
//
// Edges are indexed from 0 to 11 according to the following scheme
//
// E/W = +X/-X
// N/S = +Y/-Y
// U/D = +Z/-Z
// 0:  SB
// 1:  WB
// 2:  SW
// 3:  EB
// 4:  SE
// 5:  NB
// 6:  NW
// 7:  TS
// 8:  TW
// 9:  NE
// 10: TE
// 11: TN
//
// Draw it, it makes sense.
//
// Each triangle is represented in the topology by a set of 9 u_chars.
// The first three are the edges that its vertices lie on. Then come three
// groups of two, which are the adjacency information of the topology cell.
// Each group of two is the adjacent triangle to that edge. The first number
// is the index of the adjacent triangle, from 0 = the first triangle in the
// topology cell (if there is no adjacent triangle, a NONE_ADJACENT value of
// 100 is used). The second number is the index of the edge with respect to
// the other triangle. It can be 0, 1 or 2.
//
// So, for the cell that has inout = binary 7-0 01101001, 0x69,
// if we want to generate four polygons with edges 0,1,2 and 9,5,3 and 4,7,10
// and 11,8,6, this would be represented as the tri_topology
//
// tri_topology el = {0,1,2,  NONE,NONE,NONE, 9,5,3,  NONE,NONE,NONE,
//                    4,7,10, NONE,NONE,NONE, 11,8,6, NONE,NONE,NONE, TRI_FINISH};
//
// Note: the above example is a center-out disambiguated example. There's a
// compilation switch to set center-out and center-in disambiguation in
// topology.cpp.
#define MAX_TRI_IN_CELL 5
#define TRI_TOPOLOGY_LEN (9*MAX_TRI_IN_CELL + 1)
#define TRI_FINISH    100
#define NONE_ADJACENT 101
typedef unsigned char tri_topology[TRI_TOPOLOGY_LEN];

// The table that gives the tri_topologyset for *any* inout value of a cell.
// This determines the topology of the surface in that cell.
typedef struct _topology_struct {
  int top_type;
  tri_topology topology;
} topology_struct;

// The polygon_topology_table itself.
extern int topology_table_initialised;
extern topology_struct polygon_topology_table[256];

// Set an tri_topology to be another.
void set_tri_topology(tri_topology dest, tri_topology src);

// Reverses an tri_topology.
void reverse_tri_topology(tri_topology e, tri_topology result);

/*
 * Operations on an inout and tri_topology structure.
 */

// Rotates the cell around 120 degrees about the 0-7 diagonal of the cell.
// X -> Y, Y -> Z, Z -> X.
void rotate_cell(unsigned char inout, tri_topology e,
		 unsigned char *new_inout, tri_topology new_e);

// Reflect the cell in X.
void x_reflect_cell(unsigned char inout, tri_topology e,
		    unsigned char *new_inout, tri_topology new_e);

// Reflect the cell in Y.
void y_reflect_cell(unsigned char inout, tri_topology e,
		    unsigned char *new_inout, tri_topology new_e);

// Reflect the cell in Z.
void z_reflect_cell(unsigned char inout, tri_topology e,
		    unsigned char *new_inout, tri_topology new_e);

// Initialise the polygon topology table.
void init_topology_table(void);
