/*
 * trimem.cpp: Code for class TriMemory.
 *
 * Contains code to construct a TriMemory which is a variant of the Mirage-style
 * blocklist that's been specialised to handle strings of triangles.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Constructor: specified blocksize and maxblocks.
TriMemory::TriMemory(int blocksize, int max_memblocks)
{
  this->blocksize = blocksize;
  this->max_memblocks = max_memblocks;

  // Allocate the storage pointers.
  storage = (tri **) calloc(max_memblocks, sizeof(tri *));
  assert(storage != NULL);
  for (int i = 0; i < max_memblocks; i++)
    storage[i] = NULL;

  // Set up memory management values.
  blocks_used = 0;
  next_free = NULL;
}

// Destructor; deallocate all memory and exit.
TriMemory::~TriMemory(void)
{
  assert(storage != NULL);

  // Deallocate each block.
  for (int i = 0; i < blocks_used; i++) {
    assert(storage[i] != NULL);
    free(storage[i]);
  }

  // Deallocate the storage memory.
  free(storage);
}

// Allocate a new triagle from the storage and return a pointer to it.
tri *TriMemory::allocate(void)
{
  tri *the_block, *the_tri;
  int block_no, index_no;
  
  // Check to see if we need to allocate a new block.
  if (next_free == NULL) {
    
    if (blocks_used >= max_memblocks)
      fatal_error("TriMemory::allocate", "maximum memory size exceeded",
		    "TriMemory has a fixed limit on the number of\n"
        "blocks that can be allocated. Increase this, or reduce\n"
        "memory usage.");
    
    the_block = (tri *) calloc(blocksize, sizeof(tri));
    
    if (the_block == NULL)
      fatal_error("TriMemory::allocate", "can't allocate new block!",
		    "Tried to dynamically expand memory and failed. This\n"
        "could indicate that you are running severely low on\n"
        "memory or that there is a memory leak.");
    
    // Update the array of blocks.
    block_no = blocks_used++;
    storage[block_no] = the_block;
    
    // Now for each unused triangle in the block (which is all of them)
    // we set it to point to the next free triangle. We start from the
    // zero tri, each one pointing to the next. The last one in the block
    // points to NULL.
    //
    // This pointer is stored in stringptr as there is no data in the tri
    // right now anyway.
    for (index_no = 1; index_no < blocksize - 1; index_no++)
      the_block[index_no].stringptr = &the_block[index_no + 1];
    
    // The last element is free, but there's no free memory after it.
    the_block[blocksize - 1].stringptr = NULL;
    
    // Allocate one object for immediate use.
    the_tri = &the_block[0];
    
    // The next one is the next_free object.
    next_free = &the_block[1];
    
  } else {
    // Just use the next free object.
    the_tri = next_free;
    
    // The next free element is stored in the stringptr value.
    next_free = the_tri->stringptr;
    
#if _DEBUG
    if (next_free != NULL) {
      // Check that the free list pointer is valid.
      for (int i = 0; i < blocks_used; i++) {
        if ((next_free >= &storage[i][0]) &&
          (next_free < &storage[i][blocksize]))
          break;
      }
      
      assert(i < blocks_used);
    }
#endif
  }
  
//  if ((int) the_tri == 0x02b972f8)
//    _CrtDbgBreak();

  // Set up initial values and return the triangle.
  the_tri->string_length = 0;
  the_tri->stringptr = NULL;
  return the_tri;
}

// Deallocate a previously allocated triangle.
void TriMemory::deallocate(tri *the_tri)
{
#if _DEBUG
  int i;
    
  // Check that the object is part of our memory, and that
  // the object isn't already free.
  for (i = 0; i < blocks_used; i++) {
    assert(storage[i] != NULL);

    // Was the object allocated from this block?
    if ((the_tri >= &storage[i][0]) &&
        (the_tri <  &storage[i][blocksize])) {
      // This is a pointer to inside block i.
      break;
    }
  }
    
  // The object should have satisfied the break condition above,
  // otherwise it isn't part of our memory.
  assert(i < blocks_used);
#endif

  // Free the triangle.
  the_tri->string_length = -1;
  the_tri->stringptr = next_free;
    
  // This triangle will be next to be allocated.
  next_free = the_tri;
}

// Deallocate a string of triangles.
void TriMemory::deallocate_string(tri *the_tri)
{
  int i, delete_length;
  tri *thisptr, *nextptr;

  // Check that the string is a valid one.
  assert(the_tri != NULL);
  thisptr = the_tri;
  delete_length = the_tri->string_length;
  for (i = 0; i < delete_length; i++) {
    assert(thisptr != NULL);
    assert(thisptr->string_length == delete_length - i);
    nextptr = thisptr->stringptr;
    deallocate(thisptr);
    thisptr = nextptr;
  }
  assert(thisptr == NULL);
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
