/*
 * view.cc: Routines for displaying a view window.
 *
 * Contains code for the View class.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Handle mouse events inside the window.
void View::mouse_event(double x, double y, int button, mouse_event_type event,
		       unsigned char modifiers)
{
  if (modifiers & MODIFIER_SHIFT) {
    switch(button) {
    case 1:
      switch(event) {
      case MOUSEEVENT_PRESS:
	view_rotation_begin(x, y);
	break;
      case MOUSEEVENT_MOTION:
	view_rotation_drag(x, y);
	break;
      default:
	break;
      }
      break;
    case 2:
      switch(event) {
      case MOUSEEVENT_PRESS:
	view_translation_begin(x, y);
	break;
      case MOUSEEVENT_MOTION:
	view_translation_drag(x, y);
	break;
      default:
	break;
      }
      break;
    case 3:
      switch(event) {
      case MOUSEEVENT_PRESS:
	view_zoom_begin(x, y);
	break;
      case MOUSEEVENT_MOTION:
	view_zoom_drag(x, y);
	break;
      default:
	break;
      }
    }
  } else if (modifiers & MODIFIER_CTRL) {
    switch(button) {
    case 1:
      if ((event == MOUSEEVENT_PRESS) || (event == MOUSEEVENT_MOTION))
	set_debugging_ray(x, y);
      break;
    case 2:
      if (event == MOUSEEVENT_PRESS)
#if 0
        modeller->cycle_debugging_terminal();
#endif
      break;
    case 3:
      if (event == MOUSEEVENT_PRESS)
#if 0
        modeller->toggle_debugging_highlight();
#endif
      break;
    }
  } else {
    switch(button) {
    case 1:
      if (event == MOUSEEVENT_PRESS) {
	left_click(x, y);
      } else if (event == MOUSEEVENT_MOTION) {
	left_drag(x, y);
      } else if (event == MOUSEEVENT_RELEASE) {
	left_release();
      }
      break;
    }
  }
}

// Handle a zoom event (from a zoom button, or mousewheel).
void View::zoom_event(double z)
{
  camera_zoom *= (1 + z / 8);

  // The viewing position has changed.
  compute_camerapos();
  compute_modeltransform();
  redisplay();
}

// Rotate the camera.
void View::view_rotation_begin(double x, double y)
{
  Quaternion::set(initial_dir, camera_dir);
  initial_x = x;
  initial_y = y;
  
  // This code uses the Virtual Sphere controller, described in CHEN88:
  // "A Study in Interactive 3D Rotation Using 2D Control Devices" M. Chen,
  // S. Mountford, A. Sellen. Computer Graphics Proceedings, SIGGRAPH 88,
  // August 1988. p121-129
  
  // The initial click for the Virtual Sphere controller has been placed.
  // O is the center point of the screen, and P is the point where
  // the first click is placed. We calculate the distance OP and the
  // angle theta between it and the X axis.
  viewrotate_op = sqrt(SQR(x) + SQR(y));
  viewrotate_theta = atan2(-y, x);
  
  // Omega is the angle around the Y axis to rotate the rotation axis.
  if (viewrotate_op >= 1)
    viewrotate_omega = PI/2;
  else
    viewrotate_omega = viewrotate_op * PI/2;
}

void View::view_rotation_drag(double x, double y)
{
  vector rotated_q;
  double cursor_movedist, rotation_angle;
  quaternion rotation_quat;

  // We have the current rotation-to point for the Virtual Sphere.
  // Calculate the angle tau which this rotation vector makes with the
  // vector OP.
  Vector::assign(rotated_q,
		 x - initial_x,
		 -(y - initial_y), 0);
  Vector::rotate(Z, RAD2DEG(-viewrotate_theta), rotated_q);
  
  // Calculate the rotation axis of the quaternion.
  Vector::assign(rotation_quat, -rotated_q[Y], rotated_q[X], 0);
  Vector::rotate(Z, X, RAD2DEG(viewrotate_omega), rotation_quat);
  Vector::rotate(X, Y, RAD2DEG(viewrotate_theta), rotation_quat);
  
  // Find the distance the cursor has moved, and thus the rotation angle.
  // From Equation (5) in CHEN88.
  cursor_movedist = sqrt(SQR(x - initial_x) +
			 SQR(y - initial_y));
  rotation_angle = PI/2 * cursor_movedist *
    (1 - (1 - 0.2/PI) * (viewrotate_omega / (PI/2))
     * (1 - fabs(rotated_q[X])));
  
  // Form the rotation quaternion from the rotation axis and angle.
  Vector::normalise(rotation_quat,
		    "view_rotation_drag: zero rotation_quat");
  Vector::mult(sin(rotation_angle), rotation_quat, rotation_quat);
  rotation_quat[W] = cos(rotation_angle);
  
  // Apply this quaternion to the viewing rotation.
  Quaternion::mult(rotation_quat, initial_dir, camera_dir);
  Quaternion::inverse(camera_dir, camera_dir_inv,
		      "view_rotation_drag: zero camera_dir");

  // The viewing position has changed.
  compute_camerapos();
  compute_modeltransform();
  redisplay();
}


// Translate the camera.
void View::view_translation_begin(double x, double y)
{
  Vector::set(initial_focus, focus_position);
  initial_x = x;
  initial_y = y;     
}

void View::view_translation_drag(double x, double y)
{
  vector focus_offset;

  // The glFrustum call places screen coordinates of
  // (-aspect, -1) - (+aspect,+1) at Z = 1. The focus is at Z = camera_dist
  // so we translate the object by the movement in screen coordinates *
  // camera_dist. This is modified by a factor of camera_zoom as well.
  //
  // Also, we construct the focus offset and then rotate it by
  // camera_dir_inv.
  
  Vector::assign(focus_offset, - (x - initial_x), (y - initial_y), 0);
  Vector::mult(1.0 / camera_zoom, focus_offset, focus_offset);
  Quaternion::apply(camera_dir_inv, focus_offset, focus_offset);
  Vector::add(initial_focus, focus_offset, focus_position);

  // The viewing position has changed.
  compute_camerapos();
  compute_modeltransform();
  redisplay();
}


// Zoom the camera.
void View::view_zoom_begin(double x, double y)
{
  initial_zoom = camera_zoom;
  initial_x = x;
  initial_y = y;
}

void View::view_zoom_drag(double x, double y)
{
  camera_zoom = initial_zoom / (1 + (y - initial_y) / 2);

  // The viewing position has changed.
  compute_camerapos();
  compute_modeltransform();
  redisplay();
}

void View::compute_camerapos(void)
{
  // Set the camera position to account for the rotation change.
  Vector::assign(camera_position, 0, 0, camera_dist);
  Quaternion::apply(camera_dir, camera_position, camera_position);
  Vector::add(focus_position, camera_position, camera_position);
}


// A left-click has been detected.
void View::left_click(double x, double y)
{
  vector ray_from, ray_dir;

  position_to_ray(x, y, ray_from, ray_dir);
  if (modeller->skitterlist->current_dmd) {
    // Check to see if this click might be a manipulation against the
    // current DMD.
    if (modeller->skitterlist->dmd_click(ray_from, ray_dir)) {
      modeller->update_all_views();
      return;
    }
  }

  // Check to see if this click might be a selection operation versus
  // one of the other skitters in the SkitterList.
  if (modeller->skitterlist->skitter_select(ray_from, ray_dir)) {
    // The currently selected DMD has changed.
    modeller->update_all_views();
    return;
  }

  // This click isn't on a skitter or a DMD.
  if (modeller->skitterlist->current_dmd) {
    // Deselect the DMD.
    modeller->skitterlist->dmd_deselect();
    modeller->update_all_views();
  } else {
    // No DMD active; select an object to manipulate.
    modeller->select_along_ray(ray_from, ray_dir);
  }
}

// A left-drag has been detected.
void View::left_drag(double x, double y)
{
  vector ray_from, ray_dir;

  // Are we in the middle of a manipulation operation?
  if (modeller->skitterlist->currently_manipulating) {
    position_to_ray(x, y, ray_from, ray_dir);
    modeller->skitterlist->dmd_manipulate(ray_from, ray_dir);
  }
}

// A left-release has been detected.
void View::left_release(void)
{
  // Are we in the middle of a manipulation operation?
  if (modeller->skitterlist->currently_manipulating) {
    modeller->skitterlist->dmd_release();
    modeller->update_all_views();
  }
}

// Debugging ray things.
void View::set_debugging_ray(double x, double y)
{
  modeller->debug_line = TRUE;
#if 0
  modeller->debug_terminal_count = 0;
#endif

  position_to_ray(x, y, modeller->debug_linefrom, modeller->debug_linedir);

  modeller->update_all_views();
}

// Turn ghosts on/off.
void View::set_ghost(int new_ghost)
{
  show_ghost = new_ghost;
  update();
}

// Turn edges on/off.
void View::set_edges(int new_edges)
{
  show_edges = new_edges;
  update();
}

// Select wireframe / solid.
void View::set_wireframe(int new_wireframe)
{
  draw_wireframe = new_wireframe;
  update();
}

// Select background colour.
void View::set_bg(int is_black)
{
  black_background = is_black;
  update();
}

void View::display(void)
{
  init_for_display();
  display_axes();
  display_model();
  display_interface();
  finish_display();
}

// A global array which stores indices into star_points for the triangles that
// are used to draw the star.
int star_indices[6][3] = {{4, 0, 5},
			  {5, 1, 4},
			  {4, 2, 5},
			  {5, 3, 4},
			  {0, 2, 1},
			  {1, 3, 0}};

// Draw a star at a point.
void View::draw_star(int star_num, double star_size,
		     transform skitter_position, vector point,
		     vector star_dir, vector perp_dir)
{
  vector star_points[6], x_dir, y_dir, tri_vec[18];
  int i, j;

  assert(star_num > 0 && star_num <= 3);

  // First, construct the star's Z points (along star_dir).  
  Vector::mult(star_size / CHECK0(Vector::magnitude(star_dir),
				  "draw_star:star_dir"),
	       star_dir, star_points[4]);
  Vector::mult(-1, star_points[4], star_points[5]);

  // Next, construct the star's Y points (out perpendicular to perp_dir).
  Vector::cross(star_dir, perp_dir, y_dir);
  Vector::mult(star_size / CHECK0(Vector::magnitude(y_dir),
				  "draw_star:y_dir"),
	       y_dir, star_points[2]);
  Vector::mult(-1, star_points[2], star_points[3]);
  
  // Lastly, construct the star's X points (perpendicular to y_dir and
  // star_dir, back towards perp_dir).
  Vector::cross(y_dir, star_dir, x_dir);
  Vector::mult(star_size / CHECK0(Vector::magnitude(x_dir),
				  "draw_star:x_dir"),
	       x_dir, star_points[0]);
  Vector::mult(-1, star_points[0], star_points[1]);

  // Construct the worldspace coordinates for each of the star's points.
  for (i = 0; i < 6; i++) {
    Vector::add(point, star_points[i], star_points[i]);
    Matrix::apply(skitter_position[FORE], star_points[i], star_points[i]);
  }
  
  // Draw the triangles as necessary. Two triangles for each plane of the
  // star (from 1 to 3, specified by star_num).
  for (i = 0; i < 2*star_num; i++)
    for (j = 0; j < 3; j++)
      Vector::set(tri_vec[3*i + j], star_points[star_indices[i][j]]);

  draw_unlit_tri(2*star_num, tri_vec);
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
