/*
 * view.h: Definitions for a view window.
 *
 * Contains definitions for the View class.
 */

// Types of mouse events
enum mouse_event_type {MOUSEEVENT_PRESS, MOUSEEVENT_MOTION,
		       MOUSEEVENT_RELEASE};

enum drawpass {DRAW_SOLID, DRAW_GHOST_DEPTH, DRAW_GHOST, DRAW_EDGES};

// Modifiers on mouse events.
#define MODIFIER_NONE  0x00
#define MODIFIER_CTRL  0x80
#define MODIFIER_ALT   0x40
#define MODIFIER_SHIFT 0x20

// A ViewWindow represents a view onto the modelled scene.
class View {
public:
  // The Modeller object that holds the object which we're viewing.
  Modeller *modeller;

  // The position the camera is looking at in the scene.
  vector focus_position;

  // The camera's rotation quaternion (and its inverse)
  quaternion camera_dir, camera_dir_inv;

  // The base and current distance of the camera from the focus.
  double base_dist, camera_dist;

  // The zoom factor... 1.0 = starting point.
  double camera_zoom;

  // Position of the camera; this is derived from camera_dir and camera_dist.
  vector camera_position;

  // View toggles.
  int draw_wireframe, show_ghost, show_edges, black_background;

  // The position of the click that started a view manipulation.
  double initial_x, initial_y;

  // The initial state of the view (before the manipulation)
  vector initial_focus;
  quaternion initial_dir;
  double initial_zoom;

  // Variables for view rotation.
  double viewrotate_theta, viewrotate_op;
  double viewrotate_omega;

  // The size of the view window.
  int view_width, view_height;
  double view_aspect;

  // Resize the view window.
  virtual void resize(int width, int height) = 0;

  // The model has been updated; internal representations of the model
  // should be sync'ed with these changes.
  virtual void update(void) = 0;

  // Display the view window's contents.
  void display(void);

  // Recompute the camera position.
  void compute_camerapos(void);

  /*
   * The meat of the implemented code... begin display routines
   */

  // Translate a position on the view to a ray.
  virtual void position_to_ray(double x, double y, vector ray_from, vector ray_dir) = 0;

  // Redisplay the window; usually called after something requires
  // that we call display() again. Should be used in most cases instead
  // of display(), which may be OS-dependently tied to a redraw message.
  virtual void redisplay(void) = 0;

  // View-specific initialisation. Varies from implementation to
  // implementation.
  virtual void init_for_display(void) = 0;

  // Draw grids, axes and other scene components that aren't the object.
  virtual void display_axes(void) = 0;

  // Draw the object being modelled.
  virtual void display_model(void) = 0;
  
  // Draw the interface objects.
  virtual void display_interface(void) = 0;

  // Finish the display process. Varies from implementation to
  // implementation.
  virtual void finish_display(void) = 0;

  // Recompute the model transform for the current camera position.
  virtual void compute_modeltransform(void) = 0;

  /*
   * End main display routines.
   */

  /*
   * Primitive drawing routines that are called by the DMDs, etc.
   */

  virtual void set_drawcolor(color c) = 0;
  virtual void set_drawcolor(double r, double g, double b) = 0;

  virtual void set_drawalpha(double a) = 0;

  virtual void draw_lines(int num_lines, vector *point) = 0;
  virtual void draw_lines(int num_lines, vector **point) = 0;

  virtual void draw_unlit_tri(int num_tri, vector *point) = 0;

  /*
   * End of drawing routines.
   */

  // Draw a star at a point.
  void draw_star(int star_num, double star_size,
		 transform skitter_position, vector point,
		 vector dir, vector perp_dir);

  // Handle mouse events inside the window.
  void mouse_event(double x, double y, int button, mouse_event_type event,
		   unsigned char modifiers);

  // Handle a zoom event (from a zoom button, or mousewheel).
  void zoom_event(double z);

  // Rotate the camera.
  void view_rotation_begin(double x, double y);
  void view_rotation_drag(double x, double y);

  // Translate the camera.
  void view_translation_begin(double x, double y);
  void view_translation_drag(double x, double y);

  // Zoom the camera.
  void view_zoom_begin(double x, double y);
  void view_zoom_drag(double x, double y);

  // Left-click interactions (multi-purpose, quite complex).
  void left_click(double x, double y);
  void left_drag(double x, double y);
  void left_release(void);

  // Set the location of the debugging ray.
  void set_debugging_ray(double x, double y);

  // Turn ghosts on/off.
  void set_ghost(int new_ghost);

  // Turn edges on/off.
  void set_edges(int new_edges);

  // Select wireframe / solid.
  void set_wireframe(int new_wireframe);

  // Select background colour.
  void set_bg(int is_black);

  // Draw a collection of triangles.
  virtual void draw_collection(WorldCollection *collection,
                               int is_flipped, drawpass pass) = 0;
};

