/*
 * wglview.cc: Routines for the WGLView.
 *
 * Contains code to interface with OpenGL under Windows; specifically
 * the GLView that manages the OpenGL connection.
 */

/* Collection-safe */

#include "stdafx.h"

#ifdef _WIN32

#include "ModelApp.h"
#include "ModelDoc.h"
#include "ModelView.h"
#include "GLView.h"
#include "wglview.h"





#define SHOW_ALL_INACTIVE_SURFACES   FALSE
#define DRAW_NORMALS                 FALSE
#define DRAW_TEST_BOX                FALSE


#ifdef USE_NAMESPACES
namespace Mirage {
#endif

int global_next_drawid = 0;

/*
 * Routines for WGLView
 */

WGLView::WGLView(Modeller *m, GLView *vw)
{
  CRect currentsize;
  int j;
  double furthest_d;

  this->modeller = m;
  this->viewwindow = vw;

  modeller->bind_modelview(this);

  // Set up the initial view parameters.
  Vector::set(focus_position, zero_vector);
  Quaternion::unit(camera_dir);
  Quaternion::unit(camera_dir_inv);
  camera_zoom = 1;

  furthest_d = 0;
  for (j = X; j <= Z; j++) {
    if (modeller->maxbound[j] > -modeller->minbound[j])
      furthest_d += SQR(modeller->maxbound[j]);
    else
      furthest_d += SQR(modeller->minbound[j]);
  }
  clip_dist = sqrt(furthest_d);
  if (clip_dist > 4)
    clip_dist = 4;
  base_dist = camera_dist = 1.5 * clip_dist;
  camera_zoom = 1.0;

  // Set view flags.
  show_ghost = TRUE;
  show_edges = TRUE;
  draw_wireframe = FALSE;
  black_background = TRUE;

  // Set the GL render options.
  light0_pos[0] = +1.0;
  light0_pos[1] = -1.0;
  light0_pos[2] = +1.0;
  light0_pos[3] = +0.0;
  light1_pos[0] = -1.0;
  light1_pos[1] = +1.0;
  light1_pos[2] = -1.0;
  light1_pos[3] = +0.0;
  light_ambient[0] = (GLfloat) 0.2;
  light_ambient[1] = (GLfloat) 0.2;
  light_ambient[2] = (GLfloat) 0.2;
  light_ambient[3] = (GLfloat) 1.0;
  light_diffuse[0] = (GLfloat) 0.5;
  light_diffuse[1] = (GLfloat) 0.5;
  light_diffuse[2] = (GLfloat) 0.5;
  light_diffuse[3] = (GLfloat) 1.0;
  light_specular[0] = (GLfloat) 0.5;
  light_specular[1] = (GLfloat) 0.5;
  light_specular[2] = (GLfloat) 0.5;
  light_specular[3] = (GLfloat) 1.0;
  unselected_diffuse[0] = (GLfloat) 1.0;
  unselected_diffuse[1] = (GLfloat) 1.0;
  unselected_diffuse[2] = (GLfloat) 1.0;
  unselected_diffuse[3] = (GLfloat) 1.0;
  selected_diffuse[0] = (GLfloat) 0.3;
  selected_diffuse[1] = (GLfloat) 0.3;
  selected_diffuse[2] = (GLfloat) 1.0;
  selected_diffuse[3] = (GLfloat) 1.0;
  special_diffuse[0] = (GLfloat) 1.0;
  special_diffuse[1] = (GLfloat) 0.5;
  special_diffuse[2] = (GLfloat) 0.5;
  special_diffuse[3] = (GLfloat) 1.0;
  tri_specular[0] = (GLfloat) 0.6;
  tri_specular[1] = (GLfloat) 0.6;
  tri_specular[2] = (GLfloat) 0.6;
  tri_specular[3] = (GLfloat) 1.0;
  zero_material[0] = (GLfloat) 0.0;
  zero_material[1] = (GLfloat) 0.0;
  zero_material[2] = (GLfloat) 0.0;
  zero_material[3] = (GLfloat) 0.0;
  tri_shininess = (GLfloat) 40.0;
#if SHOW_ALL_INACTIVE_SURFACES
  ghost_color[0] = (GLfloat) 1.0;
  ghost_color[1] = (GLfloat) 0.0;
  ghost_color[2] = (GLfloat) 0.0;
  ghost_color[3] = (GLfloat) 0.6;
#else
  ghost_color[0] = (GLfloat) 0.5;
  ghost_color[1] = (GLfloat) 0.5;
  ghost_color[2] = (GLfloat) 1.0;
  ghost_color[3] = (GLfloat) 0.3;
#endif
  edge_color[0] = (GLfloat) 0.0;
  edge_color[1] = (GLfloat) 0.0;
  edge_color[2] = (GLfloat) 1.0;
  edge_color[3] = (GLfloat) 1.0;

  // Zero the display list.
  displaylist_built = FALSE;
  modelDisplayList = 0;

  // Set the size of the view window.
  viewwindow->GetClientRect(&currentsize);
  resize(currentsize.right - currentsize.left, currentsize.bottom - currentsize.top);
  compute_modeltransform();
}

WGLView::~WGLView(void)
{
  modeller->unbind_modelview(this);
}

// Resize the view window.
void WGLView::resize(int width, int height)
{
  view_width = width;
  view_height = height;
  view_aspect = (float) width / (float) height;

  glViewport(0, 0, view_width, view_height);

  compute_frustum();
}

// Compute the OpenGL viewing frustum.
void WGLView::compute_frustum(void)
{
  double xf, yf;

  // Zooming in... we modify the field of view only. Zooming out, we modify
  // the distance of the viewpoint from the scene.

  // The field of view is base_dist wide at the middle of the view frustum, with no
  // zoom applied.
  frustum_size = base_dist / camera_zoom;

  // So how wide is the frustum at the near clipping plane?
  frustum_size *= (camera_dist - clip_dist) / camera_dist;

  // Adjust in X for the view aspect of the window on the screen.
  yf = frustum_size;
  xf = frustum_size * view_aspect;

  // Locate the clipping planes.
  near_clip = camera_dist - clip_dist;
  far_clip  = camera_dist + clip_dist;

  // Set up the projection transform.
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-xf, xf, -yf, yf, near_clip, far_clip);
  
  // Change back to model view transform for rendering.
  glMatrixMode(GL_MODELVIEW);    
}

// We need to call display() again.
void WGLView::redisplay(void)
{
  // Invalidate the client window (all of it).
  viewwindow->RedrawWindow(NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);
}

// The model has been updated.
void WGLView::update(void)
{
  // We need to delete the display-list if we have one.
  if (displaylist_built) {
    glDeleteLists(modelDisplayList, 1);
    displaylist_built = FALSE;
  }

  redisplay();
}

// View-specific initialisation. OpenGL setup stuff.
void WGLView::init_for_display(void)
{
  // Clear the OpenGL framebuffer.
  if (black_background)
    glClearColor(0.0, 0.0, 0.0, 0.0);
  else
    glClearColor(1.0, 1.0, 1.0, 1.0);

  glClearStencil(0x0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

// Draw the object being modelled. Note that this sets up displaylist caches
// etc.
void WGLView::display_model(void)
{
  // Set the model transform.
  glLoadMatrixf(gl_modelmatrix);

  // Draw the object being modelled.
  if (displaylist_built) {
    // We already have a display-list cached!
    glCallList(modelDisplayList);
  } else {
    // Generate the display list (while drawing it at the same time).
    modelDisplayList = glGenLists(1);
    glNewList(modelDisplayList, GL_COMPILE_AND_EXECUTE);

    draw_model();

    glEndList();
    displaylist_built = TRUE;
  }
}

// Finish the display process. OpenGL cleanup.
void WGLView::finish_display(void)
{
  // Flip the display buffer back onto the screen.
  SwapBuffers(wglGetCurrentDC());
}

// Recompute the model transform for the current camera position.
void WGLView::compute_modeltransform(void)
{
  transform modeltransform, q_transform;

  if (camera_zoom < 1) {
    // We are zoomed out. The camera moves further back from
    // its starting point.
    camera_dist = base_dist / camera_zoom;
  } else {
    // We are zoomed in... the camera stays at the base point
    // and the frustum's size changes.
    camera_dist = base_dist;
  }
  compute_frustum();

  // The camera is pointing at focus_position from a distance of
  // camera_dist; its direction is given by camera_dir (rotation from +Z).
  //
  // Form the Mirage transform that describes the model's transformation.
  Transform::unit(modeltransform);
  Transform::shift(modeltransform,
		   -focus_position[X], -focus_position[Y], -focus_position[Z]);
  Quaternion::form_transform(camera_dir, q_transform);
  Transform::combine(modeltransform, q_transform, modeltransform);
  Transform::shift(modeltransform, 0, 0, -camera_dist);

  // Convert the Mirage transform to the GL matrix representation.
  make_glmatrix(modeltransform, gl_modelmatrix);
}

// Convert a transform from Mirage form to GL form.
void WGLView::make_glmatrix(transform t, GLfloat *gl_t)
{
  gl_t[0]  = (GLfloat) t[FORE][0][0];
  gl_t[1]  = (GLfloat) t[FORE][1][0];
  gl_t[2]  = (GLfloat) t[FORE][2][0];
  gl_t[3]  = 0.0;
  gl_t[4]  = (GLfloat) t[FORE][0][1];
  gl_t[5]  = (GLfloat) t[FORE][1][1];
  gl_t[6]  = (GLfloat) t[FORE][2][1];
  gl_t[7]  = 0.0;
  gl_t[8]  = (GLfloat) t[FORE][0][2];
  gl_t[9]  = (GLfloat) t[FORE][1][2];
  gl_t[10] = (GLfloat) t[FORE][2][2];
  gl_t[11] = 0.0;
  gl_t[12] = (GLfloat) t[FORE][0][3];
  gl_t[13] = (GLfloat) t[FORE][1][3];
  gl_t[14] = (GLfloat) t[FORE][2][3];
  gl_t[15] = 1.0;  
}

// Draw grids, axes and other scene components that aren't the object.
void WGLView::display_axes(void)
{
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);

  // Draw a white line representing the current debug-line.
  if (modeller->debug_line
#if 0
    && (!modeller->show_only_highlighted)
#endif
    ) {
    vector debug_end;

    Vector::set(debug_end, modeller->debug_linefrom);
    Vector::addfrac(100, modeller->debug_linedir, debug_end);
    glColor3f(1, 1, 1);
    glVertex3f((GLfloat) modeller->debug_linefrom[X],
	             (GLfloat) modeller->debug_linefrom[Y],
	             (GLfloat) modeller->debug_linefrom[Z]);
    glVertex3f((GLfloat) debug_end[X],
               (GLfloat) debug_end[Y],
               (GLfloat) debug_end[Z]);
  }      
  glEnd();  
}

// Draw the modelled object.
void WGLView::draw_model(void)
{
  Tracker *current_tracker;

  static int debug_disablesolidtriangles = FALSE;

  if (draw_wireframe) {
    // Set polygons to line (wireframe) mode.
    glPolygonMode(GL_FRONT, GL_LINE);
    glPolygonMode(GL_BACK,  GL_LINE);
  }

  // Set up the lights.
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glEnable(GL_LIGHT0);

  glLightfv(GL_LIGHT1, GL_POSITION, light1_pos);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT1, GL_SPECULAR, light_specular);
  glEnable(GL_LIGHT1);

  glEnable(GL_DEPTH_TEST);
  // Set up the triangles' material.
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, unselected_diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, unselected_diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, tri_specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &tri_shininess);
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, zero_material);

  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  glColor4fv(unselected_diffuse);
  currently_selected_color = FALSE;

  if (modeller->use_gouraud)
    glShadeModel(GL_SMOOTH);
  else
    glShadeModel(GL_FLAT);

  glDisable(GL_NORMALIZE);

  // Drawing solid triangles... the stencil test is enabled, and
  // all fragments pass (but may or may not modify the stencil buffer).
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS, 0x1, 0x1);

#if DRAW_TEST_BOX
  glBegin(GL_TRIANGLES);

  // -X
  glNormal3f((GLfloat) -1.0, (GLfloat)  0.0, (GLfloat)  0.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) +1.0, (GLfloat) -1.0);

  // +X
  glNormal3f((GLfloat) +1.0, (GLfloat)  0.0, (GLfloat)  0.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) -1.0, (GLfloat) +1.0);

  // -Y
  glNormal3f((GLfloat)  0.0, (GLfloat) -1.0, (GLfloat)  0.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) -1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) -1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) -1.0, (GLfloat) -1.0);

  // +Y
  glNormal3f((GLfloat)  0.0, (GLfloat) +1.0, (GLfloat)  0.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) +1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) +1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) +1.0, (GLfloat) +1.0);

  // -Z
  glNormal3f((GLfloat)  0.0, (GLfloat)  0.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) -1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) +1.0, (GLfloat) -1.0);

  // +Z
  glNormal3f((GLfloat)  0.0, (GLfloat)  0.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) -1.0, (GLfloat) -1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) +1.0, (GLfloat) +1.0);
  glVertex3f((GLfloat) +1.0, (GLfloat) -1.0, (GLfloat) +1.0);

  glEnd();
#else
  // Draw all the Collections in the octree.
  if (modeller->focused_octree != NULL)
    modeller->focused_octree->draw_collections(global_next_drawid++, this, DRAW_SOLID);

  if (modeller->csg_graph->selected_object != NULL) {
    if (show_edges) {
      // Draw all the highlighted edges of the selected object.
      // Don't write anything into the depth-buffer.
      glDepthMask(GL_FALSE);

      // No lighting calculations are applied to these lines.
      glDisable(GL_LIGHTING);

      // The stencil test is disabled.
      glDisable(GL_STENCIL_TEST);

      // Turn on polygon-offset.
      glEnable(GL_POLYGON_OFFSET_LINE);
      glPolygonOffset((GLfloat) -0.01, (GLfloat) -0.01);
      
      glColor4fv(edge_color);
      
      // Draw the selected object's highlighted split edges.
      for (current_tracker = modeller->csg_graph->selected_object->copytracker->trackerlist_head;
           current_tracker != NULL; current_tracker = current_tracker->next)
        current_tracker->copy->octree->draw_collections(global_next_drawid++, this, DRAW_EDGES);

      // Reset the draw settings.
      glDepthMask(GL_TRUE);
      glEnable(GL_LIGHTING);
      glDisable(GL_POLYGON_OFFSET_LINE);
    }

    if (show_ghost) {
      if (!draw_wireframe) {      
        // First up: we need to write the appropriate values into the depth-buffer.
        // This ignores the stencil-test.
        glDisable(GL_STENCIL_TEST);
        glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
      
        // Draw the selected object's depth values.
        for (current_tracker = modeller->csg_graph->selected_object->copytracker->trackerlist_head;
             current_tracker != NULL; current_tracker = current_tracker->next)
          current_tracker->copy->octree->draw_collections(global_next_drawid++, this, DRAW_GHOST_DEPTH);
      
        // Turn the stencil-test and color-mask back on.
        glEnable(GL_STENCIL_TEST);
        glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
      
        // Turn on polygon-offset.
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset((GLfloat) -0.01, (GLfloat) -0.01);
      
        // Set the depth-buffer to read-only mode so that we don't
        // overwrite any information already stored in it when we
        // draw the ghost triangles.
        glDepthMask(GL_FALSE);
      
        // Drawing transparent triangles... the stencil test is enabled,
        // and fragments are only drawn if the stencil is not equal to 0x1.
        // 
        // Fragments set the stencil buffer, so only one transparent layer is
        // ever drawn in a particular pixel.
        glStencilFunc(GL_NOTEQUAL, 0x1, 0x1);
        glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
      }
      
      // Set the blend function and enable blending.
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      glEnable(GL_BLEND);
      
      glColor4fv(ghost_color);
      
      // Draw the selected object. It will be masked out where it actually
      // appears in the scene.
      for (current_tracker = modeller->csg_graph->selected_object->copytracker->trackerlist_head;
           current_tracker != NULL; current_tracker = current_tracker->next)
        current_tracker->copy->octree->draw_collections(global_next_drawid++, this, DRAW_GHOST);
      
      // Restore the standard triangle-drawing settings.
      glDisable(GL_BLEND);
      glBlendFunc(GL_ONE, GL_ZERO);

      if (!draw_wireframe) {
        glStencilFunc(GL_ALWAYS, 0x1, 0x1);
        glDisable(GL_POLYGON_OFFSET_FILL);
        glDepthMask(GL_TRUE);    
      }
    }
  }
#endif

  if (draw_wireframe) {
    // Set back to normal (filled) polygon mode.
    glPolygonMode(GL_FRONT, GL_FILL);
    glPolygonMode(GL_BACK,  GL_FILL);
  }
}

int debug_specialcollx_start = -1;
int debug_specialcollx_end   = -1;

WorldCollection *debughack_specialcollx = NULL;

int debug_specialcollx_parentlevels = 0;
int debug_thiscollx_special;

// Draw a WorldCollection of triangles.
void WGLView::draw_collection(WorldCollection *collection,
                              int is_flipped, drawpass pass)
{
  if (collection->owner->is_transformed)
    setup_transform(collection->owner);

#if _DEBUG
  debug_draw_prep(&collection, is_flipped, pass);

  if (collection == NULL)
    return;
#endif
  
  if (pass == DRAW_SOLID)
    set_draw_properties(collection->owner->is_selected);

  if (pass == DRAW_EDGES) {    
    draw_edges(CUBE(collection->collx->res), collection->collx->cell);
  } else {
    draw_triangles(CUBE(collection->collx->res), collection->collx->cell,
                   collection->collx->modeller->use_gouraud, is_flipped);
  }

#if _DEBUG
#if DRAW_NORMALS
  debug_draw_normals(collection, is_flipped);
#endif
  if (debug_thiscollx_special)
    debug_draw_post(collection, is_flipped, pass);
#endif

  if (collection->owner->is_transformed)
    restore_transform();
}

// Draw the triangles from a Collection.
void WGLView::draw_triangles(int num_cells, string **cell, int use_gouraud, int is_flipped)
{
  int v, cell_index;
  string *stringptr;
  tri *drawptr;

  glBegin(GL_TRIANGLES);

  if (use_gouraud) {
    if (is_flipped) {

      for (cell_index = 0; cell_index < num_cells; cell_index++) {
        for (stringptr = cell[cell_index]; stringptr != NULL; stringptr = stringptr->stringptr) {
          for (drawptr = stringptr->head; drawptr != NULL; drawptr = drawptr->stringptr) {
            for (v = 0; v < 3; v++) {
              glNormal3f(- (GLfloat) drawptr->n[v][X],
                         - (GLfloat) drawptr->n[v][Y],
                         - (GLfloat) drawptr->n[v][Z]);
              glVertex3f((GLfloat) drawptr->v[v][X],
                         (GLfloat) drawptr->v[v][Y],
                         (GLfloat) drawptr->v[v][Z]);
            }
          }
        }
      }

    } else {

      for (cell_index = 0; cell_index < num_cells; cell_index++) {
        for (stringptr = cell[cell_index]; stringptr != NULL; stringptr = stringptr->stringptr) {
          for (drawptr = stringptr->head; drawptr != NULL; drawptr = drawptr->stringptr) {
            for (v = 0; v < 3; v++) {
              glNormal3f((GLfloat) drawptr->n[v][X],
                         (GLfloat) drawptr->n[v][Y],
                         (GLfloat) drawptr->n[v][Z]);
              glVertex3f((GLfloat) drawptr->v[v][X],
                         (GLfloat) drawptr->v[v][Y],
                         (GLfloat) drawptr->v[v][Z]);
            }
          }
        }
      }
    
    }
  } else {
    if (is_flipped) {

      for (cell_index = 0; cell_index < num_cells; cell_index++) {
        for (stringptr = cell[cell_index]; stringptr != NULL; stringptr = stringptr->stringptr) {
          glNormal3f(- (GLfloat) stringptr->norm[X],
                     - (GLfloat) stringptr->norm[Y],
                     - (GLfloat) stringptr->norm[Z]);            

          for (drawptr = stringptr->head; drawptr != NULL; drawptr = drawptr->stringptr) {
            for (v = 0; v < 3; v++)
              glVertex3f((GLfloat) drawptr->v[v][X],
                         (GLfloat) drawptr->v[v][Y],
                         (GLfloat) drawptr->v[v][Z]);
          }
        }
      }

    } else {

      for (cell_index = 0; cell_index < num_cells; cell_index++) {
        for (stringptr = cell[cell_index]; stringptr != NULL; stringptr = stringptr->stringptr) {
          glNormal3f((GLfloat) stringptr->norm[X],
                     (GLfloat) stringptr->norm[Y],
                     (GLfloat) stringptr->norm[Z]);            

          for (drawptr = stringptr->head; drawptr != NULL; drawptr = drawptr->stringptr) {
            for (v = 0; v < 3; v++)
              glVertex3f((GLfloat) drawptr->v[v][X],
                         (GLfloat) drawptr->v[v][Y],
                         (GLfloat) drawptr->v[v][Z]);
          }
        }
      }
    
    }
  }
 
  glEnd();
}
    

// Draw the triangles from a Collection.
void WGLView::draw_edges(int num_cells, string **cell)
{
  int v, cell_index;
  string *stringptr;
  tri *drawptr;

  glBegin(GL_LINES);

  for (cell_index = 0; cell_index < num_cells; cell_index++) {
    for (stringptr = cell[cell_index]; stringptr != NULL; stringptr = stringptr->stringptr) {
      for (drawptr = stringptr->head; drawptr != NULL; drawptr = drawptr->stringptr) {
        if (drawptr->highlight_edge == 0)
          continue;

        for (v = 0; v < 3; v++) {
          if (drawptr->highlight_edge & (1 << v)) {
            glVertex3f((GLfloat) drawptr->v[v][X],
                       (GLfloat) drawptr->v[v][Y],
                       (GLfloat) drawptr->v[v][Z]);
            glVertex3f((GLfloat) drawptr->v[(v + 1) % 3][X],
                       (GLfloat) drawptr->v[(v + 1) % 3][Y],
                       (GLfloat) drawptr->v[(v + 1) % 3][Z]);
          }
        }
      }
    }
  }
 
  glEnd();
}
    
// Set up the transform for a primitive to be drawn.
void WGLView::setup_transform(primtreeprimnode *node)
{
  // This primitive has been moved from its starting point.
  // If it hasn't already had its new modelmatrix calculated, then we need to do so.
  if (node->viewparam == NULL) {
    node->viewparam = malloc(16 * sizeof(GLfloat));
    if (!node->viewparam)
      fatal_error("WGLView::setup_transform", "Can't allocate modelmatrix",
                  "Out of memory.");
    make_glmatrix(node->collection_trans,
                  (GLfloat *) node->viewparam);
  }

  // Push this primitive's modelmatrix.
  //
  // NB: If the primitive has been scaled or sheared from its original position, then its
  // normals will no longer be unit vectors and GL lighting will be incorrect.
  // There is nothing we can do about this without using GL_NORMALIZE which is too slow.
  glPushMatrix();
  glMultMatrixf((GLfloat *) node->viewparam);
}

// Restore the transform.
void WGLView::restore_transform(void)
{
  // Pop this primitive's modelmatrix.
  glPopMatrix();
}

// Set the draw material, etc.
void WGLView::set_draw_properties(int is_selected)
{
  if (is_selected) {
    // Draw 0x1 in the stencil buffer,
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);

#if _DEBUG
    if (debug_thiscollx_special)
      return;
#endif

    if (!currently_selected_color) {
      glColor4fv(selected_diffuse);
      currently_selected_color = TRUE;
    }

  } else {
    // Draw 0x0 in the stencil buffer,
    glStencilOp(GL_KEEP, GL_KEEP, GL_ZERO);

#if _DEBUG
    if (debug_thiscollx_special)
      return;
#endif

    if (currently_selected_color) {
      glColor4fv(unselected_diffuse);
      currently_selected_color = FALSE;
    }
  }
}

#if _DEBUG
// Debugging... prepare to draw a WorldCollection of triangles.
void WGLView::debug_draw_prep(WorldCollection **collection_return,
                              int is_flipped, drawpass pass)
{
  int i, x, y, z, q;
  vector drawvec, drawbase, drawoffset[3];
  double d;
  WorldCollection *collection = *collection_return;

  // Is this collection flagged as "special"?
  if ((collection->collx->collx_id >= debug_specialcollx_start) &&
    (collection->collx->collx_id <= debug_specialcollx_end)) {
    
    if ((pass != DRAW_EDGES) && (pass != DRAW_GHOST_DEPTH) &&
        (debug_specialcollx_start == debug_specialcollx_end)) {
      // Draw a coloured box around it, it is the only one selected as 'special'.
      switch(pass) {
      case DRAW_SOLID:
        glDisable(GL_LIGHTING);
        glDisable(GL_COLOR_MATERIAL);
        break;
      case DRAW_GHOST:
        glDisable(GL_BLEND);
        glDisable(GL_STENCIL_TEST);
        glBlendFunc(GL_ONE, GL_ZERO);
        glDisable(GL_LIGHTING);
        glDisable(GL_COLOR_MATERIAL);
        break;
      default:
        assert(0);
      }
      glBegin(GL_LINES);
      
      Vector::set(drawbase, collection->collx->bounds[0]);
      for (q = X; q <= Z; q++)
        Vector::sub(collection->collx->bounds[(1 << q)], drawbase, drawoffset[q]);
      
      d = 1.0 / collection->collx->res;
      for (z = 0; z <= collection->collx->res; z++) {
        for (y = 0; y <= collection->collx->res; y++) {
          for (x = 0; x <= collection->collx->res; x++) {
            Vector::set(drawvec, drawbase);
            Vector::addfrac(x * d, drawoffset[X], drawvec);
            Vector::addfrac(y * d, drawoffset[Y], drawvec);
            Vector::addfrac(z * d, drawoffset[Z], drawvec);
            
            glColor3f((GLfloat) (x * d), (GLfloat) (y * d), (GLfloat) (z * d));
            
            if (x < collection->collx->res) {
              glVertex3f((GLfloat) drawvec[X],
                         (GLfloat) drawvec[Y],
                         (GLfloat) drawvec[Z]);
              glVertex3f((GLfloat) (drawvec[X] + d * drawoffset[X][X]),
                         (GLfloat) (drawvec[Y] + d * drawoffset[X][Y]),
                         (GLfloat) (drawvec[Z] + d * drawoffset[X][Z]));
            }
            if (y < collection->collx->res) {
              glVertex3f((GLfloat) drawvec[X],
                         (GLfloat) drawvec[Y],
                         (GLfloat) drawvec[Z]);
              glVertex3f((GLfloat) (drawvec[X] + d * drawoffset[Y][X]),
                         (GLfloat) (drawvec[Y] + d * drawoffset[Y][Y]),
                         (GLfloat) (drawvec[Z] + d * drawoffset[Y][Z]));
            }
            if (z < collection->collx->res) {
              glVertex3f((GLfloat) drawvec[X],
                         (GLfloat) drawvec[Y],
                         (GLfloat) drawvec[Z]);
              glVertex3f((GLfloat) (drawvec[X] + d * drawoffset[Z][X]),
                         (GLfloat) (drawvec[Y] + d * drawoffset[Z][Y]),
                         (GLfloat) (drawvec[Z] + d * drawoffset[Z][Z]));
            }
          }
        }
      }
      
      glEnd();

      switch(pass) {
      case DRAW_SOLID:
        glEnable(GL_COLOR_MATERIAL);
        glEnable(GL_LIGHTING);
        break;
      case DRAW_GHOST:
        glEnable(GL_COLOR_MATERIAL);
        glEnable(GL_LIGHTING);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);
        glEnable(GL_STENCIL_TEST);
        break;
      default:
        assert(0);
      }
    }
    
    debughack_specialcollx = collection;
    debug_thiscollx_special = TRUE;

    if (pass != DRAW_EDGES)
      glColor4fv(special_diffuse);
    
    if (debug_specialcollx_parentlevels != 0) {
      for (i = debug_specialcollx_parentlevels; (collection != NULL) && (i > 0); i--)
        collection = collection->parent;
      
      TRACE("draw parent %d of %d, which is %d\n", debug_specialcollx_parentlevels,
        debughack_specialcollx->collx->collx_id, (collection == NULL) ? -1 : collection->collx->collx_id);
    }
    
    *collection_return = collection;
  } else {
    debug_thiscollx_special = FALSE;
  }
}

// Tidy up after the debugging code.
void WGLView::debug_draw_post(WorldCollection *collection,
                              int is_flipped, drawpass pass)
{
  switch(pass) {
  case DRAW_SOLID:
    glColor4fv(unselected_diffuse);
    currently_selected_color = FALSE;
    break;
  case DRAW_GHOST:
    glColor4fv(ghost_color);
    break;
  }
}

// Draw the normals of a Collection.
void WGLView::debug_draw_normals(WorldCollection *collection, int is_flipped)
{
  int cell_index, v;
  vector tempvec;
  double tempmag;
  string *stringptr;
  tri *drawptr;

  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  glColor3f(0.0, 1.0, 0.0);

  glBegin(GL_LINES);

  for (cell_index = 0; cell_index < CUBE(collection->collx->res); cell_index++) {
    for (stringptr = collection->collx->cell[cell_index]; stringptr != NULL; stringptr = stringptr->stringptr) {
      for (drawptr = stringptr->head; drawptr != NULL; drawptr = drawptr->stringptr) {
        tempmag = Vector::magnitude(stringptr->norm);

        Vector::set(tempvec, zero_vector);
        for (v = 0; v < 3; v++)
          Vector::addfrac(1.0 / 3.0, drawptr->v[v], tempvec);

        glVertex3f((GLfloat) tempvec[X],
                   (GLfloat) tempvec[Y],
                   (GLfloat) tempvec[Z]);
        glVertex3f((GLfloat) (tempvec[X] + ((is_flipped) ? -1 : +1) * 0.1 * stringptr->norm[X] / tempmag),
                   (GLfloat) (tempvec[Y] + ((is_flipped) ? -1 : +1) * 0.1 * stringptr->norm[Y] / tempmag),
                   (GLfloat) (tempvec[Z] + ((is_flipped) ? -1 : +1) * 0.1 * stringptr->norm[Z] / tempmag));
      }
    }
  }
 
  glEnd();

  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);
}
#endif

// Draw a box (with or without transparency).
void WGLView::draw_box(box the_box, int transp)
{
#define GLVERTEX(a) {                                                                     \
  glVertex3f((GLfloat) the_box[a][X], (GLfloat) the_box[a][Y], (GLfloat) the_box[a][Z]);  \
}

#define HIGHLIGHT_LINE(a, b) {           \
  GLVERTEX(a); GLVERTEX(b);              \
}

#define HIGHLIGHT_FACE(a, b, c, d) {     \
  GLVERTEX(a); GLVERTEX(b); GLVERTEX(c); \
  GLVERTEX(c); GLVERTEX(d); GLVERTEX(a); \
}

  // Turn off lighting.
  glDisable(GL_LIGHTING);

  // Draw lines along the edges of the terminal.
  glBegin(GL_LINES);
    glColor4f(0.5, 0.5, 1.0, 1.0);
    HIGHLIGHT_LINE(0, 1);
    HIGHLIGHT_LINE(1, 3);
    HIGHLIGHT_LINE(3, 2);
    HIGHLIGHT_LINE(2, 0);
    HIGHLIGHT_LINE(4, 5);
    HIGHLIGHT_LINE(5, 7);
    HIGHLIGHT_LINE(7, 6);
    HIGHLIGHT_LINE(6, 4);
    HIGHLIGHT_LINE(0, 4);
    HIGHLIGHT_LINE(1, 5);
    HIGHLIGHT_LINE(2, 6);
    HIGHLIGHT_LINE(3, 7);
  glEnd();

  if (transp) {
    // Now set the depth-buffer to read-only mode so that we don't
    // overwrite any information already stored in it when we draw the
    // translucent cube.
    glDepthMask(GL_FALSE);

    // Set the blend function and enable blending.
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    // Draw the blue translucent triangles, with alpha of 0.2.
    glBegin(GL_TRIANGLES);
      glColor4f(0, 0, 1, (GLfloat) 0.2);
      HIGHLIGHT_FACE(0, 4, 6, 2);
      HIGHLIGHT_FACE(1, 3, 7, 5);
      HIGHLIGHT_FACE(0, 1, 5, 4);
      HIGHLIGHT_FACE(2, 6, 7, 3);
      HIGHLIGHT_FACE(0, 2, 3, 1);
      HIGHLIGHT_FACE(4, 5, 7, 6);  
    glEnd();

    // Restore the original settings.
    glDisable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ZERO);
    glDepthMask(GL_TRUE);
  }

  glEnable(GL_LIGHTING);

#undef GLVERTEX
#undef HIGHLIGHT_LINE
#undef HIGHLIGHT_FACE
}

// Translate a position on the view to a ray.
void WGLView::position_to_ray(double x, double y,
			                        vector ray_from, vector ray_dir)
{
  // Frustum vertex is at camera_dist.
  Vector::assign(ray_from, 0, 0, camera_dist);

  // Frustum is +/- frustum_size and is near_clip away.
  Vector::assign(ray_dir, frustum_size * x, -frustum_size * y, -near_clip);

  Quaternion::apply(camera_dir_inv, ray_from, ray_from);
  Quaternion::apply(camera_dir_inv, ray_dir,  ray_dir);
  Vector::add(focus_position, ray_from, ray_from); 
}

// Draw any interface objects.
void WGLView::display_interface(void)
{
#if 0
  if ((!modeller->show_only_highlighted) && (modeller->terminal_highlight))
     draw_box(modeller->highlight_terminal->terminal_extent, TRUE);
#endif

  // Time for the magical see-through UI components.
  glDepthMask(GL_FALSE);
  glDisable(GL_DEPTH_TEST);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
  draw_magic_ui(TRUE);

  glDepthMask(GL_TRUE);
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_BLEND);
  glBlendFunc(GL_ONE, GL_ZERO);

  draw_magic_ui(FALSE);
}

// A nasty hack to get around C++'s function pointer limitations.
WGLView *global_WGLView;

// Draw the UI components that are magically displayed through other objects.
void WGLView::draw_magic_ui(int use_fractional_alpha)
{
  global_WGLView = this;
  this->use_fractional_alpha = use_fractional_alpha;
  modeller->skitterlist->iterate(WGLView_draw_skitter);

  // Draw the active Skitter's DMD.
  if (modeller->skitterlist->current_dmd)
    modeller->skitterlist->current_dmd->dmd->draw_dmd(this, modeller->
						      skitterlist->current_dmd,
						      use_fractional_alpha);
}

// Draw a Skitter.
void WGLView_draw_skitter(Skitter *skitter)
{
  vector endpoint[3];
  vector arrowpt1, arrowpt2, offset_vec;
  vector *lines[6];
  double alpha;
  int ax;

  // Don't draw the Skitter that is being used as a DMD. This will be drawn
  // separately, afterwards.
  if (skitter == global_WGLView->modeller->skitterlist->current_dmd)
    return;
  
  // Find the points at the ends of these axes.
  for (ax = X; ax <= Z; ax++)
    Vector::add(skitter->location, skitter->axis_vec[ax], endpoint[ax]);

  // Find the drawing colour and alpha value.
  if (global_WGLView->modeller->skitterlist->current_dmd) {
    global_WGLView->set_drawcolor(skitter->ghosted_color);  
    if (global_WGLView->use_fractional_alpha) {
      // This skitter is currently inactive, a DMD is in use. Draw it
      // in weakly.
      alpha = skitter->ghosted_alpha;
    } else {
      // This skitter is currently inactive, a DMD is in use. It has been
      // drawn (weakly) in the fractional-alpha pass already.
      return;
    }
  } else {
    // Draw the skitter, one of many to be selected from.
    global_WGLView->set_drawcolor(skitter->normal_color);  
    alpha = (global_WGLView->use_fractional_alpha)
      ? skitter->inside_alpha : 1.0;
  }

  global_WGLView->set_drawalpha(alpha);

  for (ax = X; ax <= Z; ax++) {
    // The points of the arrow are back 0.2 units from the end of the axis.
    Vector::set(arrowpt1, endpoint[ax]);
    Vector::addfrac(-0.2, skitter->axis_vec[ax], arrowpt1);
    Vector::set(arrowpt2, arrowpt1);

    // Set the arrow points 0.05 units away from the axis. We use the other
    // two axes to create a perpendicular vector.
    Vector::sub(skitter->axis_vec[(ax+1) % 3], skitter->axis_vec[(ax+2) % 3],
		offset_vec);

    Vector::addfrac(+0.05, offset_vec, arrowpt1);
    Vector::addfrac(-0.05, offset_vec, arrowpt2);

    // Set up the lines to draw.
    lines[0] = &(skitter->location);
    lines[1] = &(endpoint[ax]);
    lines[2] = &(endpoint[ax]);
    lines[3] = &arrowpt1;
    lines[4] = &(endpoint[ax]);
    lines[5] = &arrowpt2;

    // Call the drawing routine.
    global_WGLView->draw_lines(3, lines);
  }
}

/*
 * Primitive drawing routines that are called by the DMDs, etc.
 */

void WGLView::set_drawcolor(color c)
{
  set_drawcolor(c[R], c[G], c[B]);
}

void WGLView::set_drawcolor(double r, double g, double b)
{
  glColor3f((GLfloat) r, (GLfloat) g, (GLfloat) b);
  Color::assign(current_dmd_drawingcolor, r, g, b);
}

void WGLView::set_drawalpha(double a)
{
  glColor4f((GLfloat) current_dmd_drawingcolor[R],
    	      (GLfloat) current_dmd_drawingcolor[G],
	          (GLfloat) current_dmd_drawingcolor[B], (GLfloat) a);
}

void WGLView::draw_lines(int num_lines, vector *point)
{
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);
    for (int i = 0; i < num_lines * 2; i++)
      glVertex3f((GLfloat) point[i][X],
                 (GLfloat) point[i][Y],
                 (GLfloat) point[i][Z]);
  glEnd();
}

void WGLView::draw_lines(int num_lines, vector **point)
{
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);
    for (int i = 0; i < num_lines * 2; i++)
      glVertex3f((GLfloat) (*(point[i]))[X],
                 (GLfloat) (*(point[i]))[Y],
                 (GLfloat) (*(point[i]))[Z]);
  glEnd();
}

void WGLView::draw_unlit_tri(int num_tri, vector *point)
{
  glDisable(GL_LIGHTING);
  glBegin(GL_TRIANGLES);
    for (int i = 0; i < num_tri * 3; i++)
      glVertex3f((GLfloat) point[i][X],
                 (GLfloat) point[i][Y],
                 (GLfloat) point[i][Z]);
  glEnd();
}

/*
 * End of drawing routines.
 */

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif

#endif  // #ifdef _WIN32

