/*
 * wglview.h: Definitions for the WGLView.
 *
 * Contains code to interface with OpenGL under Windows; specifically
 * the CGLView that manages the OpenGL connection.
 */

/* Collection-safe */

#ifdef _WIN32

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// A standalone function (so that function pointers can be used).
void WGLView_draw_skitter(Skitter *skitter);

// The WGLView subclasses View to interface with the Windows OpenGL.
class WGLView : public View {
public:
  // The CView subclass that we draw into.
  class GLView *viewwindow;

  // Distance of the near and far clipping planes in Z from the focus
  // position.
  double clip_dist;
  double frustum_size, near_clip, far_clip;

  // Variables for the OpenGL drawing routines.
  GLfloat gl_modelmatrix[16];
  GLfloat light0_pos[4], light1_pos[4];
  GLfloat light_ambient[4], light_diffuse[4], light_specular[3];
  GLfloat unselected_diffuse[4], selected_diffuse[4];
  GLfloat special_diffuse[4], tri_specular[4];
  GLfloat ghost_color[4], edge_color[4], zero_material[4], tri_shininess;
  int currently_selected_color, use_fractional_alpha;
  color current_dmd_drawingcolor;

  int displaylist_built;
  GLuint modelDisplayList;

  // Resize the view window.
  virtual void resize(int width, int height);

  // The model has been updated; internal representations of the model
  // should be sync'ed with these changes.
  virtual void update(void);

  /*
   * The meat of the implemented code... begin display routines
   */

  // Translate a position on the view to a ray.
  virtual void position_to_ray(double x, double y, vector ray_from, vector ray_dir);

  // Redisplay the window; calls Invalidate() in the view window,
  // which will call display() later on.
  virtual void redisplay(void);

  // View-specific initialisation. Varies from implementation to
  // implementation.
  virtual void init_for_display(void);

  // Draw grids, axes and other scene components that aren't the object.
  virtual void display_axes(void);

  // Draw the object being modelled.
  virtual void display_model(void);
  
  // Draw the interface objects.
  virtual void display_interface(void);

  // Finish the display process. Varies from implementation to
  // implementation.
  virtual void finish_display(void);

  // Recompute the model transform for the current camera position.
  virtual void compute_modeltransform(void);

  /*
   * End main display routines.
   */

  /*
   * Primitive drawing routines that are called by the DMDs, etc.
   */

  virtual void set_drawcolor(color c);
  virtual void set_drawcolor(double r, double g, double b);

  virtual void set_drawalpha(double a);

  virtual void draw_lines(int num_lines, vector *point);
  virtual void draw_lines(int num_lines, vector **point);

  virtual void draw_unlit_tri(int num_tri, vector *point);

  /*
   * End of drawing routines.
   */

  // Draw a box (with or without transparency).
  void draw_box(box the_box, int transp = FALSE);

  // Draw the UI components that are magically displayed through other objects.
  void draw_magic_ui(int use_fractional_alpha);

  // Draw the modelled object.
  void draw_model(void);

  // Draw a collection of triangles.
  virtual void draw_collection(WorldCollection *collection, 
                               int is_flipped, drawpass pass);

  // Set up and restore the transform for a primitive to be drawn.
  void setup_transform(primtreeprimnode *node);
  void restore_transform(void);

  // Set the draw material, etc.
  void set_draw_properties(int is_selected);

  // Draw the triangles from a Collection.
  void draw_triangles(int num_cells, string **cell,
                      int use_gouraud, int is_flipped);

  // Draw the highlighted edges in a Collection.
  void draw_edges(int num_cells, string **cell);

#if _DEBUG
  // Debugging... prepare to draw, and clean up after, a WorldCollection.
  void debug_draw_prep(WorldCollection **collection_return,
                       int is_flipped, drawpass pass);
  void debug_draw_post(WorldCollection *collection,
                       int is_flipped, drawpass pass);

  // Draw the normals of a Collection.
  void debug_draw_normals(WorldCollection *collection, int is_flipped);
#endif

  // Draw any interface objects.
  void draw_interface(void);

  // Compute the OpenGL viewing frustum.
  void compute_frustum(void);

  // Convert a transform from Mirage form to GL form.
  void make_glmatrix(transform t, GLfloat *gl_t);

  // Create (and bind) the view window.
  WGLView(Modeller *m, GLView *viewwindow);
  
  // Delete (and unbind) the view.
  ~WGLView();
};

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif

#endif  // #ifdef _WIN32
