/*
 * collection.cpp: Routines for class Collection.
 *
 * Contains code for the Collections, which contain a subset of the polygonization
 * of an object.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

#define WARN_EMPTYINTERVAL 1
#define WARN_NOTRIANGLES   1

// Provides unique indices.
int global_next_collection_id = 0;
int global_next_sample_id = 0;
int global_uncalculated_collections = 0;

// Add a reference to the Collection.
void Collection::add_reference(void)
{
  assert(reference_count >= 0);
  assert(reference_count < 1000);
  reference_count++;
}

// Remove a reference from the Collection. Returns TRUE if the Collection
// should be deleted now.
int Collection::remove_reference(void)
{
  assert(reference_count > 0);
  assert(reference_count < 1000);

  if (--reference_count == 0)
    return TRUE;
  else
    return FALSE;
}

// Delete the Collection.
Collection::~Collection(void)
{
  int i;

  // collx_id -1 is a sentinel that says 'we've been freed'.
  assert(collx_id != -1);
  collx_id = -1;

  // We should have no references.
  assert(reference_count == 0);

  // Remove references to all of the triangle strings that we
  // reference.
  for (i = 0; i < CUBE(res); i++) {
    if (cell[i] != NULL) {
      assert(cell[i]->reference_count > 0);
      if (--cell[i]->reference_count == 0)
        modeller->stringmem->deallocate_celllist(cell[i]);
    }
  }

  // Delete our triangle data.
  free(cell);

  // Remove the reference to our parent collection,
  // if there is one.
  if (parent_collection != NULL)
    if (parent_collection->remove_reference())
      delete parent_collection;
}


int debug_collection_construct_breakid = -1;

// Create the Collection by polygonizing a primitive surface.
Collection::Collection(primobjectnode *owner, int res, transform collectionspace,
                       box bounds, Modeller *modeller)
{
  int i;

  collx_id = global_next_collection_id++;
  sample_id = global_next_sample_id++;
  global_uncalculated_collections++;

  if (collx_id == debug_collection_construct_breakid)
    _CrtDbgBreak();

  this->res = res;
  this->owner = owner;
  this->modeller = modeller;  
  for (i = 0; i < 8; i++)
    Vector::set(this->bounds[i], bounds[i]);

  for (i = X; i <= Z; i++)
    cellsize[i] = (bounds[(1 << i)][i] - bounds[0][i]) / res;

  Transform::set(this->collectionspace, collectionspace);

  // Allocate the cell array. NB: calloc fills with zero, so all cells point to NULL.
  // The Collection is empty, at least initially. It will be filled by
  // the sampling operation.
  total_tri = 0;
  cell = (string **) calloc(CUBE(this->res), sizeof(string *));
  if (!cell)
    fatal_error("Collection::Collection", "Can't allocate cellinfo!",
                "Out of memory.");

  reference_count = 0;
  parent_collection = NULL;

  // This Collection will need to be sampled before use.
  needs_calculation = CALC_COLLX_SURFACE;
}

// Create an empty Collection that will be polygonized as part of a split operation.
Collection::Collection(Collection *parent_collection)
{
  int i;
  
  collx_id = global_next_collection_id++;
  global_uncalculated_collections++;

  if (collx_id == debug_collection_construct_breakid)
    _CrtDbgBreak();

  this->sample_id = parent_collection->sample_id;
  this->owner = parent_collection->owner;
  this->parent_collection = parent_collection;
  this->modeller = parent_collection->modeller;

  // Set up the Collection's details, mostly from the parent collection.
  this->res = parent_collection->res;
  this->owner = parent_collection->owner;
  Transform::set(this->collectionspace, parent_collection->collectionspace);
  for (i = 0; i < 8; i++)
    Vector::set(this->bounds[i], parent_collection->bounds[i]);

  for (i = X; i <= Z; i++)
    cellsize[i] = (bounds[(1 << i)][i] - bounds[0][i]) / res;

  // Allocate the cell array. NB: calloc fills with zero, so all cells point to NULL.
  // The Collection is empty, at least initially. It will be filled by
  // the split operation.
  total_tri = 0;
  cell = (string **) calloc(CUBE(this->res), sizeof(string *));
  if (!cell)
    fatal_error("Collection::Collection", "Can't allocate cellinfo!",
                "Out of memory.");

  // We are referring to our parent collection, make sure it doesn't
  // get deleted.
  this->reference_count = 0;
  parent_collection->add_reference();

  // This Collection will need to be split before use.
  needs_calculation = CALC_SPLIT;
}

// Perform all of the calculations necessary to create a Collection.
int Collection::calculate(void)
{
  switch(needs_calculation) {
  case CALC_NONE:
    return FALSE;
  case CALC_COLLX_SURFACE:
    sample_surface();
    needs_calculation = CALC_NONE;
    global_uncalculated_collections--;
    return TRUE;
  default:
    assert(0);
    return FALSE;
  }
}

// Sample the primitive's surface to create an initial polygonization.
void Collection::sample_surface(void)
{
  string **thiscell;
  unsigned char inout;
  tri_topology *t_top;
  int i, j, v[3], cell_x, cell_y, cell_z, num_tri, num_strings;
  string *this_string, *last_string;
  tri *this_tri, *last_tri, *triptr;
  pointdata *pointptr;
  int *edgeptr, start_new_string, last_string_length, this_string_length;
  vector *edgepointptr, *edgenormptr, e0, e1, last_norm, this_norm;
  
  SurfaceSampler *sampler = new SurfaceSampler(this);

  // Create a SurfaceSampler to hold all of the polygonization variables
  // and sample the surface using a grid of rays. This sets up the
  // arrays pointinfo, edge_point and edge_normal in the SurfaceSampler.
  sampler->sample();

  // Initialise the topology table if necessary.
  if (!topology_table_initialised)
    init_topology_table(); 

  // Set up pointers for the cell traversal.
  thiscell = cell;

  for (cell_z = 0; cell_z < res; cell_z++) {
    for (cell_y = 0; cell_y < res; cell_y++) {
      pointptr = sampler->pointinfo +  (0 + cell_y * (res + 1) + cell_z * SQR(res + 1));
      edgeptr = sampler->is_edge + (3 * (0 + cell_y * (res + 1) + cell_z * SQR(res + 1)));
      edgepointptr = sampler->edge_point + (3 * (0 + cell_y * (res + 1)
                                            + cell_z * SQR(res + 1)));
      edgenormptr = sampler->edge_normal + (3 * (0 + cell_y * (res + 1)
                                            + cell_z * SQR(res + 1)));

      for (cell_x = 0; cell_x < res; cell_x++) {
        inout = *pointptr;

        if ((inout != 0x00) && (inout != 0xFF)) {
          // We are currently generating the first triangle.
          last_tri = NULL;
          this_string = last_string = NULL;
          this_string_length = 0;
          num_tri = num_strings = 0;

          // Create polygons for this cell.
          t_top = &(polygon_topology_table[inout].topology);

          for (i = 0; (*t_top)[i] != TRI_FINISH; ) {
            assert(i < TRI_TOPOLOGY_LEN);

            // Make room for a new string in our buffer, containing
            // precisely one triangle.
            this_tri = modeller->trimem->allocate();

            // Copy the vertex information for this triangle.
            for (j = 0; j < 3; j++) {
            switch((*t_top)[i++]) {
              case 0:
                v[j] = 3*(0 + 0*(res+1) + 0*SQR(res+1)) + X;
                break;
              case 1:
                v[j] = 3*(0 + 0*(res+1) + 0*SQR(res+1)) + Y;
                break;
              case 2:
                v[j] = 3*(0 + 0*(res+1) + 0*SQR(res+1)) + Z;
                break;
              case 3:
                v[j] = 3*(1 + 0*(res+1) + 0*SQR(res+1)) + Y;
                break;
              case 4:
                v[j] = 3*(1 + 0*(res+1) + 0*SQR(res+1)) + Z;
                break;
              case 5:
                v[j] = 3*(0 + 1*(res+1) + 0*SQR(res+1)) + X;
                break;
              case 6:
                v[j] = 3*(0 + 1*(res+1) + 0*SQR(res+1)) + Z;
                break;
              case 7:
                v[j] = 3*(0 + 0*(res+1) + 1*SQR(res+1)) + X;
                break;
              case 8:
                v[j] = 3*(0 + 0*(res+1) + 1*SQR(res+1)) + Y;
                break;
              case 9:
                v[j] = 3*(1 + 1*(res+1) + 0*SQR(res+1)) + Z;
                break;
              case 10:
                v[j] = 3*(1 + 0*(res+1) + 1*SQR(res+1)) + Y;
                break;
              case 11:
                v[j] = 3*(0 + 1*(res+1) + 1*SQR(res+1)) + X;
                break;
              }

              assert(edgeptr[v[j]] != 0);
              Vector::set(this_tri->v[j], edgepointptr[v[j]]);
              Vector::set(this_tri->n[j], edgenormptr[v[j]]);
            }

            Vector::sub(this_tri->v[1], this_tri->v[0], e0);
            Vector::sub(this_tri->v[2], this_tri->v[0], e1);
            Vector::cross(e0, e1, this_norm);
            Vector::normalise(this_norm, "sample_surface: zero normal");

            // Now, copy the adjacency information from the topology table.
            for (j = 0; j < 3; j++) {
              this_tri->adj[j][ADJ_INDEX] = (*t_top)[i++];
              this_tri->adj[j][ADJ_VERTEX] = (*t_top)[i++];

              if (this_tri->adj[j][ADJ_INDEX] == NONE_ADJACENT)
                this_tri->adj[j][ADJ_INDEX] = -1;
            }
            this_tri->highlight_edge = 0;

            start_new_string = TRUE;
            if (last_tri != NULL) {
              Vector::sub(last_norm, this_norm, last_norm);
              if (Vector::magnitude(last_norm) < 1e-03) {
                // this_tri is coplanar with the last triangle.
                // Don't start a new string.
                start_new_string = FALSE;
              }
            }
            Vector::set(last_norm, this_norm);

            if (start_new_string) {
              if (this_string != NULL) {
                // Set up the string length of the last string.
                last_string_length = this_string_length;
                last_string = this_string;

                for (j = 0, triptr = last_string->head; triptr != NULL; triptr = triptr->stringptr, j++)
                  triptr->string_length = last_string_length - j;
              }

              // this_tri is the only triangle in its string.
              this_string_length = 1;
              assert(this_tri->stringptr == NULL);

              num_strings++;
              this_string = modeller->stringmem->allocate();
              this_string->head = this_tri;
              Vector::set(this_string->norm, this_norm);
              this_string->plane_dot = Vector::dot(this_norm, this_tri->v[0]);

              // Set up the last string in the celllist to point to this one...
              // if we are the first string, then set the cell up to point to us.
              if (last_string == NULL)
                *thiscell = this_string;
              else
                last_string->stringptr = this_string;
            } else {
              // this_tri is coplanar with the last triangle. Add it to the end of
              // the last string.
              assert(last_tri != NULL);
              last_tri->stringptr = this_tri;
              this_string_length++;
            }

            last_tri = this_tri;
            num_tri++;
          }

          if (num_tri > 0) {
            total_tri += num_tri;

            if (this_string != NULL) {
              for (j = 0, triptr = this_string->head; triptr != NULL; triptr = triptr->stringptr, j++)
                triptr->string_length = this_string_length - j;
            }

            // Set up the string.
            for (i = 0, this_string = *thiscell; this_string != NULL; this_string = this_string->stringptr, i++) {
              this_string->reference_count = (i == 0) ? 1 : -1;
              this_string->celllist_length = num_strings - i;
              this_string->num_tri = num_tri;

              num_tri -= this_string->head->string_length;
            }
            assert(num_tri == 0);
          }
        }
        pointptr++;
        thiscell++;
        edgeptr += 3;
        edgenormptr += 3;
        edgepointptr += 3;
      }
    }
  }

  // We've finished with the SurfaceSampler.
  delete sampler;
}

/*
 * SurfaceSampler
 */

SurfaceSampler::SurfaceSampler(Collection *collection)
{
  int i;

  this->collection = collection;
  res = collection->res;
  prim = collection->owner->prim;

  /*
   * Set up arrays for the polygonization process.
   */

  x_num_intersects = (int *) calloc(SQR(res+1), sizeof(int));
  y_num_intersects = (int *) calloc(SQR(res+1), sizeof(int));
  z_num_intersects = (int *) calloc(SQR(res+1), sizeof(int));
  x_world_rayfroms = (vector *) calloc(SQR(res+1), sizeof(vector));
  y_world_rayfroms = (vector *) calloc(SQR(res+1), sizeof(vector));
  z_world_rayfroms = (vector *) calloc(SQR(res+1), sizeof(vector));
  x_ray_tvals = (rayintersect *) calloc(SQR(res+1) * RAYINTBUFSIZE, sizeof(rayintersect));
  y_ray_tvals = (rayintersect *) calloc(SQR(res+1) * RAYINTBUFSIZE, sizeof(rayintersect));
  z_ray_tvals = (rayintersect *) calloc(SQR(res+1) * RAYINTBUFSIZE, sizeof(rayintersect));
  is_edge = (int *) calloc(3*CUBE(res+1), sizeof(int));
  edge_point = (vector *) calloc(3*CUBE(res+1), sizeof(vector));
  edge_normal = (vector *) calloc(3*CUBE(res+1), sizeof(vector));
  pointinfo = (pointdata *) calloc(CUBE(res+1), sizeof(pointdata));

  if (!x_num_intersects || !y_num_intersects || !z_num_intersects || 
	    !x_world_rayfroms || !y_world_rayfroms || !z_world_rayfroms || 
      !x_ray_tvals      || !y_ray_tvals      || !z_ray_tvals      ||
      !is_edge || !edge_point || !edge_normal || !pointinfo)
    fatal_error("SurfaceSampler::SurfaceSampler", "Can't allocate polygonization arrays!",
                "Out of memory.");

  // There are no edges, initially.
  memset((void *) is_edge, 0, 3*CUBE(res+1) * sizeof(int));

  // Calculate the primitive-space locations of each corner of the bounding box.
  for (i = 0; i < 8; i++)
    Matrix::apply(collection->collectionspace[AFT], collection->bounds[i], primbox[i]);

  /*
   * Construct the offsets for a single cell in PRIMITIVE SPACE. NB!! These are
   * negative to facilitate ray casting from high to low.
   */

  Vector::sub(primbox[1], primbox[0], prim_x_offset);
  Vector::sub(primbox[2], primbox[0], prim_y_offset);
  Vector::sub(primbox[4], primbox[0], prim_z_offset);
  Vector::mult(-1.0 / res, prim_x_offset, prim_x_offset);
  Vector::mult(-1.0 / res, prim_y_offset, prim_y_offset);
  Vector::mult(-1.0 / res, prim_z_offset, prim_z_offset);

  /*
   * worldbox[] is the WORLD-SPACE bounds of this terminal. Construct the
   * same offsets for WORLD SPACE.
   *
   * Note: these offsets are in collectionspace, which is currently
   * worldspace for this object, but might not be after the object is
   * manipulated, etc.
   */

  Vector::sub(collection->bounds[1], collection->bounds[0], world_x_offset);
  Vector::sub(collection->bounds[2], collection->bounds[0], world_y_offset);
  Vector::sub(collection->bounds[4], collection->bounds[0], world_z_offset);
  Vector::mult(-1.0 / res, world_x_offset, world_x_offset);
  Vector::mult(-1.0 / res, world_y_offset, world_y_offset);
  Vector::mult(-1.0 / res, world_z_offset, world_z_offset);

  // Set all the ray-intersect data to sentinel values.
  for (i = 0; i < SQR(res+1); i++) {
    x_num_intersects[i] = -1;
    y_num_intersects[i] = -1;
    z_num_intersects[i] = -1;
  }

  // Set up the dummy start- and end-of-ray intersection data.
  dummy_start_int.t = 0;
  dummy_start_int.is_entering = FALSE;
  Vector::assign(dummy_start_int.normal, 0, 0, 1);
  dummy_end_int.t = res;
  dummy_end_int.is_entering = FALSE;
  Vector::assign(dummy_end_int.normal, 0, 0, 1);
}

SurfaceSampler::~SurfaceSampler(void)
{
  free(x_num_intersects);
  free(y_num_intersects);
  free(z_num_intersects);
  free(x_world_rayfroms);
  free(y_world_rayfroms);
  free(z_world_rayfroms);
  free(x_ray_tvals);
  free(y_ray_tvals);
  free(z_ray_tvals);
  free(is_edge);
  free(edge_point);
  free(edge_normal);
  free(pointinfo);
}

void SurfaceSampler::sample(void)
{
  int i;
 
  /*
   * First step: sample the high-X,Y,Z point of the terminal using point_int.
   */

  cell_x = res;
  cell_y = res;
  cell_z = res;
  pointptr = &pointinfo[cell_x + cell_y*(res+1) + cell_z*SQR(res+1)];
  edgeptr = &is_edge[3 * (cell_x + cell_y*(res+1) + cell_z*SQR(res+1))];
  edgepointptr = &edge_point[3 * (cell_x + cell_y*(res+1)
                             + cell_z*SQR(res+1))];
  edgenormptr = &edge_normal[3 * (cell_x + cell_y*(res+1)
                             + cell_z*SQR(res+1))];

  if (prim->point_intersect(primbox[7]) == IN)
    current_in = 0x01;
  else
    current_in = 0x00;

  *pointptr = current_in;

  /*
   * Second step: cast a ray along X (from high to low) and find the
   * intersections with the primitive (if any).
   */

  ray_index = cell_z + cell_y*(res+1);
  Vector::set(prim_traverseray_from,  primbox[7]);
  Vector::set(world_traverseray_from, collection->bounds[7]);
  Vector::set(x_world_rayfroms[ray_index], world_traverseray_from);

  // Perform the ray intersection... in primitive space.
  x_num_intersects[ray_index] = prim->ray_intersect(prim_traverseray_from,
                                                    prim_x_offset,
                                                    &x_ray_tvals[ray_index * RAYINTBUFSIZE]);
  assert(x_num_intersects[ray_index] <= RAYINTBUFSIZE);

  // Find the next intersection.
  last_int = &dummy_start_int;
  current_int = &dummy_end_int;
  for (i = 0; i < x_num_intersects[ray_index]; i++)
    if ((x_ray_tvals[ray_index * RAYINTBUFSIZE + i].t > last_int->t) &&
        (x_ray_tvals[ray_index * RAYINTBUFSIZE + i].t < current_int->t))
      current_int = &x_ray_tvals[ray_index * RAYINTBUFSIZE + i];

  // Set up for the start of the traverse from high-X to low-X.
  edge_this_cell = FALSE;
  cell_x--;
  pointptr--;
  edgeptr -= 3;
  edgepointptr -= 3;
  edgenormptr -= 3;

  current_inout = (current_in) << (1 << X);

  while (cell_x >= 0) {
    // There is a ray-intersection before the end of the row.
    while ((cell_x > res - current_int->t - TINY) &&
	   (cell_x >= 0)) {
      
      // current_inout is right now the high side only.
      current_inout |= current_in;

      edgeptr[X] = FALSE;
      edgeptr[Y] = FALSE;
      edgeptr[Z] = FALSE;

      // Is there an edge in X this cell?
      if (edge_this_cell) {
        edgeval = res - last_int->t - cell_x;
        CLAMP(0, edgeval, 1);

        edgeptr[X] = TRUE;

        // Store the edge point, in worldspace.
        Vector::set(edgepointptr[X], world_traverseray_from);
        Vector::addfrac(last_int->t, world_x_offset, edgepointptr[X]);

        // Store the edge normal. Make sure to transform the
        // normal from primitive space to worldspace.
        Matrix::applytranspose(collection->collectionspace[AFT],
                               last_int->normal, edgenormptr[X]);
        Vector::normalise(edgenormptr[X], "surface_sampler: edgenormal[X]");
      }

      // Store the inout information for this cell.
      *pointptr = current_inout;

      // Set up for the next cell.
      pointptr--;
      edgeptr -= 3;
      edgepointptr -= 3;
      edgenormptr -= 3;
      cell_x--;
      edge_this_cell = FALSE;

      // Shift the inout value for this cell so it becomes the high edge
      // of the next cell.
      current_inout = (current_inout & 0x01) << (1 << X);
    }

    if (cell_x >= 0) {

      // Find the next intersection.
      last_int = current_int;
      current_int = &dummy_end_int;
      for (i = 0; i < x_num_intersects[ray_index]; i++) {
        if ((x_ray_tvals[ray_index * RAYINTBUFSIZE + i].t > last_int->t) &&
            (x_ray_tvals[ray_index * RAYINTBUFSIZE + i].t < current_int->t))
          current_int = &x_ray_tvals[ray_index * RAYINTBUFSIZE + i];
      }

      // We have found the intersection at last_int, we need to flip
      // the current inout value.
      current_in ^= 0x01;

      // The cell currently being processed has an edge in the traversal
      // direction.
      edge_this_cell = TRUE;
    }
  }


  /*
   * Third step: cast rays along Y from each point in the X row which we
   * just sampled.
   */

  Vector::set(prim_ray_from,  primbox[7]);
  Vector::set(world_ray_from, collection->bounds[7]);
  for (cell_x = res; cell_x >= 0; cell_x--) {
    cell_y = res;
    pointptr = &pointinfo[cell_x + cell_y*(res+1) + cell_z*SQR(res+1)];
    edgeptr = &is_edge[3 * (cell_x + cell_y*(res+1) + cell_z*SQR(res+1))];
    edgepointptr = &edge_point[3 * (cell_x + cell_y*(res+1)
                               + cell_z*SQR(res+1))];
    edgenormptr = &edge_normal[3 * (cell_x + cell_y*(res+1)
                               + cell_z*SQR(res+1))];
    current_inout = *pointptr;
    current_in = current_inout & 0x01;
    
    ray_index = cell_z + cell_x*(res+1);

    Vector::set(prim_traverseray_from,  prim_ray_from);
    Vector::set(world_traverseray_from, world_ray_from);
    Vector::set(y_world_rayfroms[ray_index], world_ray_from);

    // Cast the ray in primitive space.
    y_num_intersects[ray_index] = prim->ray_intersect(prim_ray_from,
                                                      prim_y_offset,
                                                      &y_ray_tvals[ray_index * RAYINTBUFSIZE]);
    assert(y_num_intersects[ray_index] <= RAYINTBUFSIZE);
	
    // Find the next intersection.
    last_int = &dummy_start_int;
    current_int = &dummy_end_int;
    for (i = 0; i < y_num_intersects[ray_index]; i++)
      if ((y_ray_tvals[ray_index * RAYINTBUFSIZE + i].t > last_int->t) &&
          (y_ray_tvals[ray_index * RAYINTBUFSIZE + i].t < current_int->t))
        current_int = &y_ray_tvals[ray_index * RAYINTBUFSIZE + i];

    // Set up for the start of the traverse from high-Y to low-Y.
    edge_this_cell = FALSE;
    cell_y--;
    pointptr -= (res+1);
    edgeptr -= 3 * (res+1);
    edgepointptr -= 3 * (res+1);
    edgenormptr -= 3 * (res+1);

    current_inout = (current_inout & 0x03) << (1 << Y);

    while (cell_y >= 0) {
      // There is a ray-intersection before the end of the row.
      while ((cell_y > res - current_int->t - TINY) &&
             (cell_y >= 0)) {

        // current_inout is right now the high Y side only.

        // Add the low-Y low-X part of the inout value; the base value of this
        // cell.
        current_inout |= current_in;

        edgeptr[X] = FALSE;
        edgeptr[Y] = FALSE;
        edgeptr[Z] = FALSE;

        if (cell_x < res) {
          // Add the low-Y high-X part of the inout value. This is the
          // base value of the cell that is +1 in X.
          xcell_in = pointptr[1] & 0x01;
          current_inout |= (xcell_in << (1 << X));

          // Is there an edge in X this cell?
          if (xcell_in != current_in) {
            // We have to use the ray in X at (cell_y, cell_z). Is this ray
            // already in the array of ray values?
            castray_index = cell_z + cell_y*(res+1);
            if (x_num_intersects[castray_index] == -1) {
              // We have to cast this ray.
              Vector::set(prim_castray_from, primbox[7]);
              Vector::addfrac(res - cell_y, prim_y_offset,
                              prim_castray_from);
              Vector::set(x_world_rayfroms[castray_index], collection->bounds[7]);
              Vector::addfrac(res - cell_y, world_y_offset,
                              x_world_rayfroms[castray_index]);

              // Cast the ray in primitive space.
              x_num_intersects[castray_index] = prim->ray_intersect(prim_castray_from,
                                                                    prim_x_offset,
                                             &x_ray_tvals[castray_index * RAYINTBUFSIZE]);
              assert(x_num_intersects[castray_index] <= RAYINTBUFSIZE);
            }
	  
            // We have the ray. Check the intersect values to find any that
            // occur along the edge of this cell. Note that cellbase_t is the
            // HIGH edge of the cell.
            cellbase_t = (res - 1) - cell_x;

            for (i = 0; i < x_num_intersects[castray_index]; i++)
              if ((x_ray_tvals[castray_index * RAYINTBUFSIZE + i].t > cellbase_t - TINY) &&
                  (x_ray_tvals[castray_index * RAYINTBUFSIZE + i].t < cellbase_t + 1 + TINY))
                break;

            if (i != x_num_intersects[castray_index]) {
              edgeval = 1 - (x_ray_tvals[castray_index * RAYINTBUFSIZE + i].t - cellbase_t);

              CLAMP(0, edgeval, 1);

              edgeptr[X] = TRUE;

              // Store the edge point in worldspace.
              Vector::set(edgepointptr[X], x_world_rayfroms[castray_index]);
              Vector::addfrac(x_ray_tvals[castray_index * RAYINTBUFSIZE + i].t,
                              world_x_offset, edgepointptr[X]);

              // Store the edge normal. Make sure to transform the
              // normal from primitive space to worldspace.
              Matrix::applytranspose(collection->collectionspace[AFT],
                                     x_ray_tvals[castray_index * RAYINTBUFSIZE + i].normal,
                                     edgenormptr[X]);
              Vector::normalise(edgenormptr[X], "surface_sampler: edgenormal[X]");

            } else {
#if WARN_EMPTYINTERVAL
              printf("*** WARNING! cell %d %d %d no X ray intersects!\n",
              cell_x, cell_y, cell_z);
#endif
            }
          } 
        } 

        // Is there an edge in Y this cell?
        if (edge_this_cell) {
          edgeval = res - last_int->t - cell_y;
          CLAMP(0, edgeval, 1);

          edgeptr[Y] = TRUE;

          // Store the edge point in worldspace.
          Vector::set(edgepointptr[Y], world_traverseray_from);
          Vector::addfrac(last_int->t, world_y_offset, edgepointptr[Y]);

          // Store the edge normal. Make sure to transform the
          // normal from primitive space to worldspace.
          Matrix::applytranspose(collection->collectionspace[AFT],
                                 last_int->normal, edgenormptr[Y]);
          Vector::normalise(edgenormptr[Y], "surface_sampler: edgenormal[Y]");
        }

        // Store the inout for this cell.
        *pointptr = current_inout;
	
        // Set up for the next cell.
        pointptr -= (res+1);
        edgeptr -= 3 * (res+1);
        edgepointptr -= 3 * (res+1);
        edgenormptr -= 3 * (res+1);
        cell_y--;
        edge_this_cell = FALSE;

        // Shift the inout value for this cell so it becomes the high edge
        // of the next cell.
        current_inout = (current_inout & 0x03) << (1 << Y);
      }

      if (cell_y >= 0) {
        // Find the next intersection.
        last_int = current_int;
        current_int = &dummy_end_int;
        for (i = 0; i < y_num_intersects[ray_index]; i++)
          if ((y_ray_tvals[ray_index * RAYINTBUFSIZE + i].t > last_int->t) &&
              (y_ray_tvals[ray_index * RAYINTBUFSIZE + i].t < current_int->t))
            current_int = &y_ray_tvals[ray_index * RAYINTBUFSIZE + i];

        // We have found the intersection at last_int, we need to flip
        // the current inout value.
        current_in ^= 0x01;

        // The cell currently being processed has an edge in the traversal
        // direction.
        edge_this_cell = TRUE;
      }
    }

    Vector::add(prim_x_offset, prim_ray_from, prim_ray_from);
    Vector::add(world_x_offset, world_ray_from, world_ray_from);
  }

  /*
   * Fourth step: cast rays along Z from each point in the XY layer which we
   * just sampled.
   */

  Vector::set(prim_ray_from,  primbox[7]);
  Vector::set(world_ray_from, collection->bounds[7]);
  for (cell_y = res; cell_y >= 0; cell_y--) {
    for (cell_x = res; cell_x >= 0; cell_x--) {
      cell_z = res;
      pointptr = &pointinfo[cell_x + cell_y*(res+1) + cell_z*SQR(res+1)];
      edgeptr = &is_edge[3 * (cell_x + cell_y*(res+1) + cell_z*SQR(res+1))];
      edgepointptr = &edge_point[3 * (cell_x + cell_y*(res+1)
				                        + cell_z*SQR(res+1))];
      edgenormptr = &edge_normal[3 * (cell_x + cell_y*(res+1)
                                + cell_z*SQR(res+1))];
      current_inout = *pointptr;
      current_in = current_inout & 0x01;

      ray_index = cell_x + cell_y*(res+1);

      Vector::set(prim_traverseray_from,  prim_ray_from);
      Vector::set(world_traverseray_from, world_ray_from);
      Vector::set(z_world_rayfroms[ray_index], world_ray_from);

      // Cast the ray in primitive space.
      z_num_intersects[ray_index] = prim->ray_intersect(prim_ray_from,
                                                        prim_z_offset,
                                     &z_ray_tvals[ray_index * RAYINTBUFSIZE]);
      assert(z_num_intersects[ray_index] <= RAYINTBUFSIZE);
      
      // Find the next intersection.
      last_int = &dummy_start_int;
      current_int = &dummy_end_int;
      for (i = 0; i < z_num_intersects[ray_index]; i++)
        if ((z_ray_tvals[ray_index * RAYINTBUFSIZE + i].t > last_int->t) &&
            (z_ray_tvals[ray_index * RAYINTBUFSIZE + i].t < current_int->t))
          current_int = &z_ray_tvals[ray_index * RAYINTBUFSIZE + i];

      // Set up for the start of the traverse from high-Z to low-Z.
      edge_this_cell = FALSE;
      cell_z--;
      pointptr -= SQR(res+1);
      edgeptr -= 3 * SQR(res+1);
      edgepointptr -= 3 * SQR(res+1);
      edgenormptr -= 3 * SQR(res+1);

      current_inout = (current_inout & 0x0F) << (1 << Z);
      
      while (cell_z >= 0) {
        // There is a ray-intersection before the end of the row.
        while ((cell_z > res - current_int->t - TINY) &&
               (cell_z >= 0)) {

          // current_inout is right now the high Z side only.

          // Add the low-Z low-Y low-X part of the inout value; the base
          // value of this cell.
          current_inout |= current_in;

          edgeptr[X] = FALSE;
          edgeptr[Y] = FALSE;
          edgeptr[Z] = FALSE;

          if (cell_y < res) {
            // Add the low-Z high-Y part of the inout value; the base and
            // high-X values of the cell that is +1 in Y.
            ycell_in = pointptr[res + 1] & 0x03;
            current_inout |= (ycell_in << (1 << Y));

            // Is there an edge in Y this cell?
            if ((ycell_in & 0x01) != current_in) {
              // We have to use the ray in Y at (cell_x, cell_z). Is this ray
              // already in the array of ray values?
              castray_index = cell_z + cell_x*(res+1);
              if (y_num_intersects[castray_index] == -1) {
                // We have to cast this ray.
                Vector::set(prim_castray_from, primbox[7]);
                Vector::addfrac(res - cell_z, prim_z_offset,
                                prim_castray_from);
                Vector::addfrac(res - cell_x, prim_x_offset,
                                prim_castray_from);

                Vector::set(y_world_rayfroms[castray_index], collection->bounds[7]);
                Vector::addfrac(res - cell_z, world_z_offset,
                                y_world_rayfroms[castray_index]);
                Vector::addfrac(res - cell_x, world_x_offset,
                y_world_rayfroms[castray_index]);

                // Cast the ray in primitive space.
                y_num_intersects[castray_index] = prim->ray_intersect(prim_castray_from,
                                                                      prim_y_offset,
                                             &y_ray_tvals[castray_index * RAYINTBUFSIZE]);
                assert(y_num_intersects[castray_index] <= RAYINTBUFSIZE);
              }
	  
              // We have the ray. Check the intersect values to find any that
              // occur along the edge of this cell. Note that cellbase_t is the
              // HIGH edge of the cell.
              cellbase_t = (res - 1) - cell_y;
              for (i = 0; i < y_num_intersects[castray_index]; i++) {
                if ((y_ray_tvals[castray_index * RAYINTBUFSIZE + i].t > cellbase_t - TINY) &&
                    (y_ray_tvals[castray_index * RAYINTBUFSIZE + i].t < cellbase_t + 1 + TINY))
                  break;
              }
              if (i != y_num_intersects[castray_index]) {
                edgeval = 1 - (y_ray_tvals[castray_index * RAYINTBUFSIZE + i].t - cellbase_t);
                CLAMP(0, edgeval, 1);

                edgeptr[Y] = TRUE;

                // Store the edge point in worldspace.
                Vector::set(edgepointptr[Y], y_world_rayfroms[castray_index]);
                Vector::addfrac(y_ray_tvals[castray_index * RAYINTBUFSIZE + i].t,
                                world_y_offset,
                                edgepointptr[Y]);

                // Store the edge normal. Make sure to transform
                // the normal from primitive space to worldspace.
                Matrix::applytranspose(collection->collectionspace[AFT],
                                       y_ray_tvals[castray_index * RAYINTBUFSIZE + i].normal,
                                       edgenormptr[Y]);
                Vector::normalise(edgenormptr[Y], "surface_sampler: edgenormal[Y]");
              } else {
#if WARN_EMPTYINTERVAL
                printf("*** WARNING! cell %d %d %d no Y ray intersects!\n",
                       cell_x, cell_y, cell_z);
#endif
              }
            } 
          } 
		   
          if (cell_x < res) {
            // Add the low-Z low-Y high-X part of the inout value; the base
            // value of the cell that is +1 in X.
            xcell_in = pointptr[1] & 0x01;
            current_inout |= (xcell_in << (1 << X));

            // Is there an edge in X this cell?
            if (xcell_in != current_in) {
              // We have to use the ray in X at (cell_y, cell_z). Is this ray
              // already in the array of ray values?
              castray_index = cell_z + cell_y*(res+1);
              if (x_num_intersects[castray_index] == -1) {
                // We have to cast this ray.
                Vector::set(prim_castray_from, primbox[7]);
                Vector::addfrac(res - cell_z, prim_z_offset,
                                prim_castray_from);
                Vector::addfrac(res - cell_y, prim_y_offset,
                                prim_castray_from);

                Vector::set(x_world_rayfroms[castray_index], collection->bounds[7]);
                Vector::addfrac(res - cell_z, world_z_offset,
                                x_world_rayfroms[castray_index]);
                Vector::addfrac(res - cell_y, world_y_offset,
                                x_world_rayfroms[castray_index]);

                // Cast the ray in primitive space.
                x_num_intersects[castray_index] = prim->ray_intersect(prim_castray_from,
                                                                      prim_x_offset,
                                              &x_ray_tvals[castray_index * RAYINTBUFSIZE]);
                assert(x_num_intersects[castray_index] <= RAYINTBUFSIZE);
              }
	  
              // We have the ray. Check the intersect values to find any that
              // occur along the edge of this cell. Note that cellbase_t is the
              // HIGH edge of the cell.
              cellbase_t = (res - 1) - cell_x;
              for (i = 0; i < x_num_intersects[castray_index]; i++) {
                if ((x_ray_tvals[castray_index * RAYINTBUFSIZE + i].t > cellbase_t - TINY) &&
                    (x_ray_tvals[castray_index * RAYINTBUFSIZE + i].t < cellbase_t + 1 + TINY))
                  break;
              }
              if (i != x_num_intersects[castray_index]) {
                edgeval = 1 - (x_ray_tvals[castray_index * RAYINTBUFSIZE + i].t - cellbase_t);
                CLAMP(0, edgeval, 1);

                edgeptr[X] = TRUE;

                // Store the edge point in worldspace.
                Vector::set(edgepointptr[X], x_world_rayfroms[castray_index]);
                Vector::addfrac(x_ray_tvals[castray_index * RAYINTBUFSIZE + i].t,
                                world_x_offset,
                                edgepointptr[X]);

                // Store the edge position and normal. Make sure to transform
                // the normal from primitive space to worldspace.
                Matrix::applytranspose(collection->collectionspace[AFT],
                                       x_ray_tvals[castray_index * RAYINTBUFSIZE + i].normal,
                                       edgenormptr[X]);
                Vector::normalise(edgenormptr[X], "surface_sampler: edgenormal[X]");
              } else {
#if WARN_EMPTYINTERVAL
                printf("*** WARNING! cell %d %d %d no Y ray intersects!\n",
                       cell_x, cell_y, cell_z);
#endif
              }
            }
          }
 
          // Is there an edge in Z this cell?
          if (edge_this_cell) {
            edgeval = res - last_int->t - cell_z;
            CLAMP(0, edgeval, 1);

            edgeptr[Z] = TRUE;
	    
            // Store the edge point in worldspace.
            Vector::set(edgepointptr[Z], world_traverseray_from);
            Vector::addfrac(last_int->t, world_z_offset, edgepointptr[Z]);

            // Store the edge position and normal. Make sure to transform
            // the normal from primitive space to worldspace.
            Matrix::applytranspose(collection->collectionspace[AFT],
                                   last_int->normal,
                                   edgenormptr[Z]);
            Vector::normalise(edgenormptr[Z], "surface_sampler: edgenormal[Z]");
          }
 
          // Store the inout for this cell.
          *pointptr = current_inout;

          // Set up for the next cell.
          pointptr -= SQR(res+1);
          edgeptr -= 3 * SQR(res+1);
          edgepointptr -= 3 * SQR(res+1);
          edgenormptr -= 3 * SQR(res+1);
          cell_z--;
          edge_this_cell = FALSE;

          // Shift the inout value for this cell so it becomes the high edge
          // of the next cell.
          current_inout = (current_inout & 0x0F) << (1 << Z);
        }
	
        if (cell_z >= 0) {
          // Find the next intersection.
          last_int = current_int;
          current_int = &dummy_end_int;
          for (i = 0; i < z_num_intersects[ray_index]; i++)
            if ((z_ray_tvals[ray_index * RAYINTBUFSIZE + i].t > last_int->t) &&
                (z_ray_tvals[ray_index * RAYINTBUFSIZE + i].t < current_int->t))
              current_int = &z_ray_tvals[ray_index * RAYINTBUFSIZE + i];

          // We have found the intersection at last_int, we need to flip
          // the current inout value.
          current_in ^= 0x01;

          // The cell currently being processed has an edge in the traversal
          // direction.
          edge_this_cell = TRUE;
        }
      }
      Vector::add(prim_x_offset, prim_ray_from, prim_ray_from);
      Vector::add(world_x_offset, world_ray_from, world_ray_from);
    }
    Vector::addfrac(-(res+1), prim_x_offset, prim_ray_from);
    Vector::addfrac(-(res+1), world_x_offset, world_ray_from);
    Vector::add(prim_y_offset, prim_ray_from, prim_ray_from);
    Vector::add(world_y_offset, world_ray_from, world_ray_from);
  }
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
