/*
 * collection.h: Definitions for class Collection.
 *
 * Contains code for the Collections, which contain a subset of the polygonization
 * of an object. Also SurfaceSampler, which samples an implicit surface.
 */

// pointdata holds the in/out status of the corners of the current cell.
typedef unsigned char pointdata;

enum calctype {CALC_NONE = 0, CALC_COLLX_SURFACE, CALC_WC_FROM_COLLX, CALC_SPLIT};

// The Collection class encapsulates the polygons that make up the surface of
// a primitive within some region. These polygons may be the actual surface,
// or a subtracted-away surface. They are created by a grid-based sampling scheme,
// but this grid is not necessarily aligned to the worldspace octree grid.
class Collection {
public:
  int collx_id, sample_id;

  int res;         // number of cells on a side
  string **cell;   // 3D array, CUBE(res). A list of strings of triangles or NULL for every cell.
  int total_tri;   // number of triangles referred to by this Collection

  // The primitive whose surface we have polygonized.
  primobjectnode *owner;

  // Our boundaries in the space that the primitive was originally subdivided 
  // into collections within (collectionspace).
  box bounds;

  // The size of our cells in collectionspace.
  vector cellsize;

  // The transformation from primitivespace to collectionspace.
  transform collectionspace;

  // Housekeeping stuff.
  int reference_count;
  Collection *parent_collection;
  Modeller *modeller;

  // Have we been calculated yet?
  calctype needs_calculation;
  
  // Add a reference to the Collection.
  void add_reference(void);

  // Remove a reference from the Collection. Returns TRUE if the Collection should
  // be deleted now.
  int remove_reference(void);

  // Sample the primitive that we contain.
  void sample_surface(void);

  // Perform all of the calculations necessary to create a WorldCollection.
  int calculate(void);

  // Create the Collection by polygonizing a primitive surface.
  Collection(primobjectnode *surfacenode, int res, transform collectionspace,
             box bounds, Modeller *modeller);

  // Create an empty Collection that will be polygonized as part of a split operation.
  Collection(Collection *parent_collection);

  // Delete the Collection.
  ~Collection(void);
};

// Size of the ray-intersection buffer for the SurfaceSampler's raycasting.
const int RAYINTBUFSIZE = 100;

// The SurfaceSampler samples an implicit surface in a grid fashion.
// Its information may then be used to construct the polygonization for a
// Collection.
class SurfaceSampler {
public:
  Collection *collection;

  // Variables used to construct the sampling.
  int res;
  box primbox;
  vector prim_x_offset, prim_y_offset, prim_z_offset;
  vector world_x_offset, world_y_offset, world_z_offset;
  int cell_x, cell_y, cell_z, ray_index, edge_this_cell;
  int castray_index;
  pointdata *pointinfo, *pointptr;
  double edgeval;
  unsigned char current_in, current_inout;
  unsigned char xcell_in, ycell_in;
  double cellbase_t;
  rayintersect *last_int, *current_int, dummy_start_int, dummy_end_int;
  vector prim_ray_from, prim_castray_from, prim_traverseray_from;
  vector world_ray_from, world_traverseray_from;
  Primitive *prim;
  int *x_num_intersects, *y_num_intersects, *z_num_intersects;
  rayintersect *x_ray_tvals, *y_ray_tvals, *z_ray_tvals;
  vector *x_world_rayfroms, *y_world_rayfroms, *z_world_rayfroms;
  int *is_edge;
  vector *edge_point, *edge_normal;
  int *edgeptr;
  vector *edgepointptr, *edgenormptr;

  // Create the SurfaceSampler.
  SurfaceSampler(Collection *collection);

  // Sample the surface.
  void sample(void);

  // Free the memory used in the sampling.
  ~SurfaceSampler();
};
