/*
 * copytracker.h: Definitions for class CopyTracker and supporting code.
 *
 * Contains code to create and maintain a CopyTracker, which tracks the
 * number of instances of a objectnode that are present in the Primtree
 * associated with its CSGgraph.
 */

/* Collection-safe */

// A Tracker class contains a reference to a particular copy of the
// objectnode.
class Tracker {
public:
  // The CopyTracker that this is attached to.
  CopyTracker *owner;

  // The value of the CopyTracker's timestamp variable when this Tracker was
  // added to it.
  int timestamp;

  // The primtreenode that is the copy.
  primtreenode *copy;

  // Pointers to the next and previous Trackers in the linked list.
  Tracker *prev, *next;

  // Construct a new Tracker node.
  Tracker(CopyTracker *owner, primtreenode *copy)
  {
    this->owner = owner;
    this->copy = copy;

    prev = next = NULL;
  }
};

// Each CopyTracker is associated with a particular objectnode of the
// CSGgraph. It tracks the number of primtreenodes that represent the
// objectnode.
class CopyTracker {
public:
  // The objectnode whose copies this tracks.
  objectnode *original;

  // A timestamp variable that is used to differentiate between 'new' and 'old'
  // copies.
  int timestamp;

  // Total number of copies that have been made of this object, and
  // a linked list that holds the Tracker classes for each copy
  // that is made of this objectnode.
  int num_copies;
  Tracker *trackerlist_head, *trackerlist_tail;

  // Add a new Tracker to the CopyTracker.
  Tracker *add_new_tracker(primtreenode *new_primnode);

  // Remove a Tracker from the CopyTracker.
  void remove_tracker(Tracker *killme);

  // Advance the timestamp variable.
  void advance_timestamp(void);

  // Compare the CopyTracker's timestamp to that of a Tracker. If this
  // returns TRUE, the Tracker has been added since the timestamp change; if
  // this returns FALSE, the Tracker is an old one, pre-timestamp-change.
  int has_current_timestamp(Tracker *tracker);

  // Called when a primtreenode is deleted, gets rid of its tracker.
  // Returns TRUE if its objectnode has no primtreenode equivalents left,
  // and needs to be deleted.
  int notify_primtreenode_deletion(Tracker *tracker);

  // Select or deselect the copies of an object.
  void select_copies(void);
  void deselect_copies(void);

  // Construct a new CopyTracker, completely empty.
  CopyTracker(objectnode *original) {
    this->original = original;
    num_copies = 0;
    trackerlist_head = trackerlist_tail = NULL;
    timestamp = 0;
  }

  // Delete the CopyTracker (and deallocate all its Trackers).
  ~CopyTracker(void)
  {
    Tracker *current, *last;

    for (current = trackerlist_head; current != NULL ;) {
      last = current;
      current = last->next;
      delete last;
    }
  }      
  
};
