/*
 * skitterlist.cc: Routines for class SkitterList and supporters.
 *
 * The SkitterList class contains a list of all the skitters that are currently
 * visible in the modeller; these are the coordinate systems for each of the
 * transformations on the selected object.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Create the Skitter.
Skitter::Skitter(int node_index, TransformListNode *trans, transform parent_position)
{
  this->node_index = node_index;

  Transform::set(this->parent_position, parent_position);

  transformation = trans;
  transformation->trans->get_color(normal_color, &inside_alpha,
				   ghosted_color, &ghosted_alpha);
  size = 1.0;
  update();

  dmd = NULL;
}

// If the transformation changes, the Skitter must too.
void Skitter::update(void)
{
  int ax;

  // Find the position matrix at the end of the transformation.
  Transform::combine(transformation->cumulative, parent_position, position);

  // Find the position matrix at the start of the transformation.
  if (transformation->prev == NULL)
    Transform::set(withouttrans_position, parent_position);
  else
    Transform::combine(transformation->prev->cumulative, parent_position,
		       withouttrans_position);

  // Where is the Skitter?
  Matrix::apply(position[FORE], zero_vector, location);

  // Next step: where does this position matrix map the unit vectors to?
  Vector::assign(axis_vec[X], size, 0.0, 0.0);
  Vector::assign(axis_vec[Y], 0.0, size, 0.0);
  Vector::assign(axis_vec[Z], 0.0, 0.0, size);

  for (ax = X; ax <= Z; ax++)
    Matrix::applynoshift(position[FORE], axis_vec[ax], axis_vec[ax]);
}

// Delete the Skitter.
Skitter::~Skitter(void)
{
  if (dmd)
    delete dmd;
}

// Change the object that is selected to have its skitters tracked
// by the SkitterList.
void SkitterList::purge(void)
{
  if (selected_node == NULL)
    return;

  // Delete all the SkitterList's Skitter tracking structures.
  for (int i = 0; i < num_copies; i++) {
    for (int j = 0; j < num_skitters; j++)
      delete copies[i].skitter[j];
    free(copies[i].skitter);
  }
  free(copies);

  num_copies = num_skitters = 0;
  selected_node = NULL;

  // Call to dmd_deselect() will force transview updating
  dmd_deselect();
}

// Add an objectnode to be tracked.
void SkitterList::new_selection(objectnode *node)
{
  Tracker *tptr;
  TransformListNode *tlptr;
  primtreenode *parent_node;
  transform parent_position;
  int i, j;

  if (selected_node != NULL)
    purge();

  selected_node = node;

  // Allocate the copies.
  assert(node != NULL);
  assert(node->copytracker != NULL);
  num_copies = node->copytracker->num_copies;
  copies = (CopySkitters *) calloc(num_copies, sizeof(CopySkitters));
  if (!copies)
    fatal_error("SkitterList::new_selection", "Can't allocate copies!",
		"Out of memory.");

  assert(node->transformlist != NULL);
  num_skitters = node->transformlist->num_transforms;

  // For each copy, set up the Skitters.
  for (i = 0, tptr = node->copytracker->trackerlist_head; tptr != NULL;
       i++, tptr = tptr->next) {
    copies[i].copy = tptr->copy;
    
    // Set up the Skitters for this copy.
    assert(copies[i].copy != NULL);
    parent_node = copies[i].copy->parent;
    if (parent_node == NULL)
      Transform::unit(parent_position);
    else
      Transform::set(parent_position, parent_node->position);

    // Allocate the skitter array.
    copies[i].skitter = (Skitter **) calloc(num_skitters, sizeof(Skitter *));
    if (!copies[i].skitter)
      fatal_error("SkitterList::new_selection", "Can't allocate skitters!",
		  "Out of memory.");

    for (j = 0, tlptr = node->transformlist->head; tlptr != NULL;
	 j++, tlptr = tlptr->next) {
      copies[i].skitter[j] = new Skitter(j, tlptr, parent_position);
    }

    // We must have created the appropriate number of Skitters.
    assert(j == num_skitters);
  }

  // We must have created Skitters for each copy... the number of copies must
  // match the number of copies in the CopyTracker.
  assert(i == num_copies);

  modeller->update_all_transviews();
}

// Update all Skitters; re-read their transformations' positions.
void SkitterList::update_all_skitters(void)
{
  int i, j;
  
  for (i = 0; i < num_copies; i++)
    for (j = 0; j < num_skitters; j++)
      copies[i].skitter[j]->update();
}

// Add or remove transformations from those that are tracked by the
// SkitterList. Should be called after any changes in the transformations
// in the TransformList.
void SkitterList::update_skitterlist(void)
{
  Skitter ***new_skitter;
  int i, new_num_skitters, new_index, old_index;
  TransformListNode *tlptr, *search_ptr, *temp_tlptr;
  transform this_parent_position;
  primtreenode *this_parent;
  Tracker *tptr;

  if (selected_node == NULL)
    return;

  // This routine does not handle changes in the copies that are made of the
  // selected object. The solution will have to be to delete the SkitterList
  // and then construct a new one. This must be handled elsewhere.
  //
  // Actually, the only way that copies could be created of an object would
  // be to link it into something, and that requires changing the selected
  // object, so the SkitterList will be purged anyway. Not a problem.
  assert(num_copies == selected_node->copytracker->num_copies);

  // We are going to have to create new Skitter lists for each copy.
  new_num_skitters = selected_node->transformlist->num_transforms;

  new_skitter = (Skitter ***) calloc(num_copies, sizeof(Skitter **));
  if (!new_skitter)
    fatal_error("SkitterList::update_skitterlist", "Can't allocate newskitter",
                "Out of memory.");
  for (i = 0; i < num_copies; i++) {
    new_skitter[i] = (Skitter **) calloc(new_num_skitters, sizeof(Skitter **));
    if ((new_num_skitters > 0) && (!new_skitter[i]))
      fatal_error("SkitterList::update_skitterlist",
                  "Can't allocate new_skitter[i]",
                  "Out of memory.");
  }

  // Now that we have the new skitter-tracking lists set up, it's time to
  // traverse the new TransformList, checking each transformation to see that
  // we have a Skitter set up appropriately.
  for (new_index = old_index = 0, tlptr = selected_node->transformlist->head;
       tlptr != NULL; ) {
    // Do we have the correct Skitters set up in our lists?
    if ((old_index < num_skitters) && (copies[0].skitter[old_index]->transformation == tlptr)) {
      // Yes, we have Skitters set up that point to this transformation.
      // Add them to the new set, and go to the next one.
      for (i = 0; i < num_copies; i++) {
        new_skitter[i][new_index] = copies[i].skitter[old_index];
        new_skitter[i][new_index]->node_index = new_index;
        new_skitter[i][new_index]->update();
      }
      new_index++;
      old_index++;
      tlptr = tlptr->next;
    } else {
      // Either the old transformation has been deleted, and the new
      // transformation is simply the next one from the list, or the new
      // transformation is a new one that has been added, and the old
      // transformation may be found after it in the list.
      // 
      // We search for the transformation that the Skitters _are_ tracking.
      // If we find it, then this is a new transformation that's been added,
      // and we need to create new Skitters for it between here and the
      // transformation that the old Skitters are tracking. If we don't find
      // it, then the transformation that the Skitters are tracking has been
      // deleted, and we should move to the next set of old Skitters.
      //
      // Note that if a transformation is moved to be later in the list,
      // then the 'new' transformations that are found between the old
      // position and the new position are treated as newly created. And then
      // the 'old' skitters for these transformations are deleted, once the
      // new position of the moved transformation is found.
      //
      // If a transformation is moved to be earlier in the list, then it is
      // treated as a new creation when first encountered, all the leapfrogged
      // transformations are used unchanged, and the old transformation is
      // deleted when encountered. See, it works... so why don't you believe
      // this, Chris?
      if (old_index >= num_skitters) {
        // Well, there are no old Skitters that could be tracking this!
        // We need to create new Skitters anyway, as if we could find the
        // transformation.
        temp_tlptr = tlptr;    // any non-NULL value would do
      } else {
        // Do the search.
        search_ptr = copies[0].skitter[old_index]->transformation;
        for (temp_tlptr = tlptr; temp_tlptr != NULL;
             temp_tlptr = temp_tlptr->next)
          if (temp_tlptr == search_ptr)
            break;
      }
      
      if (temp_tlptr == NULL) {
        // Okay, we couldn't find the transformation that the old Skitters
        // are tracking. This means that they need to be deleted.
        // We stay at the same place in the new list.
        for (i = 0; i < num_copies; i++)
          delete copies[i].skitter[old_index];
        old_index++;
      } else {
        // We found the transformation that the old Skitters are tracking,
        // or there are no more old Skitters. In either case,
        // we don't delete them, we just add new Skitters to handle this
        // transformation.
        for (tptr = selected_node->copytracker->trackerlist_head, i = 0;
             tptr != NULL; i++, tptr = tptr->next) {
          // We need to find the parent_position for this copy. If there were
          // any old Skitters, then we can use their parent_positions.
          // Otherwise we need to find them from the primtree.
          if (num_skitters > 0) {
            Transform::set(this_parent_position,
                           copies[i].skitter[0]->parent_position);
          } else {
            this_parent = tptr->copy->parent;
            if (this_parent == NULL)
              Transform::unit(this_parent_position);
            else
              Transform::set(this_parent_position, this_parent->position);
          }

          // Create the Skitter in question.
          new_skitter[i][new_index] = new Skitter(new_index, tlptr, this_parent_position);      
        }
          
        // Sanity check: we must have created num_copies new Skitters.
        assert(i == num_copies);
        
        // Go to the next Skitter to be created.
        new_index++;
        tlptr = tlptr->next;
      }
    }
  }

  // If we didn't finish traversing the old set of skitters yet, go through
  // them and delete them all.
  for ( ; old_index < num_skitters; old_index++)
    for (i = 0; i < num_copies; i++)
      delete copies[i].skitter[old_index];
  
  // We have now created a new set of lists that track the new Skitters in
  // the scene. Update our member variables.
  for (i = 0; i < num_copies; i++) {
    // The array may need resizing.
    if (num_skitters != new_num_skitters) {
      copies[i].skitter = (Skitter **) realloc((void *) copies[i].skitter,
					       new_num_skitters *
					       sizeof(Skitter *));
      if ((new_num_skitters > 0) && (copies[i].skitter == NULL))
        fatal_error("SkitterList::update_skitterlist",
                    "Can't realloc skitter array!",
                    "Out of memory.");
    }
    
    if (new_num_skitters > 0) {
      // Copy the data.
      memcpy((void *) copies[i].skitter, (void *) new_skitter[i],
	     new_num_skitters * sizeof(Skitter *));

      // We don't need the intermediate variables any more.
      free(new_skitter[i]);
    }
  }
  free(new_skitter);

  num_skitters = new_num_skitters;

  // The SkitterList has changed, so we don't have the DMD selected any more.
  // Can't call dmd_deselect() because that updates the transviews, and they
  // will do that themselves once the move / create / delete operation is
  // complete.
  current_dmd = NULL;
  currently_manipulating = FALSE;
}

// Iterate over the Skitters in the list.
void SkitterList::iterate(void (*iterator)(Skitter *))
{
  int i, j;

  for (i = 0; i < num_copies; i++)
    for (j = 0; j < num_skitters; j++)
      iterator(copies[i].skitter[j]);
}

// Create an empty SkitterList.
SkitterList::SkitterList(Modeller *modeller)
{
  this->modeller = modeller;
  selected_node = NULL;
  
  num_copies = num_skitters = 0;
  copies = NULL;

  current_dmd = NULL;
  currently_manipulating = FALSE;
}

// Check to see if a selection click might be on a skitter.
int SkitterList::skitter_select(vector ray_from, vector ray_dir)
{
  int i, j;
  double dist, closest_dist = +INFINITY;
  Skitter *closest = NULL;
  
  // Query each skitter to see if it might be being selected.
  for (i = 0; i < num_copies; i++) {
    for (j = 0; j < num_skitters; j++) {
      if (copies[i].skitter[j] != current_dmd) {
        dist = copies[i].skitter[j]->select_click(ray_from, ray_dir);
        if (dist < closest_dist) {
          closest = copies[i].skitter[j];
          closest_dist = dist;
        }
      }
    }
  }

  // If we got close to a Skitter, select it.
  if (closest_dist < 0.01) {
    assert(closest != NULL);
    current_dmd = closest;
    current_dmd->do_select();

    modeller->update_all_transviews();

    return TRUE;
  } else {
    return FALSE;
  }
}

// Check to see if a selection click might be a manipulation of the DMD.
int SkitterList::dmd_click(vector ray_from, vector ray_dir)
{
  int returnval;
  vector local_from, local_dir;

  assert(!currently_manipulating);
  assert(current_dmd != NULL);
  assert(current_dmd->dmd != NULL);

  // Transform the ray into the DMD's coordinate space. NB: we use the
  // Skitter's withouttrans_position as DMDs have their coordinate space
  // before application of the Skitter's transformation.
  Matrix::apply(current_dmd->withouttrans_position[AFT],
		ray_from, local_from);
  Matrix::applynoshift(current_dmd->withouttrans_position[AFT],
		       ray_dir, local_dir);

  returnval = current_dmd->dmd->check_manipulate(local_from, local_dir);

  if (current_dmd->dmd->is_being_manipulated)
    currently_manipulating = TRUE;

  return returnval;
}

// Manipulate the DMD.
void SkitterList::dmd_manipulate(vector ray_from, vector ray_dir)
{
  vector local_from, local_dir;

  assert(currently_manipulating);
  assert(current_dmd != NULL);
  assert(current_dmd->dmd != NULL);

  // Transform the ray into the DMD's coordinate space. NB: we use the
  // Skitter's withouttrans_position as DMDs have their coordinate space
  // before application of the Skitter's transformation.
  Matrix::apply(current_dmd->withouttrans_position[AFT],
                ray_from, local_from);
  Matrix::applynoshift(current_dmd->withouttrans_position[AFT],
                       ray_dir, local_dir);

  current_dmd->dmd->manipulate(local_from, local_dir);

  // Update the changes in the TransformList.
  selected_node->transformlist->update_transform(current_dmd->transformation);
  selected_node->transformlist->propagate_matrices();
  update_all_skitters();
  modeller->update_all_transviews();
  modeller->csg_graph->update_transformed_object(selected_node);
  modeller->update_all_views();
}

// Release the DMD.
void SkitterList::dmd_release(void)
{
  assert(currently_manipulating);
  assert(current_dmd != NULL);
  assert(current_dmd->dmd != NULL);

  current_dmd->dmd->release();
  currently_manipulating = FALSE;
}

// DMD selection.
void SkitterList::dmd_select(Skitter *skitter)
{
  skitter->do_select();

  current_dmd = skitter;
  currently_manipulating = FALSE;

  modeller->update_all_transviews();
}

void SkitterList::dmd_deselect(void)
{
  current_dmd = NULL;
  currently_manipulating = FALSE;

  modeller->update_all_transviews();
}

// Delete the SkitterList.
SkitterList::~SkitterList(void)
{
  // Delete _everything_.
  purge();
}

// Could this click select the Skitter?
double Skitter::select_click(vector ray_from, vector ray_dir)
{
  vector local_from, local_dir;
  double u_v, u_u, v_v, t, ax_dist, dist;
  double closest_dist = +INFINITY;

  // First, transform the ray into the Skitter's local space.
  Matrix::apply(position[AFT], ray_from, local_from);
  Matrix::applynoshift(position[AFT], ray_dir, local_dir);

  // A little vector algebra will reveal that the closest a ray comes to the
  // origin is at t = - (u . v) / (v . v), and the distance at this point
  // is d = (u . u) + t (u . v).

  // First, check against the X axis. Perform calculations in the YZ plane.
  u_u = SQR(local_from[Y]) + SQR(local_from[Z]);
  v_v = SQR(local_dir[Y]) + SQR(local_dir[Z]);
  u_v = local_from[Y]*local_dir[Y] + local_from[Z]*local_dir[Z];
  
  if (fabs(v_v) > TINY) {
    t = - u_v / v_v;
    
    // Find out how far along the axis this closest-approach is.
    ax_dist = local_from[X] + t * local_dir[X];
    if ((ax_dist > 0) && (ax_dist < size)) {
      // How far from the axis is the approach?
      dist = fabs(u_u + t * u_v);
      if (dist < closest_dist)
        closest_dist = dist;
    }
  }
  
  // First, check against the Y axis. Perform calculations in the XZ plane.
  u_u = SQR(local_from[X]) + SQR(local_from[Z]);
  v_v = SQR(local_dir[X]) + SQR(local_dir[Z]);
  u_v = local_from[X]*local_dir[X] + local_from[Z]*local_dir[Z];
  
  if (fabs(v_v) > TINY) {
    t = - u_v / v_v;
    
    // Find out how far along the axis this closest-approach is.
    ax_dist = local_from[Y] + t * local_dir[Y];
    if ((ax_dist > 0) && (ax_dist < size)) {
      // How far from the axis is the approach?
      dist = fabs(u_u + t * u_v);
      if (dist < closest_dist)
        closest_dist = dist;
    }
  }
  
  // First, check against the Z axis. Perform calculations in the XY plane.
  u_u = SQR(local_from[X]) + SQR(local_from[Y]);
  v_v = SQR(local_dir[X]) + SQR(local_dir[Y]);
  u_v = local_from[X]*local_dir[X] + local_from[Y]*local_dir[Y];
  
  if (fabs(v_v) > TINY) {
    t = - u_v / v_v;
    
    // Find out how far along the axis this closest-approach is.
    ax_dist = local_from[Z] + t * local_dir[Z];
    if ((ax_dist > 0) && (ax_dist < size)) {
      // How far from the axis is the approach?
      dist = fabs(u_u + t * u_v);
      if (dist < closest_dist)
        closest_dist = dist;
    }
  }

  // Return the closest that the ray came to any of the axes.
  return closest_dist;
}

// Select the Skitter.
void Skitter::do_select(void)
{
  if (dmd == NULL) {
    // We need to create a DMD for this Skitter, from its Transformation.
    dmd = transformation->trans->create_dmd();
  } else {
    // We already have a DMD, it probably doesn't need updating.
  }
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
