/*
 * skitterlist.h: Definitions for class SkitterList.
 *
 * The SkitterList class contains a list of all the skitters that are currently
 * visible in the modeller; these are the coordinate systems for each of the
 * transformations on the selected object.
 */

/* Collection-safe */

class Skitter {
public:
  // Which parent CSGgraph node is this skitter associated with?
  transform parent_position;

  // The parameters of the skitter.
  transform withouttrans_position;
  transform position;
  vector location;
  vector axis_vec[3];
  color normal_color, ghosted_color;
  double inside_alpha, ghosted_alpha;
  double size;

  // The DMD that's attached to the Skitter. This may be NULL if the
  // skitter hasn't been selected yet.
  DMD *dmd;

  // The transformation that generates these parameters.
  int node_index;
  TransformListNode *transformation;

  // Create the Skitter.
  Skitter(int node_index, TransformListNode *trans, transform parent_position);

  // Could this click select the Skitter? How close does the ray pass?
  double select_click(vector ray_from, vector ray_dir);

  // Select the Skitter.
  void do_select(void);

  // If the transformation changes, the Skitter must too.
  void update(void);

  // Delete the Skitter.
  ~Skitter(void);
};

// A structure that tracks the skitters associated with one copy of the 
// object.
typedef struct CopySkitters_ {
  primtreenode *copy;
  Skitter **skitter;
} CopySkitters;

class SkitterList {
public:
  int num_copies;
  int num_skitters;
  CopySkitters *copies;

  Modeller *modeller;

  objectnode *selected_node;

  // Skitter that is currently selected for a DMD.
  Skitter *current_dmd;

  // Are we currently manipulating the DMD in a mouse-drag operation?
  int currently_manipulating;

  // Change the object that is selected to have its skitters tracked
  // by the SkitterList.
  void purge(void);
  void new_selection(objectnode *node);

  // Update all Skitters; re-read their transformations' positions.
  void update_all_skitters(void);

  // Add or remove transformations from those that are tracked by the
  // SkitterList. Should be called after any changes in the transformations
  // in the TransformList.
  void update_skitterlist(void);

  // Iterate over the Skitters in the list.
  void iterate(void (*iterator)(Skitter *));

  // Check to see if a selection click might be on a skitter.
  int skitter_select(vector ray_from, vector ray_to);

  // DMD selection.
  void dmd_select(Skitter *skitter);
  void dmd_deselect(void);

  // Manipulation of the DMD.
  int dmd_click(vector ray_from, vector ray_to);
  void dmd_manipulate(vector ray_from, vector ray_to);
  void dmd_release(void);

  // Create an empty SkitterList.
  SkitterList(Modeller *modeller);

  // Delete the SkitterList.
  ~SkitterList(void);
};
