/*
 * splittracker.cpp: Routines for class SplitTracker.
 *
 * Contains code for the SplitTracker, which keeps track of which WorldCollections in
 * an octree need to be split.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

splittrackernode *debug_splittrackernode = (splittrackernode *) 0xDEADBEEF;
int debug_splittracker_breakcollx = -1;
int debug_splittracker_breaksample = -1;
extern int debug_incidentstatus;

// Find the splittrackernode for a particular WorldCollection.
splittrackernode *SplitTracker::find_node(int sample_id, primtreeprimnode *primptr)
{
  splittrackernode *splitptr;
  
  // Find whether we are already tracking this collection, or another
  // collection with the same sample_id (base parent collection).
  for (splitptr = head; splitptr != NULL; splitptr = splitptr->next)
    if ((splitptr->sample_id == sample_id) && (splitptr->collection->owner == primptr))
      break;

  return splitptr;
}

// Create a new (uninitialised) node in the list.
splittrackernode *SplitTracker::create_node(void)
{
  splittrackernode *splitptr;

  num_nodes++;

  // We are not tracking this collection. We must create
  // a new (empty) splittrackernode for it.
  splitptr = (splittrackernode *) calloc(1, sizeof(splittrackernode));
  if (!splitptr)
    fatal_error("SplitTracker::create_node", "Can't allocate new splittrackernode",
                "Out of memory.");

  if (splitptr == debug_splittrackernode)
    _CrtDbgBreak();

  // Add this node into the list.
  splitptr->next = head;
  if (splitptr->next == NULL)
    tail = splitptr;
  else
    splitptr->next->prev = splitptr;

  splitptr->prev = NULL;
  head = splitptr;

  return splitptr;
}

// Remove a particular node from the collectionlist.
void SplitTracker::remove_node(splittrackernode *kill_me)
{
  assert(num_nodes > 0);
  num_nodes--;

  if (kill_me == debug_splittrackernode)
    _CrtDbgBreak();

  if (kill_me->prev == NULL)
    head = kill_me->next;
  else
    kill_me->prev->next = kill_me->next;

  if (kill_me->next == NULL)
    tail = kill_me->prev;
  else
    kill_me->next->prev = kill_me->prev;

  delete_node(kill_me);
}

// Delete all nodes belonging to a particular primnode.
// This is called because we detect that a primnode has been
// moved, so all its collections will be present in different
// terminals anyway.
void SplitTracker::delete_nodes(primtreeprimnode *primptr)
{
  splittrackernode *splitptr, *nextptr;

  for (splitptr = head; splitptr != NULL; splitptr = nextptr) {
    nextptr = splitptr->next;

    if (splitptr->collection->owner == primptr)
      delete_node(splitptr);
  }
}

// Deallocate a splittrackernode.
void SplitTracker::delete_node(splittrackernode *kill_me)
{
  int i, j;
  solidleafnode *leaf;

  if (kill_me->collection->remove_reference())
    delete kill_me->collection;
  
  if (kill_me->child)
    if (kill_me->child->remove_reference())
      delete kill_me->child;

  if (kill_me->num_deleted_terminals < kill_me->num_terminals) {
    for (i = 0; i < kill_me->num_terminals; i++) {
      if (!kill_me->terminalbuf[i].is_deleted) {
        leaf = kill_me->terminalbuf[i].our_leafnode;
        if (leaf != NULL) {
          for (j = 0; j < leaf->num_collections; j++) {
            if (leaf->collectionbuf[j].splitptr == kill_me)
              leaf->collectionbuf[j].splitptr = NULL;
          }
        }
      }
    }
  } else {
    assert(kill_me->num_deleted_terminals == kill_me->num_terminals);
  }

  if (kill_me->terminalbuf != NULL)
    free(kill_me->terminalbuf);

  free(kill_me);
}

// Register that a WorldCollection overlaps a solidterminal which has other
// WorldCollections in it.
void SplitTracker::register_terminal(WorldCollection *collection,
                                     int is_flipped,
                                     SolidOctree *our_octree,
                                     int terminal_id,
                                     solidterminal *our_terminal,
                                     solidleafnode *our_leafnode,
                                     SolidOctree *opponent_octree, int is_left,
                                     solidterminal *opposing_surface)
{
  int i, free_space;
  splittrackernode *splitnode = find_node(collection->collx->sample_id, collection->owner);

  if (splitnode == debug_splittrackernode)
    _CrtDbgBreak();

  if (splitnode == NULL) {
    // This sample ID is a new one to us. Create a new node for it.
    splitnode = create_node();

    splitnode->newly_overwritten = FALSE;
    splitnode->sample_id = collection->collx->sample_id;
    splitnode->collection = collection;
    collection->add_reference();

    splitnode->is_flipped = is_flipped;
    splitnode->is_left = is_left;
    splitnode->keep_inside_surface = (is_left) ? owner->left_in_is_active
                                               : owner->right_in_is_active;
    splitnode->num_terminals = splitnode->num_deleted_terminals = 0;
    splitnode->terminalbufsize = 0;
    splitnode->terminalbuf = NULL;

    splitnode->child = NULL;
  } else {
    // Is this a new collection based off the same sample ID?
    if (splitnode->collection != collection) {
      // This should only happen once per split op... 
      if (splitnode->newly_overwritten)
        TRACE("Warning... splitnode sample_id %d w/ collx %d being overwritten by collx %d... no children!\n",
              splitnode->sample_id, splitnode->collection->collx->collx_id, collection->collx->collx_id);
      assert(splitnode->is_left == is_left);

      // Replace it.
      if (splitnode->collection->remove_reference())
        delete splitnode->collection;

      splitnode->newly_overwritten = TRUE;
      splitnode->collection = collection;
      splitnode->collection->add_reference();

      splitnode->is_flipped = is_flipped;

      if (splitnode->child) {
        if (splitnode->child->remove_reference())
          delete splitnode->child;
        splitnode->child = NULL;
      }
    }
  }

  // Check to see if this terminal has already been registered.
  free_space = -1;
  for (i = 0; i < splitnode->num_terminals; i++) {
    if (splitnode->terminalbuf[i].is_deleted) {
      // This is a deleted space.
      free_space = i;
    } else {
      // Check to see if this is our terminal, or at least at the same position
      // in the octree structure.
      if ((our_terminal != NULL) &&    (splitnode->terminalbuf[i].our_terminal != NULL) &&
          (our_terminal->terminal_id == splitnode->terminalbuf[i].our_terminal->terminal_id))
        break;
    }
  }
    
  if ((i == splitnode->num_terminals) && (free_space != -1)) {
    // Overwrite a deleted terminal instead of growing the buffer.
    i = free_space;
    splitnode->num_deleted_terminals--;
  }

  // i is now the index of the terminalbuffer to write into. We may be
  // overwriting a previous terminal entry, or we may be adding onto the
  // end and need the buffer to be expanded.
  if (splitnode->terminalbufsize <= i) {
    splitnode->terminalbufsize += terminal_bufsize_increment;
    splitnode->terminalbuf = (terminaldata *) realloc((void *) splitnode->terminalbuf,
      splitnode->terminalbufsize
      * sizeof(terminaldata));
    if (!splitnode->terminalbuf)
      fatal_error("SplitTracker::register_terminal", "Can't expand terminalbuf",
      "Out of memory");
  }
    
  // Register this terminal.    
  splitnode->terminalbuf[i].is_deleted = FALSE;
  splitnode->terminalbuf[i].update_id = this_update_id;
  splitnode->terminalbuf[i].terminal_id = terminal_id;
  splitnode->terminalbuf[i].our_terminal = our_terminal;
  splitnode->terminalbuf[i].our_leafnode = our_leafnode;
  splitnode->terminalbuf[i].our_octree   = our_octree;
  splitnode->terminalbuf[i].opponent_octree = opponent_octree;    
  splitnode->terminalbuf[i].opposing_surface = opposing_surface;

  if (i >= splitnode->num_terminals)
    splitnode->num_terminals++;
}

// Create a new SplitTracker.
SplitTracker::SplitTracker(SolidOctree *owner)
{
  this->owner = owner;

  num_nodes = 0;
  head = tail = NULL;

  this_update_id = 0;

  visitedsize = 1 << (3 * owner->modeller->octree_level);
  visited_this_update = (int *) calloc(visitedsize, sizeof(int));
}
  
// Delete the SplitTracker.
SplitTracker::~SplitTracker(void)
{
  splittrackernode *splitptr, *nextcoll;

  // Deallocate our linked-list chains.
  for (splitptr = head; splitptr != NULL; splitptr = nextcoll) {
    nextcoll = splitptr->next;
    delete_node(splitptr);
  }

  free(visited_this_update);
}

// Prepare for an update pass.
void SplitTracker::start_new_update(void)
{
  this_update_id++;

  memset((void *) visited_this_update, 0, visitedsize * sizeof(int));
}

// Register that we visited a section of the octree this update.
void SplitTracker::register_visit(int terminal_id, int num_levels)
{
  for (int i = terminal_id << (3 * num_levels);
       i < (terminal_id + 1) << (3 * num_levels); i++)
    visited_this_update[i] = TRUE;
}

// Perform all the splits.
void SplitTracker::go_split(void)
{
  splittrackernode *splitptr, *nextptr;
  int i, num_splits, terminal_splits, splitbufsize, num_keeps;
  splitterminal *splitbuf;

  splitbufsize = 0;
  splitbuf = NULL;

  for (splitptr = head; splitptr != NULL; splitptr = nextptr) {
    nextptr = splitptr->next;

    if ((splitptr == debug_splittrackernode) ||
        ((debug_incidentstatus == 0) && ((splitptr->collection->collx->collx_id == debug_splittracker_breakcollx) ||
                                         (splitptr->collection->collx->sample_id == debug_splittracker_breaksample))))
      _CrtDbgBreak();

    // Clear the newly_overwritten flag at this, the end of the split pass.
    splitptr->newly_overwritten = FALSE;

    num_keeps = terminal_splits = num_splits = 0;
  
    for (i = 0; i < splitptr->num_terminals; i++) {
      if (splitptr->terminalbuf[i].is_deleted)
        continue;

      if ((splitptr->terminalbuf[i].update_id < this_update_id) &&
          (visited_this_update[splitptr->terminalbuf[i].terminal_id])) {
        // This entry is from an old update, and we visited this terminal this
        // update and didn't find anything worth adding. So we need to delete
        // this entry.
        splitptr->terminalbuf[i].is_deleted = TRUE;
        splitptr->num_deleted_terminals++;
        continue;
      }

      // Resize our split buffer if necessary.
      if (num_splits >= splitbufsize) {
        splitbufsize += terminal_bufsize_increment;
        splitbuf = (splitterminal *) realloc((void *) splitbuf,
                                                splitbufsize * sizeof(splitterminal));
        if (!splitbuf)
          fatal_error("SplitTracker::go_split", "Can't grow splitbuf",
                      "Out of memory");
      }
          
      if ((splitbuf[num_splits].terminal =
           splitptr->terminalbuf[i].opposing_surface) == NULL) {
        // This terminal was up against either EMPTY or FULL and
        // doesn't have a surface here. Instead this terminal's worth of
        // our surface is just kept.
        num_keeps++;
      } else {
        // This terminal is up against another terminal and will
        // need to be split against it. 
        terminal_splits++;
      }

      splitbuf[num_splits].terminal_id = splitptr->terminalbuf[i].terminal_id;
      splitbuf[num_splits].cellid_overlap = -1;
      splitbuf[num_splits].our_opposing_leafnode = splitptr->terminalbuf[i].our_leafnode;
      num_splits++;
    }

    if (splitptr->num_terminals <= splitptr->num_deleted_terminals) {
      // There are no terminals attached to this node; they have all been
      // deleted. We don't need this node any more.
      remove_node(splitptr);
      continue;
    }

    // Discard the previous child, if there was one.
    if (splitptr->child) {
      if (splitptr->child->remove_reference())
        delete splitptr->child;
      splitptr->child = NULL;
    }
    
    if (terminal_splits > 0) {
      // Create a child collection, split as some fragment of this WorldCollection.
      splitptr->child = new WorldCollection(num_splits, splitbuf,
                                            splitptr->keep_inside_surface, splitptr->collection);
    } else if (num_keeps > 0) {
      // There were terminals indicating that we should insert this WorldCollection
      // unchanged.
      splitptr->child = splitptr->collection;
    } else {
      // There were only terminals indicating that we should discard this Collection.
      splitptr->child = NULL;
    }
    
    if (splitptr->child) {
      // Part or all of this WorldCollection is being kept as active. Add a reference
      // to it (for this node) and then add it to all of the terminals which we
      // were present for.
      splitptr->child->add_reference();
    }

    // We now must traverse all of our terminals where there is information stored
    // about this Collection, and update it accordingly.
    //
    // NB: we don't need to set the changed flag on these terminals because
    // any changes have to have occurred in the part of the collection that
    // doesn't overlap that terminal, otherwise they would have been changed
    // themselves!
    //
    // Also NB: we don't add to any terminals that are DISCARD, only those that are
    // surfaces or KEEP.
    for (i = 0; i < splitptr->num_terminals; i++) {
      if ((!splitptr->terminalbuf[i].is_deleted) &&
          (splitptr->terminalbuf[i].our_leafnode != NULL)) {
        assert(splitptr->terminalbuf[i].our_octree == owner);

        if (splitptr->child != NULL) {
          // Add a reference to our child.
          //
          // Note that we may have to overwrite the last pass's active child
          // (prev_activechild) in some leafnodes (i.e. those that didn't get
          // traversed this pass around, due to not being changed).
          owner->add_collection(splitptr->child,
                                splitptr->terminalbuf[i].our_leafnode,
                                splitptr);
        } else {
          // We have no active child. This means that not only must we remove any
          // collections that share our sample_id from our terminals, but we must also
          // remove our references to these terminals (since we won't be notified of
          // their deletion).
          owner->remove_collections(splitptr->sample_id,
                                    splitptr->terminalbuf[i].our_leafnode,
                                    splitptr);
          splitptr->terminalbuf[i].is_deleted = TRUE;
          splitptr->num_deleted_terminals++;
        }
      }
    }
  }

  if (splitbuf != NULL)
    free(splitbuf);
}

// Print out the SplitTracker.
void SplitTracker::print(void)
{
  splittrackernode *splitptr;
  int i;

  TRACE("*** SplitTracker print!!\n");
  TRACE("  belongs to 0x%08X, octree #%d\n", (int) this->owner->represented_object, this->owner->represented_object->index);
  TRACE("  contains %d splitnodes\n", num_nodes);

  for (splitptr = head; splitptr != NULL; splitptr = splitptr->next) {
    if (splitptr == debug_splittrackernode)
      _CrtDbgBreak();
    
    // Check pointers.
    if (splitptr->prev)
      assert(splitptr->prev->next == splitptr);
    else
      assert(head == splitptr);
    if (splitptr->next)
      assert(splitptr->next->prev == splitptr);
    else
      assert(tail == splitptr);

    TRACE("  + splittrackernode 0x%08X, collx %d [s/i %d] (0x%08X), prim %d\n", (int) splitptr, splitptr->collection->collx->collx_id,
          splitptr->collection->collx->sample_id, (int) splitptr->collection, splitptr->collection->owner->index);
    assert(splitptr->collection->collx->sample_id == splitptr->sample_id);
    TRACE("    [%s%s%s] child 0x%08X, %d terminals (%d del) in bufsize %d\n",
      splitptr->is_flipped ? "F|" : "", splitptr->is_left ? "L|" : "R|", splitptr->keep_inside_surface ? "K/IN" : "D/IN",
      (int) splitptr->child, splitptr->num_terminals, splitptr->num_deleted_terminals, splitptr->terminalbufsize);
    
    for (i = 0; i < splitptr->num_terminals; i++) {
      TRACE("    - tbuf[%d] ", i);
      if (splitptr->terminalbuf[i].is_deleted)
        TRACE("deleted\n");
      else {
        if (splitptr->terminalbuf[i].our_leafnode != NULL) {
          TRACE("octree 0x%08X, terminal 0x%08X, leafnode 0x%08X",
                (int) splitptr->terminalbuf[i].our_octree, (int) splitptr->terminalbuf[i].our_terminal,
                (int) splitptr->terminalbuf[i].our_leafnode);
        } else {
          TRACE("NULL terminal/leafnode/octree");
        }
        if (splitptr->terminalbuf[i].opposing_surface != NULL) {
          TRACE(", opposing surface terminal 0x%08X", (int) splitptr->terminalbuf[i].opposing_surface);
        } else {
          TRACE(", opposing KEEP terminal");
        }
        TRACE(" [octree 0x%08X]\n", (int) splitptr->terminalbuf[i].opponent_octree);
      }        
    }
  }
  TRACE("*** end!!\n");
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
