/*
 * splittracker.h: Definitions for class SplitTracker.
 *
 * Contains code for the SplitTracker, which keeps track of which WorldCollections in
 * an octree need to be split.
 */

// A terminaldata structure contains the information stored in a splittrackernode
// for a particular solidterminal.
typedef struct terminaldata_ {
  int is_deleted;
  int update_id, terminal_id;
  SolidOctree *our_octree, *opponent_octree;
  solidterminal *our_terminal;
  solidleafnode *our_leafnode;

  // What we're opposed to in this terminal. May be either a surface or a
  // keep terminal (opposing_surface == NULL).
  solidterminal *opposing_surface;
} terminaldata;

// A splittrackernode contains information about a particular WorldCollection
// that is part of our octree. More specifically it is attached to a
// particular sampled surface, so its sample_id is the static part.
typedef struct splittrackernode_ {
  int sample_id, newly_overwritten;
  WorldCollection *collection;
  int is_flipped, is_left, keep_inside_surface;

  int num_terminals, num_deleted_terminals, terminalbufsize;
  terminaldata *terminalbuf;

  WorldCollection *child;
  
  struct splittrackernode_ *next, *prev;
} splittrackernode;


const int terminal_bufsize_increment = 10;

typedef struct splitterminal_ {
  solidterminal *terminal; // May be NULL to indicate not a terminal, but KEEP instead
  solidleafnode *our_opposing_leafnode;
  int cellid_overlap, terminal_id;
} splitterminal;

// A SplitTracker keeps track of the WorldCollections that are present in an
// octree, and which terminals they are present in.
class SplitTracker {
public:
  SolidOctree *owner;

  // The current update_id that we are doing.
  int this_update_id;

  // Whether or not a particular terminal has been visited this update.
  int visitedsize;
  int *visited_this_update;

  int num_nodes;
  splittrackernode *head, *tail;

  // Create a new splittrackernode.
  splittrackernode *create_node(void);

  // Find the splittrackernode for a particular WorldCollection. If one does not
  // exist, returns NULL.
  splittrackernode *find_node(int sample_id, primtreeprimnode *primptr);

  // Remove a particular node from the collectionlist.
  void remove_node(splittrackernode *kill_me);

  // Deallocate a splittrackernode.
  void delete_node(splittrackernode *kill_me);

  // Delete all nodes that belong to a particular primnode.
  // This is called because we detect that a primnode has been
  // moved, so all its collections will be present in different
  // terminals anyway.
  void delete_nodes(primtreeprimnode *primptr);

  // Prepare for an update pass.
  void start_new_update(void);

  // Register a Collection as being present in a terminal node.
  void register_terminal(WorldCollection *collection, int is_flipped, SolidOctree *our_octree,
                         int terminal_id, solidterminal *our_terminal, solidleafnode *our_leafnode,
                         SolidOctree *opponent_octree, int is_left, solidterminal *opposing_surface);

  // Register that we visited a section of the octree this update.
  void register_visit(int terminal_id, int num_levels);

  // Perform all the splits.
  void go_split(void);

  // Print out the SplitTracker.
  void print(void);

  // Create a new SplitTracker.
  SplitTracker(SolidOctree *owner);

  // Delete the SplitTracker.
  ~SplitTracker(void);
};
