/*
 * stringmem.cpp: Code for class StringMemory.
 *
 * Contains code to construct a StringMemory which is a variant of the Mirage-style
 * blocklist that's been specialised to handle celllists of strings.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Constructor: specified blocksize and maxblocks.
StringMemory::StringMemory(Modeller *owner, int blocksize, int max_memblocks)
{
  this->owner = owner;
  this->blocksize = blocksize;
  this->max_memblocks = max_memblocks;

  // Allocate the storage pointers.
  storage = (string **) calloc(max_memblocks, sizeof(string *));
  assert(storage != NULL);
  for (int i = 0; i < max_memblocks; i++)
    storage[i] = NULL;

  // Set up memory management values.
  blocks_used = 0;
  next_free = NULL;
}

// Destructor; deallocate all memory and exit.
StringMemory::~StringMemory(void)
{
  assert(storage != NULL);

  // Deallocate each block.
  for (int i = 0; i < blocks_used; i++) {
    assert(storage[i] != NULL);
    free(storage[i]);
  }

  // Deallocate the storage memory.
  free(storage);
}

// Allocate a new triagle from the storage and return a pointer to it.
string *StringMemory::allocate(void)
{
  string *the_block, *the_string;
  int block_no, index_no;
  
  // Check to see if we need to allocate a new block.
  if (next_free == NULL) {
    
    if (blocks_used >= max_memblocks)
      fatal_error("StringMemory::allocate", "maximum memory size exceeded",
		    "StringMemory has a fixed limit on the number of\n"
        "blocks that can be allocated. Increase this, or reduce\n"
        "memory usage.");
    
    the_block = (string *) calloc(blocksize, sizeof(string));
    
    if (the_block == NULL)
      fatal_error("StringMemory::allocate", "can't allocate new block!",
		    "Tried to dynamically expand memory and failed. This\n"
        "could indicate that you are running severely low on\n"
        "memory or that there is a memory leak.");
    
    // Update the array of blocks.
    block_no = blocks_used++;
    storage[block_no] = the_block;
    
    // Now for each unused string in the block (which is all of them)
    // we set it to point to the next free string. We start from the
    // zero string, each one pointing to the next. The last one in the block
    // points to NULL.
    //
    // This pointer is stored in stringptr as there is no data in the string
    // right now anyway.
    for (index_no = 1; index_no < blocksize - 1; index_no++)
      the_block[index_no].stringptr = &the_block[index_no + 1];
    
    // The last element is free, but there's no free memory after it.
    the_block[blocksize - 1].stringptr = NULL;
    
    // Allocate one object for immediate use.
    the_string = &the_block[0];
    
    // The next one is the next_free object.
    next_free = &the_block[1];
    
  } else {
    // Just use the next free object.
    the_string = next_free;
    
    // The next free element is stored in the stringptr value.
    next_free = the_string->stringptr;
    
#if _DEBUG
    if (next_free != NULL) {
      // Check that the free list pointer is valid.
      for (int i = 0; i < blocks_used; i++) {
        if ((next_free >= &storage[i][0]) &&
          (next_free < &storage[i][blocksize]))
          break;
      }
      
      assert(i < blocks_used);
    }
#endif
  }
  
  // Set up initial values and return the string.
  the_string->celllist_length = 0;
  the_string->stringptr = NULL;
  return the_string;
}

// Deallocate a previously allocated string.
void StringMemory::deallocate(string *the_string)
{
#if _DEBUG
  int i;
    
  // Check that the object is part of our memory, and that
  // the object isn't already free.
  for (i = 0; i < blocks_used; i++) {
    assert(storage[i] != NULL);

    // Was the object allocated from this block?
    if ((the_string >= &storage[i][0]) &&
        (the_string <  &storage[i][blocksize])) {
      // This is a pointer to inside block i.
      break;
    }
  }
    
  // The object should have satisfied the break condition above,
  // otherwise it isn't part of our memory.
  assert(i < blocks_used);
#endif

  // Deallocate the string's triangles.
  owner->trimem->deallocate_string(the_string->head);

  // Free the string.
  the_string->celllist_length = -1;
  the_string->stringptr = next_free;
    
  // This string will be next to be allocated.
  next_free = the_string;
}

// Deallocate a celllist of strings.
void StringMemory::deallocate_celllist(string *the_string)
{
  int i, delete_length;
  string *thisptr, *nextptr;

  // Check that the string is a valid one.
  assert(the_string != NULL);
  assert(the_string->reference_count == 0);
  thisptr = the_string;
  delete_length = the_string->celllist_length;
  for (i = 0; i < delete_length; i++) {
    assert(thisptr != NULL);
    assert(thisptr->celllist_length == delete_length - i);
    assert((i == 0) || (thisptr->reference_count == -1));
    nextptr = thisptr->stringptr;
    deallocate(thisptr);
    thisptr = nextptr;
  }
  assert(thisptr == NULL);
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
