/*
 * transformation.cc: Routines for class ObjectTransformation and subclasses.
 *
 * Contains code to define the various kinds of transformations that may be
 * applied to an object.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Get the transform matrix for the translation.
void ObjectTranslation::get_transform(transform t)
{
  Transform::unit(t);

  t[FORE][X][3] =  trans_amount[X];
  t[FORE][Y][3] =  trans_amount[Y];
  t[FORE][Z][3] =  trans_amount[Z];
  t[AFT ][X][3] = -trans_amount[X];
  t[AFT ][Y][3] = -trans_amount[Y];
  t[AFT ][Z][3] = -trans_amount[Z];  
}

// Get the color of this transformation (for drawing its skitter)
void ObjectTranslation::get_color(color main_col, double *inside_alpha,
				  color ghosted_color, double *ghosted_alpha)
{
  // A nice forest green.
  Color::assign(main_col, 0.133, 0.545, 0.133);
  Color::assign(ghosted_color, 0.244, 1.000, 0.244);

  *inside_alpha = 0.5;
  *ghosted_alpha = 0.3;
}

// Create a new Direct Manipulation Device for the translation.
DMD *ObjectTranslation::create_dmd(void)
{
  return new TranslationDMD(this);
}

// Get the name of the transformation.
void ObjectTranslation::get_name(char *namestr)
{
  strcpy(namestr, "Translation");
}

// Get a string that describes the parameters of the transformation.
void ObjectTranslation::get_params(char *paramstr)
{
  sprintf(paramstr, "(%.2f %.2f %.2f)", trans_amount[X], trans_amount[Y], trans_amount[Z]);
}

// Get the transform matrix for the scale.
void ObjectScale::get_transform(transform t)
{
  double scale_inv = 1.0 / CHECK0(scale_amount,
				  "ObjectScale: zero scale factor");

  Transform::unit(t);

  t[FORE][X][X] = scale_amount;
  t[FORE][Y][Y] = scale_amount;
  t[FORE][Z][Z] = scale_amount;
  t[AFT ][X][X] = scale_inv;
  t[AFT ][Y][Y] = scale_inv;
  t[AFT ][Z][Z] = scale_inv;
}

// Get the color of this transformation (for drawing its skitter)
void ObjectScale::get_color(color main_col, double *inside_alpha, 
			    color ghosted_color, double *ghosted_alpha)
{
  // A bright yellow.
  Color::assign(main_col, 1.0, 1.0, 0.0);
  Color::assign(ghosted_color, 1.0, 1.0, 0.0);

  *inside_alpha = 0.5;
  *ghosted_alpha = 0.3;
}

// Create a new Direct Manipulation Device for the scale.
DMD *ObjectScale::create_dmd(void)
{
  return new ScaleDMD(this);
}

// Get the name of the transformation.
void ObjectScale::get_name(char *namestr)
{
  strcpy(namestr, "Scale");
}

// Get a string that describes the parameters of the transformation.
void ObjectScale::get_params(char *paramstr)
{
  sprintf(paramstr, "%.2f", scale_amount);
}

// Get the transform matrix for the rotation.
void ObjectRotation::get_transform(transform t)
{
  vector debug;

  Vector::assign(debug, 1, 0, 0);
  Quaternion::apply(rotate_q, debug, debug);
  Quaternion::form_transform(rotate_q, t);
}

// Get the color of this transformation (for drawing its skitter)
void ObjectRotation::get_color(color main_col, double *inside_alpha,
			       color ghosted_color, double *ghosted_alpha)
{
  // A purple.
  Color::assign(main_col, 0.627, 0.125, 0.941);
  Color::assign(ghosted_color, 0.627, 0.125, 0.941);
  
  *inside_alpha = 0.5;
  *ghosted_alpha = 0.3;
}

// Create a new Direct Manipulation Device for the rotation.
DMD *ObjectRotation::create_dmd(void)
{
  return new RotationDMD(this);
}

// Get the name of the transformation.
void ObjectRotation::get_name(char *namestr)
{
  strcpy(namestr, "Rotation");
}

// Get a string that describes the parameters of the transformation.
void ObjectRotation::get_params(char *paramstr)
{
  double rot_deg, axis_scale;
  vector rot_axis;

  if (rotate_q[W] >= 1.0) {
    rot_deg = 0;
    Vector::set(rot_axis, zero_vector);
  } else if (rotate_q[W] <= -1.0) {
    rot_deg = 360;
    Vector::set(rot_axis, zero_vector);
  } else {
    rot_deg = RAD2DEG(2 * acos(rotate_q[W]));
    axis_scale = sqrt(1.0 - SQR(rotate_q[W]));
    Vector::mult(1.0 / axis_scale, rotate_q, rot_axis);
  }

  sprintf(paramstr, "%.1f deg, axis (%.2f %.2f %.2f)", 
          rot_deg, rot_axis[X], rot_axis[Y], rot_axis[Z]);
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
