/*
 * transformation.h: Definitions for class ObjectTransformation and subclasses.
 *
 * Contains code to define the various kinds of transformations that may be
 * applied to an object.
 */

// Types of transformations.
typedef enum {TRANS_TRANSLATION, TRANS_SCALE, TRANS_ROTATION} trans_type;

// All transformations subclass ObjectTransformation.
class ObjectTransformation {
public:
  // Get the transform matrix for this transformation.
  virtual void get_transform(transform t) = 0;

  // Get the color of this transformation (for drawing its skitter)
  virtual void get_color(color main_col, double *inside_alpha,
			 color ghosted_color, double *ghosted_alpha) = 0;

  // Create a new Direct Manipulation Device for the transformation.
  virtual DMD *create_dmd(void) = 0;

  // Get the name of the transformation.
  virtual void get_name(char *namestr) = 0;

  // Get a string that describes the parameters of the transformation.
  virtual void get_params(char *paramstr) = 0;
};

// A translation.
class ObjectTranslation : public ObjectTransformation {
public:
  vector trans_amount;

  // Get the transform matrix for the translation.
  virtual void get_transform(transform t);

  // Get the color of this transformation (for drawing its skitter)
  virtual void get_color(color main_col, double *inside_alpha,
			 color ghosted_color, double *ghosted_alpha);

  // Create a new Direct Manipulation Device for the transformation.
  virtual DMD *create_dmd(void);

  // Get the name of the transformation.
  virtual void get_name(char *namestr);

  // Get a string that describes the parameters of the transformation.
  virtual void get_params(char *paramstr);

  // Create an identity translation.
  ObjectTranslation(void) {
    Vector::set(trans_amount, zero_vector);
  }

  // Set the parameters of the translation.
  void set_translation(double x, double y, double z) {
    Vector::assign(trans_amount, x, y, z);
  }
  void set_translation(vector t) {
    Vector::set(trans_amount, t);
  }  

  // Create a translation of the specified amount.
  ObjectTranslation(double x, double y, double z) {
    Vector::assign(trans_amount, x, y, z);
  }
  ObjectTranslation(vector t) {
    Vector::set(trans_amount, t);
  }
};

// A scale.
class ObjectScale : public ObjectTransformation {
public:
  double scale_amount;

  // Get the transform matrix for the scale.
  virtual void get_transform(transform t);

  // Get the color of this transformation (for drawing its skitter)
  virtual void get_color(color main_col, double *inside_alpha,
			 color ghosted_color, double *ghosted_alpha);

  // Create a new Direct Manipulation Device for the transformation.
  virtual DMD *create_dmd(void);

  // Get the name of the transformation.
  virtual void get_name(char *namestr);

  // Get a string that describes the parameters of the transformation.
  virtual void get_params(char *paramstr);

  // Create an identity scale.
  ObjectScale(void) {
    scale_amount = 1;
  }

  // Set the scale factor.
  void set_scale(double s) {
    scale_amount = s;
  }

  // Create a scale of the specified amount.
  ObjectScale(double s) {
    scale_amount = s;
  }
};

// A rotation.
class ObjectRotation : public ObjectTransformation {
public:
  quaternion rotate_q;

  // Get the transform matrix for the scale.
  virtual void get_transform(transform t);

  // Get the color of this transformation (for drawing its skitter)
  virtual void get_color(color main_col, double *inside_alpha,
			 color ghosted_color, double *ghosted_alpha);

  // Create a new Direct Manipulation Device for the transformation.
  virtual DMD *create_dmd(void);

  // Get the name of the transformation.
  virtual void get_name(char *namestr);

  // Get a string that describes the parameters of the transformation.
  virtual void get_params(char *paramstr);

  // Create an identity rotation.
  ObjectRotation(void) {
    Quaternion::unit(rotate_q);
  }

  // Set the rotation. Various methods.
  void set_rotation(quaternion q) {
    Quaternion::set(rotate_q, q);
  }
  void set_rotation(vector axis, double rot_rad) {
    Vector::normalise(axis, "ObjectRotation: set_rotation, zero axis");
    Vector::mult(sin(rot_rad/2), axis, rotate_q);
    rotate_q[W] = cos(rot_rad/2);
  }
  void set_rotation(axis from, axis to, double rot_rad) {
    vector from_ax, to_ax, rotation_axis;

    Vector::set(from_ax, zero_vector);
    Vector::set(to_ax, zero_vector);
    from_ax[from] = 1.0;
    to_ax[to] = 1.0;
    Vector::cross(from_ax, to_ax, rotation_axis);

    set_rotation(rotation_axis, rot_rad);
  }

  // Create a rotation of the specified amount.
  ObjectRotation(vector axis, double rot_rad) {
    set_rotation(axis, rot_rad);
  }
  ObjectRotation(axis from, axis to, double rot_rad) {
    set_rotation(from, to, rot_rad);
  }
};

