/*
 * transformlist.cc: Routines for class TransformList and supporting code.
 *
 * Contains code to create and maintain a set of transformations that are
 * being applied to an object.
 */

/* Collection-safe */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Get the total transformation.
void TransformList::get_position(transform t)
{
  if (tail == NULL)
    Transform::unit(t);
  else
    Transform::set(t, tail->cumulative);
}

// Update the transformation matrix at a node.
void TransformList::update_transform(TransformListNode *n)
{
  assert(n->trans != NULL);
  n->trans->get_transform(n->matrix);
}

// Propagate the transformation matrices, starting from a node.
void TransformList::propagate_matrices(void)
{
  if (head != NULL)
    propagate_matrices(head);
}

void TransformList::propagate_matrices(TransformListNode *n)
{
  for (TransformListNode *ptr = n; ptr != NULL; ptr = ptr->next) {
    if (ptr->prev == NULL)
      Transform::set(ptr->cumulative, ptr->matrix);
    else
      Transform::combine(ptr->matrix, ptr->prev->cumulative, ptr->cumulative);
  }
}

// Insert a node into the list. Used internally by add_transformation and
// move_transformation.
void TransformList::insert_node(TransformListNode *newnode,
				TransformListNode *pos)
{
  newnode->prev = pos;

  if (newnode->prev == NULL) {
    newnode->next = head;
    head = newnode;
  } else {
    newnode->next = newnode->prev->next;
    newnode->prev->next = newnode;
  }

  if (newnode->next == NULL)
    tail = newnode;
  else
    newnode->next->prev = newnode;

  num_transforms++;
}

// Remove a node from the list. Used internally by delete_transformation and
// move_transformation.
void TransformList::remove_node(TransformListNode *n)
{
  if (n->prev == NULL)
    head = n->next;
  else
    n->prev->next = n->next;

  if (n->next == NULL)
    tail = n->prev;
  else
    n->next->prev = n->prev;

  assert(num_transforms > 0);
  num_transforms--;
}

// Add a transformation to the list, immediately after the given node.
// NULL means at the head.
void TransformList::add_transformation(ObjectTransformation *objtrans,
				       TransformListNode *n)
{
  TransformListNode *newnode;

  if (!(newnode = (TransformListNode *) calloc(1, sizeof(TransformListNode))))
    fatal_error("TransformList::add_transformation", "Can't allocate newnode!",
		"Out of memory.");

  assert(objtrans != NULL);
  newnode->trans = objtrans;
  objtrans->get_transform(newnode->matrix);

  insert_node(newnode, n);

  propagate_matrices(newnode);
}

// Move a transformation to a new position in the list, immediately after
// new_pos. NULL means at the head.
void TransformList::move_transformation(TransformListNode *n,
					TransformListNode *new_pos)
{
  assert(n != NULL);
  
  // Are we already at the right position?
  if ((new_pos == n->prev) || (new_pos == n))
    return;

  // First, remove the node.
  remove_node(n);

  // Then, insert the node at its new position.
  insert_node(n, new_pos);

  // Propagate the matrices... this is only necessary from whichever of
  // the old-position and the new-position is closer to the head of the list.
  // However this is non-trivial to work out, and we don't want to have to
  // propagate twice, so we just go from the start of the list.
  propagate_matrices();
}

// Delete a transformation.
void TransformList::delete_transformation(TransformListNode *n)
{
  assert(n != NULL);
  remove_node(n);

  assert(n->trans != NULL);
  delete n->trans;

  free(n);
}

// Print out the list.
void TransformList::print(void)
{
  TransformListNode *ptr;

  printf("Print TransformList (with %d transforms):\n", num_transforms);
  if (head == NULL)
    printf("head NULL\n");
  else
    printf("head: 0x%08X\n", (unsigned int) head);
  if (tail == NULL)
    printf("tail NULL\n");
  else
    printf("tail: 0x%08X\n", (unsigned int) tail);

  for (ptr = head; ptr != NULL; ptr = ptr->next) {
    printf("node at 0x%08X: prev ", (unsigned int) ptr);

    if (ptr->prev == NULL)
      printf("NULL, next ");
    else
      printf("0x%08X, next ", (unsigned int) ptr->prev);

    if (ptr->next == NULL)
      printf("NULL\n");
    else
      printf("0x%08X\n", (unsigned int) ptr->next);

    printf("  ObjectTransformation 0x%08X\n", (unsigned int) ptr->trans);
  }
}

// Create a non-empty TransformList.
TransformList::TransformList(int num_trans, ObjectTransformation **trans_array)
{
  int i;

  num_transforms = 0;
  head = tail = NULL;

  assert((num_trans == 0) || (trans_array != NULL));
  for (i = 0; i < num_trans; i++)
    add_transformation(trans_array[i], NULL);
}

// Delete the TransformList.
TransformList::~TransformList(void)
{
  TransformListNode *ptr, *next;

  for (ptr = head; ptr != NULL; ptr = next) {
    next = ptr->next;

    assert(ptr->trans != NULL);
    delete ptr->trans;

    free(ptr);
  }
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
