/*
 * wgraphview.cc: Routines for WObjectGraphView.
 *
 * Contains code to draw the graph that has been laid out by the GraphLayout,
 * within a Windows view.
 */

/* Collection-safe */

#include "stdafx.h"

#ifdef _WIN32

#include "ModelApp.h"
#include "ModelDoc.h"
#include "CGraphView.h"
#include "wgraphview.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

WObjectGraphView::WObjectGraphView(Modeller *m, CGraphView *graphview)
{
  this->modeller = m;
  this->graphview = graphview;

  num_nodes = 0;
  nodearray = NULL;
  nodearray_size = 0;
  nodearray_blocksize = 10;

  num_links = 0;
  linkarray = NULL;
  linkarray_size = 0;
  linkarray_blocksize = 10;

  viewstate = VS_SELECT;

  modeller->bind_graphview(this);
}

// Clear the canvas.
void WObjectGraphView::clear_canvas(void)
{
  num_nodes = num_links = 0;
}

// Add a node to the canvas. If this is selectable, then it must have
// a non-NULL selection_node. NB: the linktype indicators use this, too.
void WObjectGraphView::add_node(int xpos, int ypos, int size, int selected,
     char *internal_text,
     char *external_text,
     objectnode *selection_node)
{
  nodeinfo *newnode;

  newnode = new_nodeinfo();

  newnode->pos[X] = xpos;
  newnode->pos[Y] = ypos;
  newnode->size = size;
  newnode->is_selected = selected;
  newnode->objectgraphptr = selection_node;

  if (internal_text == NULL) {
    strcpy(newnode->internaltext, "");
  } else {
    assert(strlen(internal_text) < MAX_OBJECTGRAPH_LABELSIZE);
    strcpy(newnode->internaltext, internal_text);
  }

  if (external_text == NULL) {
    strcpy(newnode->externaltext, "");
  } else {
    assert(strlen(external_text) < MAX_OBJECTGRAPH_LABELSIZE);
    strcpy(newnode->externaltext, external_text);
  }
}


// Add a link to the canvas.
void WObjectGraphView::add_link(int num_segs, int *seg_coord,
			    int is_ghost, int has_arrow)
{
  int i;
  linkinfo *newlink;

  newlink = new_linkinfo();

  assert((num_segs > 0) && (num_segs <= MAX_OBJECTGRAPH_LINKSEGS));

  newlink->num_segs = num_segs;
  for (i = 0; i < num_segs; i++) {
    newlink->segpos[i][X] = seg_coord[2*i + X];
    newlink->segpos[i][Y] = seg_coord[2*i + Y];
  }
  newlink->is_ghost = is_ghost;
  newlink->has_arrow = has_arrow;
}

// Called by the layout engine when the drawing pass is completed.
void WObjectGraphView::finish_drawing(void)
{
  graphview->Invalidate();
}

// Delete the WObjectGraphView.
WObjectGraphView::~WObjectGraphView()
{
  modeller->unbind_graphview(this);

  if (nodearray)
    free(nodearray);
  if (linkarray)
    free(linkarray);
}

// Get a new pointer into nodearray
nodeinfo *WObjectGraphView::new_nodeinfo(void)
{
  if (num_nodes >= nodearray_size) {
    // Allocate another block.
    nodearray_size += nodearray_blocksize;

    nodearray = (nodeinfo *) realloc((void *) nodearray, nodearray_size * sizeof(nodeinfo));
    if (!nodearray)
      fatal_error("WObjectGraphView::new_nodeinfo", "Can't expand nodearray!",
                                                    "Out of memory.");
  }

  return &nodearray[num_nodes++];
}

// Get a new pointer into linkarray.
linkinfo *WObjectGraphView::new_linkinfo(void)
{
  if (num_links >= linkarray_size) {
    // Allocate another block.
    linkarray_size += linkarray_blocksize;

    linkarray = (linkinfo *) realloc((void *) linkarray, linkarray_size * sizeof(linkinfo));
    if (!linkarray)
      fatal_error("WObjectGraphView::new_linkinfo", "Can't expand linkarray!",
                                                    "Out of memory.");
  }

  return &linkarray[num_links++];
}

// Find the size of the objects on the canvas.
void WObjectGraphView::get_size(int *xsize, int *ysize)
{
  int xmax, ymax, i, j;
  nodeinfo *node;
  linkinfo *link;

  xmax = ymax = 10;

  for (i = 0, node = nodearray; i < num_nodes; i++, node++) {
    if (node->pos[X] + node->size / 2  + 10> xmax)
      xmax = node->pos[X] + node->size / 2 + 10;
    if (node->pos[Y] + node->size / 2 + 10> ymax)
      ymax = node->pos[Y] + node->size / 2 + 10;
  }

  for (i = 0, link = linkarray; i < num_links; i++, link++) {
    for (j = 0; j < link->num_segs; j++) {
      if (link->segpos[j][X] + 10 > xmax)
        xmax = link->segpos[j][X] + 10;
      if (link->segpos[j][Y] + 10 > ymax)
        ymax = link->segpos[j][Y] + 10;
    }
  }

  *xsize = xmax;
  *ysize = ymax;
}

// The user has clicked on the canvas to select a node.
void WObjectGraphView::selection_click(int x, int y)
{
  int i, xpos, ypos;
  nodeinfo *node;

  // Iterate through all the nodes and see if we might have
  // clicked on one to select it.
  for (i = 0, node = nodearray; i < num_nodes; i++, node++) {
    xpos = x - node->pos[0];
    ypos = y - node->pos[1];

    if ((xpos > - node->size/2) && (xpos < + node->size/2) &&
        (ypos > - node->size/2) && (ypos < + node->size/2)) {
      // This node was clicked upon.
      switch(viewstate) {
      case VS_SELECT:
        modeller->do_select(node->objectgraphptr);
        return;
      case VS_LINK_PLUS:
        modeller->do_link_copy(node->objectgraphptr, PLUS);
        break;
      case VS_LINK_MINUS:
        modeller->do_link_copy(node->objectgraphptr, MINUS);
        break;
      case VS_LINK_AND:
        modeller->do_link_copy(node->objectgraphptr, AND);
        break;
      default:
        assert(0);
      }
    }
  }

  // Return to the selection state.
  if (viewstate == VS_SELECT)
    modeller->do_select(NULL);
  else
    // Stop the link operation.
    modeller->setgraphviewstate(VS_SELECT);
}

// Set the state of the graphview.
void WObjectGraphView::setgraphviewstate(graphviewstate state)
{
  viewstate = state;

  if (state == VS_SELECT) {
    // This is being called from the modeller code
    // rather than a link-button, so we need to update
    // the link-buttons and return them to their normal state.
    graphview->GetDocument()->OnSetSelectViewState();
  }
}

#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif

#endif  // #ifndef _WIN32
