/*
 * worldcollection.cpp: Routines for class WorldCollection.
 *
 * Contains code for the WorldCollections, which track a given instance
 * in worldspace of a Collection.
 */

// #include "includelist.h"
#include "Stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Provides a unique index for split events.
int global_split_id_counter = 0;
int global_uncalculated_worldcollections = 0;

int debug_split_breakid = -1;
int debug_split_breaksample = -1;
int debug_split_breaksplitid = -1;
int debug_cell_break[3] = {2, 3, 2};

#define DEBUG_ONETERMINAL        FALSE
#define CELL_EDGE_TOLERANCE    1e-05

// Add a reference to the WorldCollection.
void WorldCollection::add_reference(void)
{
  assert(reference_count >= 0);
  assert(reference_count < 1000);
  reference_count++;
}

// Remove a reference from the WorldCollection. Returns TRUE if the WorldCollection
// should be deleted now.
int WorldCollection::remove_reference(void)
{
  assert(reference_count > 0);
  assert(reference_count < 1000);

  if (--reference_count == 0)
    return TRUE;
  else
    return FALSE;
}

// Delete the WorldCollection.
WorldCollection::~WorldCollection(void)
{
  // We should have no references.
  assert(reference_count == 0);

  // Remove references to all of the celllists that we reference.
  for (int i = 0; i < CUBE(collx->res); i++) {
    if (cell[i] != NULL) {
      assert(cell[i]->reference_count > 0);
      if (--(cell[i]->reference_count) == 0)
        modeller->worldstringmem->deallocate_celllist(cell[i]);
    }
  }

  // Delete our triangle data.
  free(cell);

  // Remove the reference to our Collection.
  if (collx->remove_reference())
    delete collx;

  // If we have a calcsplit terminal buffer, free it.
  if (needs_calculation == CALC_SPLIT) {
    assert(calcsplit_terminal != NULL);
    free(calcsplit_terminal);
  }
}

// from collection.cpp
extern int debug_collection_construct_breakid;

// Construct a WorldCollection by polygonizing a primitive surface for a
// given octree in a leafnode.
WorldCollection::WorldCollection(primtreeprimnode *owner,
                                 primobjectnode *surfacenode,
                                 int res, transform collectionspace,
                                 box bounds, Modeller *modeller)
{
  // This routine is only called on un-transformed primitive nodes.
  assert(owner->is_transformed == FALSE);
  
  this->modeller = modeller;
  this->owner = owner;
  reference_count = 0;
  global_uncalculated_worldcollections++;

  last_draw_id = -1;

  collx = new Collection(surfacenode, res, collectionspace, bounds, modeller);
  collx->add_reference();

  this->parent = NULL;

  if (collx->collx_id == debug_collection_construct_breakid)
    _CrtDbgBreak();

  // Set up our cell array.
  cell = (worldstring **) calloc(CUBE(res), sizeof(worldstring *));
  if (!cell)
    fatal_error("WorldCollection::WorldCollection", "Can't allocate cell array!",
                "Out of memory.");

  // Will be calculated from the polygonized primitive surface as necessary.
  needs_calculation = CALC_WC_FROM_COLLX;
}

// Create a new WorldCollection from a pre-existing **primitive** Collection.
WorldCollection::WorldCollection(primtreeprimnode *owner, Collection *collx)
{
  if (collx->collx_id == debug_collection_construct_breakid)
    _CrtDbgBreak();

  modeller = collx->modeller;
  this->owner = owner;
  this->collx = collx;
  collx->add_reference();
  global_uncalculated_worldcollections++;

  reference_count = 0;
  last_draw_id = -1;

  // This must be a primitive Collection that we're linked to.
  assert(collx->parent_collection == NULL);
  this->parent = NULL;

  // Set up our cell array.
  cell = (worldstring **) calloc(CUBE(collx->res), sizeof(worldstring *));
  if (!cell)
    fatal_error("WorldCollection::WorldCollection", "Can't allocate cell array!",
                "Out of memory.");

  // Will be calculated from the polygonized primitive surface as necessary.
  needs_calculation = CALC_WC_FROM_COLLX;
}

// Create an new WorldCollection that is the split fragment of some parent.
WorldCollection::WorldCollection(int num_terminals, splitterminal *splitbuf, int keep_inside,
                                 WorldCollection *parent_collection)
{
  reference_count = 0;
  this->parent = parent_collection;
  owner = parent_collection->owner;
  modeller = parent_collection->modeller;
  global_uncalculated_worldcollections++;
  last_draw_id = -1;

  // Allocate the cell array.
  cell = (worldstring **) calloc(CUBE(parent_collection->collx->res), sizeof(worldstring *));
  if (!cell)
    fatal_error("WorldCollection::WorldCollection", "Can't allocate cell array!",
                "Out of memory.");

  // Create a new, blank Collection for us.
  collx = new Collection(parent_collection->collx);
  collx->add_reference();

  if (collx->collx_id == debug_collection_construct_breakid)
    _CrtDbgBreak();

  // Our parent will be split to find us, if necessary.
  needs_calculation = CALC_SPLIT;
  calcsplit_num_terminals = num_terminals;
  calcsplit_terminal = (splitterminal *) malloc(num_terminals * sizeof(splitterminal));
  assert(calcsplit_terminal != NULL);
  memcpy((void *) calcsplit_terminal, (void *) splitbuf, num_terminals * sizeof(splitterminal));
  calcsplit_keep_inside = keep_inside;
}

// Perform all of the calculations necessary to create a WorldCollection.
int WorldCollection::calculate(int cell_index)
{
  int i;

  switch(needs_calculation) {
  case CALC_NONE:
    return FALSE;

  case CALC_WC_FROM_COLLX:
    collx->calculate();
    for (i = 0; i < CUBE(collx->res); i++)
      if (collx->cell[i] != NULL)
        cell[i] = make_worldspace_cell(collx->cell[i]);
    needs_calculation = CALC_NONE;
    global_uncalculated_worldcollections--;
    return TRUE;

  case CALC_SPLIT:
    assert(parent != NULL);
    assert(collx->needs_calculation == CALC_SPLIT);
    parent->calculate(cell_index);

    if (cell_index != -1) {
      if (parent->cell[cell_index] == NULL) {
        // Our parent has no triangles in this cell, so we don't need
        // to calculate after all.
        return FALSE;
      }
    }

    parent->split(calcsplit_num_terminals, calcsplit_terminal,
                  this, calcsplit_keep_inside);

    needs_calculation = CALC_NONE;
    collx->needs_calculation = CALC_NONE;
    free(calcsplit_terminal);
    global_uncalculated_worldcollections--;
    return TRUE;

  default:
    assert(0);
    return FALSE;
  }
}

// Transform a cell of triangle strings from collectionspace to worldspace.
worldstring *WorldCollection::make_worldspace_cell(string *cs_str)
{
  worldstring *ws_str;
  int q;

  if (cs_str == NULL)
    return NULL;

  ws_str = modeller->worldstringmem->allocate();

  ws_str->celllist_length = cs_str->celllist_length;
  ws_str->reference_count = (cs_str->reference_count == -1) ? -1 : 1;

  ws_str->plane_dot = cs_str->plane_dot;
  if (owner->is_transformed) {
    Matrix::applytranspose(owner->collection_trans[AFT], cs_str->norm, ws_str->norm);
    for (q = X; q <= Z; q++)
      ws_str->plane_dot += owner->collection_trans[FORE][q][3] * ws_str->norm[q];
  } else {
    Vector::set(ws_str->norm, cs_str->norm);
  }

  // Make the rest of the celllist.
  ws_str->stringptr = make_worldspace_cell(cs_str->stringptr);

  return ws_str;
}

// Split this Collection against another object. The other object is represented
// by a set of solidterminalnodes, each covering some part of the collection.
int WorldCollection::split(int num_terminals, splitterminal *terminalbuf,
                            WorldCollection *child, int keep_inside)
{
  int j, keep_tri, inout_result, cell_index;
  int some_kept, some_discarded, split_any;
  int debug_break = FALSE;
  vector pointsample;
  splitterminal *overlap_keep;

  this_split_id = global_split_id_counter++;
  some_kept = some_discarded = split_any = FALSE;

  if ((this ->collx->sample_id == debug_split_breaksample) ||
      (this ->collx->collx_id  == debug_split_breakid) ||
      (child->collx->collx_id  == debug_split_breakid) ||
      (this_split_id           == debug_split_breaksplitid)) {
    debug_break = TRUE;
    _CrtDbgBreak();
  }

  this->num_split_terminals = num_terminals;
  this->split_terminal = terminalbuf;
  this->keep_inside = keep_inside;

  // First: Precompute information for each terminal.
  precompute_terminal_info();

  // Initialise the split buffers.
  splitbufsize = 0;
  split_string = NULL;

  // For each cell in our Collection...
  cell_index = 0;

  for (cell_z = 0; cell_z < collx->res; cell_z++) {
    for (cell_y = 0; cell_y < collx->res; cell_y++) {
      for (cell_x = 0; cell_x < collx->res; cell_x++, cell_index++) {
#if _DEBUG
        if ((debug_break) &&
          (cell_x == debug_cell_break[X]) &&
          (cell_y == debug_cell_break[Y]) && 
          (cell_z == debug_cell_break[Z]))
          _CrtDbgBreak();
#endif
        
        if (cell[cell_index] == NULL) {
          // This cell contains no triangles. We don't need to do anything as the child is created empty.
          assert(child       ->cell[cell_index]  == NULL);
          assert(child->collx->cell[cell_index]  == NULL);
          continue;
        }
        
        // Clear the split buffer.
        num_split = 0;
        thiscell_overlaps = 0;
        
        // For each terminal, if this cell could overlap that Terminal then
        // add triangles from it.
        overlap_keep = current_terminal = NULL;
        traverse_extent(cell_index, &overlap_keep, &current_terminal);
        
        if (current_terminal == NULL) {
          // No split terminals overlap our position.
          if (overlap_keep == NULL) {
            // There are no terminals that overlap our position.
            // We must be outside the modeller's workspace.
            keep_tri = FALSE;
          } else {
            // This part of our Collection is just against FULL or EMPTY and is being kept as a result.
            keep_tri = TRUE;            
          }
        } else if (num_split > 0) {
          // Perform the split. This stores split triangles in split_in and split_out
          // and their collections.
          if (modeller->trisplitter->split_cell(cell_index, this, child, num_split, split_string,
                                                (keep_inside) ? -1 : +1, &keep_tri)) {
            // The cell was split, and split triangles have been inserted. Continue.
            split_any = TRUE;
            continue;
          } else {
            // The cell was not split. All triangles are either kept or discarded based on the
            // value inserted into keep_tri by split_cell().
          }
        } else {
          // We need to point-intersect against the Terminal which contains
          // this part of the Collection. Pick the midpoint of the first triangle from our cell and
          // sample there.
          //
          // NB: this point lies on the surface and so is guaranteed to overlap any
          // relevant terminals. If we picked some other point in the cell (that didn't lie on the
          // surface), it would not necessarily have the relevant terminals being passed in to us,
          // as they might not be flagged as CHANGED in the octree pass.
          //
          // NB also: We can safely sample against current_terminal only because there are no triangles
          // in this area... thus all the terminals must give us the same (homogeneous) result across this
          // cell.
          Vector::set(pointsample, zero_vector);
          for (j = 0; j < 3; j++)
            Vector::addfrac(1.0 / 3.0, this->collx->cell[cell_index]->head->v[j], pointsample);
          if (owner->is_transformed)
            Matrix::apply(owner->collection_trans[FORE], pointsample, pointsample);
          inout_result = world_point_intersect(current_terminal->terminal->rcsggraph, pointsample);
          
          keep_tri = (keep_inside) ? inout_result : !inout_result;
        }
        
        // We should never get here if there are no triangles in this cell.
        assert(cell[cell_index] != NULL);
        assert(collx->cell[cell_index] != NULL);
        
        if (keep_tri) {
          some_kept = TRUE;
          
          child->cell[cell_index] = cell[cell_index];
          assert(child->cell[cell_index]->reference_count > 0);
          child->cell[cell_index]->reference_count++;                                        
          
          child->collx->cell[cell_index] = collx->cell[cell_index];
          assert(child->collx->cell[cell_index]->reference_count > 0);
          child->collx->cell[cell_index]->reference_count++;                                 
          
          child->collx->total_tri += child->collx->cell[cell_index]->num_tri;             
        } else {
          some_discarded = TRUE;
        }
      }
    }
  }

  // Deallocate our entry in the primsplitinfo stack.
  modeller->dealloc_primsplitinfo(psi_index);

  // Deallocate the extents list.
  deallocate_extent(extent);

  // Free the split buffer.
  if (splitbufsize > 0)
    free(split_string);

  return ((split_any) || (some_kept && some_discarded));
}

// Traverse an extentlist looking for terminals which overlap the
// present cell.
void WorldCollection::traverse_extent(int cell_index, splitterminal **overlap_keep, splitterminal **overlap_split)
{
  traverse_extent(cell_index, overlap_keep, overlap_split, Z, extent);
}

// Traverse an extentlist looking for terminals which overlap the
// present cell.
void WorldCollection::traverse_extent(int cell_index, splitterminal **overlap_keep, splitterminal **overlap_split,
                                      int ax, extentlist *list)
{
  double min_bound, max_bound;
  extentlist *node;

  if (list == NULL)
    return;

  // Work out whereabouts in this axis the current cell extends to.
  min_bound = ws_zerocorner[ax] + cell_x * ws_cell_step[X][ax]
                                + cell_y * ws_cell_step[Y][ax]
                                + cell_z * ws_cell_step[Z][ax];
  max_bound = min_bound + singlecell_maxstep[ax];
  min_bound += singlecell_minstep[ax];

  // Find the first node in the chain that overlaps our mininum
  // bound by at least CELL_EDGE_TOLERANCE.
  for (node = list; node != NULL; node = node->next) {
    if (node->value - min_bound > CELL_EDGE_TOLERANCE) {
      // Our minimum bounds lie below (overlapping) this value by
      // at least the tolerance value.
      break;
    }
  }

  // Look at every node in the chain that overlaps our bounds.
  for ( ; node != NULL; node = node->next) {
    if (ax == X) {
      if (node->termptr != NULL) {
        // There is a terminal here which we overlap.
        node->termptr->cellid_overlap = cell_index;
        thiscell_overlaps++;
        
        if (node->termptr->terminal != NULL) {
          *overlap_split = node->termptr;
          insert_triangles(node->termptr->terminal->rcsggraph, FALSE, cell_x, cell_y, cell_z);
        } else {
          *overlap_keep = node->termptr;
        }
      }
    } else {
      // Traverse the next layer down.
      traverse_extent(cell_index, overlap_keep, overlap_split, ax - 1, node->layer);
    }

    if (max_bound - node->value < CELL_EDGE_TOLERANCE) {
      // There is less than the tolerance value worth of our bounds remaining once we
      // pass this value. So we can stop.
      return;
    }
  }
}

// Precompute information for the terminals that we will be splitting against.
void WorldCollection::precompute_terminal_info(void)
{
  int i, j, k, q;
#if _DEBUG
  box ws_bounds;
#endif

  psi_index = modeller->alloc_primsplitinfo();

  // Calculate whereabouts this WorldCollection lies in worldspace.
  // FIXME: when we merge Collections with WorldCollections this will all be
  // somewhat unnecessary...
  Matrix::apply(owner->collection_trans[FORE], collx->bounds[0], ws_zerocorner);
  for (q = X; q <= Z; q++) {
    Matrix::apply(owner->collection_trans[FORE], collx->bounds[1 << q], ws_cell_step[q]);
    Vector::sub(ws_cell_step[q], ws_zerocorner, ws_cell_step[q]);
    Vector::mult(1.0 / collx->res, ws_cell_step[q], ws_cell_step[q]);
  }

  // Now that we know how far a single cell's step is in X, Y and Z, we need to
  // work out the minimum and maximum possible combinations.
  Vector::set(singlecell_minstep, zero_vector);
  Vector::set(singlecell_maxstep, zero_vector);
  for (j = X; j <= Z; j++) {
    for (k = X; k <= Z; k++) {
      ((ws_cell_step[j][k] < 0) ? singlecell_minstep : singlecell_maxstep)[k] += ws_cell_step[j][k];
    }
  }

  extent = NULL;

  // Now for every terminal, we need to insert it into our extents structure.
  for (i = 0; i < num_split_terminals; i++) {
#if _DEBUG
    // NB: this **must** be the same as the one in octree.cpp
#define TERMINAL_FACETOUCH_TOLERANCE  1e-04

    // We need to check that this Terminal really does overlap this Collection.
    // Calculate where the collection's points lie with respect to the terminal, and do
    // a voxel-intersection test.
    //
    // NB: in order to use cube_voxel_intersect we need to scale the points so that the
    // terminal lies from -1 to +1 instead of 0 to 1. Also we want to be conservative with this;
    // we don't want terminals that just touch faces to be tested against. So we add another scaling to
    // make the terminalbox a little smaller.
    for (j = 0; j < 8; j++) {
      Matrix::apply(owner->collection_trans[FORE], collx->bounds[j], ws_bounds[j]);

      for (k = X; k <= Z; k++)
        ws_bounds[j][k] = (2 * (ws_bounds[j][k] - modeller->terminalcorner[split_terminal[i].terminal_id][k])
                  / modeller->terminalsize[k]) - 1;

      Vector::mult(1.0 + TERMINAL_FACETOUCH_TOLERANCE, ws_bounds[j], ws_bounds[j]);
    }

    if (cube_voxel_intersect(ws_bounds) == OUT)
      TRACE("collx_id %d terminals %d... terminal %d does not overlap bounds\n", collx->collx_id, num_split_terminals, i);
#endif

    // Insert this terminal into the extents framework.
    insert_into_extents(&split_terminal[i]);

    // Traverse the RCSGgraph and for each primitive, work out the transformation
    // from OUR collectionspace into ITS collectionspace. This lets us work out
    // quickly whereabouts 
    if (split_terminal[i].terminal != NULL)
      compute_splitinfo(split_terminal[i].terminal->rcsggraph);
  }
}

// Insert a terminal's worldspace extents into our extents framework.
void WorldCollection::insert_into_extents(splitterminal *termptr)
{
  extent = insert_into_extents(termptr, Z, extent);
}

  // Insert a terminal's worldspace extents into our extents framework.
extentlist *WorldCollection::insert_into_extents(splitterminal *termptr, int ax, extentlist *list)
{
  int found;
  double min_bound, max_bound, delta;
  extentlist *prev_node, *next_node, *new_node;

  min_bound = modeller->terminalcorner[termptr->terminal_id][ax];
  max_bound = min_bound + modeller->terminalsize[ax];

  // Look for a node which contains the start of our interval. If there isn't one, create one.
  prev_node = NULL;
  found = FALSE;
  for (next_node = list; next_node != NULL; prev_node = next_node, next_node = next_node->next) {
    delta = min_bound - next_node->value;
    if (fabs(delta) < 1e-04) {
      found = TRUE;
      break;
    } else if (delta < 0) {
      break;
    }
  }

  if (found) {
    // There is a node for the start of our interval already.
    prev_node = next_node;
    next_node = next_node->next;
  } else {
    // We have found the place where the node that marks the start of our interval
    // should be inserted... just after prev_node.
    new_node = allocate_extent();
    if (prev_node == NULL)
      list = new_node;
    else
      prev_node->next = new_node;
    
    new_node->value = min_bound;
    new_node->layer = NULL;
    new_node->next = next_node;

    prev_node = new_node;
  }

  // Look for a node which contains the end of our interval. If there isn't one, create one.
  found = FALSE;
  for ( ; next_node != NULL; prev_node = next_node, next_node = next_node->next) {
    delta = max_bound - next_node->value;
    if (fabs(delta) < 1e-04) {
      found = TRUE;
      break;
    } else if (delta < 0) {
      break;
    }
  }

  if (found) {
    // There is a node for the end of our interval already.
    if (ax > X) {
      // Add this terminal into the relevant layer.
      next_node->layer = insert_into_extents(termptr, ax - 1, next_node->layer);
    } else {
      // We have found an entry for us to put our terminal in.
      assert(next_node->termptr == NULL);
      next_node->termptr = termptr;
    }
  } else {
    // We have found the place where the node that marks the end of our interval
    // should be inserted... just after prev_node.
    new_node = allocate_extent();
    if (prev_node == NULL)
      list = new_node;
    else
      prev_node->next = new_node;
    
    new_node->value = max_bound;
    if (ax > X)
      new_node->layer = insert_into_extents(termptr, ax - 1, NULL);
    else
      new_node->termptr = termptr;
    new_node->next = next_node;
  }

  return list;
}

// Allocate or deallocate an extentlist.
extentlist *WorldCollection::allocate_extent(void)
{
  extentlist *new_node = (extentlist *) calloc(1, sizeof(extentlist));
  if (!new_node)
    fatal_error("WorldCollection::allocate_extent", "Can't allocate new extentlist",
                "Out of memory.");
  return new_node;
}

void WorldCollection::deallocate_extent(extentlist *kill_me, int ax)
{
  if (kill_me == NULL)
    return;

  if (ax > X)
    deallocate_extent(kill_me->layer, ax - 1);

  deallocate_extent(kill_me->next, ax);
  free(kill_me);
}

// Point-intersect w.r.t the RCSGgraph; used from the TriSplitter when there's
// a situation that just can't be resolved any other way.
//
// This is an expensive call, because it has to determine first which terminal to
// intersect against, and then actually do the intersection.
int WorldCollection::world_point_intersect(int cell_index, vector worldpt)
{
  int i, j, found_keep = FALSE;
  vector location;

  for (i = 0; i < num_split_terminals; i++) {
    if (split_terminal[i].cellid_overlap != cell_index)
      continue;

    // Calculate this point's location relative to the terminal's location (in worldspace).
    for (j = X; j <= Z; j++) {
      location[j] = worldpt[j] - modeller->terminalcorner[split_terminal[i].terminal_id][j];
      if ((location[j] < 0) || (location[j] > modeller->terminalsize[j])) {
        // This terminal doesn't overlap.
        break;
      }
    }

    if (j < Z) {
      // We didn't finish the loop, this terminal doesn't overlap, continue looking.
      continue;
    }

    // This area is marked as wholly KEEP. Keep looking to see if
    // there are any real terminals here, though.
    if (split_terminal[i].terminal == NULL) {
      found_keep = TRUE;
      continue;
    }

    // This terminal overlaps the specified point, so we can use it to sample with.
    break;
  }

  if (i >= num_split_terminals) {
    if (found_keep) {
      // No split terminals overlap this point... at least, none with an actual
      // terminal pointer. We return the KEEP value that we found earlier.
      return keep_inside;
    } else {
      // NO terminals overlap this point. Perhaps it lies outside
      // the modeller's workspace. We discard it anyway.
      return !keep_inside;
    }
  } else {
    return world_point_intersect(split_terminal[i].terminal->rcsggraph, worldpt);
  }
}


int WorldCollection::world_point_intersect(solidrcsggraph rcsggraph, vector worldpt)
{
  int returnval;

  if (RCSG_ISLEAF(rcsggraph.flags)) {
    vector primpt;

    Matrix::apply(rcsggraph.leaf->prim->position[AFT], worldpt, primpt);

    if (rcsggraph.leaf->prim->prim->point_intersect(primpt) == IN)
      returnval = TRUE;
    else
      returnval = FALSE;

  } else {
    int left_result;

    left_result = world_point_intersect(rcsggraph.rcsg->left, worldpt);

    switch(rcsggraph.rcsg->link) {
    case PLUS:
      if (left_result)
        returnval = TRUE;
      else
        returnval = world_point_intersect(rcsggraph.rcsg->right, worldpt);
      break;
    case MINUS:
      if (left_result)
        returnval = !world_point_intersect(rcsggraph.rcsg->right, worldpt);
      else
        returnval = FALSE;
      break;
    case AND:
      if (left_result)
        returnval = world_point_intersect(rcsggraph.rcsg->right, worldpt);
      else
        returnval = FALSE;
      break;
    default:
      assert(0);
    }
  }

  if (RCSG_ISFLIPPED(rcsggraph.flags))
    returnval = !returnval;

  return returnval;
}

int debug_inserttriangles_collxid_break = -1;

// Insert the triangles from a particular surface that overlap the current cell
// under consideration.
void WorldCollection::insert_triangles(solidrcsggraph rcsggraph,
                                       int is_flipped, int cell_x, int cell_y, int cell_z)
{
  if (RCSG_ISFLIPPED(rcsggraph.flags))
    is_flipped = !is_flipped;

  if (RCSG_ISLEAF(rcsggraph.flags)) {
    primsplitinfo *info = rcsggraph.leaf->prim->splitinfo[psi_index];
    int i;

    assert(info != NULL);

    // Calculate the offset of the current cell is in the collectionspace of this primitive.
    Vector::mult(cell_x, info->cs_x_cell, info->cell_min);
    Vector::addfrac(cell_y, info->cs_y_cell, info->cell_min);
    Vector::addfrac(cell_z, info->cs_z_cell, info->cell_min);

    // Determine the min and max bounds of this cell in collectionspace.
    Vector::add(info->zero_max, info->cell_min, info->cell_max);
    Vector::add(info->zero_min, info->cell_min, info->cell_min);

    // Insert the triangles from each collection in this leafnode.
    for (i = 0; i < rcsggraph.leaf->num_collections; i++) {
#if _DEBUG
      if (rcsggraph.leaf->collectionbuf[i].collection->collx->collx_id == debug_inserttriangles_collxid_break)
        _CrtDbgBreak();
#endif

      insert_triangles(rcsggraph.leaf->collectionbuf[i].collection, is_flipped);
    }
  } else {
    insert_triangles(rcsggraph.rcsg->left,
                     is_flipped, cell_x, cell_y, cell_z);

    // Subtracted surfaces insert with triangles flipped.
    if (rcsggraph.rcsg->link == MINUS)
      is_flipped = !is_flipped;

    insert_triangles(rcsggraph.rcsg->right,
                     is_flipped, cell_x, cell_y, cell_z);
  }
}

void WorldCollection::insert_triangles(WorldCollection *collection, int is_flipped)
{
  int j, insertcell_x, insertcell_y, insertcell_z, insertcell_index;
  int cell_min[3], cell_max[3];
  string *cs_str;
  worldstring *ws_str;

  // Calculate where the cell lies with respect to all three axes.
  for (j = X; j <= Z; j++) {
    cell_min[j] = (int) (((collection->owner->splitinfo[psi_index]->cell_min[j] - collection->collx->bounds[0][j])
                                / collection->collx->cellsize[j]) + CELL_EDGE_TOLERANCE);
    if (cell_min[j] >= collection->collx->res)
      return;
    else if (cell_min[j] < 0)
      cell_min[j] = 0;

    cell_max[j] = (int) (((collection->owner->splitinfo[psi_index]->cell_max[j] - collection->collx->bounds[0][j])
                               / collection->collx->cellsize[j]) - CELL_EDGE_TOLERANCE);
    if (cell_max[j] < cell_min[j])
      return;
    else if (cell_max[j] >= collection->collx->res)
      cell_max[j] = collection->collx->res - 1;
  }

#if _DEBUG
  for (j = X; j <= Z; j++) {
    assert((cell_min[j] < collx->res) && (cell_min[j] >= 0));
    assert((cell_max[j] <= collx->res) && (cell_max[j] >= 0));
  }
#endif

  // If we're here, then we need to know about triangles inside our area.
  // Go calculate them, if we haven't already.
  collection->calculate();

  for (insertcell_z = cell_min[Z]; insertcell_z <= cell_max[Z]; insertcell_z++) {
    for (insertcell_y = cell_min[Y]; insertcell_y <= cell_max[Y]; insertcell_y++) {

      insertcell_index = cell_min[X] + insertcell_y * collection->collx->res
                                     + insertcell_z * SQR(collection->collx->res);

      for (insertcell_x = cell_min[X]; insertcell_x <= cell_max[X]; insertcell_x++, insertcell_index++) {
        if ((cs_str = collection->collx->cell[insertcell_index]) == NULL) {
          assert(collection->cell[insertcell_index] == NULL);
          continue;
        }
        for (ws_str = collection->cell[insertcell_index]; cs_str != NULL;
             cs_str = cs_str->stringptr, ws_str = ws_str->stringptr) {
          add_split_string(collection, cs_str, ws_str, is_flipped);
        }
        assert(ws_str == NULL);
      }
    }
  }
}

// Insert a string of triangles from a Collection into the list of
// triangles to be split against.
void WorldCollection::add_split_string(WorldCollection *collection, string *cs_str, worldstring *ws_str, int flip_tri)
{
  if (num_split >= splitbufsize) {
    splitbufsize += split_stringbuf_increment;

    split_string  = (splitstring *) realloc((void *) split_string,  splitbufsize * sizeof(splitstring));

    if (!split_string)
      fatal_error("WorldCollection::add_split_string", "Can't grow splitbuf",
                  "Out of memory.");
  }

  split_string[num_split].owner   = collection;
  split_string[num_split].flipped = flip_tri;
  split_string[num_split].cs_str  = cs_str;
  split_string[num_split].ws_str  = ws_str;
  num_split++;
}

// Traverse an RCSGgraph and precompute split data for all primitives and collections
// that it contains.
void WorldCollection::compute_splitinfo(solidrcsggraph rcsggraph)
{
  int i, j;
  double this_val;
  primsplitinfo *info;

  if (RCSG_ISLEAF(rcsggraph.flags)) {
    primtreeprimnode *p = rcsggraph.leaf->prim;

    if ((info = p->splitinfo[psi_index]) != NULL) {
      if (info->split_id == this_split_id) {
        // We've already got up-to-date information.
        return;
      }
    } else {
      // Create a new primsplitinfo node for this primitive.
      info = (primsplitinfo *) malloc(sizeof(primsplitinfo));
      if (!info)
        fatal_error("Collection::compute_splitinfo", "Can't allocate primsplitinfo",
                    "Out of memory.");

      p->splitinfo[psi_index] = info;
    }

    info->split_id = this_split_id;

    if (p->is_transformed) {
      // Take the zero-corner and cell offsets into collectionspace for this group of Collections.
      Matrix::apply(p->collection_trans[AFT], ws_zerocorner, info->zero_corner);
      Matrix::applynoshift(p->collection_trans[AFT], ws_cell_step[X], info->cs_x_cell);
      Matrix::applynoshift(p->collection_trans[AFT], ws_cell_step[Y], info->cs_y_cell);
      Matrix::applynoshift(p->collection_trans[AFT], ws_cell_step[Z], info->cs_z_cell);
    } else {
      Vector::set(info->zero_corner, ws_zerocorner);
      Vector::set(info->cs_x_cell, ws_cell_step[X]);
      Vector::set(info->cs_y_cell, ws_cell_step[Y]);
      Vector::set(info->cs_z_cell, ws_cell_step[Z]);
    }

    // Calculate the minimum and maximum offset in each axis for a cell.
    for (j = X; j <= Z; j++) {
      info->zero_min[j] = 0;
      info->zero_max[j] = 0;

      for (i = 1; i < 8; i++) {
        this_val = ((i & 0x1) ? info->cs_x_cell[j] : 0) +
                   ((i & 0x2) ? info->cs_y_cell[j] : 0) +
                   ((i & 0x4) ? info->cs_z_cell[j] : 0);
        if (this_val < info->zero_min[j])
          info->zero_min[j] = this_val;
        else if (this_val > info->zero_max[j])
          info->zero_max[j] = this_val;
      }
    }
    Vector::add(info->zero_corner, info->zero_min, info->zero_min);
    Vector::add(info->zero_corner, info->zero_max, info->zero_max);

  } else {
    // Compute info for the RCSGgraph's children.
    compute_splitinfo(rcsggraph.rcsg->left);
    compute_splitinfo(rcsggraph.rcsg->right);
  }
}



#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif

