/*
 * worldcollection.h: Definitions for class WorldCollection and supporting code.
 *
 * Contains code for the WorldCollections, which track a given instance
 * in worldspace of a Collection.
 */

// The information that needs to be stored in a primtreeprimnode for
// a split operation.
typedef struct primsplitinfo_ {
  // The split instance that this relates to.
  int split_id;

  // Collection-space coordinates for points in the Collection that's being split.
  // The zero corner, row, column, layer increments, the indices of the
  // max and min corners in each direction, and the max and min extent of the
  // zero cell and the current cell in collectionspace.
  vector zero_corner;
  vector cs_x_cell, cs_y_cell, cs_z_cell;
  vector zero_min, zero_max, cell_min, cell_max;
} primsplitinfo;

const int collection_tribuf_blocksize = 600;

// An extentlist is used for keeping track of whereabouts the terminals are
// that overlap our WorldCollection.
typedef struct extentlist_ {
  // The value in whichever dimension we're tracking.
  double value;

  // Holds either a Terminal or another layer of extentlists.
  union {
    struct splitterminal_ *termptr;
    struct extentlist_ *layer;
  };

  // Pointer to the next item in the extentlist.
  struct extentlist_ *next;
} extentlist;

const int split_stringbuf_increment = 20;

// A WorldCollection tracks a given instance in worldspace
// of a particular Collection. It holds the worldspace coordinates
// for all of its triangles.
class WorldCollection {
public:
  Modeller *modeller;

  // Our Collection.
  Collection *collx;

  // Parent WorldCollection (only used for debugging)
  class WorldCollection *parent;

  // The primtreeprimnode to whom this instance belongs.
  primtreeprimnode *owner;

  // NB: res and total_tri are the same as the corresponding Collection.

  worldstring **cell;     // 3D array, CUBE(res). A celllist of worldstrings of triangles or NULL for every cell.

  int reference_count;

  // Have we been calculated yet?
  calctype needs_calculation;

  // If we are to be split, the set of terminals to split us against.
  int calcsplit_num_terminals, calcsplit_keep_inside;
  struct splitterminal_ *calcsplit_terminal;

  // The ID of the last draw-pass to draw this collection. Needed so that we don't render ourselves
  // multiple times.
  int last_draw_id;

  /*
   * Split-pass variables and functions
   */

  // Temporary variables for the instance that we're currently splitting.
  extentlist *extent;
  int this_split_id, keep_inside, thiscell_overlaps, psi_index, cell_x, cell_y, cell_z;
  primtreeprimnode *split_owner;
  class TriSplitter *trisplitter;
  int num_split_terminals;
  struct splitterminal_ *split_terminal;
  struct splitterminal_ *current_terminal;
  vector ws_zerocorner, ws_cell_step[3], singlecell_minstep, singlecell_maxstep;
  int num_split, splitbufsize;
  struct splitstring_ *split_string;

  // Split this Collection against another object. The other object is represented
  // by a set of solidterminalnodes, each covering some part of the collection.
  //
  // Returns TRUE if the surface is in fact split.
  int split(int num_terminals, struct splitterminal_ *splitbuf, WorldCollection *child, int keep_inside);

  // Precompute information for the terminals that we will be splitting against.
  void precompute_terminal_info(void);

  // Insert a terminal's worldspace extents into our extents framework.
  void insert_into_extents(struct splitterminal_ *termptr);
  extentlist *insert_into_extents(struct splitterminal_ *termptr, int ax, extentlist *list);

  // Traverse an extentlist looking for terminals which overlap the
  // present cell.
  void traverse_extent(int cell_index, struct splitterminal_ **overlap_keep, struct splitterminal_ **overlap_split);
  void traverse_extent(int cell_index, struct splitterminal_ **overlap_keep, struct splitterminal_ **overlap_split,
                       int ax, extentlist *list);

  // Allocate or deallocate an extentlist.
  extentlist *allocate_extent(void);
  void deallocate_extent(extentlist *kill_me, int ax = Z);

  // Traverse an rcsggraph and precompute split data for all primitives and collections
  // that it contains.
  void compute_splitinfo(solidrcsggraph rcsggraph);

  // Insert the triangles from a particular surface that overlap the current cell
  // under consideration. Also returns in/out status.
  void insert_triangles(solidrcsggraph rcsggraph, int is_flipped, int cell_x, int cell_y, int cell_z);
  void insert_triangles(WorldCollection *collection, int is_flipped);

  // Insert a string of triangles from a Collection into the list of
  // triangles to be split against.
  void add_split_string(WorldCollection *collection, string *tri_string, worldstring *ws_str, int flip_tri);

  // Point-intersect w.r.t the RCSGgraph.
  int world_point_intersect(int cell_index, vector worldpt);
  int world_point_intersect(solidrcsggraph rcsggraph, vector worldpt);

  /*
   * End of split-pass variables and functions
   */

  // Perform all of the calculations necessary to create a WorldCollection.
  int calculate(int cell_index = -1);

  // Copy a cell full of triangles from collectionspace to our worldspace.
  worldstring *make_worldspace_cell(string *cs_str);

  // Add a reference to the WorldCollection.
  void add_reference(void);

  // Remove a reference from the WorldCollection. Returns TRUE if the WorldCollection should
  // be deleted now.
  int remove_reference(void);

  // Construct a WorldCollection by polygonizing a primitive surface for a
  // given octree in a leafnode.
  WorldCollection(primtreeprimnode *owner,
                  primobjectnode *surfacenode,
                  int res, transform collectionspace,
                  box bounds, Modeller *modeller);

  // Create a new WorldCollection from a pre-existing **primitive** Collection.
  WorldCollection(primtreeprimnode *owner, Collection *collx);

  // Create an new WorldCollection that is the split fragment of some parent.
  WorldCollection(int num_terminals, struct splitterminal_ *splitbuf, int keep_inside,
                  WorldCollection *parent_collection);

  // Delete the WorldCollection.
  ~WorldCollection(void);
};
