/*
 * transview.cc: Routines for TransView.
 *
 * Contains platform-independent code to interact with the transformation views.
 */

/* Collection-safe */

#include "stdafx.h"

#ifdef USE_NAMESPACES
namespace Mirage {
#endif

// Update the transformation view so that the correct
// transformation (if any) is selected.
void TransView::update(void)
{
  SkitterList *skitterlist = modeller->skitterlist;
  TransformList *tlist;
  TransformListNode *tlnode;
  int i;

  // We aren't in the middle of a drag operation.
  movetoline = -1;

  if (skitterlist->selected_node == NULL) {
    // Nothing is selected. Therefore there are no
    // transformations for us to display.
    num_trans = 0;
    selected_trans = -1;
  } else {
    assert(skitterlist->num_copies > 0);

    // Get the transformation list.
    tlist = skitterlist->selected_node->transformlist;
    assert(tlist != NULL);

    // Which transformation is selected?
    if (skitterlist->current_dmd)
      selected_trans = (num_trans - 1) - skitterlist->current_dmd->node_index;
    else
      selected_trans = -1;

    num_trans = tlist->num_transforms;
    assert(num_trans < MAX_TRANSVIEW_SIZE);

    // For each transformation, get its name and parameters for display.
    for (i = 0, tlnode = tlist->head; i < num_trans; i++, tlnode = tlnode->next) {
      if (((num_trans - 1) - i) == selected_trans)
        assert(tlnode == skitterlist->current_dmd->transformation);

      // NB: We list transformations in the reverse order to the one
      // in which the transformationlist stores them.
      tlnode->trans->get_name  (namestring [(num_trans - 1) - i]);
      tlnode->trans->get_params(paramstring[(num_trans - 1) - i]);
    }
  }

  // Redraw the transformation view window.
  redraw();
}

// Check to see if a move operation needs to
// be performed (after a mouse-up).
void TransView::check_move(void)
{
  TransformList *tlist;
  TransformListNode *movenode, *new_position;
  int i, tlist_index;

  if (movetoline == -1) {
    // We aren't doing a move operation.
    return;
  }

  if ((modeller->skitterlist->selected_node == NULL) ||
      (modeller->skitterlist->current_dmd == NULL)) {
    // Nothing to move.
    return;
  }

  // We have to move the transformation at
  // selected_trans, to the position indicated by
  // movetoline.
  tlist = modeller->skitterlist->selected_node->transformlist;
  movenode = modeller->skitterlist->current_dmd->transformation;
  
  // movetoline is the boundary between lines movetoline-1 and
  // movetoline. Because the top of the transview list is the
  // tail of the transformlist, this means that the node
  // should be inserted between nodes (num_trans - 1) - movetoline
  // and (num_trans - 1) - (movetoline - 1).
  
  if ((movetoline == selected_trans) || (movetoline == selected_trans + 1)) {
    // We are trying to move a transformation to its current place.
    // This is not a valid move operation. Do nothing.
    return;
  }
  
  // move_transformation requires the node that is before the position
  // to move to, i.e. the node with index (num_trans - 1) - movetoline.
  
  assert((movetoline >= 0) && (movetoline <= tlist->num_transforms));
  if (movetoline == num_trans) {
    // Move to the end of the transview list, which
    // is the head of the transformlist.
    new_position = NULL;
  } else {
    tlist_index = (num_trans - 1) - movetoline;
    for (i = 0, new_position = tlist->head;
    ((i < tlist_index) && (new_position != NULL));
    i++, new_position = new_position->next)
      ;
    assert(new_position != NULL);
  }
  
  // Perform the move operation.
  tlist->move_transformation(movenode, new_position);
  
  // What index has the moved node gone to?
  if (movetoline < selected_trans)
    selected_trans = movetoline;
  else
    selected_trans = movetoline - 1;
  movetoline = -1;
  
  // Propagate the new transformation ordering through the
  // TransformationList.
  tlist->propagate_matrices();
  
  // Shift the Skitters around.
  modeller->skitterlist->update_skitterlist();
  
  // Select this node.
  modeller->skitterlist->dmd_select(modeller->skitterlist->copies[0].skitter[(num_trans - 1) - selected_trans]);
  
  // Update the polygonization.
  modeller->csg_graph->update_transformed_object(modeller->skitterlist->selected_node);
  modeller->update_all_views();  
}

// Delete the currently selected transformation.
void TransView::delete_trans(void)
{
  // Nothing selected to delete?
  if (selected_trans == -1)
    return;

  // Delete the transformation.
  modeller->skitterlist->selected_node->transformlist->delete_transformation(modeller->skitterlist->current_dmd->transformation);

  // Nothing selected any more.
  selected_trans = -1;

  // Propagate the new transformation ordering through the
  // TransformationList.
  modeller->skitterlist->selected_node->transformlist->propagate_matrices();

  // Shift the Skitters around.
  modeller->skitterlist->update_skitterlist();
  modeller->skitterlist->dmd_deselect();

  // Update the polygonization.
  modeller->csg_graph->update_transformed_object(modeller->skitterlist->selected_node);
  modeller->update_all_views();  
}

// Add a new transformation.
void TransView::add_trans(trans_type ttype)
{
  switch(ttype) {
  case TRANS_TRANSLATION:
    add_trans(new ObjectTranslation);
    break;
  case TRANS_ROTATION:
    add_trans(new ObjectRotation);
    break;
  case TRANS_SCALE:
    add_trans(new ObjectScale);
    break;
  default:
    assert(0);
  }
}


void TransView::add_trans(ObjectTransformation *new_trans)
{
  TransformList *tlist;
  TransformListNode *tlptr;
  int i;

  tlist = modeller->skitterlist->selected_node->transformlist;

  // We want to either transform everything (if no transformation is selected)
  // or transform everything "after" the selected transformation.
  if (selected_trans == -1) {
    // Transform everything, by adding the new transformation
    // after everything.
    for (tlptr = tlist->head; (tlptr != NULL) && (tlptr->next != NULL); tlptr = tlptr->next)
      ;
  } else {
    // We want to insert immediately after the selected transformation.
    // This is (num_trans - 1) - selected_trans in the chain.
    for (i = 0, tlptr = tlist->head; i < (num_trans - 1) - selected_trans; tlptr = tlptr->next)
      assert(tlptr != NULL);
  }

  // Insert the new transformation at the right place (directly after tlptr).
  tlist->add_transformation(new_trans, tlptr);

  // Propagate the new transformation's effect through the
  // TransformationList.
  tlist->propagate_matrices();

  // Shift the Skitters around.
  modeller->skitterlist->update_skitterlist();
  
  // Select the new transformation. It will have moved
  // everything one down the list, meaning that selected_trans
  // is the same as it was.
  if (selected_trans == -1)
    selected_trans = 0;

  // Select this transformation. NB: num_trans hasn't been incremented yet (it will be
  // updated in update_all_transviews, called from dmd_select) so we don't use 
  // (num_trans - 1) - selected_trans but instead we use num_trans - selected_trans.
  modeller->skitterlist->dmd_select(modeller->skitterlist->copies[0].skitter[num_trans - selected_trans]);

  // Update the polygonization.
  modeller->csg_graph->update_transformed_object(modeller->skitterlist->selected_node);
  modeller->update_all_views();  
}


#ifdef USE_NAMESPACES
} // namespace Mirage {
#endif
